/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/BaseModule.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core;

import modules.wireless.partition.Partition;
import core.sync.Lp;
import core.util.CustomLogger;
import core.util.EventLogger;

/**
 * Base class for all simulation modules.
 * 
 * @author Patrick Peschlow
 * @author Tobias Honecker
 * 
 * @version 1.0
 */
public abstract class BaseModule {
	/**
	 * The unique ID of this module. Assigned automatically during
	 * initialization.
	 */
	protected final int id_;

	/**
	 * An ID assigned by the programmer. Should be unique within the class of
	 * the module.
	 */
	protected final int number_;

	/**
	 * The LP which contains this BaseModule. During initialization, this is the
	 * InitLp. This field is transient because the LP should not be included by
	 * the serialization mechanism in case a module is transferred to a remote
	 * LP over the network.
	 */
	protected transient Lp lp_;

	/**
	 * Default constructor. Registers the module to the {@link core.ModuleRoot},
	 * obtaining a unique module ID.
	 * 
	 * @param number
	 *            the class-internal programmer-assigned number for this module.
	 */
	public BaseModule(int number) {
		number_ = number;
		lp_ = ModuleRoot.getInitLp();
		id_ = ModuleRoot.getRoot().registerModule(this, number);
		if (CustomLogger.isDebugEnabled()) {
			CustomLogger.debug("Registered module " + this);
		}
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public final boolean equals(Object o) {
		if (o == this) {
			return true;
		}
		if (o == null || !(o instanceof BaseModule)) {
			return false;
		}
		return id_ == ((BaseModule) o).id_;
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public final int hashCode() {
		return id_;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "Module " + id_ + " (" + getClass().getSimpleName() + "["
				+ number_ + "])";
	}

	/**
	 * Provides the module ID of this module.
	 * 
	 * @return the module ID of this module.
	 */
	public final int getId() {
		return id_;
	}

	/**
	 * Provides the number of this module.
	 * 
	 * @return the number of this module.
	 */
	public final int getNumber() {
		return number_;
	}

	/**
	 * Sets the LP which contains this module.
	 * 
	 * @param lp
	 *            the LP which contains this module.
	 */
	public final void setLp(Lp lp) {
		lp_ = lp;
	}

	/**
	 * Executes an event. Called by the LP.
	 * 
	 * @param execTime
	 *            the execution time (timestamp) of the event.
	 * @param srcModuleId
	 *            the source module ID of the event.
	 * @param eventId
	 *            the ID of the event.
	 * @param parameter
	 *            the parameter of the event.
	 */
	@SuppressWarnings("rawtypes")
	public final void execute(long execTime, int srcModuleId, Enum eventId,
			Object parameter) {
		if (EventLogger.isEnabled()) {
			lp_.trace(execTime, number_, eventId, parameter);
		}
		activate(execTime, srcModuleId, eventId, parameter);
	}

	/**
	 * Convenience method. Asks the ModuleRoot for the module ID of a
	 * combination of module class and number.
	 * 
	 * @param moduleClass
	 *            the class of the module.
	 * @param number
	 *            the number of the module.
	 * 
	 * @return the corresponding module ID.
	 */
	@SuppressWarnings("rawtypes")
	public final int findModuleId(Class moduleClass, int number) {
		return ModuleRoot.getRoot().findModuleId(moduleClass, number);
	}

	/**
	 * Schedules a new event.
	 * 
	 * @param execTime
	 *            the execution time for the event.
	 * @param dstModuleId
	 *            the module ID of the module handling this event.
	 * @param eventId
	 *            the event ID.
	 * @param parameter
	 *            the event parameter.
	 */
	@SuppressWarnings("rawtypes")
	public final void scheduleNewEvent(long execTime, int dstModuleId,
			Enum eventId, Object parameter) {
		assert execTime >= 0 : "Event timestamp < 0... perhaps an overflow?";
		assert lp_ != null;
		// Special cases to ensure identical results of sequential and parallel
		// simulations.
		if (eventId == Partition.Events.LPPROPAGATION
				|| eventId == Partition.Events.END_LPPROPAGATION) {
			lp_.createNewEvent(execTime, dstModuleId, id_, eventId, parameter,
					0);
		} else {
			final int priority = lp_.getAndIncrementPriority();
			lp_.createNewEvent(execTime, dstModuleId, id_, eventId, parameter,
					priority);
		}
	}

	/**
	 * Schedules a new event.
	 * 
	 * @param execTime
	 *            the execution time for the event.
	 * @param dstModuleId
	 *            the module ID of the module handling this event.
	 * @param eventId
	 *            the event ID.
	 * @param parameter
	 *            the event parameter.
	 * @param schedTime
	 *            the scheduling time of the event.
	 */
	@SuppressWarnings("rawtypes")
	public final void scheduleNewEventWithSchedTime(long execTime,
			int dstModuleId, Enum eventId, Object parameter, long schedTime) {
		assert execTime >= 0 : "Event timestamp < 0... perhaps an overflow?";
		assert lp_ != null;
		// Special cases to ensure identical results of sequential and parallel
		// simulations.
		if (eventId == Partition.Events.LPPROPAGATION
				|| eventId == Partition.Events.END_LPPROPAGATION) {
			lp_.createNewEvent(execTime, dstModuleId, id_, eventId, parameter,
					0, schedTime);
		} else {
			final int priority = lp_.getAndIncrementPriority();
			lp_.createNewEvent(execTime, dstModuleId, id_, eventId, parameter,
					priority, schedTime);
		}
	}

	/**
	 * Schedules a new event.
	 * 
	 * @param execTime
	 *            the execution time for the event.
	 * @param dstModuleId
	 *            the module ID of the module handling this event.
	 * @param eventId
	 *            the event ID.
	 * @param parameter
	 *            the event parameter.
	 * @param srcModuleId
	 *            the source module ID of the event.
	 */
	@SuppressWarnings("rawtypes")
	public final void scheduleNewEventWithSrcModuleId(long execTime,
			int dstModuleId, Enum eventId, Object parameter, int srcModuleId) {
		assert execTime >= 0 : "Event timestamp < 0... perhaps an overflow?";
		assert lp_ != null;
		// Special cases to ensure identical results of sequential and parallel
		// simulations.
		if (eventId == Partition.Events.LPPROPAGATION
				|| eventId == Partition.Events.END_LPPROPAGATION) {
			lp_.createNewEvent(execTime, dstModuleId, srcModuleId, eventId,
					parameter, 0);
		} else {
			final int priority = lp_.getAndIncrementPriority();
			lp_.createNewEvent(execTime, dstModuleId, srcModuleId, eventId,
					parameter, priority);
		}
	}

	/**
	 * Schedules a new event.
	 * 
	 * @param execTime
	 *            the execution time for the event.
	 * @param dstModuleId
	 *            the module ID of the module handling this event.
	 * @param eventId
	 *            the event ID.
	 * @param parameter
	 *            the event parameter.
	 * @param srcModuleId
	 *            the source module ID of the event.
	 * @param schedTime
	 *            the scheduling time of the event.
	 */
	@SuppressWarnings("rawtypes")
	public final void scheduleNewEventWithSrcModuleIdAndSchedTime(
			long execTime, int dstModuleId, Enum eventId, Object parameter,
			int srcModuleId, long schedTime) {
		assert execTime >= 0 : "Event timestamp < 0... perhaps an overflow?";
		assert lp_ != null;
		// Special cases to ensure identical results of sequential and parallel
		// simulations.
		if (eventId == Partition.Events.LPPROPAGATION
				|| eventId == Partition.Events.END_LPPROPAGATION) {
			lp_.createNewEvent(execTime, dstModuleId, srcModuleId, eventId,
					parameter, 0, schedTime);
		} else {
			final int priority = lp_.getAndIncrementPriority();
			lp_.createNewEvent(execTime, dstModuleId, srcModuleId, eventId,
					parameter, priority, schedTime);
		}
	}

	/**
	 * Called by ModuleRoot after all instances of classes extending BaseModule
	 * have been created. It should be used to establish connections between the
	 * modules, schedule initial events, etc.
	 */
	public abstract void init();

	/**
	 * Called whenever an event should be handled by an instance of this class,
	 * either by the scheduler and LP via execute() or directly by another
	 * module.
	 * 
	 * @param execTime
	 *            the time of the event.
	 * @param srcModuleId
	 *            the unique ModuleId of the module which scheduled the event.
	 * @param eventId
	 *            the unique EventId of the event type.
	 * @param parameter
	 *            the parameter of the event (which can be null for some
	 *            events).
	 */
	@SuppressWarnings("rawtypes")
	public abstract void activate(long execTime, int srcModuleId, Enum eventId,
			Object parameter);
}
