/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/Config.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core;

import jist.swans.Constants;
import jist.swans.field.Fading;
import jist.swans.field.PathLoss;
import modules.wireless.scenario.NodeConfig;
import modules.wireless.scenario.NodeConfig.ScenarioType;
import modules.wireless.scenario.SimulationCallback;
import rng.DefaultRandomManager;
import rng.RandomManager;
import rng.RandomStream;
import core.scheduler.BinaryTournamentQueue;
import core.scheduler.LadderQueue;
import core.scheduler.ListQueue;
import core.scheduler.ShiftedHeapQueue;
import core.sync.DistributedSimController;
import core.sync.SeqSimController;
import core.sync.ParallelSimController;
import core.util.CustomLogger;
import core.util.EventLogger;

/**
 * Configuration class which allows (and requires) to set parameters before the
 * simulation starts.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class Config {
	/**
	 * Checks whether the simulation is already running (then parameters must
	 * not be changed anymore).
	 */
	private static boolean isRunning = false;

	/**
	 * Performs a sanity check before the simulation starts and exits the
	 * application of an invalid parameter setting is discovered.
	 */
	public static void sanityCheck() {
		if (simType == SimType.SEQUENTIAL && lookahead != Lookahead.NONE) {
			if (CustomLogger.isWarningEnabled()) {
				CustomLogger
						.warning("For sequential simulations, you have to use Lookahead.NONE. Setting lookahead to Lookahead.NONE");
			}
			lookahead = Lookahead.NONE;
			usesProtocolLookahead = false;
		}
		if (simType == SimType.PARALLEL || simType == SimType.DISTRIBUTED) {
			if (numPartitions < 2) {
				if (barrier != BarrierType.SINGLE_PARTITION) {
					if (CustomLogger.isWarningEnabled()) {
						CustomLogger
								.warning("For parallel simulations with one partition, you have to use single partition barrier. Setting barrier to Barrier.SINGLE_PARTITION.");
					}
					barrier = BarrierType.SINGLE_PARTITION;
				}
			} else {
				if (barrier == BarrierType.SINGLE_PARTITION) {
					if (CustomLogger.isWarningEnabled()) {
						CustomLogger
								.warning("For parallel simulations with two or more partitions, you cannot use single partition barrier. Setting barrier to Barrier.TOURNAMENT.");
					}
					barrier = BarrierType.TOURNAMENT;
				}
			}
			if (scheduler == Scheduler.LADDER) {
				if (CustomLogger.isWarningEnabled()) {
					CustomLogger
							.warning("Ladder queue scheduler does not work with parallel simulation. Setting scheduler to Scheduler.ShiftedHeap.");
				}
				scheduler = Scheduler.SHIFTEDHEAP;
			}
		}
		if (random == null) {
			if (CustomLogger.isWarningEnabled()) {
				CustomLogger
						.warning("Forgot to set random manager type! Setting random manager to RNGSTREAM.");
			}
			random = Random.RNGSTREAM;
		}
		if (randomSeed == -1) {
			if (CustomLogger.isWarningEnabled()) {
				CustomLogger
						.warning("Forgot to set random seed! Setting seed to 1.");
			}
			randomSeed = 1;
		}
		if (propagation == Propagation.INDIVIDUAL && usesLazyMac) {
			configError("Lazy MAC currently works only with common propagation delay.");
		}
		if (lookahead == Lookahead.PROTOCOL && usesLazyMac) {
			configError("Lazy MAC currently works only with event lookahead.");
		}
		if (lookahead == Lookahead.PROTOCOL
				&& NodeConfig.getScenarioType() == ScenarioType.MESH) {
			if (CustomLogger.isWarningEnabled()) {
				CustomLogger
						.warning("The implementation of protocol lookahead for mesh scenarios is still buggy and may be improved. Run at your own risk.");
			}
		}
		if (endTime <= startTime) {
			configError("The simulation end time has to be greater than start time.");
		}
		isRunning = true;
		NodeConfig.simulationStarts();
	}

	/**
	 * Simulation type.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	@SuppressWarnings("rawtypes")
	public static enum SimType {
		/**
		 * Sequential simulation.
		 */
		SEQUENTIAL(SeqSimController.class),
		/**
		 * Parallel simulation on a multi-core or multi-processor machine.
		 */
		PARALLEL(ParallelSimController.class),
		/**
		 * Distributed simulation on interconnected hosts (may include parallel
		 * simulation.
		 */
		DISTRIBUTED(DistributedSimController.class);

		/**
		 * The class of the simulation controller.
		 */
		final Class simCtrlClass_;

		/**
		 * Constructor.
		 * 
		 * @param clazz
		 *            the class of the simulation controller.
		 */
		private SimType(Class clazz) {
			simCtrlClass_ = clazz;
		}

		/**
		 * Provides the class of the simulation controller.
		 * 
		 * @return the class of the simulation controller.
		 */
		public Class getSimControllerClass() {
			return simCtrlClass_;
		}
	}

	/**
	 * Barrier type used in parallel simulations.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	public static enum BarrierType {
		/**
		 * Butterfly barrier.
		 */
		BUTTERFLY,
		/**
		 * Dissemination barrier.
		 */
		DISSEMINATION,
		/**
		 * Tournament barrier.
		 */
		TOURNAMENT,
		/**
		 * Dummy barrier used for "parallel" simulation with a single partition
		 * only.
		 */
		SINGLE_PARTITION;
	}

	/**
	 * Lookahead strategy used in parallel/distributed simulations.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	public static enum Lookahead {
		/**
		 * Protocol lookahead on MAC/PHY layers.
		 */
		PROTOCOL,
		/**
		 * Event lookahead.
		 */
		EVENT,
		/**
		 * No lookahead.
		 */
		NONE,
		/**
		 * Ideal Simulation Protocol (method to collect ideal lookahead, used
		 * for testing purposes only).
		 */
		ISP;
	}

	/**
	 * Scheduler type used by the LPs.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	@SuppressWarnings("rawtypes")
	public static enum Scheduler {
		/**
		 * Binary tournament queue.
		 */
		BINARYTOURNAMENT(BinaryTournamentQueue.class),
		/**
		 * Sorted linked list queue.
		 */
		LIST(ListQueue.class),
		/**
		 * Binary heap queue.
		 */
		SHIFTEDHEAP(ShiftedHeapQueue.class),
		/**
		 * Ladder queue.
		 */
		LADDER(LadderQueue.class);

		/**
		 * The class of the event queue.
		 */
		final Class eventQueueClass_;

		/**
		 * Constructor.
		 * 
		 * @param clazz
		 *            the class of the event queue.
		 */
		private Scheduler(Class clazz) {
			eventQueueClass_ = clazz;
		}

		/**
		 * Provides the class of the event queue.
		 * 
		 * @return the class of the event queue.
		 */
		public Class getEventQueueClass() {
			return eventQueueClass_;
		}
	}

	/**
	 * Random generator source package used.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	@SuppressWarnings("rawtypes")
	public static enum Random {
		/**
		 * Random manager using the RngStream class.
		 */
		RNGSTREAM(DefaultRandomManager.class);

		/**
		 * The class of the random manager.
		 */
		final Class randomManagerClass_;

		/**
		 * Constructor.
		 * 
		 * @param clazz
		 *            the class of the random manager.
		 */
		private Random(Class clazz) {
			randomManagerClass_ = clazz;
		}

		/**
		 * Provides the class of the random manager.
		 * 
		 * @return the class of the random manager.
		 */
		public Class getRandomManagerClass() {
			return randomManagerClass_;
		}
	}

	/**
	 * Propagation type used.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	public static enum Propagation {
		/**
		 * Individual propagation delay calculation based on transmitter and
		 * receiver positions. The smallest possible delay is assumed to be a
		 * nanosecond.
		 */
		INDIVIDUAL(Constants.NANO_SECOND),
		/**
		 * Common propagation delay for all transmissions. The common delay is
		 * assumed to be a microsecond.
		 */
		COMMON(Constants.MICRO_SECOND);

		/**
		 * The minimum (or: common) propagation delay.
		 */
		final long propagationDelay_;

		/**
		 * Constructor.
		 * 
		 * @param propagationDelay
		 *            the minimum (or: common) propagation delay.
		 */
		private Propagation(long propagationDelay) {
			propagationDelay_ = propagationDelay;
		}

		/**
		 * Provides the minimum (or: common) propagation delay.
		 * 
		 * @return the minimum (or: common) propagation delay.
		 */
		public long getDelay() {
			return propagationDelay_;
		}
	}

	/**
	 * Pathloss model used.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	@SuppressWarnings("rawtypes")
	public static enum PathlossModel {
		/**
		 * No pathloss.
		 */
		NONE(PathLoss.None.class) {
			/**
			 * No random numbers needed.
			 */
			public void setupRngs() {
				return;
			}
		},
		/**
		 * Free space pathloss.
		 */
		FREESPACE(PathLoss.FreeSpace.class) {
			/**
			 * No random numbers needed.
			 */
			public void setupRngs() {
				return;
			}
		},
		/**
		 * Two-ray ground pathloss.
		 */
		TWORAY(PathLoss.TwoRay.class) {
			/**
			 * No random numbers needed.
			 */
			public void setupRngs() {
				return;
			}
		},
		/**
		 * Shadowing model.
		 */
		SHADOWING(PathLoss.Shadowing.class) {
			/**
			 * Random stream initialization.
			 */
			public void setupRngs() {
				RandomStream[] pathlossRNGs = new RandomStream[numVirtualPartitions];
				for (int i = 0; i < numVirtualPartitions; i++) {
					pathlossRNGs[i] = RandomManager.getInstance()
							.getNormalSubstream(0.0, PathLoss.Shadowing.std_dB);
				}
				PathLoss.setRNGs(pathlossRNGs);
			}
		};

		/**
		 * The class of the pathloss model.
		 */
		final Class pathlossClass_;

		/**
		 * Constructor.
		 * 
		 * @param clazz
		 *            the class of the pathloss model.
		 */
		private PathlossModel(Class clazz) {
			pathlossClass_ = clazz;
		}

		/**
		 * Provides the class of the pathloss model.
		 * 
		 * @return the class of the pathloss model.
		 */
		public Class getPathlossClass() {
			return pathlossClass_;
		}

		/**
		 * Sets up the random number streams to be used by the model (if
		 * applicable).
		 */
		public abstract void setupRngs();
	}

	/**
	 * Fading model used.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	@SuppressWarnings("rawtypes")
	public static enum FadingModel {
		/**
		 * No fading.
		 */
		NONE(Fading.None.class) {
			/**
			 * No random numbers needed.
			 */
			public void setupRngs() {
				return;
			}
		},
		/**
		 * Rayleigh fading.
		 */
		RAYLEIGH(Fading.Rayleigh.class) {
			/**
			 * Random stream initialization.
			 */
			public void setupRngs() {
				RandomStream[] fadingRNGs = new RandomStream[numVirtualPartitions];
				for (int i = 0; i < numVirtualPartitions; i++) {
					fadingRNGs[i] = RandomManager.getInstance()
							.getUniform01Substream();
				}
				Fading.setRNGs(fadingRNGs);
			}
		},
		/**
		 * Rician fading.
		 */
		RICIAN(Fading.Rician.class) {
			/**
			 * Random stream initialization.
			 */
			public void setupRngs() {
				RandomStream[] fadingRNGs = new RandomStream[numVirtualPartitions];
				for (int i = 0; i < numVirtualPartitions; i++) {
					fadingRNGs[i] = RandomManager.getInstance()
							.getUniform01Substream();
				}
				Fading.setRNGs(fadingRNGs);
			}
		};

		/**
		 * The class of the fading model.
		 */
		final Class fadingClass_;

		/**
		 * Constructor.
		 * 
		 * @param clazz
		 *            the class of the fading model.
		 */
		private FadingModel(Class clazz) {
			fadingClass_ = clazz;
		}

		/**
		 * Provides the class of the fading model.
		 * 
		 * @return the class of the fading model.
		 */
		public Class getFadingClass() {
			return fadingClass_;
		}

		/**
		 * Sets up the random number streams to be used by the model (if
		 * applicable).
		 */
		public abstract void setupRngs();
	}

	/**
	 * Node movement implementation used.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	public static enum Movement {
		/**
		 * Continuous movement. Exact node positions are calculated on demand.
		 */
		CONTINUOUS,
		/**
		 * Discrete movement. Node positions are updated in fixed intervals (and
		 * assumed to stay the same during the intervals).
		 */
		DISCRETE;
	}

	/**
	 * Name of the (event) trace file of the simulation.
	 */
	private static String tracefile = null;

	/**
	 * Simulation type.
	 */
	private static SimType simType = null;

	/**
	 * Barrier type.
	 */
	private static BarrierType barrier = null;

	/**
	 * Lookahead strategy.
	 */
	private static Lookahead lookahead = null;

	/**
	 * Convenience variable. Whether the parallel/distributed simulation uses
	 * the protocol lookahead strategy.
	 */
	private static boolean usesProtocolLookahead = false;

	/**
	 * The scheduler type.
	 */
	private static Scheduler scheduler = null;

	/**
	 * The global random seed of the simulation.
	 */
	private static int randomSeed = -1;

	/**
	 * The random number generator source package.
	 */
	private static Random random = null;

	/**
	 * Propagation model.
	 */
	private static Propagation propagation = null;

	/**
	 * Pathloss model.
	 */
	private static PathlossModel pathlossModel = null;

	/**
	 * Fading model.
	 */
	private static FadingModel fadingModel = null;

	/**
	 * BER table.
	 */
	private static String berTableFile = "jist/swans/radio/default.BERtable";

	/**
	 * Movement type.
	 */
	private static Movement movement = null;

	/**
	 * Name of the movement file (BonnMotion).
	 */
	private static String movementFile = null;

	/**
	 * Interval of node position updates (if the Discrete strategy is used).
	 */
	private static long movementInterval = Constants.SECOND;

	/**
	 * Whether the Lazy MAC strategy is used.
	 */
	private static boolean usesLazyMac = false;

	/**
	 * Whether to collect ISP (Ideal Simulation Protocol) data in a parallel
	 * simulation.
	 */
	private static boolean collectDataISP = false;

	/**
	 * The number of partitions/LPs used in a parallel/distributed simulation.
	 */
	private static int numPartitions = -1;

	/**
	 * The number of virtual partitions used. Controls the assignment of random
	 * number streams to nodes and is important for repeatability.
	 */
	private static int numVirtualPartitions = -1;

	/**
	 * The start time of the simulation (cannot be configured).
	 */
	private static final long startTime = 0;

	/**
	 * The end time of the simulation.
	 */
	private static long endTime = -1;

	/**
	 * Callback for final processing / output after the simulation.
	 */
	private static SimulationCallback callback;

	/**
	 * Sets the name of the trace file.
	 * 
	 * @param name
	 *            the file name.
	 */
	public static void setTracefile(String name) {
		if (isRunning) {
			invalidConfigChange(name);
		}
		tracefile = name;
		EventLogger.setEnabled(true);
		EventLogger.setOutput(tracefile);
		System.out.println("EventLogger writing event trace to " + tracefile);
	}

	/**
	 * Sets the simulation type.
	 * 
	 * @param arg
	 *            the simulation type.
	 * @param numPart
	 *            the number of partitions.
	 * @param numVirtualPart
	 *            the number of virtual partitions (which is the number of
	 *            random number generators used, to ensure reproducibility
	 *            between sequential and parallel simulations). Can be ignored
	 *            unless results have to be directly compared.
	 */
	public static void setSimType(SimType arg, int numPart, int numVirtualPart) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		if (numPart < 1) {
			configError("numPartitions must be positive!");
		}
		if (arg == SimType.SEQUENTIAL && numPart > 1) {
			if (CustomLogger.isWarningEnabled()) {
				CustomLogger
						.warning("For sequential simulations, you have to use one partition! Setting numPartitions to 1.");
			}
			numPart = 1;
		}
		if (numPart > numVirtualPart) {
			configError("numPartitions must not be larger than numVirtualPartitions!");
		}
		if (numVirtualPart % numPart != 0) {
			configError("numVirtualPartitions % numPartitions must be zero!");
		}
		Config.simType = arg;
		Config.numPartitions = numPart;
		Config.numVirtualPartitions = numVirtualPart;
	}

	/**
	 * Sets the barrier type.
	 * 
	 * @param arg
	 *            the barrier type.
	 */
	public static void setBarrier(BarrierType arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		barrier = arg;
	}

	/**
	 * Sets the lookahead strategy.
	 * 
	 * @param arg
	 *            the lookahead strategy.
	 */
	public static void setLookahead(Lookahead arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		lookahead = arg;
		if (arg == Lookahead.PROTOCOL) {
			usesProtocolLookahead = true;
		}
	}

	/**
	 * Sets the scheduler type.
	 * 
	 * @param arg
	 *            the scheduler type.
	 */
	public static void setScheduler(Scheduler arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		scheduler = arg;
	}

	/**
	 * Sets the random seed.
	 * 
	 * @param arg
	 *            the random seed.
	 */
	public static void setRandomSeed(int arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		randomSeed = arg;
	}

	/**
	 * Sets the random manager type.
	 * 
	 * @param arg
	 *            the random manager type.
	 */
	public static void setRandom(Random arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		random = arg;
	}

	/**
	 * Sets the propagation model.
	 * 
	 * @param arg
	 *            the propagation model.
	 */
	public static void setPropagation(Propagation arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		propagation = arg;
	}

	/**
	 * Sets the pathloss model.
	 * 
	 * @param arg
	 *            the pathloss model.
	 */
	public static void setPathlossModel(PathlossModel arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		pathlossModel = arg;
	}

	/**
	 * Sets the fading model.
	 * 
	 * @param arg
	 *            the fading model.
	 */
	public static void setFadingModel(FadingModel arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		fadingModel = arg;
	}

	/**
	 * Sets the BER table file name.
	 * 
	 * @param arg
	 *            the BER table file name.
	 */
	public static void setBerTableFile(String arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		berTableFile = arg;
	}

	/**
	 * Sets the movement strategy.
	 * 
	 * @param arg
	 *            the movement strategy.
	 */
	public static void setMovement(Movement arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		movement = arg;
	}

	/**
	 * Sets the movement file name (BonnMotion).
	 * 
	 * @param arg
	 *            the movement file name.
	 */
	public static void setMovementFile(String arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		movementFile = arg;
	}

	/**
	 * Sets the interval of node position updates (for Discrete movement
	 * strategy).
	 * 
	 * @param arg
	 *            the interval of node position updates.
	 */
	public static void setMovementInterval(long arg) {
		if (arg <= 0) {
			configError("Movement interval has to be positive: " + arg);
		}
		if (isRunning) {
			invalidConfigChange(arg);
		}
		movementInterval = arg;
	}

	/**
	 * Whether to use the Lazy MAC strategy.
	 * 
	 * @param arg
	 *            whether to use Lazy MAC.
	 */
	public static void setLazyMac(boolean arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		usesLazyMac = arg;
	}

	/**
	 * Sets the simulation callback.
	 * 
	 * @param arg
	 *            the simulation callback.
	 */
	public static void setSimulationCallback(SimulationCallback arg) {
		assert callback == null;
		callback = arg;
	}

	/**
	 * Whether to collect ISP (Ideal Simulation Protocol) data in a parallel
	 * simulation.
	 */
	public static void collectDataISP() {
		if (isRunning) {
			invalidConfigChange("collectDataISP");
		}
		collectDataISP = true;
	}

	/**
	 * Sets the end time of the simulation.
	 * 
	 * @param arg
	 *            the end time.
	 */
	public static void setEndTime(long arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		endTime = arg;
	}

	/**
	 * Provides the simulation type.
	 * 
	 * @return the simulation type.
	 */
	public static SimType getSimType() {
		if (simType == null) {
			configError("Forgot to set sim type by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return simType;
	}

	/**
	 * Provides the barrier type.
	 * 
	 * @return the barrier type.
	 */
	public static BarrierType getBarrier() {
		if (barrier == null) {
			configError("Forgot to set barrier by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return barrier;
	}

	/**
	 * Provides the lookahead strategy used.
	 * 
	 * @return the lookahead strategy used.
	 */
	public static Lookahead getLookahead() {
		if (lookahead == null) {
			configError("Forgot to set lookahead by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return lookahead;
	}

	/**
	 * Whether protocol lookahead is used.
	 * 
	 * @return whether protocol lookahead is used.
	 */
	public static boolean usesProtocolLookahead() {
		return usesProtocolLookahead;
	}

	/**
	 * Provides the scheduler type of the simulation.
	 * 
	 * @return the scheduler type of the simulation.
	 */
	public static Scheduler getScheduler() {
		if (scheduler == null) {
			configError("Forgot to set scheduler by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return scheduler;
	}

	/**
	 * Provides the random seed of the simulation.
	 * 
	 * @return the random seed of the simulation.
	 */
	public static int getRandomSeed() {
		if (randomSeed == -1) {
			configError("Forgot to set random seed by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return randomSeed;
	}

	/**
	 * Provides the random manager of the simulation.
	 * 
	 * @return the random manager of the simulation.
	 */
	public static Random getRandom() {
		if (random == null) {
			configError("Forgot to set random manager class by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return random;
	}

	/**
	 * Provides the propagation model of the simulation.
	 * 
	 * @return the propagation model of the simulation.
	 */
	public static Propagation getPropagation() {
		if (propagation == null) {
			configError("Forgot to set propagation by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return propagation;
	}

	/**
	 * Provides the propagation model of the simulation.
	 * 
	 * @return the propagation model of the simulation.
	 */
	public static PathlossModel getPathlossModel() {
		if (pathlossModel == null) {
			configError("Forgot to set path loss model by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return pathlossModel;
	}

	/**
	 * Provides the fading model of the simulation.
	 * 
	 * @return the fading model of the simulation.
	 */
	public static FadingModel getFadingModel() {
		if (fadingModel == null) {
			configError("Forgot to set fading model by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return fadingModel;
	}

	/**
	 * Provides the BER table file.
	 * 
	 * @return the name of the BER table file.
	 */
	public static String getBerTableFile() {
		return berTableFile;
	}

	/**
	 * Provides the movement strategy used by the simulation.
	 * 
	 * @return the movement strategy used by the simulation.
	 */
	public static Movement getMovement() {
		if (movement == null) {
			configError("Forgot to set movement by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return movement;
	}

	/**
	 * Provides the movement file used by the simulation.
	 * 
	 * @return the name of movement file used by the simulation.
	 */
	public static String getMovementFile() {
		if (movementFile == null) {
			configError("Forgot to set movement file by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return movementFile;
	}

	/**
	 * Provides the movement interval used by the simulation in case of Discrete
	 * position update.
	 * 
	 * @return the movement interval used by the simulation in case of Discrete
	 *         position update.
	 */
	public static long getMovementInterval() {
		return movementInterval;
	}

	/**
	 * Provides the (optional) simulation callback.
	 * 
	 * @return the simulation callback.
	 */
	public static SimulationCallback getCallback() {
		return callback;
	}

	/**
	 * Whether the Lazy MAC strategy is used.
	 * 
	 * @return whether the Lazy MAC strategy is used.
	 */
	public static boolean usesLazyMac() {
		return usesLazyMac;
	}

	/**
	 * Whether the simulation collects ISP (Ideal Simulation Protocol) data.
	 * 
	 * @return whether the simulation collects ISP data.
	 */
	public static boolean isCollectingISP() {
		return collectDataISP;
	}

	/**
	 * Provides the number of partitions of the simulation.
	 * 
	 * @return the number of partitions of the simulation.
	 */
	public static int getNumPartitions() {
		if (numPartitions == -1) {
			configError("Forgot to set number of partitions by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return numPartitions;
	}

	/**
	 * Provides the number of virtual partitions of the simulation.
	 * 
	 * @return the number of virtual partitions of the simulation.
	 */
	public static int getNumVirtualPartitions() {
		if (numVirtualPartitions == -1) {
			configError("Forgot to set number of partitions by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return numVirtualPartitions;
	}

	/**
	 * Provides the start time of the simulation.
	 * 
	 * @return the start time of the simulation.
	 */
	public static long getStartTime() {
		return startTime;
	}

	/**
	 * Provides the end time of the simulation.
	 * 
	 * @return the end time of the simulation.
	 */
	public static long getEndTime() {
		if (endTime == -1) {
			configError("Forgot to set end time by calling class "
					+ Config.class.getSimpleName() + " ?");
		}
		return endTime;
	}

	/**
	 * Prints an error message in case of a configuration error.
	 * 
	 * @param msg
	 *            the error message.
	 */
	private static void configError(String msg) {
		String output = "Critical configuration error! " + msg + "... exiting!";
		System.out.println(output);
		System.err.println(output);
		System.exit(-1);
	}

	/**
	 * Prints an error message in case of an attempt to change the configuration
	 * during the simulation.
	 * 
	 * @param arg
	 *            the modified configuration parameter.
	 */
	private static void invalidConfigChange(Object arg) {
		throw new RuntimeException(
				"Attempt to change configuration parameters during the simulation: "
						+ arg);
	}

}
