/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/Event.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core;

import java.io.Serializable;

/**
 * A simulation event. The equals(), hashCode() and compareTo() methods are
 * consistent.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class Event implements Comparable<Event>, Serializable {
	/**
	 * Default ID.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Whether simple LP-scope counter-based event priorities are used (4th
	 * order).
	 */
	private static final boolean USE_COUNTER_PRIORITY = true;

	/**
	 * Whether the source module of an event is used as its priority (3rd
	 * order).
	 */
	private static final boolean USE_SRC_PRIORITY = true;

	/**
	 * Whether the destination module of an event is used as its priority (2nd
	 * order).
	 */
	private static final boolean USE_DST_PRIORITY = true;

	/**
	 * Whether the scheduling time of an event is used as its priority (1st
	 * order).
	 */
	private static final boolean USE_SCHEDTIME_PRIORITY = true;

	/**
	 * The time at which the event will occur.
	 */
	private final long execTime_;

	/**
	 * The time at which the event was scheduled.
	 */
	private final long schedTime_;

	/**
	 * The globally unique ID of the event type.
	 */
	@SuppressWarnings("rawtypes")
	private final Enum id_;

	/**
	 * The ID of the module which scheduled the event.
	 */
	private final int srcModuleId_;

	/**
	 * The ID of the module which handles the event.
	 */
	private final int dstModuleId_;

	/**
	 * The (optional) parameter for the event. The object scheduling this event
	 * has to make sure that the parameter is compatible with the event type.
	 */
	private final Object parameter_;

	/**
	 * A priority which may be used as a second criterion for ordering events.
	 */
	private final int priority_;

	/**
	 * Creates a new instance of <code>Event</code> with the given parameters.
	 * 
	 * @param execTime
	 *            the time of the event.
	 * @param schedTime
	 *            the time at which it the event scheduled.
	 * @param dstModuleId
	 *            the Id of the object responsible for handling the event.
	 * @param srcModuleId
	 *            the ID of the module which scheduled the event.
	 * @param id
	 *            the unique ID of the event type.
	 * @param parameter
	 *            the optional parameter for the event.
	 */
	@SuppressWarnings("rawtypes")
	public Event(long execTime, long schedTime, int dstModuleId,
			int srcModuleId, Enum id, Object parameter) {
		execTime_ = execTime;
		schedTime_ = schedTime;
		dstModuleId_ = dstModuleId;
		srcModuleId_ = srcModuleId;
		id_ = id;
		parameter_ = parameter;
		priority_ = 0;
	}

	/**
	 * Constructor with additional priority parameter.
	 * 
	 * @param execTime
	 *            the time of the event.
	 * @param schedTime
	 *            the time at which it the event scheduled.
	 * @param dstModuleId
	 *            the Id of the object responsible for handling the event.
	 * @param srcModuleId
	 *            the ID of the module which scheduled the event.
	 * @param id
	 *            the unique ID of the event type.
	 * @param parameter
	 *            the optional parameter for the event.
	 * @param priority
	 *            the priority value for the event.
	 */
	@SuppressWarnings("rawtypes")
	public Event(long execTime, long schedTime, int dstModuleId,
			int srcModuleId, Enum id, Object parameter, int priority) {
		execTime_ = execTime;
		schedTime_ = schedTime;
		dstModuleId_ = dstModuleId;
		srcModuleId_ = srcModuleId;
		id_ = id;
		parameter_ = parameter;
		priority_ = priority;
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object o) {
		if (o == this) {
			return true;
		}
		if (o == null || !(o instanceof Event)) {
			return false;
		}
		final Event otherEvent = (Event) o;
		if (execTime_ != otherEvent.execTime_) {
			return false;
		}
		if (schedTime_ != otherEvent.schedTime_) {
			return false;
		}
		if (dstModuleId_ != otherEvent.dstModuleId_) {
			return false;
		}
		if (srcModuleId_ != otherEvent.srcModuleId_) {
			return false;
		}
		if (id_ != otherEvent.id_) {
			return false;
		}
		if (parameter_ != null) {
			return (otherEvent.parameter_ != null && parameter_
					.equals(otherEvent.parameter_));
		}
		return (otherEvent.parameter_ == null);
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		int result = 17;
		result = 37 * result + (int) (execTime_ ^ (execTime_ >>> 32));
		result = 37 * result + (int) (schedTime_ ^ (schedTime_ >>> 32));
		result = 37 * result + dstModuleId_;
		result = 37 * result + srcModuleId_;
		result = 37 * result + id_.hashCode();
		result = 37 * result + (parameter_ == null ? 0 : parameter_.hashCode());
		return result;
	}

	/**
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public int compareTo(Event otherEvent) {
		if (execTime_ < otherEvent.execTime_) {
			return -1;
		}
		if (execTime_ > otherEvent.execTime_) {
			return 1;
		}
		if (schedTime_ < otherEvent.schedTime_) {
			return -1;
		}
		if (schedTime_ > otherEvent.schedTime_) {
			return 1;
		}
		if (dstModuleId_ < otherEvent.dstModuleId_) {
			return -1;
		}
		if (dstModuleId_ > otherEvent.dstModuleId_) {
			return 1;
		}
		if (srcModuleId_ < otherEvent.srcModuleId_) {
			return -1;
		}
		if (srcModuleId_ > otherEvent.srcModuleId_) {
			return 1;
		}
		final int compareId = id_.compareTo(otherEvent.id_);
		if (compareId != 0) {
			return compareId;
		}
		if (parameter_ != null) {
			if (otherEvent.parameter_ != null) {
				final int hashCode = parameter_.hashCode();
				final int otherHashCode = otherEvent.parameter_.hashCode();
				if (hashCode < otherHashCode) {
					return -1;
				}
				if (hashCode > otherHashCode) {
					return 1;
				}
				return 0;
			}
			return -1;
		}
		if (otherEvent.parameter_ != null) {
			return 1;
		}
		return 0;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "(" + execTime_ + "," + schedTime_ + ","
				+ ModuleRoot.getRoot().findModule(dstModuleId_).toString()
				+ ","
				+ ModuleRoot.getRoot().findModule(srcModuleId_).toString()
				+ "," + id_ + ","
				+ (parameter_ == null ? "null" : parameter_.toString()) + ")";
	}

	/**
	 * Whether this event has a smaller timestamp than (i.e., is to be scheduled
	 * for execution before) the specified event.
	 * 
	 * @param otherEvent
	 *            the event to be compared to this event.
	 * @return whether this event is to be scheduled before the other event.
	 */
	public boolean hasSmallerTimestampThan(Event otherEvent) {
		if (USE_SRC_PRIORITY && execTime_ == otherEvent.execTime_) {
			if (USE_DST_PRIORITY && srcModuleId_ == otherEvent.srcModuleId_) {
				if (USE_SCHEDTIME_PRIORITY
						&& dstModuleId_ == otherEvent.dstModuleId_) {
					if (USE_COUNTER_PRIORITY
							&& schedTime_ == otherEvent.schedTime_) {
						return priority_ < otherEvent.priority_;
					}
					return schedTime_ < otherEvent.schedTime_;
				}
				return dstModuleId_ < otherEvent.dstModuleId_;
			}
			return srcModuleId_ < otherEvent.srcModuleId_;
		}
		return execTime_ < otherEvent.execTime_;
	}

	/**
	 * Provides the timestamp of this event.
	 * 
	 * @return the timestamp of this event.
	 */
	public long getExecTime() {
		return execTime_;
	}

	/**
	 * Provides the scheduling time of this event.
	 * 
	 * @return the scheduling time of this event.
	 */
	public long getSchedTime() {
		return schedTime_;
	}

	/**
	 * Provides the ID of the module handling this event.
	 * 
	 * @return the ID of the module handling this event.
	 */
	public int getDstModuleId() {
		return dstModuleId_;
	}

	/**
	 * Provides the ID of the module that scheduled this event.
	 * 
	 * @return the ID of the module that scheduled this event.
	 */
	public int getSrcModuleId() {
		return srcModuleId_;
	}

	/**
	 * Provides the ID of this event.
	 * 
	 * @return the ID of this event.
	 */
	@SuppressWarnings("rawtypes")
	public Enum getId() {
		return id_;
	}

	/**
	 * Provides the parameter of this event.
	 * 
	 * @return the parameter of this event.
	 */
	public Object getParameter() {
		return parameter_;
	}

	/**
	 * Provides the priority of this event.
	 * 
	 * @return the priority of this event.
	 */
	public int getPriority() {
		return priority_;
	}
}
