/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/ModuleRoot.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core;

import gnu.trove.TIntIntHashMap;
import gnu.trove.TIntObjectHashMap;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import core.sync.InitLp;
import core.util.CustomLogger;

/**
 * Provides all instances of classes extending BaseModule with unique IDs.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class ModuleRoot {
	/**
	 * The global <code>InitLp</code> object.
	 */
	private static InitLp thisInitLp;

	/**
	 * The <code>ModuleRoot</code> object.
	 */
	private static ModuleRoot root;

	/**
	 * Provides a reference to (and instantiates, if necessary) the InitLp of
	 * the simulation.
	 * 
	 * @return the unique instance of InitLp.
	 */
	public static InitLp getInitLp() {
		if (ModuleRoot.thisInitLp == null) {
			if (CustomLogger.isDebugEnabled()) {
				CustomLogger.debug("ModuleRoot:getInitLp() == null");
			}
			ModuleRoot.thisInitLp = new InitLp();
		}
		return ModuleRoot.thisInitLp;
	}

	/**
	 * Provides a reference to (and instantiates, if necessary) the ModuleRoot.
	 * 
	 * @return the unique instance of ModuleRoot.
	 */
	public static ModuleRoot getRoot() {
		if (ModuleRoot.root == null) {
			ModuleRoot.root = new ModuleRoot();
		}
		return ModuleRoot.root;
	}

	/**
	 * Key: Module ID. Value: The corresponding BaseModule.
	 */
	private final List<BaseModule> moduleIdToModule_;

	/**
	 * Key: Module ID. Value: Class of that module.
	 */
	@SuppressWarnings("rawtypes")
	private final List<Class> moduleIdToClass_;

	/**
	 * Set of all module classes that occur in the simulation.
	 */
	@SuppressWarnings("rawtypes")
	private final Set<Class> classes_;

	/**
	 * Key: Module class. Value: Map of node ID to module ID.
	 */
	@SuppressWarnings("rawtypes")
	private final Map<Class, TIntIntHashMap> classAndInternalIdToModuleId_;

	/**
	 * Key: LP ID. Value: List of modules to be simulated by the LP.
	 */
	private final TIntObjectHashMap<List<BaseModule>> lpIdToModules_;

	/**
	 * Counts the number of registered modules.
	 */
	private int numRegisteredModules_;

	/**
	 * Constructor.
	 */
	@SuppressWarnings("rawtypes")
	ModuleRoot() {
		moduleIdToModule_ = new ArrayList<BaseModule>();
		moduleIdToClass_ = new ArrayList<Class>();
		classes_ = new HashSet<Class>();
		classAndInternalIdToModuleId_ = new HashMap<Class, TIntIntHashMap>();
		lpIdToModules_ = new TIntObjectHashMap<List<BaseModule>>();
		numRegisteredModules_ = 0;
	}

	/**
	 * Registers instances of classes extending BaseModule. This method is
	 * automatically called by the standard constructor of
	 * <core>BaseModule</core>.
	 * 
	 * @param module
	 *            a reference to the module being registered.
	 * @param classInternalId
	 *            the class-internal number (node ID) of the module.
	 * @return the newly assigned unique module ID for the module.
	 */
	@SuppressWarnings("rawtypes")
	public int registerModule(BaseModule module, int classInternalId) {
		assert moduleIdToModule_.size() == numRegisteredModules_;
		assert moduleIdToClass_.size() == numRegisteredModules_;
		assert moduleIdToModule_.size() == numRegisteredModules_;

		// Store class of this module.
		final Class moduleClass = module.getClass();
		if (!classes_.contains(moduleClass)) {
			classes_.add(moduleClass);
			classAndInternalIdToModuleId_
					.put(moduleClass, new TIntIntHashMap());
		}
		final int newModuleId = numRegisteredModules_;
		moduleIdToModule_.add(module);
		moduleIdToClass_.add(moduleClass);
		classAndInternalIdToModuleId_.get(moduleClass).put(classInternalId,
				newModuleId);
		numRegisteredModules_++;
		return newModuleId;
	}

	/**
	 * Called to allocate the specified module to the specified ID.
	 * 
	 * @param module
	 *            the module to allocate.
	 * @param lpId
	 *            the ID of the LP to simulate the module.
	 */
	public void allocateModuleToLp(BaseModule module, int lpId) {
		if (!lpIdToModules_.containsKey(lpId)) {
			lpIdToModules_.put(lpId, new ArrayList<BaseModule>());
		}
		lpIdToModules_.get(lpId).add(module);
	}

	/**
	 * Provides the list of all modules of the specified LP.
	 * 
	 * @param lpId
	 *            the ID of the LP.
	 * @return the list of all modules of the specified LP.
	 */
	public List<BaseModule> getModulesOfLp(int lpId) {
		assert lpIdToModules_.containsKey(lpId);
		assert !lpIdToModules_.get(lpId).isEmpty();
		return lpIdToModules_.get(lpId);
	}

	/**
	 * Called when all modules have been created and registered.
	 */
	public void init() {
		// Initialize all known modules (for, e.g., scheduling initial events or
		// establishing connections).
		for (BaseModule module : moduleIdToModule_) {
			if (module != null) {
				module.init();
			}
		}

	}

	/**
	 * Provides a list of all modules involved in the simulation.
	 * 
	 * @return a list of all modules involved in the simulation.
	 */
	public List<BaseModule> getModuleList() {
		return moduleIdToModule_;
	}

	/**
	 * Provides the module ID of the module with the specified node ID and the
	 * specified class.
	 * 
	 * @param moduleClass
	 *            the class of the module.
	 * @param internalClassId
	 *            the class-internal ID of the module.
	 * @return the module ID of the specified module.
	 */
	@SuppressWarnings("rawtypes")
	public int findModuleId(Class moduleClass, int internalClassId) {
		return classAndInternalIdToModuleId_.get(moduleClass).get(
				internalClassId);
	}

	/**
	 * Provides a reference to a module for a given module ID.
	 * 
	 * @param moduleId
	 *            the module ID.
	 * @return a reference to the corresponding module.
	 */
	public BaseModule findModule(int moduleId) {
		return moduleIdToModule_.get(moduleId);
	}
}
