/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: distribution/CommLayer.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.distribution;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;

import core.util.CustomLogger;


/**
 * Blocking communication layer for distributed simulation using a central
 * controller.
 * 
 * @author Patrick Peschlow
 * @author Martin Geuer
 * 
 * @version 1.0
 */
public class CommLayer {
	/**
	 * The socket of this communication layer.
	 */
	private Socket socket_ = null;

	/**
	 * The input stream of this channel.
	 */
	private ObjectInputStream ois_;

	/**
	 * The output stream of this channel.
	 */
	private ObjectOutputStream oos_;

	/**
	 * Creates a blocking network communication.
	 * 
	 * @param socket
	 *            The socket to listen on.
	 */
	CommLayer(Socket socket) {
		socket_ = socket;
		try {
			oos_ = new ObjectOutputStream(socket_.getOutputStream());
			ois_ = new ObjectInputStream(socket_.getInputStream());
		} catch (IOException e) {
			if (CustomLogger.isFatalEnabled()) {
				CustomLogger.fatal("Error during stream initialization", e);
			}
		}
	}

	/**
	 * Reads an object from the output stream.
	 * 
	 * @return the object read..
	 */
	public Object readObjectBlocking() {
		Object o = null;
		try {
			o = ois_.readObject();
		} catch (IOException e) {
			if (CustomLogger.isFatalEnabled()) {
				CustomLogger.fatal("Error while reading object: " + o, e);
			}
		} catch (ClassNotFoundException e) {
			if (CustomLogger.isFatalEnabled()) {
				CustomLogger.fatal("Error while reading object: " + o, e);
			}
		}
		return o;
	}

	/**
	 * Writes an object to the output stream.
	 * 
	 * @param o
	 *            the object to write.
	 */
	public void writeObject(Object o) {
		try {
			oos_.writeObject(o);
			oos_.flush();
			oos_.reset();
		} catch (IOException e) {
			if (CustomLogger.isFatalEnabled()) {
				CustomLogger.fatal("Error while writing object: " + o, e);
			}
		}

	}

	/**
	 * Called to shutdown the socket communication.
	 */
	void shutdown() {
		try {
			if (!socket_.isOutputShutdown()) {
				socket_.shutdownOutput();
			}
			if (!socket_.isInputShutdown()) {
				socket_.shutdownInput();
			}
			if (!socket_.isClosed()) {
				socket_.close();
			}
			ois_.close();
			oos_.close();
		} catch (IOException e) {
			if (CustomLogger.isWarningEnabled()) {
				CustomLogger.warning("Exception during shutdown.", e);
			}
		}
	}
}