/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: distribution/CreateCommand.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.distribution;

import java.io.Serializable;

/**
 * Initial message that the server sends to the distributed simulation clients.
 * 
 * @author Patrick Peschlow
 * @author Daniel Heupel
 * 
 * @version 1.0
 */
public class CreateCommand implements Serializable {
	/**
	 * Default ID.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Total number of LPs.
	 */
	private final int numLps_;

	/**
	 * Init string containing the init file.
	 */
	private final String initString_;

	/**
	 * ID of the first LP of this client.
	 */
	private final int firstLp_;

	/**
	 * Number of local LPs of this client.
	 */
	private final int numLocalLps_;

	/**
	 * Constructor.
	 * 
	 * @param numLps
	 *            total number of LPs.
	 * @param initString
	 *            init file string.
	 * @param firstLp
	 *            ID of the first LP of this client.
	 * @param numLocalLps
	 *            number of local LPs of this client.
	 */
	CreateCommand(int numLps, String initString, int firstLp, int numLocalLps) {
		numLps_ = numLps;
		initString_ = initString;
		firstLp_ = firstLp;
		numLocalLps_ = numLocalLps;
	}

	/**
	 * Provides the number of LPs.
	 * 
	 * @return the number of LPs.
	 */
	int getNumLps() {
		return numLps_;
	}

	/**
	 * Provides the init string.
	 * 
	 * @return the init string.
	 */
	String getInitString() {
		return initString_;
	}

	/**
	 * Provides the ID of the first LP of this client.
	 * 
	 * @return the ID of the first LP of this client.
	 */
	int getFirstLp() {
		return firstLp_;
	}

	/**
	 * Provides the number of local LPs of this client.
	 * 
	 * @return the number of local LPs of this client.
	 */
	int getNumLocalLps() {
		return numLocalLps_;
	}
}
