/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: distribution/HostIdentifier.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.distribution;

import java.io.Serializable;

/**
 * Unique identifier of a client.
 * 
 * @author Patrick Peschlow
 * @author Martin Geuer
 * 
 * @version 1.0
 */
public class HostIdentifier implements Serializable {
	/**
	 * Default ID.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Numerical ID. Used to generate unique host identifiers.
	 */
	private final int id_;

	/**
	 * Name of the host.
	 */
	private final String hostName_;

	/**
	 * Constructor.
	 * 
	 * @param host
	 *            The name of the host.
	 * @param id
	 *            A numerical ID.
	 */
	HostIdentifier(String host, int id) {
		hostName_ = host;
		id_ = id;
	}

	/**
	 * Provides the ID.
	 * 
	 * @return the ID.
	 */
	int getId() {
		return id_;
	}

	/**
	 * Provides the host name.
	 * 
	 * @return the host name.
	 */
	String getName() {
		return hostName_;
	}

	/**
	 * @see Object#equals(Object)
	 */
	@Override
	public String toString() {
		return hostName_ + ":" + id_;
	}

	/**
	 * @see Object#equals(Object)
	 */
	@Override
	public boolean equals(Object other) {
		if (other == this) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (other instanceof HostIdentifier) {
			final HostIdentifier o = (HostIdentifier) other;
			if (o.id_ == id_) {
				if (hostName_.equalsIgnoreCase(o.hostName_)) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		int result = 17;
		result = 37 * result + id_;
		result = 37 * result + hostName_.hashCode();
		return result;
	}
}
