/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: distribution/SimClient.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.distribution;

import groovy.lang.Binding;
import groovy.lang.GroovyShell;

import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;

import modules.wireless.scenario.CreateWorld;
import core.Config;
import core.ModuleRoot;
import core.sync.DistributedSimController;
import core.util.CustomLogger;
import core.util.EventLogger;

/**
 * Used with distributed simulation to set up the simulation on remote hosts.
 * 
 * @author Patrick Peschlow
 * @author Thomas Saftig
 * @author Daniel Heupel
 * 
 * @version 1.0
 */
public class SimClient {
	/**
	 * Name of the host initializing the simulation.
	 */
	private final String initHost_;

	/**
	 * Port on which to reach the initiating host.
	 */
	private final int initPort_;

	/**
	 * Number of local LPs.
	 */
	private int numLocalLps_ = 0;

	/**
	 * Activates the local <code>RemoteInitController</code>.
	 * 
	 * @param port
	 *            Port on which to reach the initializing host.
	 * @param initHost
	 *            Name of the initializing host.
	 * @param numLps
	 *            Proposed number of LPs for this host.
	 */
	public SimClient(int port, String initHost, int numLps) {
		initHost_ = initHost;
		initPort_ = port;
		numLocalLps_ = numLps;
	}

	/**
	 * Sets up the communication to the remote hosts. Receives the init data for
	 * the simulation. Starts the simulation.
	 */
	public void runSimulation() {
		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("Connecting to server " + initHost_ + " on port "
					+ initPort_ + " (registering " + numLocalLps_ + " LPs)");
		}

		CommLayer myCommLayer = connectToInitHost(initHost_, initPort_);

		InetAddress localHost = null;
		try {
			localHost = InetAddress.getLocalHost();
		} catch (UnknownHostException e) {
			if (CustomLogger.isFatalEnabled()) {
				CustomLogger.fatal("Localhost unknown!", e);
			}
		}
		myCommLayer.writeObject(localHost.getHostName());

		if (numLocalLps_ == 0) {
			numLocalLps_ = Runtime.getRuntime().availableProcessors();
		}

		myCommLayer.writeObject(numLocalLps_);

		CreateCommand createCommand = (CreateCommand) myCommLayer
				.readObjectBlocking();

		assert createCommand.getNumLocalLps() == numLocalLps_;
		final int firstLp = createCommand.getFirstLp();

		// Evaluate init script.
		new GroovyShell(new Binding()).evaluate(createCommand.getInitString());

		// Create all modules of the simulation.
		new CreateWorld().setupSimulationDistributed(firstLp, numLocalLps_);

		if (CustomLogger.isDebugEnabled()) {
			CustomLogger.debug("awaitClients(" + firstLp + "): Evaluate done");
		}

		// Create local root and initialize all modules.
		ModuleRoot.getRoot().init();

		// Instantiate the simulation controller.
		DistributedSimController simCtrl = new DistributedSimController(
				myCommLayer, firstLp, numLocalLps_);

		// Initiate the simulation.
		simCtrl.startSim();

		// user defined actions at the end of a simulations
		if (Config.getCallback() != null) {
			Config.getCallback().endSimulation(simCtrl);
		}

		myCommLayer.writeObject(simCtrl.getStatsPacket());
		myCommLayer.shutdown();

		EventLogger.closeAll();
	}

	/**
	 * Connects to the init host (the server).
	 * 
	 * @param host
	 *            the hostname.
	 * @param port
	 *            the remote port.
	 * @return the new communication layer.
	 */
	private CommLayer connectToInitHost(String host, int port) {
		Socket socket = null;
		do {
			try {
				socket = new Socket(host, port);
			} catch (UnknownHostException e) {
				if (CustomLogger.isFatalEnabled()) {
					CustomLogger
							.fatal("SimController Host unknown: " + host, e);
				}
			} catch (IOException e) {
				if (CustomLogger.isDebugEnabled()) {
					CustomLogger.debug("HostCtrl: SimCtrl probably not up yet");
				}
				try {
					Thread.sleep(2000);
				} catch (InterruptedException e1) {
					if (CustomLogger.isDebugEnabled()) {
						CustomLogger.debug("Could not sleep. Trying again...");
					}
				}
			}
		} while (socket == null);
		return new CommLayer(socket);
	}
}