/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: distribution/SimMaster.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.distribution;

import java.io.IOException;
import java.net.ServerSocket;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import modules.wireless.scenario.NodeConfig;
import modules.wireless.scenario.NodeConfig.ScenarioType;
import core.Config;
import core.Event;
import core.util.CustomLogger;

/**
 * Used with distributed simulation to set up the simulation on the initial host
 * (the server).
 * 
 * @author Patrick Peschlow
 * @author Thomas Saftig
 * @author Daniel Heupel
 * 
 * @version 1.0
 */
public class SimMaster {
	/**
	 * LP map
	 */
	private final Map<HostIdentifier, Integer> numLocalLps_ = new HashMap<HostIdentifier, Integer>();

	/**
	 * Port on which the server listens for connections.
	 */
	private final int initPort_;

	/**
	 * Map of host identifiers to comm layers.
	 */
	private final Map<HostIdentifier, CommLayer> hostConnections_ = new HashMap<HostIdentifier, CommLayer>();

	/**
	 * Counts the number of synchronization points.
	 */
	private long numSyncs_ = 0;

	/**
	 * Constructor.
	 * 
	 * @param initPort
	 *            port on which to open the server socket.
	 */
	public SimMaster(int initPort) {
		initPort_ = initPort;
	}

	/**
	 * Called by initiator.
	 * 
	 * @param initString
	 *            the scenario init file as a string.
	 */
	public void runSimulation(String initString) {
		awaitConnections();

		distributeScenario(initString);

		final int numClients = hostConnections_.size();
		CommLayer[] allCommLayers = new CommLayer[numClients];
		int i = 0;
		for (CommLayer commLayer : hostConnections_.values()) {
			allCommLayers[i++] = commLayer;
		}
		while (true) {
			// Collect safe time bounds and event lists from clients.
			long globalSafeTimeBound = Long.MAX_VALUE;
			List<Event> allEvents = new ArrayList<Event>();
			for (CommLayer commLayer : allCommLayers) {
				BarrierPacket incomingPacket = (BarrierPacket) commLayer
						.readObjectBlocking();
				final long safeTimeBound = incomingPacket.getSafeTimeBound();
				if (safeTimeBound < globalSafeTimeBound) {
					globalSafeTimeBound = safeTimeBound;
				}
				allEvents.addAll(incomingPacket.getEvents());
			}

			if (CustomLogger.isDebugEnabled()) {
				if (numSyncs_ % 100 == 0) {
					CustomLogger.debug("safeTimeBound: " + globalSafeTimeBound);
				}
			}

			// Distribute global safe time bound and complete event list.
			BarrierPacket outgoingPacket = new BarrierPacket(
					globalSafeTimeBound, allEvents);
			for (CommLayer commLayer : allCommLayers) {
				commLayer.writeObject(outgoingPacket);
			}
			numSyncs_++;

			// Are we finished?
			if (globalSafeTimeBound == Long.MAX_VALUE) {
				break;
			}
		}

		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("Simulation over. Printing statistics...");
		}

		// Collect and output statistics.
		StatsPacket collectedStatsPacket = new StatsPacket();
		for (CommLayer commLayer : allCommLayers) {
			StatsPacket clientStatsPacket = (StatsPacket) commLayer
					.readObjectBlocking();
			updateStatsPacket(collectedStatsPacket, clientStatsPacket);
			commLayer.shutdown();
		}
		printStatistics(collectedStatsPacket);
	}

	/**
	 * Distributes the scenario information to the clients.
	 * 
	 * @param initString
	 *            the scenario init file as a string.
	 */
	private void distributeScenario(String initString) {
		// Distribute simulation
		int currLp = 0;
		for (HostIdentifier hi : hostConnections_.keySet()) {
			CommLayer commLayer = hostConnections_.get(hi);
			CreateCommand createCommand_ = new CreateCommand(
					Config.getNumPartitions(), initString, currLp,
					numLocalLps_.get(hi));
			commLayer.writeObject(createCommand_);
			currLp += numLocalLps_.get(hi);

		}
		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("Sent scenario information to clients.");
		}
	}

	/**
	 * Waits for connection from enough clients to deliver the requested number
	 * of LPs.
	 * 
	 * @return a map of host identifiers to the communication layers to be used
	 *         for communicating with the clients.
	 */
	private Map<HostIdentifier, CommLayer> awaitConnections() {
		// Establish client connections
		ServerSocket serverSocket = null;
		try {
			serverSocket = new ServerSocket(initPort_);
		} catch (IOException e) {
			if (CustomLogger.isFatalEnabled()) {
				CustomLogger.fatal("Error while creating server socket!", e);
			}
		}
		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("Listening on port " + initPort_);
		}
		int numOfferedLps = 0;
		while (numOfferedLps < Config.getNumPartitions()) {
			try {
				// Wait for new client
				CommLayer commLayer = new CommLayer(serverSocket.accept());
				String host = (String) commLayer.readObjectBlocking();
				if (CustomLogger.isDebugEnabled()) {
					CustomLogger.debug("received connection from " + host);
				}
				// Get possible LP count of client
				int numRemoteLps = ((Integer) commLayer.readObjectBlocking())
						.intValue();
				// Create unique HostIdentifier
				HostIdentifier modifiedHi = new HostIdentifier(host,
						numOfferedLps);
				if ((numOfferedLps + numRemoteLps) > Config.getNumPartitions()) {
					numRemoteLps = Config.getNumPartitions() - numOfferedLps;
				}
				hostConnections_.put(modifiedHi, commLayer);
				numLocalLps_.put(modifiedHi, numRemoteLps);
				numOfferedLps += numRemoteLps;
				if (CustomLogger.isInfoEnabled()) {
					CustomLogger.info("Registered host: "
							+ modifiedHi.toString() + " (with "
							+ numLocalLps_.get(modifiedHi) + " LPs)");
				}
			} catch (IOException e) {
				if (CustomLogger.isFatalEnabled()) {
					CustomLogger.fatal("Error while accepting connections", e);
				}
			}
		}

		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("Total number of clients: "
					+ hostConnections_.size());
		}
		return hostConnections_;
	}

	/**
	 * Updates the simulation statistics with data sent by a client.
	 * 
	 * @param myStatsPacket
	 *            the local statistics packet that collects all statistics.
	 * @param rcvdStatsPacket
	 *            the remote statistics packet with a part of the statistics.
	 */
	private void updateStatsPacket(StatsPacket myStatsPacket,
			StatsPacket rcvdStatsPacket) {
		myStatsPacket.tcpRecv += rcvdStatsPacket.tcpRecv;
		myStatsPacket.udpRecv += rcvdStatsPacket.udpRecv;
		myStatsPacket.tcpSend += rcvdStatsPacket.tcpSend;
		myStatsPacket.udpSend += rcvdStatsPacket.udpSend;
		myStatsPacket.throughNet += rcvdStatsPacket.throughNet;
		myStatsPacket.radioInc += rcvdStatsPacket.radioInc;
		myStatsPacket.radioDrop += rcvdStatsPacket.radioDrop;
		myStatsPacket.captSignals += rcvdStatsPacket.captSignals;
		myStatsPacket.macBytesSend += rcvdStatsPacket.macBytesSend;
		myStatsPacket.macQueueOffered += rcvdStatsPacket.macQueueOffered;
		myStatsPacket.macQueueDropped += rcvdStatsPacket.macQueueDropped;
		myStatsPacket.directOffered += rcvdStatsPacket.directOffered;
		myStatsPacket.transmissions += rcvdStatsPacket.transmissions;
		myStatsPacket.broadcasts += rcvdStatsPacket.broadcasts;
		myStatsPacket.unicasts += rcvdStatsPacket.unicasts;
		myStatsPacket.routingFwBroadcasts += rcvdStatsPacket.routingFwBroadcasts;
		myStatsPacket.routingRcvBroadcasts += rcvdStatsPacket.routingRcvBroadcasts;
		myStatsPacket.transmissions802_11 += rcvdStatsPacket.transmissions802_11;
		myStatsPacket.transmissions802_11s += rcvdStatsPacket.transmissions802_11s;
		myStatsPacket.dummyTransmissions += rcvdStatsPacket.dummyTransmissions;
		myStatsPacket.hwmpDrops += rcvdStatsPacket.hwmpDrops;
		myStatsPacket.numBeacons += rcvdStatsPacket.numBeacons;
	}

	/**
	 * Prints the collected statistics at the end of the simulation.
	 * 
	 * @param statsPacket
	 *            the statistics packet containing all statistics.
	 */
	private void printStatistics(StatsPacket statsPacket) {
		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("radio inc/dropped:       "
					+ statsPacket.radioInc + " / " + statsPacket.radioDrop);
			CustomLogger.info("radio drop%:             "
					+ (statsPacket.radioDrop * 100.0 / statsPacket.radioInc));
			CustomLogger.info("captured signals:        "
					+ statsPacket.captSignals);
			CustomLogger.info("tcp send/receive:        " + statsPacket.tcpSend
					+ " / " + statsPacket.tcpRecv);
			CustomLogger.info("udp send/receive:        " + statsPacket.udpSend
					+ " / " + statsPacket.udpRecv);
			CustomLogger.info("throughput (net):        "
					+ statsPacket.throughNet + " bytes/sec");
			CustomLogger.info("bytes sent (mac):        "
					+ statsPacket.macBytesSend);
			CustomLogger.info("MacQueueOffered:         "
					+ statsPacket.macQueueOffered);
			CustomLogger.info("MacQueueDropped:         "
					+ statsPacket.macQueueDropped);
			CustomLogger.info("DirectOffered:           "
					+ statsPacket.directOffered);
			CustomLogger.info("transmissions:           "
					+ statsPacket.transmissions);
			CustomLogger.info("broadcasts:              "
					+ statsPacket.broadcasts);
			CustomLogger.info("unicasts:                "
					+ statsPacket.unicasts);
			CustomLogger
					.info("broadcast%:              "
							+ (statsPacket.broadcasts * 100.0 / statsPacket.transmissions));
			if (NodeConfig.getScenarioType() == ScenarioType.MANET) {
				CustomLogger.info("routing fwBroadcasts:    "
						+ statsPacket.routingFwBroadcasts);
				CustomLogger.info("routing rcvBroadcasts:   "
						+ statsPacket.routingRcvBroadcasts);
			} else if (NodeConfig.getScenarioType() == ScenarioType.MESH) {
				CustomLogger.info("802.11 transmissions:    "
						+ statsPacket.transmissions802_11);
				CustomLogger.info("802.11s transmissions:   "
						+ statsPacket.transmissions802_11s);
				CustomLogger
						.info("802.11s transmissions%:  "
								+ 100.0
								* statsPacket.transmissions802_11s
								/ (statsPacket.transmissions802_11s + statsPacket.transmissions802_11));
				CustomLogger.info("Dummy transmissions:     "
						+ statsPacket.dummyTransmissions);
				CustomLogger.info("HWMP Queue Drops:        "
						+ statsPacket.hwmpDrops);
				CustomLogger.info("number of beacons:       "
						+ statsPacket.numBeacons);
			}
			CustomLogger.info("number of syncs:         " + numSyncs_);
		}
	}

}
