/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/scheduler/BinaryTournamentQueue.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package core.scheduler;

import core.Event;

/**
 * Event queue based on complete binary trees (CBTs) and binary tournaments.
 * Inspired by the paper:<br>
 * M. Marin: "On the Pending Event Set and Binary Tournaments". In Proceedings
 * of the 10th European Simulation Symposium (ESS '98), 1998.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class BinaryTournamentQueue implements EventQueue {
	/**
	 * The initial length of the internal array.
	 */
	private static final int INIT_LENGTH = 100;

	/**
	 * The array containing all nodes of the tree, starting at index 1.
	 */
	private int[] cbt_;

	/**
	 * The array mapping leaf IDs to their array indices in cbt_, starting at
	 * index (and ID) 1.
	 */
	private int[] leaf_;

	/**
	 * The array mapping leaf IDs to the actual events, starting at index (and
	 * ID) 1.
	 */
	private Event[] events_;

	/**
	 * The current number of events in this tree.
	 */
	private int numLeaves_;

	/**
	 * List to store currently unused indices.
	 */
	private int[] list_;

	/**
	 * The number of currently unused indices.
	 */
	private int numUnused_;

	/**
	 * Constructor.
	 */
	public BinaryTournamentQueue() {
		cbt_ = new int[2 * INIT_LENGTH];
		cbt_[0] = 0;
		leaf_ = new int[INIT_LENGTH];
		leaf_[0] = 0;
		events_ = new Event[INIT_LENGTH];
		events_[0] = null;
		numLeaves_ = 0;
		list_ = new int[INIT_LENGTH];
		numUnused_ = 0;
	}

	/**
	 * @see core.scheduler.EventQueue#size()
	 */
	public int size() {
		return numLeaves_;
	}

	/**
	 * @see core.scheduler.EventQueue#isEmpty()
	 */
	public boolean isEmpty() {
		return numLeaves_ == 0;
	}

	/**
	 * @see core.scheduler.EventQueue#insert(core.Event)
	 */
	public void insert(Event event) {
		if (numLeaves_ == events_.length - 1) {
			doubleCapacity();
		}
		// int newLeafId = numLeaves_ + 1;
		int newLeafId;
		if (numUnused_ == 0) {
			newLeafId = numLeaves_ + 1;
		} else {
			newLeafId = list_[numUnused_ - 1];
			// PP: Can probably be optimized away, but it not important.
			list_[numUnused_--] = 0;
		}
		events_[newLeafId] = event;

		// Find first leaf node.
		int firstLeafIndex = numLeaves_;
		int firstleafId = cbt_[firstLeafIndex];
		// Make first leaf node its left son.
		int leftSonIndex = firstLeafIndex * 2;
		cbt_[leftSonIndex] = firstleafId;
		leaf_[firstleafId] = leftSonIndex;
		// Make new node its right son.
		int rightSonIndex = leftSonIndex + 1;
		cbt_[rightSonIndex] = newLeafId;
		leaf_[newLeafId] = rightSonIndex;

		while (true) {
			if (firstleafId == 0
					|| events_[firstleafId].hasSmallerTimestampThan(event)) {
				break;
			}
			// Update parent's priority.
			cbt_[firstLeafIndex] = newLeafId;
			firstLeafIndex /= 2;
			firstleafId = cbt_[firstLeafIndex];
		}

		numLeaves_++;
	}

	/**
	 * @see core.scheduler.EventQueue#removeFirst()
	 */
	public Event removeFirst() {
		assert !isEmpty();
		// Get minimum event.
		final int minLeafId = cbt_[1];
		final Event minEvent = events_[minLeafId];
		events_[minLeafId] = null;
		if (numLeaves_ == 1) {
			cbt_[1] = 0;
			leaf_[minLeafId] = 0;
			list_[numUnused_++] = minLeafId;
		} else if (numLeaves_ == 2) {
			final int remainingLeafId = minLeafId == cbt_[2] ? cbt_[3]
					: cbt_[2];
			cbt_[1] = remainingLeafId;
			leaf_[remainingLeafId] = 1;
			cbt_[2] = 0;
			cbt_[3] = 0;
			leaf_[minLeafId] = 0;
			list_[numUnused_++] = minLeafId;
		} else {
			// Find last leaf.
			final int maxLeafIndex = 2 * numLeaves_ - 1;
			// And its parent node.
			final int parentIndex = maxLeafIndex / 2;
			// Replace parent by its smaller priority child.
			final int smallerLeafId = cbt_[parentIndex];
			leaf_[smallerLeafId] = parentIndex;
			// Find the larger node (to move it away).
			final int leafToMoveId = (smallerLeafId != cbt_[maxLeafIndex]) ? cbt_[maxLeafIndex]
					: cbt_[maxLeafIndex - 1];
			// Delete the two former leaves.
			cbt_[maxLeafIndex] = 0;
			cbt_[maxLeafIndex - 1] = 0;
			// Find at which index the min event has been.
			int minEventIndex = leaf_[minLeafId];
			// Update with moved away node.
			cbt_[minEventIndex] = leafToMoveId;
			leaf_[leafToMoveId] = minEventIndex;
			// Clear ID of min event.
			leaf_[minLeafId] = 0;
			list_[numUnused_++] = minLeafId;
			updateCbt(minEventIndex);
		}
		numLeaves_--;
		return minEvent;
	}

	/**
	 * @see core.scheduler.EventQueue#peekFirst()
	 */
	public Event peekFirst() {
		assert !isEmpty();
		return events_[cbt_[1]];
	}

	/**
	 * Updates the CBT for the specified index.
	 * 
	 * @param index
	 *            the index where to update the CBT.
	 */
	private void updateCbt(int index) {
		do {
			index /= 2;
			final int leftChild = cbt_[2 * index];
			final int rightChild = cbt_[2 * index + 1];
			cbt_[index] = events_[leftChild]
					.hasSmallerTimestampThan(events_[rightChild]) ? leftChild
					: rightChild;
		} while (index != 1);
	}

	/**
	 * Doubles the capacity of the internal array.
	 */
	private void doubleCapacity() {
		int[] cbt = new int[cbt_.length * 2];
		System.arraycopy(cbt_, 0, cbt, 0, cbt_.length);
		cbt_ = cbt;
		int[] leaf = new int[leaf_.length * 2];
		System.arraycopy(leaf_, 0, leaf, 0, leaf_.length);
		leaf_ = leaf;
		Event[] events = new Event[events_.length * 2];
		System.arraycopy(events_, 0, events, 0, events_.length);
		events_ = events;
		int[] list = new int[list_.length * 2];
		System.arraycopy(list_, 0, list, 0, list_.length);
		list_ = list;
	}

	/**
	 * Dumps the internal data structures (for debugging).
	 */
	@SuppressWarnings("unused")
	private void print() {
		System.out.print("cbt_ = ");
		for (int index = 1; index < cbt_.length; index++) {
			System.out.print(cbt_[index] + " ");
		}
		System.out.println();
		System.out.print("leaf_ = ");
		for (int index = 1; index < leaf_.length; index++) {
			System.out.print(leaf_[index] + " ");
		}
		System.out.println();
		System.out.print("events_ = ");
		for (int index = 1; index < events_.length; index++) {
			System.out.print(events_[index] + " ");
		}
		System.out.println();
	}

}
