/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/scheduler/EventScheduler.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.scheduler;

import core.Event;
import core.util.CustomLogger;

/**
 * Scheduler that contains the event queue.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class EventScheduler {
	/**
	 * The event queue.
	 */
	private EventQueue queue_;

	/**
	 * Constructor.
	 * 
	 * @param queueClass
	 *            the class of the requested event queue instance.
	 */
	@SuppressWarnings("rawtypes")
	public EventScheduler(Class queueClass) {
		try {
			queue_ = (EventQueue) queueClass.newInstance();
		} catch (Exception e) {
			if (CustomLogger.isFatalEnabled()) {
				CustomLogger.fatal("Unknown event queue class!", e);
			}
		}
	}

	/**
	 * Provides the timestamp of the next (smallest-timestamped) event in the
	 * queue.
	 * 
	 * @return the smallest timestamp of all events in the queue.
	 */
	public long getNextEventTime() {
		return queue_.isEmpty() ? Long.MAX_VALUE : queue_.peekFirst()
				.getExecTime();
	}

	/**
	 * Adds an event to the event queue.
	 * 
	 * @param event
	 *            the event to be added.
	 */
	public void add(Event event) {
		queue_.insert(event);
	}

	/**
	 * Removes the next event from the event queue.
	 * 
	 * @return the next event from the queue.
	 */
	public Event removeNextEvent() {
		return queue_.removeFirst();
	}
}
