/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/scheduler/LadderQueue.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.scheduler;

import java.util.ArrayDeque;
import java.util.Arrays;

import core.Event;

/**
 * Event queue based on the ladder queue. Based on the paper:
 * 
 * W. T. Tang, R. S. M. Goh, I. L.-J. Thng: "Ladder Queue: An O(1) Priority
 * Queue Structure for Large-scale Discrete Event Simulation". In ACM TOMACS
 * 15(3), 175-204, 2005.
 * 
 * Note: PP: The implementation currently works only with sequential simulation.
 * This is because of an inherent assumption of the ladder queue: no events are
 * scheduled with a timestamp smaller than the timestamp of the event most
 * recently removed from the queue. In sequential simulation, this assumption
 * obviously holds. In parallel simulation, it does not hold. In my view, there
 * is no easy and performant way to fix this.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class LadderQueue implements EventQueue {
	/**
	 * Constant used for clarity.
	 */
	private static final int NOT_FOUND = -1;

	/**
	 * The maximum ID for rungs to be used (since the rung IDs start at 0, there
	 * will be at most MAX_RUNG_ID + 1 rungs).
	 */
	private static final int MAX_RUNG_ID = 7;

	/**
	 * If the number of events in a bucket or the "bottom" list exceeds this
	 * threshold, then a spawning action would be initiated.
	 */
	private static final int THRESH = 50;

	/**
	 * The array of rungs (the "ladder").
	 */
	private final Rung[] rungs_;

	/**
	 * The "top" list.
	 */
	private final Top top_;

	/**
	 * The sorted "bottom" list (from which events are removed by
	 * removeFirst()).
	 */
	private final EventQueue bottom_;

	/**
	 * The maximum index of all rungs currently in use (i.e., the current number
	 * of rungs minus one).
	 */
	private int curRungId_;

	/**
	 * The current total number of events in this queue.
	 */
	private int size_;

	/**
	 * Constructor.
	 */
	LadderQueue() {
		rungs_ = new Rung[MAX_RUNG_ID + 1];
		top_ = new Top();
		// bottom_ = new SortedListQueue();
		bottom_ = new ShiftedHeapQueue();
		curRungId_ = -1;
		size_ = 0;
	}

	/**
	 * @see core.scheduler.EventQueue#size()
	 */
	public int size() {
		return size_;
	}

	/**
	 * @see core.scheduler.EventQueue#isEmpty()
	 */
	public boolean isEmpty() {
		return size_ == 0;
	}

	/**
	 * @see core.scheduler.EventQueue#insert(core.Event)
	 */
	public void insert(Event event) {
		size_++;
		// Insert into top?
		if (top_.add(event)) {
			return;
		}
		// Find the correct rung for the event.
		for (int rungId = 0; rungId <= curRungId_; rungId++) {
			if (rungs_[rungId].fits(event)) {
				rungs_[rungId].add(event);
				return;
			}
		}
		// Did not find a rung. The event has to be inserted into bottom.
		if (bottom_.size() >= THRESH) {
			// Bottom too full. Create a new rung above.
			assert hasRung();
			assert curRungId_ < MAX_RUNG_ID : "Too many rungs created!";
			curRungId_++;
			rungs_[curRungId_] = rungs_[curRungId_ - 1].createRungFromBottom(
					bottom_, rungs_[curRungId_]);
			// Insert the new event to be scheduled into the new rung.
			rungs_[curRungId_].add(event);
		} else {
			// Bottom still has space. Simply insert the event.
			bottom_.insert(event);
		}
	}

	/**
	 * @see core.scheduler.EventQueue#removeFirst()
	 */
	public Event removeFirst() {
		size_--;
		if (bottom_.isEmpty()) {
			fillBottom();
		}
		return bottom_.removeFirst();
	}

	/**
	 * @see core.scheduler.EventQueue#peekFirst()
	 */
	public Event peekFirst() {
		if (bottom_.isEmpty()) {
			fillBottom();
		}
		return bottom_.peekFirst();
	}

	/**
	 * Moves events into "bottom".
	 */
	private void fillBottom() {
		// The bottom is empty and has to be filled again, with events from the
		// lowest non-empty rung or the top.
		if (hasRung()) {
			// We currently have one or more rungs.
			final int lowestBucketId = recurseRung();
			if (lowestBucketId != NOT_FOUND) {
				// A non-empty bucket in a rung has been found, so move its
				// events to the bottom.
				rungs_[curRungId_].moveBucketToBottom(lowestBucketId, bottom_);
				return;
			}
			// Else: no rung left, so fall through and take events from top.
		}
		assert !hasRung();
		if (top_.moveToBottom(bottom_)) {
			return;
		}
		assert !hasRung();
		// Create first rung and move all events from top into it.
		rungs_[0] = top_.createFirstRung();
		curRungId_++;
		// Move events from the first bucket of the new rung into bottom.
		final int lowestBucketId = recurseRung();
		assert lowestBucketId != NOT_FOUND;
		rungs_[curRungId_].moveBucketToBottom(lowestBucketId, bottom_);
	}

	/**
	 * Finds the next non-empty bucket of the ladder.
	 * 
	 * @return the index of the next non-empty bucket (of the current rung) if
	 *         such a bucket exists, otherwise {@link #NOT_FOUND}.
	 */
	private int recurseRung() {
		while (true) {
			// Find next non-empty bucket from lowest rung.
			int bucketId = rungs_[curRungId_].getCurrentBucket();
			final int rungSize = rungs_[curRungId_].getNumBuckets();
			while (bucketId < rungSize
					&& rungs_[curRungId_].getNumEvents(bucketId) == 0) {
				bucketId++;
				rungs_[curRungId_].shift();
			}
			if (bucketId == rungSize) {
				// No event in any bucket of this rung found.
				curRungId_--;
				if (curRungId_ >= 0) {
					// Still a rung left, so try that one.
					continue;
				}
				// No rung left, so return.
				return NOT_FOUND;
			}
			if (rungs_[curRungId_].getNumEvents(bucketId) >= THRESH
					&& curRungId_ < MAX_RUNG_ID) {
				// The bucket found is too full, so a new rung is created.
				assert curRungId_ < MAX_RUNG_ID : "Too many rungs created!";
				curRungId_++;
				rungs_[curRungId_] = rungs_[curRungId_ - 1]
						.createRungFromBucket(bucketId, rungs_[curRungId_]);
				// Continue with the newly created rung.
				continue;
			}
			// A non-empty bucket that is not too full has been found, so return
			// its ID.
			return bucketId;
		}
	}

	/**
	 * Returns whether there is currently at least one rung in use.
	 * 
	 * @return <code>true</code> if the number of rungs currently used is
	 *         nonzero, <code>false</code> otherwise.
	 */
	private boolean hasRung() {
		return curRungId_ >= 0;
	}

	/**
	 * Implements a rung.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private static class Rung {
		/**
		 * The buckets of this rung.
		 */
		private final ArrayDeque<Event>[] buckets_;

		/**
		 * The starting timestamp threshold of the first bucket of this rung.
		 * Used for calculating the bucket index when inserting an event.
		 */
		private long start_;

		/**
		 * The starting timestamp threshold of the first valid bucket of this
		 * rung at which subsequent dequeue operations will start. Also the
		 * minimum timestamp threshold of events which can be enqueued in this
		 * rung.
		 */
		private long curMin_;

		/**
		 * The index of the first non-empty bucket of this rung. Shifted to the
		 * right whenever a bucket of this rung is moved to bottom or to another
		 * rung.
		 */
		private int curBucket_;

		/**
		 * The bucket width of this rung.
		 */
		private long bucketwidth_;

		/**
		 * Constructor.
		 * 
		 * @param size
		 *            the number of buckets of this rung.
		 * @param start
		 *            the starting timestamp of this rung (lower bound for
		 *            events to be enqueued).
		 * @param curMin
		 *            the current minimum timestamp bound for events in this
		 *            rung.
		 * @param bucketwidth
		 *            the width of each bucket of this rung.
		 */
		@SuppressWarnings("unchecked")
		private Rung(int size, long start, long curMin, long bucketwidth) {
			reset(start, curMin, bucketwidth);
			buckets_ = new ArrayDeque[size];
			for (int i = 0; i < buckets_.length; i++) {
				buckets_[i] = new ArrayDeque<Event>(THRESH);
			}
		}

		/**
		 * Resets an existing rung (avoids creating a new rung object and its
		 * buckets).
		 * 
		 * @param start
		 *            the starting timestamp of this rung (lower bound for
		 *            events to be enqueued).
		 * @param curMin
		 *            the current minimum timestamp bound for events in this
		 *            rung.
		 * @param bucketwidth
		 *            the width of each bucket of this rung.
		 */
		private void reset(long start, long curMin, long bucketwidth) {
			start_ = start;
			curMin_ = curMin;
			curBucket_ = 0;
			bucketwidth_ = bucketwidth;
		}

		/**
		 * Decides whether the event may be enqueued into this rung.
		 * 
		 * @param event
		 *            the event.
		 * @return <code>true</code> if the event fits into this rung,
		 *         <code>false</code> otherwise.
		 */
		private boolean fits(Event event) {
			return event.getExecTime() >= curMin_;
		}

		/**
		 * Provides the ID of the current (i.e., the first non-empty) bucket of
		 * this rung.
		 * 
		 * @return the bucket ID.
		 */
		private int getCurrentBucket() {
			return curBucket_;
		}

		/**
		 * Provides the number of buckets of this rung.
		 * 
		 * @return the number of buckets of this rung.
		 */
		private int getNumBuckets() {
			return buckets_.length;
		}

		/**
		 * Provides the number of events of a specific bucket of this rung.
		 * 
		 * @param bucketId
		 *            the bucket ID.
		 * @return the number of events in the bucket.
		 */
		private int getNumEvents(int bucketId) {
			return buckets_[bucketId].size();
		}

		/**
		 * Moves all events of the specified bucket into the "bottom" queue.
		 * 
		 * @param bucketId
		 *            the ID of the bucket.
		 * @param bottom
		 *            a reference to the "bottom" queue.
		 */
		private void moveBucketToBottom(int bucketId, EventQueue bottom) {
			final ArrayDeque<Event> bucket = buckets_[bucketId];
			while (!bucket.isEmpty()) {
				bottom.insert(bucket.pollFirst());
			}
			shift();
		}

		/**
		 * Shifts the current bucket of this rung to the next bucket.
		 */
		private void shift() {
			curMin_ += bucketwidth_;
			curBucket_++;
		}

		/**
		 * Provides the bucket ID for an event to be enqueued in this rung.
		 * 
		 * @param ts
		 *            the timestamp of the event.
		 * @return the ID of the bucket where the event is to be enqueued.
		 */
		private int getBucketId(long ts) {
			return (int) ((ts - start_) / bucketwidth_);
		}

		/**
		 * Creates a new rung from this rung.
		 * 
		 * @param bucketId
		 *            the ID of the bucket that becomes the next rung.
		 * @param rung
		 *            a reference to the new rung (could be null).
		 * @return a reference to the new rung.
		 */
		private Rung createRungFromBucket(int bucketId, Rung rung) {
			// Set all values for the rung.
			long bucketwidth = bucketwidth_ / THRESH;
			if (bucketwidth * THRESH < bucketwidth_) {
				bucketwidth++;
			}
			// Has this rung been used before? If so, then reuse it.
			if (rung == null) {
				rung = new Rung(THRESH, curMin_, curMin_, bucketwidth);
			} else {
				rung.reset(curMin_, curMin_, bucketwidth);
			}
			// Move events from bucket to the new rung.
			final ArrayDeque<Event> bucket = buckets_[bucketId];
			while (!bucket.isEmpty()) {
				rung.add(bucket.pollFirst());
			}
			shift();
			return rung;
		}

		/**
		 * Creates a new rung from the bottom queue with the help of this rung.
		 * 
		 * @param bottom
		 *            a reference to the "bottom" queue.
		 * @param rung
		 *            a reference to the new rung (could be null).
		 * @return a reference to the new rung.
		 */
		private Rung createRungFromBottom(EventQueue bottom, Rung rung) {
			// Set all values for the rung.
			final long minTS = bottom.peekFirst().getExecTime();
			// Note: The following works only with sequential simulation.
			long bucketwidth = bucketwidth_ / THRESH;
			if (bucketwidth * THRESH < bucketwidth_) {
				bucketwidth++;
			}
			long val = start_;
			while (minTS >= val + bucketwidth_) {
				val += bucketwidth_;
			}
			// Has this rung been used before? If so, then reuse it.
			if (rung == null) {
				rung = new Rung(THRESH, val, val, bucketwidth);
			} else {
				rung.reset(val, val, bucketwidth);
			}
			// Move events from bottom into the new rung.
			while (!bottom.isEmpty()) {
				rung.add(bottom.removeFirst());
			}
			return rung;
		}

		/**
		 * Adds event to a bucket of this rung.
		 * 
		 * @param event
		 *            the event
		 */
		private void add(Event event) {
			buckets_[getBucketId(event.getExecTime())].add(event);
		}

		/**
		 * @see java.lang.Object#toString()
		 */
		public String toString() {
			return Arrays.toString(buckets_);
		}
	}

	/**
	 * The top array of the ladder queue.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private static class Top {
		/**
		 * The initial size of the top_ array.
		 */
		private static final int TOP_INIT_SIZE = 50;

		/**
		 * The unsorted "top" list.
		 */
		private ArrayDeque<Event> events_;

		/**
		 * The minimum timestamp of all events in top_. Its value is updated as
		 * events are enqueued into top_.
		 */
		private long topMin_;

		/**
		 * The maximum timestamp of all events in top_. Its value is updated as
		 * events are enqueued into top_.
		 */
		private long topMax_;

		/**
		 * The minimum timestamp threshold of events that must be enqueued in
		 * top_.
		 */
		private long topStart_;

		/**
		 * Constructor.
		 */
		private Top() {
			events_ = new ArrayDeque<Event>(TOP_INIT_SIZE);
			topMin_ = Long.MAX_VALUE;
			topMax_ = Long.MIN_VALUE;
			topStart_ = 0;
		}

		/**
		 * Tries to add an event to the "top" queue. The event is only added if
		 * its timestamp is larger than the minimum timestamp required by the
		 * "top" queue.
		 * 
		 * @param event
		 *            the event to add.
		 * @return whether the event was added to the "top" queue.
		 */
		private boolean add(Event event) {
			final long ts = event.getExecTime();
			if (ts >= topStart_) {
				if (events_.isEmpty()) {
					topMax_ = ts;
					topMin_ = ts;
				} else {
					if (ts > topMax_) {
						topMax_ = ts;
					} else if (ts < topMin_) {
						topMin_ = ts;
					}
				}
				events_.add(event);
				return true;
			}
			return false;
		}

		/**
		 * Checks for a special case where the "top" queue contains only
		 * simultaneous events. In that case, the events are directly moved into
		 * the "bottom" queue without creating a rung.
		 * 
		 * @param bottom
		 *            a reference to the "bottom" queue.
		 * @return whether the events have been moved into the "bottom" queue.
		 */
		private boolean moveToBottom(EventQueue bottom) {
			if (topMax_ == topMin_) {
				// Special case: only events with the same timestamp in top, so
				// directly move them into bottom.
				while (!events_.isEmpty()) {
					bottom.insert(events_.pollFirst());
				}
				// Reset top information.
				topStart_ = topMax_;
				topMin_ = Long.MAX_VALUE;
				topMax_ = Long.MIN_VALUE;
				return true;
			}
			return false;
		}

		/**
		 * Creates a new rung from the top queue.
		 * 
		 * @return the first rung of the ladder.
		 */
		private Rung createFirstRung() {
			// Always create rung since its bucket number is variable.
			// Set all values for the rung.
			final int numEvents = events_.size();
			final long width = topMax_ - topMin_;
			long bucketwidth = width / (numEvents - 1);
			if (bucketwidth * (numEvents - 1) < width) {
				bucketwidth++;
			}
			Rung rung = new Rung(numEvents, topMin_, topMin_, bucketwidth);
			// Move events into the new rung.
			while (!events_.isEmpty()) {
				rung.add(events_.pollFirst());
			}
			// Reset top information.
			topStart_ = topMin_ + numEvents * bucketwidth;
			topMin_ = Long.MAX_VALUE;
			topMax_ = Long.MIN_VALUE;
			return rung;
		}
	}
}
