/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/scheduler/ListQueue.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.scheduler;

import core.Event;

/**
 * Event queue based on a simple linked list implementation. Recommended for
 * testing purposes only.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class ListQueue implements EventQueue {
	/**
	 * The head of the list.
	 */
	private ListElement head_ = new ListHead();

	/**
	 * The tail of the list.
	 */
	private ListElement tail_ = new ListTail();

	/**
	 * The current size of the list.
	 */
	private int size_ = 0;

	/**
	 * Constructor.
	 */
	public ListQueue() {
		head_.next_ = tail_;
		tail_.previous_ = head_;
	}

	/**
	 * @see core.scheduler.EventQueue#size()
	 */
	public int size() {
		return size_;
	}

	/**
	 * @see core.scheduler.EventQueue#isEmpty()
	 */
	public boolean isEmpty() {
		return size_ == 0;
	}

	/**
	 * @see core.scheduler.EventQueue#insert(core.Event)
	 */
	public void insert(Event event) {
		ListElement newElement = new ListElement(event);
		ListElement currentElement = tail_.previous_;
		while (newElement.isSmallerThan(currentElement)) {
			currentElement = currentElement.previous_;
		}
		// Now newElement is larger than (or equal to) currentElement.
		currentElement.next_.previous_ = newElement;
		newElement.next_ = currentElement.next_;
		newElement.previous_ = currentElement;
		currentElement.next_ = newElement;
		size_++;
	}

	/**
	 * @see core.scheduler.EventQueue#removeFirst()
	 */
	public Event removeFirst() {
		Event event = head_.next_.event_;
		head_.next_ = head_.next_.next_;
		head_.next_.previous_ = head_;
		size_--;
		return event;
	}

	/**
	 * @see core.scheduler.EventQueue#peekFirst()
	 */
	public Event peekFirst() {
		assert size_ > 0;
		return head_.next_.event_;
	}

	/**
	 * List element holding an event.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private static class ListElement {
		/**
		 * The encapsulated event.
		 */
		private Event event_;

		/**
		 * The next element in the list.
		 */
		private ListElement next_;

		/**
		 * The previous element in the list.
		 */
		private ListElement previous_;

		/**
		 * Constructor.
		 */
		private ListElement() {
			// Empty.
		}

		/**
		 * Constructor.
		 * 
		 * @param event
		 *            the event to encapsulate.
		 */
		private ListElement(Event event) {
			event_ = event;
		}

		/**
		 * For comparison with other elements.
		 * 
		 * @param otherElement
		 *            the element to compare to.
		 * @return <code>true</code> if this element is smaller than the other
		 *         one, otherwise <code>false</code>.
		 */
		private boolean isSmallerThan(ListElement otherElement) {
			if (otherElement instanceof ListHead) {
				return false;
			}
			if (otherElement instanceof ListTail) {
				return true;
			}
			return event_.hasSmallerTimestampThan(otherElement.event_);
		}
	}

	/**
	 * Convenience class.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private static class ListHead extends ListElement {
		// Empty.
	}

	/**
	 * Convenience class.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private static class ListTail extends ListElement {
		// Empty.
	}
}
