/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/scheduler/ShiftedHeapQueue.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.scheduler;

import core.Event;

/**
 * Event queue based on an optimized binary heap implementation.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class ShiftedHeapQueue implements EventQueue {
	/**
	 * The initial length of the internal array.
	 */
	private static final int INIT_LENGTH = 100;

	/**
	 * The current number of events on the heap.
	 */
	private int size_;

	/**
	 * The internal array.
	 */
	private Event[] events_;

	/**
	 * Helper variable for halving the array size.
	 */
	private int halveSize_;

	/**
	 * Constructor.
	 */
	public ShiftedHeapQueue() {
		size_ = 0;
		events_ = new Event[INIT_LENGTH + 1];
	}

	/**
	 * @see core.scheduler.EventQueue#size()
	 */
	public int size() {
		return size_;
	}

	/**
	 * @see core.scheduler.EventQueue#isEmpty()
	 */
	public boolean isEmpty() {
		return size_ == 0;
	}

	/**
	 * @see core.scheduler.EventQueue#insert(core.Event)
	 */
	public void insert(Event event) {
		if (size_ == events_.length - 1) {
			doubleCapacity();
		}

		// Percolate up.
		int hole = ++size_;
		for (; hole > 1 && event.hasSmallerTimestampThan(events_[hole / 2]); hole /= 2) {
			events_[hole] = events_[hole / 2];
		}
		events_[hole] = event;
		// Potentially faster percolate up algorithm. May be used as an
		// alternative.
		/*
		 * int hole = ++size_; int parent = hole / 2; while(hole > 1 &&
		 * event.hasSmallerTimestampThan(events_[parent])) { events_[hole] =
		 * events_[parent]; hole = parent; parent = hole / 2; } events_[hole] =
		 * event;
		 */
	}

	/**
	 * @see core.scheduler.EventQueue#removeFirst()
	 */
	public Event removeFirst() {
		assert !isEmpty();

		Event minItem = events_[1];
		events_[1] = events_[size_];
		events_[size_--] = null;
		percolateDown(1);

		if (size_ < halveSize_) {
			halveCapacity();
		}
		return minItem;
	}

	/**
	 * @see core.scheduler.EventQueue#peekFirst()
	 */
	public Event peekFirst() {
		return events_[1];
	}

	/**
	 * Internal method to percolate down in the heap.
	 * 
	 * @param hole
	 *            the index at which the percolate begins.
	 */
	private void percolateDown(int hole) {
		// Percolate down.
		int child;
		Event tmp = events_[hole];
		for (; hole * 2 <= size_; hole = child) {
			child = hole * 2;
			if (child != size_
					&& events_[child + 1]
							.hasSmallerTimestampThan(events_[child])) {
				child++;
			}
			if (events_[child].hasSmallerTimestampThan(tmp)) {
				events_[hole] = events_[child];
			} else {
				break;
			}
		}
		events_[hole] = tmp;

		// Possibly faster percolate down.
		/*
		 * Event tmp = events_[hole]; int child = hole * 2; while (child <=
		 * size_) { if (child != size_ && events_[child +
		 * 1].hasSmallerTimestampThan(events_[child])) { child++; } if
		 * (events_[child].hasSmallerTimestampThan(tmp)) { events_[hole] =
		 * events_[child]; hole = child; child = hole * 2; } else { break; } }
		 * events_[hole] = tmp;
		 */
	}

	/**
	 * Doubles the capacity of the internal array.
	 */
	private void doubleCapacity() {
		Event[] events = new Event[events_.length * 2 + 1];
		System.arraycopy(events_, 1, events, 1, size_);
		events_ = events;
		halveSize_ = events_.length / 4;
	}

	/**
	 * Halves the capacity of the internal array.
	 */
	private void halveCapacity() {
		int newLength = events_.length / 2;
		if (newLength < INIT_LENGTH) {
			halveSize_ = 0;
			return;
		}
		assert size_ <= newLength;
		Event[] events = new Event[newLength];
		System.arraycopy(events_, 1, events, 1, size_);
		events_ = events;
		halveSize_ = events_.length / 4;
	}
}
