/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/sync/ConcurrentEventArray.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.sync;

import java.util.ArrayList;
import java.util.List;

import core.Event;

/**
 * Data structure used to store events sent from one LP/partition to another. It
 * does not sort the events but keeps FIFO order.
 * 
 * The events are stored into an array which has an initially fixed size. Should
 * the capacity be reached before the receiving LP collects the events, an
 * overflow list is used.
 * 
 * @author Patrick Peschlow
 * @author Martin Geuer
 * 
 * @version 1.0
 */
public class ConcurrentEventArray {
	/**
	 * Default initial size.
	 */
	private static final int INIT_SIZE = 100;

	/**
	 * The array holding the events.
	 */
	private final Event[] events_;

	/**
	 * The position of the next event in the array.
	 */
	private int nextEvent_ = 0;

	/**
	 * The position of the next free position in the array.
	 */
	private int nextFreePosition_ = 0;

	/**
	 * An array where events are stored whenever the main array can not hold any
	 * more events.
	 */
	private List<Event> overflow_ = new ArrayList<Event>();

	/**
	 * Constructor.
	 * 
	 * @param initialSize
	 *            The initial size of the array.
	 */
	ConcurrentEventArray(int initialSize) {
		events_ = new Event[initialSize];
	}

	/**
	 * Constructor. Uses a default size for the array.
	 */
	ConcurrentEventArray() {
		this(INIT_SIZE);
	}

	/**
	 * Returns the next event or <code>null</code> if no event is available.
	 * 
	 * @return the next event.
	 */
	Event peek() {
		return events_[nextEvent_];
	}

	/**
	 * Adds an event to the array. If the array is full, the overflow array is
	 * used. The access to the overflow array is implemented using synchronized
	 * blocks because it is so rarely (usually: never) used (thanks to event
	 * bundling, there are not many events exchanged between LPs).
	 * 
	 * @param event
	 *            the event to add.
	 */
	void add(Event event) {
		if (events_[nextFreePosition_] == null) {
			if (fillArray() && events_[nextFreePosition_] == null) {
				events_[nextFreePosition_] = event;
				nextFreePosition_++;
				if (nextFreePosition_ == events_.length) {
					nextFreePosition_ = 0;
				}
			} else {
				synchronized (overflow_) {
					overflow_.add(event);
				}
			}
		} else {
			synchronized (overflow_) {
				overflow_.add(event);
			}
		}
	}

	/**
	 * Transfers events from the overflow array to the main array if possible.
	 * Implemented using synchronized blocks because it is so rarely (usually:
	 * never) used (thanks to event bundling, there are not many events
	 * exchanged between LPs).
	 * 
	 * @return whether the overflow array is empty after this operation.
	 */
	boolean fillArray() {
		if (overflow_.isEmpty()) {
			return true;
		}
		synchronized (overflow_) {
			while (!overflow_.isEmpty() && events_[nextFreePosition_] == null) {
				events_[nextFreePosition_] = overflow_.remove(0);
				nextFreePosition_ = (nextFreePosition_ + 1) % events_.length;
			}
		}
		return overflow_.isEmpty();
	}

	/**
	 * Removes the first event from the array.
	 * 
	 * @return the next event.
	 */
	Event poll() {
		if (events_[nextEvent_] == null) {
			fillArray();
		}
		final Event e = events_[nextEvent_];
		events_[nextEvent_] = null;
		nextEvent_++;
		if (nextEvent_ == events_.length) {
			nextEvent_ = 0;
		}
		return e;
	}

	/**
	 * Checks whether an event is available in this structure.
	 * 
	 * @return <code>true</code> if an event is available, otherwise
	 *         <code>false</code>.
	 */
	boolean isEmpty() {
		return events_[nextEvent_] == null && overflow_.isEmpty();
	}

	/**
	 * Provides the number of events that are currently in the array.
	 * 
	 * @return the number of events in the array.
	 */
	int size() {
		if (nextFreePosition_ >= nextEvent_) {
			return nextFreePosition_ - nextEvent_;
		}
		return events_.length - nextEvent_ + nextFreePosition_;
	}
}
