/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/sync/DistributedSimController.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.sync;

import gnu.trove.TIntIntHashMap;
import gnu.trove.TIntObjectHashMap;

import java.util.ArrayList;
import java.util.List;

import jist.swans.mac.Mac802_11;
import core.BaseModule;
import core.Config;
import core.Config.BarrierType;
import core.Config.Lookahead;
import core.Event;
import core.ModuleRoot;
import core.distribution.BarrierPacket;
import core.distribution.CommLayer;
import core.distribution.StatsPacket;
import core.scheduler.EventScheduler;
import core.util.CustomLogger;
import edu.bonn.cs.net.jbarrier.barrier.Barrier;
import edu.bonn.cs.net.jbarrier.barrier.ButterflyBarrier;
import edu.bonn.cs.net.jbarrier.barrier.DisseminationBarrier;
import edu.bonn.cs.net.jbarrier.barrier.GenericReductor;
import edu.bonn.cs.net.jbarrier.barrier.TournamentBarrier;

/**
 * Implements a distributed centralized barrier synchronization mechanism on top
 * of the parallel conservative window-based synchronization mechanism. This
 * class has been designed for distributed wireless network simulations using
 * HiPWiNS, but is should be easily adaptable to other types of simulation
 * scenarios.
 * 
 * @author Patrick Peschlow
 * @author Martin Geuer
 * @author Andreas Voss
 * @author Daniel Heupel
 * 
 * @version 1.0
 */
public class DistributedSimController extends AbstractSimController implements
		LpBasedSimController {
	/**
	 * Key: LP ID, value: the corresponding LP.
	 */
	private final DistributedLp[] lpIdToLp_;

	/**
	 * The number of LPs.
	 */
	private final int numLps_;

	/**
	 * The number of threads (local LPs) used.
	 */
	private final int numThreads_;

	/**
	 * The barrier action.
	 */
	private final Runnable barrierAction_;

	/**
	 * An array holding all threads.
	 */
	private final DistributedThread[] threads_;

	/**
	 * Key: LP ID. Value: Thread ID.
	 */
	private final int[] lpIdToThreadId_;

	/**
	 * Key: Thread ID. Value: LP ID.
	 */
	private final int[] threadIdToLpId_;

	/**
	 * ID of first local LP
	 */
	private final int lpIdShift_;

	/**
	 * Key: LP ID. Value: Whether the LP is managed by a thread on this host.
	 */
	private final boolean[] isLocalLp_;

	/**
	 * Used by clients
	 */
	private final CommLayer commLayer_;

	/**
	 * Packet used to send statistics at the end of the simulation.
	 */
	private final StatsPacket statsPacket_ = new StatsPacket();

	/**
	 * Collects outgoing events to be sent over the network.
	 */
	private final List<Event> outgoingNetworkEvents_ = new ArrayList<Event>();

	/**
	 * The safe time bound for the current window (calculated in each barrier).
	 */
	private long safeTimeBound_ = 0;

	/**
	 * Counts the number of synchronization points.
	 */
	private long numSyncs_ = 0;

	/**
	 * Constructor.
	 * 
	 * @param commLayer
	 *            the communication layer used by this controller.
	 * @param firstLocalLpId
	 *            the offset of LP IDs on this host.
	 * @param numThreads
	 *            the number of threads (local LPs) to run on this host.
	 */
	public DistributedSimController(CommLayer commLayer, int firstLocalLpId,
			int numThreads) {
		commLayer_ = commLayer;
		lpIdShift_ = firstLocalLpId;
		numThreads_ = numThreads;
		// Create LPs and assign modules to LPs.
		final List<BaseModule> modules = ModuleRoot.getRoot().getModuleList();
		numLps_ = Config.getNumPartitions();
		lpIdToLp_ = new DistributedLp[numLps_];
		isLocalLp_ = new boolean[numLps_];
		threadIdToLpId_ = new int[numLps_];
		lpIdToThreadId_ = new int[numLps_];
		threads_ = new DistributedThread[numThreads_];
		TIntIntHashMap moduleIdToLpId = new TIntIntHashMap(modules.size());
		int lastLocalLpId = firstLocalLpId + numThreads_ - 1;
		for (int lpId = firstLocalLpId; lpId <= lastLocalLpId; lpId++) {
			final List<BaseModule> thisLpModules = ModuleRoot.getRoot()
					.getModulesOfLp(lpId);
			EventScheduler eventScheduler = new EventScheduler(Config
					.getScheduler().getEventQueueClass());
			DistributedLp lp = new DistributedLp(lpId, eventScheduler, this);
			TIntObjectHashMap<BaseModule> thisLpModuleMap = new TIntObjectHashMap<BaseModule>(
					thisLpModules.size());
			for (BaseModule module : thisLpModules) {
				module.setLp(lp);
				final int moduleId = module.getId();
				thisLpModuleMap.put(moduleId, module);
				moduleIdToLpId.put(moduleId, lpId);
			}
			lp.setModuleMap(thisLpModuleMap);
			assert lpIdToLp_[lpId] == null;
			lpIdToLp_[lpId] = lp;
		}
		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("DistributedSimController: Created "
					+ numThreads_ + " LPs (with IDs " + firstLocalLpId + " to "
					+ lastLocalLpId + ")");
		}
		// Schedule initial events.
		InitLp initLp = ModuleRoot.getInitLp();
		TIntObjectHashMap<List<Event>> eventsMap = initLp.getEventsMap();
		for (BaseModule module : modules) {
			final int moduleId = module.getId();
			if (eventsMap.containsKey(moduleId)) {
				for (Event event : eventsMap.get(moduleId)) {
					getLp(moduleIdToLpId.get(moduleId)).getScheduler().add(
							event);
				}
			}
		}
		// Set up threads and barrier.
		switch (Config.getLookahead()) {
		case EVENT:
			if (Config.getBarrier() == BarrierType.TOURNAMENT)
				barrierAction_ = new TournamentEventBarrierAction();
			else
				barrierAction_ = new EventLookaheadBarrierAction();
			break;
		case PROTOCOL:
			barrierAction_ = new ProtocolLookaheadBarrierAction();
			break;
		case NONE:
			barrierAction_ = new NoLookaheadBarrierAction();
			break;
		default:
			throw new RuntimeException("Bad barrier action type!");
		}

		switch (Config.getBarrier()) {
		case BUTTERFLY:
			assert (numThreads_ & (numThreads_ - 1)) == 0;
		case DISSEMINATION:
		case TOURNAMENT:
		case SINGLE_PARTITION:
			break;
		default:
			throw new RuntimeException("Bad barrier type!");
		}
	}

	/**
	 * Provides the statistics packet to fill in at the end of the simulation.
	 * 
	 * @return the statistics packet.
	 */
	public StatsPacket getStatsPacket() {
		return statsPacket_;
	}

	/**
	 * Provides the number of synchronization points (for statistics at the end
	 * of the simulation).
	 * 
	 * @return the number of synchronization points.
	 */
	public long getNumSyncs() {
		return numSyncs_;
	}

	/**
	 * @see core.sync.LpBasedSimController#sendEvent(core.Event, int)
	 */
	@Override
	public void sendEvent(Event event, int srcLpId) {
		assert event.getDstModuleId() == -1;
		outgoingNetworkEvents_.add(event);
	}

	/**
	 * @see core.sync.LpBasedSimController#getSafeTimeBound()
	 */
	@Override
	public long getSafeTimeBound() {
		return safeTimeBound_;
	}

	/**
	 * @see core.sync.AbstractSimController#prepareSim()
	 */
	protected void prepareSim() {
		// Create the barrier object.
		Barrier barrier;
		if (numThreads_ == 1) {
			barrier = new SinglePartitionBarrier(barrierAction_);
		} else {
			switch (Config.getBarrier()) {
			case DISSEMINATION:
				barrier = new DisseminationBarrier(numThreads_, barrierAction_,
						new TwofoldReduction());
				break;
			case BUTTERFLY:
				barrier = new ButterflyBarrier(numThreads_, barrierAction_,
						new TwofoldReduction());
				break;
			case TOURNAMENT:
				if (Config.getLookahead() == Lookahead.EVENT) {
					barrier = new TournamentBarrier(numThreads_,
							barrierAction_, new TournamentEventReduction());
				} else {
					barrier = new TournamentBarrier(numThreads_,
							barrierAction_, new TwofoldReduction());
				}
				break;
			case SINGLE_PARTITION:
				barrier = new SinglePartitionBarrier(barrierAction_);
				break;
			default:
				throw new RuntimeException("Unknown barrier algorithm type!");
			}
		}

		if (CustomLogger.isDebugEnabled()) {
			CustomLogger
					.debug("DistributedSimController: prepareSim(), numThreads:"
							+ numThreads_);
		}

		// Assign the LPs to threads and create the threads.
		for (int threadId = 0; threadId < numThreads_; threadId++) {
			final int lpId = (lpIdShift_ + threadId);
			final DistributedLp lp = getLp(lpId);
			lpIdToThreadId_[lpId] = threadId;
			threadIdToLpId_[threadId] = lpId;
			isLocalLp_[lpId] = true;
			if (Config.getBarrier() == BarrierType.TOURNAMENT
					&& Config.getLookahead() == Lookahead.EVENT) {
				threads_[threadId] = new TournamentEventDistributedBarrierThread(
						this, numThreads_, threadId, lp, barrier);
			} else {
				threads_[threadId] = new DistributedThread(this, numThreads_,
						threadId, lp, barrier);
			}
			if (CustomLogger.isDebugEnabled()) {
				CustomLogger.debug("threadId:" + threadId + ",lpId:" + lpId
						+ ",lpIdToThreadId_[lpId]:" + lpIdToThreadId_[lpId]);
			}
		}
	}

	/**
	 * @see AbstractSimController#runSim()
	 */
	protected void runSim() {
		// Start simulation threads.
		for (DistributedThread thread : threads_) {
			thread.start();
		}
		// Wait until done.
		for (DistributedThread thread : threads_) {
			try {
				thread.join();
			} catch (InterruptedException e) {
				if (CustomLogger.isWarningEnabled()) {
					CustomLogger.warning(
							"Thread has been interrupted during execution!", e);
				}
			}
		}
	}

	/**
	 * @see core.sync.AbstractSimController#finishSim()
	 */
	protected void finishSim() {
		if (CustomLogger.isInfoEnabled()) {
			long numExecutedEvents = 0;
			for (int lpId = 0; lpId < numLps_; lpId++) {
				if (isLocalLp_[lpId]) {
					numExecutedEvents += getLp(lpId).getNumExecutedEvents();
				}
			}
			for (int lpId = 0; lpId < numLps_; lpId++) {
				if (isLocalLp_[lpId]) {
					final DistributedLp lp = getLp(lpId);
					CustomLogger.info("LP "
							+ lp.getId()
							+ " executed events: "
							+ lp.getNumExecutedEvents()
							+ " ("
							+ Math.round((double) lp.getNumExecutedEvents()
									/ (double) numExecutedEvents * 100.0)
							+ "%)");
				}
			}
			CustomLogger
					.info("DistributedBarrierSimController - Total number of executed events: "
							+ numExecutedEvents);
		}
	}

	/**
	 * Provides the LP to a specified LP ID.
	 * 
	 * @param lpId
	 *            the LP ID.
	 * @return the corresponding LP.
	 */
	protected DistributedLp getLp(int lpId) {
		return lpIdToLp_[lpId];
	}

	/**
	 * Returns the thread ID to the specified LP ID.
	 * 
	 * @param lpId
	 *            the LP ID.
	 * @return the thread ID.
	 */
	protected int getThreadId(int lpId) {
		return lpIdToThreadId_[lpId];
	}

	/**
	 * Provides the thread to a specified thread ID.
	 * 
	 * @param threadId
	 *            the thread ID.
	 * @return the corresponding thread.
	 */
	protected DistributedThread getThread(int threadId) {
		return threads_[threadId];
	}

	/**
	 * Shut down all threads.
	 */
	private void shutdown() {
		for (int i = 0; i < threads_.length; i++) {
			threads_[i].shutdown();
		}
	}

	/**
	 * Performs the distributed barrier with the server as soon as the parallel
	 * barrier has been completed.
	 */
	private void performDistributedBarrier() {
		// Send safe time bound and events to server.
		BarrierPacket outgoingPacket = new BarrierPacket(safeTimeBound_,
				outgoingNetworkEvents_);
		commLayer_.writeObject(outgoingPacket);

		// Receive global safe time bound and all events.
		BarrierPacket incomingPacket = (BarrierPacket) commLayer_
				.readObjectBlocking();
		safeTimeBound_ = incomingPacket.getSafeTimeBound();
		List<Event> events = incomingPacket.getEvents();

		// Distribute events to my LPs.
		for (int threadId = 0; threadId < threads_.length; threadId++) {
			getThread(threadId).setIncomingNetworkEvents(events);
		}

		// Clear for next iteration.
		outgoingNetworkEvents_.clear();

		// Are we finished?
		if (safeTimeBound_ == Long.MAX_VALUE) {
			shutdown();
		}
	}

	/**
	 * Barrier action in case of event lookahead.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private class EventLookaheadBarrierAction implements Runnable {
		/**
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			++numSyncs_;
			long net = threads_[0].getNET();
			for (int i = 1; i < threads_.length; i++) {
				net = Math.min(net, threads_[i].getNET());
			}
			long ntet = threads_[0].getNTET();
			for (int i = 1; i < threads_.length; i++) {
				ntet = Math.min(ntet, threads_[i].getNTET());
			}
			safeTimeBound_ = Math.min(ntet, net + Mac802_11.RX_TX_TURNAROUND)
					+ Config.getPropagation().getDelay();

			if (safeTimeBound_ < 0) {
				safeTimeBound_ = Long.MAX_VALUE;
			}
			performDistributedBarrier();
		}
	}

	/**
	 * Barrier action in case of protocol lookahead.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private class ProtocolLookaheadBarrierAction implements Runnable {
		/**
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			++numSyncs_;
			long net = threads_[0].getNET();
			for (int i = 1; i < threads_.length; i++) {
				net = Math.min(net, threads_[i].getNET());
			}
			long ntet = threads_[0].getNTET();
			for (int i = 1; i < threads_.length; i++) {
				ntet = Math.min(ntet, threads_[i].getNTET());
			}
			safeTimeBound_ = Math.max(ntet, net)
					+ Config.getPropagation().getDelay();
			if (safeTimeBound_ < 0) {
				safeTimeBound_ = Long.MAX_VALUE;
			}
			performDistributedBarrier();
		}
	}

	/**
	 * Barrier action in case of no advanced lookahead usage.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private class NoLookaheadBarrierAction implements Runnable {
		/**
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			++numSyncs_;
			long net = threads_[0].getNET();
			for (int i = 1; i < threads_.length; i++) {
				net = Math.min(net, threads_[i].getNET());
			}
			long ntet = threads_[0].getNTET();
			for (int i = 1; i < threads_.length; i++) {
				ntet = Math.min(ntet, threads_[i].getNTET());
			}
			safeTimeBound_ = net + Config.getPropagation().getDelay();
			if (safeTimeBound_ < 0) {
				safeTimeBound_ = Long.MAX_VALUE;
			}
			performDistributedBarrier();
		}
	}

	/**
	 * Optimized barrier method for tournament barriers with event lookahead.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private class TournamentEventBarrierAction implements Runnable {
		/**
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			++numSyncs_;
			final long globalMin = ((TournamentEventDistributedBarrierThread) threads_[0])
					.getTourHelperMinValue();
			safeTimeBound_ = globalMin + Config.getPropagation().getDelay();
			if (safeTimeBound_ < 0) {
				safeTimeBound_ = Long.MAX_VALUE;
			}
			performDistributedBarrier();
		}
	}

	/**
	 * Two-fold reduction of NET and NTET values.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	class TwofoldReduction implements GenericReductor {
		/**
		 * @see edu.bonn.cs.net.jbarrier.barrier.GenericReductor#reduce(int,
		 *      int)
		 */
		@Override
		public void reduce(int threadId1, int threadId2) {
			final DistributedThread bth1 = threads_[threadId1];
			final DistributedThread bth2 = threads_[threadId2];
			final long net = bth2.getNET();
			final long ntet = bth2.getNTET();
			if (bth1.getNET() > net) {
				bth1.setNET(net);
			}
			if (bth1.getNTET() > ntet) {
				bth1.setNTET(ntet);
			}
		}
	}

	/**
	 * Specialized reduction for simulations with tournament barrier and event
	 * lookahead.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	class TournamentEventReduction implements GenericReductor {
		/**
		 * @see edu.bonn.cs.net.jbarrier.barrier.GenericReductor#reduce(int,
		 *      int)
		 */
		@Override
		public void reduce(int threadId1, int threadId2) {
			final TournamentEventDistributedBarrierThread bth1 = (TournamentEventDistributedBarrierThread) threads_[threadId1];
			final TournamentEventDistributedBarrierThread bth2 = (TournamentEventDistributedBarrierThread) threads_[threadId2];
			final long helper = bth2.getTourHelperMinValue();
			if (bth1.getTourHelperMinValue() > helper) {
				bth1.setTourHelperMinValue(helper);
			}
		}
	}

}
