/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/sync/DistributedThread.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.sync;

import java.util.ArrayList;
import java.util.List;

import modules.wireless.partition.DistributedPartition;

import jist.swans.mac.Mac802_11;
import core.Event;
import edu.bonn.cs.net.jbarrier.barrier.Barrier;

/**
 * A thread executing an LP in a distributed simulation.
 * 
 * @author Patrick Peschlow
 * @author Martin Geuer
 * @author Daniel Heupel
 * 
 * @version 1.0
 */
class DistributedThread extends ParallelThread {
	/**
	 * The list of incoming events delivered via the network in every window.
	 */
	private List<Event> incomingNetworkEvents_;

	/**
	 * Constructor.
	 * 
	 * @param simCtrl
	 *            the SimController.
	 * @param numThreads
	 *            the number of threads.
	 * 
	 * @param threadId
	 *            the unique id of this thread.
	 * @param lp
	 *            the LP run by this thread.
	 * @param barrier
	 *            the barrier object used by this thread.
	 */
	protected DistributedThread(LpBasedSimController simCtrl, int numThreads,
			int threadId, DistributedLp lp, Barrier barrier) {
		super(simCtrl, numThreads, threadId, lp, barrier);
		incomingNetworkEvents_ = new ArrayList<Event>();
	}

	/**
	 * Hands the thread a reference to the (shared) list of incoming network
	 * events.
	 * 
	 * @param events
	 *            the list of incoming network events.
	 */
	protected void setIncomingNetworkEvents(List<Event> events) {
		incomingNetworkEvents_ = events;
	}

	/**
	 * Schedules the network events received during the last window.
	 */
	protected void startScheduling() {
		// We don't need to consider incomingChannels_ in a distributed
		// simulation, as they are not used.
		for (Event event : incomingNetworkEvents_) {
			assert event.getDstModuleId() == DistributedPartition.DIST_PARTITION_BROADCAST;
			lp_.receiveEvent(copy(event, lp_.getId()));
		}
	}

	/**
	 * Copies an event, updating its destination module ID to a new value.
	 * 
	 * @param event
	 *            the event to copy.
	 * @param dstModuleId
	 *            the new destination module ID.
	 * @return the new event.
	 */
	private Event copy(Event event, int dstModuleId) {
		Event copy = new Event(event.getExecTime(), event.getSchedTime(),
				dstModuleId, event.getSrcModuleId(), event.getId(),
				event.getParameter(), event.getPriority());
		return copy;
	}
}

/**
 * Optimized version of the thread class for simulations with tournament barrier
 * and event lookahead.
 * 
 * @author Patrick Peschlow
 * @author Andreas Voss
 * @author Daniel Heupel
 * 
 * @version 1.0
 */
class TournamentEventDistributedBarrierThread extends DistributedThread {
	/**
	 * Helper value for the optimized tournament barrier with event lookahead.
	 */
	private long tourHelperMinValue_;

	/**
	 * Constructor.
	 * 
	 * @param simCtrl
	 *            the SimController.
	 * @param numThreads
	 *            the number of threads.
	 * 
	 * @param threadId
	 *            the unique id of this thread.
	 * @param lp
	 *            the LP run by this thread.
	 * @param barrier
	 *            the barrier object used by this thread.
	 */
	protected TournamentEventDistributedBarrierThread(
			DistributedSimController simCtrl, int numThreads, int threadId,
			DistributedLp lp, Barrier barrier) {
		super(simCtrl, numThreads, threadId, lp, barrier);
	}

	/**
	 * @see Thread#run()
	 */
	@Override
	public final void run() {
		while (!shutdown_) {
			startScheduling();
			startWorking();
			// Perform local minimum calculation already here.
			tourHelperMinValue_ = (net_ == Long.MAX_VALUE) ? ntet_ : Math.min(
					ntet_, net_ + Mac802_11.RX_TX_TURNAROUND);
			barrier_.await(threadId_);
		}
	}

	/**
	 * Provides the helper value used in the reduction.
	 * 
	 * @return the tourHelperMinValue
	 */
	public long getTourHelperMinValue() {
		return tourHelperMinValue_;
	}

	/**
	 * Sets the helper value used in the reduction.
	 * 
	 * @param tourHelperMinValue
	 *            the new value.
	 */
	public void setTourHelperMinValue(long tourHelperMinValue) {
		tourHelperMinValue_ = tourHelperMinValue;
	}
}
