/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/sync/InitLp.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.sync;

import gnu.trove.TIntObjectHashMap;

import java.util.ArrayList;
import java.util.List;

import core.BaseModule;
import core.Config;
import core.Event;

/**
 * A dummy LP implementation, which is only used to collect initial events at
 * the beginning of the simulation.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class InitLp implements Lp {
	/**
	 * Key: Module ID. Value: The initial events of that module.
	 */
	private final TIntObjectHashMap<List<Event>> eventsMap_;

	/**
	 * Constructor.
	 */
	public InitLp() {
		eventsMap_ = new TIntObjectHashMap<List<Event>>();
	}

	/**
	 * @see core.sync.Lp#setModuleMap(gnu.trove.TIntObjectHashMap)
	 */
	@Override
	public void setModuleMap(TIntObjectHashMap<BaseModule> moduleMap) {
		assert false;
	}

	/**
	 * @see core.sync.Lp#createNewEvent(long, int, int, java.lang.Enum,
	 *      java.lang.Object, int)
	 */
	@SuppressWarnings("rawtypes")
	@Override
	public void createNewEvent(long execTime, int dstModuleId, int srcModuleId,
			Enum eventId, Object parameter, int priority) {
		createNewEvent(execTime, dstModuleId, srcModuleId, eventId, parameter,
				priority, Config.getStartTime());
	}

	/**
	 * @see core.sync.Lp#createNewEvent(long, int, int, java.lang.Enum,
	 *      java.lang.Object, int, long)
	 */
	@SuppressWarnings("rawtypes")
	@Override
	public void createNewEvent(long execTime, int dstModuleId, int srcModuleId,
			Enum eventId, Object parameter, int priority, long schedTime) {
		assert srcModuleId == dstModuleId : "All initial events have to be internal events.";
		if (execTime <= Config.getEndTime()) {
			Event event = new Event(execTime, schedTime, dstModuleId,
					srcModuleId, eventId, parameter, priority);
			if (!eventsMap_.containsKey(dstModuleId)) {
				eventsMap_.put(dstModuleId, new ArrayList<Event>());
			}
			eventsMap_.get(dstModuleId).add(event);
		}
	}

	/**
	 * @see core.sync.Lp#trace(long, int, java.lang.Enum, java.lang.Object)
	 */
	@SuppressWarnings("rawtypes")
	@Override
	public void trace(long time, int moduleNumber, Enum eventId,
			Object eventParam) {
		assert false;
	}

	/**
	 * @see core.sync.Lp#getAndIncrementPriority()
	 */
	@Override
	public int getAndIncrementPriority() {
		return 0;
	}

	/**
	 * Provides a map of all initial events of all modules.
	 * 
	 * @return a map of all initial events of all modules.
	 */
	public TIntObjectHashMap<List<Event>> getEventsMap() {
		return eventsMap_;
	}
}