/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/sync/ParallelLp.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.sync;

import gnu.trove.TIntObjectHashMap;
import jist.swans.radio.RadioNoise;
import modules.wireless.partition.Partition;
import core.BaseModule;
import core.Config;
import core.Event;
import core.scheduler.EventQueue;
import core.scheduler.EventScheduler;
import core.scheduler.ShiftedHeapQueue;
import core.util.EventLogger;

/**
 * An LP for use with parallel ParallelSimController.
 * 
 * @author Patrick Peschlow
 * @author Martin Geuer
 * 
 * @version 1.0
 */
public class ParallelLp implements Lp {
	/**
	 * The ID of this LP.
	 */
	protected final int lpId_;

	/**
	 * The SimController this LP belongs to.
	 */
	protected final LpBasedSimController simCtrl_;

	/**
	 * The event scheduler of this LP.
	 */
	private final EventScheduler scheduler_;

	/**
	 * An additional event queue for transmit events required in case we use
	 * event lookahead calculation.
	 */
	private final EventQueue transmitQueue_;

	/**
	 * EventLogger for this LP.
	 */
	private final transient EventLogger logger_;

	/**
	 * The timestamp of the last executed event (= scheduling time for new
	 * events).
	 */
	private long clock_ = 0;

	/**
	 * The number of events executed by this LP so far.
	 */
	private long numExecutedEvents_ = 0;

	/**
	 * Key: Module ID. Value: The corresponding module.
	 */
	protected TIntObjectHashMap<BaseModule> moduleMap_;

	/**
	 * A counter for generating event priorities.
	 */
	private int priorityCounter_ = 0;

	/**
	 * The minimal execution time of an event which was sent to another LP in
	 * this window.
	 */
	protected long minSendExecTime_ = Long.MAX_VALUE;

	/**
	 * Constructor.
	 * 
	 * @param lpId
	 *            the LP ID.
	 * @param scheduler
	 *            the scheduler.
	 * @param simCtrl
	 *            the SimController.
	 */
	ParallelLp(int lpId, EventScheduler scheduler, LpBasedSimController simCtrl) {
		lpId_ = lpId;
		scheduler_ = scheduler;
		simCtrl_ = simCtrl;
		logger_ = EventLogger.getInstance(lpId_);
		transmitQueue_ = (Config.getLookahead() == Config.Lookahead.EVENT ? new ShiftedHeapQueue()
				: null);
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "LP_" + lpId_;
	}

	/**
	 * @see core.sync.Lp#setModuleMap(gnu.trove.TIntObjectHashMap)
	 */
	@Override
	public void setModuleMap(TIntObjectHashMap<BaseModule> moduleMap) {
		moduleMap_ = moduleMap;
		assert Partition.class.isAssignableFrom(moduleMap_.get(lpId_)
				.getClass()) : "LP ID is not equal to partition ID. This may lead to unexpected effects!";
	}

	/**
	 * @see core.sync.Lp#createNewEvent(long, int, int, java.lang.Enum,
	 *      java.lang.Object, int)
	 */
	@SuppressWarnings("rawtypes")
	@Override
	public void createNewEvent(long execTime, int dstModuleId, int srcModuleId,
			Enum eventId, Object parameter, int priority) {
		createNewEvent(execTime, dstModuleId, srcModuleId, eventId, parameter,
				priority, clock_);
	}

	/**
	 * @see core.sync.Lp#createNewEvent(long, int, int, java.lang.Enum,
	 *      java.lang.Object, int, long)
	 */
	@SuppressWarnings("rawtypes")
	@Override
	public void createNewEvent(long execTime, int dstModuleId, int srcModuleId,
			Enum eventId, Object parameter, int priority, long schedTime) {
		// Only schedule events smaller or equal to the simulation end time.
		if (execTime > Config.getEndTime()) {
			return;
		}

		final Event event = new Event(execTime, schedTime, dstModuleId,
				srcModuleId, eventId, parameter, priority);

		if (moduleMap_.containsKey(dstModuleId)) {
			receiveEvent(event);
		} else {
			assert eventId == Partition.Events.LPPROPAGATION;
			minSendExecTime_ = Math.min(minSendExecTime_, event.getExecTime());
			simCtrl_.sendEvent(event, lpId_);
		}
	}

	/**
	 * @see core.sync.Lp#trace(long, int, java.lang.Enum, java.lang.Object)
	 */
	@SuppressWarnings("rawtypes")
	@Override
	public void trace(long time, int moduleNumber, Enum eventId,
			Object eventParam) {
		assert logger_ != null : "Logger is null!";
		logger_.trace(time, moduleNumber, eventId, eventParam);
	}

	/**
	 * @see core.sync.Lp#getAndIncrementPriority()
	 */
	@Override
	public int getAndIncrementPriority() {
		return priorityCounter_++;
	}

	/**
	 * Provides the ID of this LP.
	 * 
	 * @return the LP ID.
	 */
	protected int getId() {
		return lpId_;
	}

	/**
	 * Provides the scheduler of this LP.
	 * 
	 * @return the scheduler.
	 */
	protected EventScheduler getScheduler() {
		return scheduler_;
	}

	/**
	 * Provides the number of executed events of this LP.
	 * 
	 * @return the number of executed events.
	 */
	protected long getNumExecutedEvents() {
		return numExecutedEvents_;
	}

	/**
	 * Provides the NET (timestamp of the next event) currently scheduled at
	 * this LP.
	 * 
	 * @return the timestamp of the next event.
	 */
	protected long getNextEventTimestamp() {
		return scheduler_.getNextEventTime();
	}

	/**
	 * Immediately schedules an event. This method may only be called when the
	 * event has already reached its destination LP for sure.
	 * 
	 * @param event
	 *            the event to schedule at this LP.
	 */
	protected void receiveEvent(Event event) {
		scheduler_.add(event);
		if (Config.getLookahead() == Config.Lookahead.EVENT) {
			if (event.getId() == RadioNoise.Events.TRANSMIT) {
				transmitQueue_.insert(event);
			}
		}
	}

	/**
	 * Allows the LP to run until the specified simulation time value.
	 * 
	 * @param time
	 *            the time value.
	 */
	protected void run(long time) {
		minSendExecTime_ = Long.MAX_VALUE;
		while (getNextEventTimestamp() < time) {
			executeEvent(scheduler_.removeNextEvent());
		}
	}

	/**
	 * Tests whether this LP has a safe event to execute.
	 * 
	 * @param time
	 *            the safe time up to which events can be executed.
	 * @return <code>true</code> whether the LP has an event to execute,
	 *         otherwise <code>false</code>.
	 */
	boolean hasSafeEvents(long time) {
		return (scheduler_.getNextEventTime() < time);
	}

	/**
	 * Returns the minimal execution time of an event sent to another LP during
	 * the last window.
	 * 
	 * @return the <code>minSendExecTime_</code>.
	 */
	long getMinSendExecTime() {
		return minSendExecTime_;
	}

	/**
	 * Provides the next known timestamp of a transmit event.
	 * 
	 * @return the next known timestamp of a transmit event.
	 */
	long getNextTransmitEventTimestamp() {
		return transmitQueue_.isEmpty() ? Long.MAX_VALUE : transmitQueue_
				.peekFirst().getExecTime();
	}

	/**
	 * Executes an event.
	 * 
	 * @param event
	 *            the event to execute.
	 */
	private void executeEvent(Event event) {
		assert event != null;
		if (Config.getLookahead() == Config.Lookahead.EVENT) {
			if (event.getId() == RadioNoise.Events.TRANSMIT) {
				final Event transmit = transmitQueue_.removeFirst();
				assert transmit.getExecTime() == event.getExecTime();
			}
		}

		assert event.getExecTime() >= clock_ : "execTime: "
				+ event.getExecTime() + " >= clock_: " + clock_ + " diff: "
				+ (event.getExecTime() - clock_);
		clock_ = event.getExecTime();

		moduleMap_.get(event.getDstModuleId()).execute(event.getExecTime(),
				event.getSrcModuleId(), event.getId(), event.getParameter());
		numExecutedEvents_++;
	}
}
