/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/sync/TPCBarrierSimController.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.sync;

import edu.bonn.cs.net.jbarrier.barrier.Barrier;
import edu.bonn.cs.net.jbarrier.barrier.ButterflyBarrier;
import edu.bonn.cs.net.jbarrier.barrier.DisseminationBarrier;
import edu.bonn.cs.net.jbarrier.barrier.GenericReductor;
import edu.bonn.cs.net.jbarrier.barrier.TournamentBarrier;
import gnu.trove.TIntObjectHashMap;
import gnu.trove.TLongArrayList;

import java.io.BufferedReader;
import java.io.FileReader;
import java.util.List;

import jist.swans.Constants;
import jist.swans.mac.Mac802_11;
import core.BaseModule;
import core.Config;
import core.Config.BarrierType;
import core.Config.Lookahead;
import core.Event;
import core.ModuleRoot;
import core.scheduler.EventScheduler;
import core.util.CustomLogger;

/**
 * Implements a conservative window-based synchronization mechanism. This class
 * has been designed for parallel wireless network simulations using HiPWiNS,
 * but is should be easily adaptable to other types of simulation scenarios.
 * 
 * @author Patrick Peschlow
 * @author Martin Geuer
 * @author Andreas Voss
 * 
 * @version 1.0
 */
public class ParallelSimController extends AbstractSimController implements
		LpBasedSimController {
	/**
	 * Key: LP ID. Value: The corresponding LP.
	 */
	private final ParallelLp[] lpIdToLp_;

	/**
	 * Key: Module ID. Value: The corresponding LP ID.
	 */
	private final int[] moduleIdToLpId_;

	/**
	 * The number of LPs.
	 */
	private final int numLps_;

	/**
	 * The number of threads used.
	 */
	private final int numThreads_;

	/**
	 * The barrier action.
	 */
	private final Runnable barrierAction_;

	/**
	 * An array holding all threads.
	 */
	private final ParallelThread[] threads_;

	/**
	 * The safe time bound for the current window (calculated in each barrier).
	 */
	private long safeTimeBound_ = 0;

	/**
	 * Counts the number of synchronization points.
	 */
	private long numSyncs_ = 0;

	/**
	 * Used for storing all timestamps of transmit events (for ISP).
	 */
	private long[] transmitTimes_;

	/**
	 * Used for running ISP simulations.
	 */
	private int ispCurrentIndex_ = 0;

	/**
	 * Constructor.
	 */
	public ParallelSimController() {
		// Create LPs and assign modules to LPs.
		final List<BaseModule> modules = ModuleRoot.getRoot().getModuleList();
		final int numModules = modules.size();
		numLps_ = Config.getNumPartitions();
		lpIdToLp_ = new ParallelLp[numLps_];
		moduleIdToLpId_ = new int[numModules];
		for (int lpId = 0; lpId < numLps_; lpId++) {
			final List<BaseModule> thisLpModules = ModuleRoot.getRoot()
					.getModulesOfLp(lpId);
			EventScheduler eventScheduler = new EventScheduler(Config
					.getScheduler().getEventQueueClass());
			ParallelLp lp = new ParallelLp(lpId, eventScheduler, this);
			TIntObjectHashMap<BaseModule> thisLpModuleMap = new TIntObjectHashMap<BaseModule>(
					thisLpModules.size());
			for (BaseModule module : thisLpModules) {
				module.setLp(lp);
				final int moduleId = module.getId();
				thisLpModuleMap.put(moduleId, module);
				moduleIdToLpId_[moduleId] = lpId;
			}
			lp.setModuleMap(thisLpModuleMap);
			assert lpIdToLp_[lpId] == null;
			lpIdToLp_[lpId] = lp;
		}
		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("LpBasedSimController: Created " + numLps_
					+ " LPs.");
		}
		// Schedule initial events.
		InitLp initLp = ModuleRoot.getInitLp();
		TIntObjectHashMap<List<Event>> eventsMap = initLp.getEventsMap();
		for (BaseModule module : modules) {
			final int moduleId = module.getId();
			if (eventsMap.containsKey(moduleId)) {
				for (Event event : eventsMap.get(moduleId)) {
					final ParallelLp lp = getLp(getLpId(moduleId));
					lp.getScheduler().add(event);
				}
			}
		}
		// Set up threads and barrier.
		numThreads_ = Config.getNumPartitions();
		threads_ = new ParallelThread[numThreads_];
		switch (Config.getLookahead()) {
		case EVENT:
			if (Config.getBarrier() == BarrierType.TOURNAMENT) {
				barrierAction_ = new TournamentEventBarrierAction();
			} else {
				barrierAction_ = new EventLookaheadBarrierAction();
			}
			break;
		case PROTOCOL:
			barrierAction_ = new ProtocolLookaheadBarrierAction();
			break;
		case ISP:
			barrierAction_ = new ISPLookaheadBarrierAction();
			break;
		case NONE:
			barrierAction_ = new NoLookaheadBarrierAction();
			break;
		default:
			throw new RuntimeException("Bad barrier action type!");
		}

		switch (Config.getBarrier()) {
		case BUTTERFLY:
			assert (numThreads_ & (numThreads_ - 1)) == 0;
		case DISSEMINATION:
		case TOURNAMENT:
		case SINGLE_PARTITION:
			break;
		default:
			throw new RuntimeException("Bad barrier type!");
		}
	}

	/**
	 * Provides the number of synchronization points (for statistics at the end
	 * of the simulation).
	 * 
	 * @return the number of synchronization points.
	 */
	public long getNumSyncs() {
		return numSyncs_;
	}

	/**
	 * Provides the LP ID to the specified module ID.
	 * 
	 * @param moduleId
	 *            the module ID.
	 * @return the ID of the LP that manages the module.
	 */
	public int getLpId(int moduleId) {
		return moduleIdToLpId_[moduleId];
	}

	/**
	 * @see core.sync.LpBasedSimController#sendEvent(core.Event, int)
	 */
	@Override
	public void sendEvent(Event event, int srcLpId) {
		threads_[getLpId(event.getDstModuleId())].receiveEvent(event, srcLpId);
	}

	/**
	 * @see core.sync.LpBasedSimController#getSafeTimeBound()
	 */
	@Override
	public long getSafeTimeBound() {
		return safeTimeBound_;
	}

	/**
	 * @see core.sync.AbstractSimController#prepareSim()
	 */
	@Override
	protected void prepareSim() {
		// Create the barrier object.
		Barrier barrier;
		switch (Config.getBarrier()) {
		case DISSEMINATION:
			barrier = new DisseminationBarrier(numThreads_, barrierAction_,
					new TwofoldReduction());
			break;
		case BUTTERFLY:
			barrier = new ButterflyBarrier(numThreads_, barrierAction_,
					new TwofoldReduction());
			break;
		case TOURNAMENT:
			if (Config.getLookahead() == Lookahead.EVENT) {
				barrier = new TournamentBarrier(numThreads_, barrierAction_,
						new TournamentEventReduction());
			} else {
				barrier = new TournamentBarrier(numThreads_, barrierAction_,
						new TwofoldReduction());
			}
			break;
		case SINGLE_PARTITION:
			barrier = new SinglePartitionBarrier(barrierAction_);
			break;
		default:
			throw new RuntimeException("Unknown barrier algorithm type!");
		}

		// Assign the LPs to threads and create the threads.
		assert numLps_ == numThreads_;
		for (int threadId = 0; threadId < numThreads_; threadId++) {
			// In parallel simulation, LP ID is equal to thread ID.
			final int lpId = threadId;
			final ParallelLp lp = getLp(lpId);
			if (Config.getBarrier() == BarrierType.TOURNAMENT
					&& Config.getLookahead() == Lookahead.EVENT) {
				threads_[threadId] = new TournamentEventBarrierThread(this,
						numThreads_, threadId, lp, barrier);
			} else {
				threads_[threadId] = new ParallelThread(this, numThreads_,
						threadId, lp, barrier);
			}
		}

		// Do we use the Ideal Simulation Protocol?
		if (Config.getLookahead() == Lookahead.ISP) {
			readDataISP();
		}
	}

	/**
	 * @see core.sync.AbstractSimController#runSim()
	 */
	@Override
	protected void runSim() {
		// Start simulation threads.
		for (ParallelThread thread : threads_) {
			thread.start();
		}
		// Wait until done.
		for (ParallelThread thread : threads_) {
			try {
				thread.join();
			} catch (InterruptedException e) {
				if (CustomLogger.isWarningEnabled()) {
					CustomLogger.warning(
							"Thread has been interrupted during execution!", e);
				}
			}
		}
	}

	/**
	 * @see core.sync.AbstractSimController#finishSim()
	 */
	@Override
	protected void finishSim() {
		if (CustomLogger.isInfoEnabled()) {
			long numExecutedEvents = 0;
			for (int lpId = 0; lpId < numLps_; lpId++) {
				numExecutedEvents += getLp(lpId).getNumExecutedEvents();
			}
			for (int lpId = 0; lpId < numLps_; lpId++) {
				final ParallelLp lp = getLp(lpId);
				CustomLogger.info("LP "
						+ lp.getId()
						+ " executed events: "
						+ lp.getNumExecutedEvents()
						+ " ("
						+ Math.round((double) lp.getNumExecutedEvents()
								/ (double) numExecutedEvents * 100.0) + "%)");
			}
			CustomLogger
					.info("TPCBarrierSimController - Total number of executed events: "
							+ numExecutedEvents);
		}
	}

	/**
	 * Provides the LP to a specified LP ID.
	 * 
	 * @param lpId
	 *            the LP ID.
	 * @return the corresponding LP.
	 */
	protected ParallelLp getLp(int lpId) {
		return lpIdToLp_[lpId];
	}

	/**
	 * Provides the thread to a specified thread ID.
	 * 
	 * @param threadId
	 *            the thread ID.
	 * @return the corresponding thread.
	 */
	protected ParallelThread getThread(int threadId) {
		return threads_[threadId];
	}

	/**
	 * Shuts down all threads.
	 */
	private void shutdown() {
		for (int i = 0; i < threads_.length; i++) {
			threads_[i].shutdown();
		}
	}

	/**
	 * Read in transmit event timestamp data from a file (for ISP runs only).
	 */
	private void readDataISP() {
		String line;
		try {
			BufferedReader br = new BufferedReader(new FileReader(
					"transmit_times"));
			line = br.readLine();
			br.close();
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		String[] transmitTimesStrings = line.split(" ");
		final int numTimes = transmitTimesStrings.length;
		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("Read in " + numTimes + " transmit times.");
		}
		TLongArrayList times = new TLongArrayList(numTimes + 1);
		long last = -1;
		for (int i = 0; i < numTimes; i++) {
			final long current = Long.parseLong(transmitTimesStrings[i]);
			if (current != last) {
				times.add(current);
				last = current;
			}
		}
		final int numUniqueTimes = times.size();
		times.add(Long.MAX_VALUE - Constants.NANO_SECOND);
		transmitTimes_ = times.toNativeArray();
		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("After removing duplicates: " + numUniqueTimes
					+ " transmit times.");
		}
	}

	/**
	 * Barrier action in case of event lookahead.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private class EventLookaheadBarrierAction implements Runnable {
		/**
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			++numSyncs_;
			final long net = threads_[0].getNET();
			if (net == Long.MAX_VALUE) {
				shutdown();
				return;
			}
			final long ntet = threads_[0].getNTET();
			safeTimeBound_ = Math.min(ntet, net + Mac802_11.RX_TX_TURNAROUND)
					+ Config.getPropagation().getDelay();
		}
	}

	/**
	 * Barrier action in case of protocol lookahead.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private class ProtocolLookaheadBarrierAction implements Runnable {
		/**
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			++numSyncs_;
			final long net = threads_[0].getNET();
			if (net == Long.MAX_VALUE) {
				shutdown();
				return;
			}
			final long nsp = threads_[0].getNTET();
			safeTimeBound_ = Math.max(nsp, net)
					+ Config.getPropagation().getDelay();
		}
	}

	/**
	 * Barrier action in case of ISP.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private class ISPLookaheadBarrierAction implements Runnable {
		/**
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			++numSyncs_;
			final long net = threads_[0].getNET();
			if (net == Long.MAX_VALUE) {
				shutdown();
				return;
			}
			safeTimeBound_ = transmitTimes_[ispCurrentIndex_++]
					+ Constants.NANO_SECOND;
		}
	}

	/**
	 * Barrier action in case of no advanced lookahead usage.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private class NoLookaheadBarrierAction implements Runnable {
		/**
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			++numSyncs_;
			final long net = threads_[0].getNET();
			if (net == Long.MAX_VALUE) {
				shutdown();
				return;
			}
			safeTimeBound_ = net + Config.getPropagation().getDelay();
		}
	}

	/**
	 * Optimized barrier method for tournament barriers with event lookahead.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private class TournamentEventBarrierAction implements Runnable {
		/**
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			final long globalMin = ((TournamentEventBarrierThread) threads_[0])
					.getTourHelperMinValue();
			++numSyncs_;
			if (globalMin == Long.MAX_VALUE) {
				shutdown();
				return;
			}
			safeTimeBound_ = globalMin + Config.getPropagation().getDelay();

		}
	}

	/**
	 * Two-fold reduction of NET and NTET values.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	class TwofoldReduction implements GenericReductor {
		/**
		 * @see edu.bonn.cs.net.jbarrier.barrier.GenericReductor#reduce(int,
		 *      int)
		 */
		@Override
		public void reduce(int threadId1, int threadId2) {
			final ParallelThread bth1 = threads_[threadId1];
			final ParallelThread bth2 = threads_[threadId2];
			final long net = bth2.getNET();
			final long ntet = bth2.getNTET();
			if (bth1.getNET() > net) {
				bth1.setNET(net);
			}
			if (bth1.getNTET() > ntet) {
				bth1.setNTET(ntet);
			}
		}
	}

	/**
	 * Specialized reduction for simulations with tournament barrier and event
	 * lookahead.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	class TournamentEventReduction implements GenericReductor {
		/**
		 * @see edu.bonn.cs.net.jbarrier.barrier.GenericReductor#reduce(int,
		 *      int)
		 */
		@Override
		public void reduce(int threadId1, int threadId2) {
			final TournamentEventBarrierThread bth1 = (TournamentEventBarrierThread) threads_[threadId1];
			final TournamentEventBarrierThread bth2 = (TournamentEventBarrierThread) threads_[threadId2];
			final long helper = bth2.getTourHelperMinValue();
			if (bth1.getTourHelperMinValue() > helper) {
				bth1.setTourHelperMinValue(helper);
			}
		}
	}
}
