/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/sync/TPCBarrierThread.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.sync;

import jist.swans.Constants;
import jist.swans.mac.Mac802_11;
import modules.wireless.partition.Partition;
import core.Config;
import core.Config.Lookahead;
import core.Event;
import edu.bonn.cs.net.jbarrier.barrier.Barrier;

/**
 * A thread executing an LP in a parallel simulation.
 * 
 * @author Patrick Peschlow
 * @author Martin Geuer
 * 
 * @version 1.0
 */
class ParallelThread extends Thread {
	/**
	 * The SimController.
	 */
	protected final LpBasedSimController simCtrl_;

	/**
	 * The total number of threads.
	 */
	protected final int numThreads_;

	/**
	 * The unique ID of this thread.
	 */
	protected final int threadId_;

	/**
	 * The barrier object used.
	 */
	protected final Barrier barrier_;

	/**
	 * The LP/partition of this thread.
	 */
	protected final ParallelLp lp_;

	/**
	 * Channels for incoming events from the other threads.
	 */
	protected final ConcurrentEventArray[] incomingChannels_;

	/**
	 * The local next event time (NET) estimate of this thread.
	 */
	protected long net_;

	/**
	 * The local next transmit event time (NTET) estimate of this thread.
	 */
	protected long ntet_;

	/**
	 * A flag which indicates whether the simulation is over. Does NOT have to
	 * be declared volatile because it may only be set during the barrier action
	 * (and updates are thus enforced by synchronization piggybacking via the
	 * signaling flags of the threads within the barrier).
	 */
	protected boolean shutdown_ = false;

	/**
	 * Constructor.
	 * 
	 * @param simCtrl
	 *            the SimController.
	 * @param numThreads
	 *            the number of threads.
	 * 
	 * @param threadId
	 *            the unique id of this thread.
	 * @param lp
	 *            the LP run by this thread.
	 * @param barrier
	 *            the barrier object used by this thread.
	 */
	protected ParallelThread(LpBasedSimController simCtrl, int numThreads,
			int threadId, ParallelLp lp, Barrier barrier) {
		super("Threadpool-Worker-" + threadId);
		simCtrl_ = simCtrl;
		numThreads_ = numThreads;
		threadId_ = threadId;
		lp_ = lp;
		barrier_ = barrier;
		// Channel initialization.
		incomingChannels_ = new ConcurrentEventArray[numThreads];
		for (int i = 0; i < incomingChannels_.length; ++i) {
			if (i != threadId_) {
				incomingChannels_[i] = new ConcurrentEventArray();
			}
		}
	}

	/**
	 * @see Thread#run()
	 */
	@Override
	public void run() {
		while (!shutdown_) {
			startScheduling();
			startWorking();
			barrier_.await(threadId_);
		}
	}

	/**
	 * Shuts down this thread.
	 */
	protected final void shutdown() {
		shutdown_ = true;
	}

	/**
	 * Receives an incoming event, which will be scheduled at the beginning of
	 * the next window.
	 * 
	 * @param event
	 *            the event.
	 * @param threadId
	 *            the ID of the thread that sends the event.
	 */
	protected void receiveEvent(Event event, int threadId) {
		assert threadId != threadId_;
		incomingChannels_[threadId].add(event);
	}

	/**
	 * Provides the current next event time (NET) estimate of this thread.
	 * 
	 * @return the next event time.
	 */
	protected final long getNET() {
		return net_;
	}

	/**
	 * Sets the current next event time (NET).
	 * 
	 * @param net
	 *            the next NET.
	 */
	protected final void setNET(long net) {
		net_ = net;
	}

	/**
	 * Provides the current next transmit event time (NTET) estimate of this
	 * thread.
	 * 
	 * @return the next transmit event time.
	 */
	protected final long getNTET() {
		return ntet_;
	}

	/**
	 * Sets the current next transmit event time (NTET).
	 * 
	 * @param ntet
	 *            the next NTET.
	 */
	protected final void setNTET(long ntet) {
		ntet_ = ntet;
	}

	/**
	 * Schedules all events that were received during the last window.
	 */
	protected void startScheduling() {
		for (int i = 0; i < incomingChannels_.length; ++i) {
			if (i != threadId_) {
				while (!incomingChannels_[i].isEmpty()) {
					final Event event = incomingChannels_[i].poll();
					lp_.receiveEvent(event);
				}
			}
		}
	}

	/**
	 * Executes the LP up to (but not including) the safeTimeBound. In the
	 * process, all required information for the next barrier is computed.
	 */
	protected void startWorking() {
		final long safeTimeBound = simCtrl_.getSafeTimeBound();
		final ParallelLp lp = lp_;
		if (lp.hasSafeEvents(safeTimeBound)) {
			// lp.run() immediately sets minSendExecTime to Long.MAX_VALUE.
			lp.run(safeTimeBound);
			net_ = Math
					.min(lp.getMinSendExecTime(), lp.getNextEventTimestamp());
		} else {
			// Important because the NET could have become smaller in the last
			// window.
			net_ = lp.getNextEventTimestamp();
		}
		// Estimate next safe time window for this LP. Assumptions:
		// - Every LP contains exactly one partition.
		// - No zero-delay events scheduled between LPs.
		if (safeTimeBound <= Config.getEndTime()) {
			if (Config.getLookahead() == Lookahead.PROTOCOL) {
				ntet_ = Partition.getPartition(lp.getId())
						.calculateNextSyncPoint(
								safeTimeBound - Constants.NANO_SECOND);
			} else if (Config.getLookahead() == Lookahead.EVENT) {
				ntet_ = lp.getNextTransmitEventTimestamp();
			}
		}
	}
}

/**
 * Optimized version of the thread class for simulations with tournament barrier
 * and event lookahead.
 * 
 * @author Patrick Peschlow
 * @author Andreas Voss
 * 
 * @version 1.0
 */
class TournamentEventBarrierThread extends ParallelThread {
	/**
	 * Helper value for the optimized tournament barrier with event lookahead.
	 */
	private long tourHelperMinValue_;

	/**
	 * Constructor.
	 * 
	 * @param simCtrl
	 *            the SimController.
	 * @param numThreads
	 *            the number of threads.
	 * 
	 * @param threadId
	 *            the unique id of this thread.
	 * @param lp
	 *            the LP run by this thread.
	 * @param barrier
	 *            the barrier object used by this thread.
	 */
	protected TournamentEventBarrierThread(LpBasedSimController simCtrl,
			int numThreads, int threadId, ParallelLp lp, Barrier barrier) {
		super(simCtrl, numThreads, threadId, lp, barrier);
	}

	/**
	 * @see Thread#run()
	 */
	@Override
	public final void run() {
		while (!shutdown_) {
			startScheduling();
			startWorking();
			// Perform local minimum calculation already here.
			tourHelperMinValue_ = (net_ == Long.MAX_VALUE) ? ntet_ : Math.min(
					ntet_, net_ + Mac802_11.RX_TX_TURNAROUND);
			barrier_.await(threadId_);
		}
	}

	/**
	 * Provides the helper value used in the reduction.
	 * 
	 * @return the tourHelperMinValue
	 */
	public long getTourHelperMinValue() {
		return tourHelperMinValue_;
	}

	/**
	 * Sets the helper value used in the reduction.
	 * 
	 * @param tourHelperMinValue
	 *            the new value.
	 */
	public void setTourHelperMinValue(long tourHelperMinValue) {
		tourHelperMinValue_ = tourHelperMinValue;
	}
}
