/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/sync/SeqLp.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.sync;

import gnu.trove.TIntObjectHashMap;
import core.BaseModule;
import core.Config;
import core.Event;
import core.scheduler.EventScheduler;
import core.util.EventLogger;

/**
 * An LP implementation used with <code>SeqSimController</code>. There is only a
 * single LP which contains all modules of the simulation.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class SeqLp implements Lp {
	/**
	 * The number of events executed by this LP.
	 */
	protected long numExecutedEvents_ = 0;

	/**
	 * The event scheduler of this LP.
	 */
	protected EventScheduler scheduler_;

	/**
	 * The exec time of the last executed event (= sched time for new events).
	 */
	protected long clock_ = 0;

	/**
	 * Key: module ID. Value: the corresponding module.
	 */
	protected TIntObjectHashMap<BaseModule> moduleMap_;

	/**
	 * A counter for generating event priorities.
	 */
	protected int priorityCounter_ = 0;

	/**
	 * EventLogger for this LP
	 */
	protected EventLogger logger_;

	/**
	 * Constructor.
	 * 
	 * @param eventScheduler
	 *            the scheduler for this LP.
	 */
	public SeqLp(EventScheduler eventScheduler) {
		scheduler_ = eventScheduler;
		logger_ = EventLogger.getInstance(SeqSimController.LPID);
	}

	/**
	 * @see core.sync.Lp#createNewEvent(long, int, int, java.lang.Enum,
	 *      java.lang.Object, int)
	 */
	@SuppressWarnings("rawtypes")
	@Override
	public void createNewEvent(long execTime, int dstModuleId, int srcModuleId,
			Enum eventId, Object parameter, int priority) {
		createNewEvent(execTime, dstModuleId, srcModuleId, eventId, parameter,
				priority, clock_);
	}

	/**
	 * @see core.sync.Lp#createNewEvent(long, int, int, java.lang.Enum,
	 *      java.lang.Object, int, long)
	 */
	@SuppressWarnings("rawtypes")
	@Override
	public void createNewEvent(long execTime, int dstModuleId, int srcModuleId,
			Enum eventId, Object parameter, int priority, long schedTime) {
		if (execTime <= Config.getEndTime()) {
			Event event = new Event(execTime, schedTime, dstModuleId,
					srcModuleId, eventId, parameter, priority);
			scheduler_.add(event);
		}
	}

	/**
	 * @see core.sync.Lp#trace(long, int, java.lang.Enum, java.lang.Object)
	 */
	@SuppressWarnings("rawtypes")
	@Override
	public void trace(long time, int nodeId, Enum eventId, Object eventParam) {
		assert logger_ != null : "Logger is null.";
		logger_.trace(time, nodeId, eventId, eventParam);
	}

	/**
	 * @see core.sync.Lp#setModuleMap(gnu.trove.TIntObjectHashMap)
	 */
	@Override
	public void setModuleMap(TIntObjectHashMap<BaseModule> moduleMap) {
		moduleMap_ = moduleMap;
	}

	/**
	 * @see core.sync.Lp#getAndIncrementPriority()
	 */
	@Override
	public int getAndIncrementPriority() {
		return priorityCounter_++;
	}

	/**
	 * Defines what the LP does when it runs.
	 * 
	 * @param endTime
	 *            the end time of the simulation.
	 */
	void runUntil(long endTime) {
		while (scheduler_.getNextEventTime() <= endTime) {
			executeEvent(scheduler_.removeNextEvent());
		}
	}

	/**
	 * Provides the total number of events the LP executed so far.
	 * 
	 * @return the total number of events the LP executed so far.
	 */
	long getNumExecutedEvents() {
		return numExecutedEvents_;
	}

	/**
	 * Executes an event.
	 * 
	 * @param event
	 *            the event to execute.
	 */
	private void executeEvent(Event event) {
		assert event != null;
		assert event.getExecTime() >= clock_;
		clock_ = event.getExecTime();
		final BaseModule handler = moduleMap_.get(event.getDstModuleId());
		handler.execute(event.getExecTime(), event.getSrcModuleId(),
				event.getId(), event.getParameter());
		numExecutedEvents_++;
	}

}
