/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: core/sync/SeqSimController.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.sync;

import gnu.trove.TIntObjectHashMap;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.List;

import core.BaseModule;
import core.Config;
import core.Event;
import core.ModuleRoot;
import core.scheduler.EventScheduler;
import core.util.CustomLogger;

/**
 * A SimController for sequential simulation. It uses a single LP which contains
 * all simulation modules and runs until the simulation is finished.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class SeqSimController extends AbstractSimController {
	/**
	 * The ID of the single LP of the simulation (which is 0).
	 */
	static final int LPID = 0;

	/**
	 * Collects transmit times (for parallel simulation using ISP).
	 */
	private static StringBuffer transmitEventTimes;

	/**
	 * Registers a transmit time (for ISP).
	 * 
	 * @param time
	 *            the transmit time.
	 */
	public static void registerTransmitTime(long time) {
		assert Config.isCollectingISP();
		transmitEventTimes.append(time + " ");
	}

	/**
	 * The single LP that contains the whole simulation scenario.
	 */
	protected final SeqLp lp_;

	/**
	 * Constructor.
	 */
	public SeqSimController() {
		EventScheduler eventScheduler = new EventScheduler(Config
				.getScheduler().getEventQueueClass());
		// Create single LP.
		lp_ = new SeqLp(eventScheduler);
		// Add events to the scheduler.
		InitLp initLp = ModuleRoot.getInitLp();
		TIntObjectHashMap<List<Event>> eventsMap = initLp.getEventsMap();
		final List<BaseModule> modules = ModuleRoot.getRoot().getModuleList();
		final int numModules = modules.size();
		TIntObjectHashMap<BaseModule> moduleMap = new TIntObjectHashMap<BaseModule>(
				numModules);
		for (BaseModule module : modules) {
			module.setLp(lp_);
			final int moduleId = module.getId();
			if (eventsMap.containsKey(moduleId)) {
				for (Event event : eventsMap.get(moduleId)) {
					eventScheduler.add(event);
				}
			}
			moduleMap.put(moduleId, module);
		}
		lp_.setModuleMap(moduleMap);
	}

	/**
	 * @see core.sync.AbstractSimController#prepareSim()
	 */
	protected void prepareSim() {
		if (Config.isCollectingISP()) {
			transmitEventTimes = new StringBuffer(1000000);
		}
	}

	/**
	 * @see core.sync.AbstractSimController#runSim()
	 */
	protected void runSim() {
		assert lp_ != null;
		lp_.runUntil(Config.getEndTime());
	}

	/**
	 * @see core.sync.AbstractSimController#finishSim()
	 */
	protected void finishSim() {
		final long executedEvents = lp_.getNumExecutedEvents();
		if (CustomLogger.isInfoEnabled()) {
			CustomLogger
					.info("SeqSimController: Total number of executed events: "
							+ executedEvents);
		}
		if (Config.isCollectingISP()) {
			try {
				BufferedWriter buffer = new BufferedWriter(new FileWriter(
						"transmit_times"));
				buffer.write(transmitEventTimes.toString());
				buffer.flush();
				buffer.close();
			} catch (IOException e) {
				throw new RuntimeException("Transmit times file error!!!");
			}
		}
	}
}
