/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: util/CustomLogger.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.util;

/**
 * General logger for the simulator.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class CustomLogger {
	/**
	 * Debug levels. Convention: debug levels are listed in increasing order.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	public static enum Level {
		/**
		 * Everything is logged.
		 */
		ALL,
		/**
		 * Debugging output is logged.
		 */
		DEBUG,
		/**
		 * Informational output is logged.
		 */
		INFO,
		/**
		 * Warning output is logged.
		 */
		WARNING,
		/**
		 * Fatal errors are logged.
		 */
		FATAL;
	}

	/**
	 * Prefix for debugging output.
	 */
	private static final String DEBUG_PREFIX = Level.DEBUG.name() + " - ";

	/**
	 * Prefix for informational output.
	 */
	private static final String INFO_PREFIX = Level.INFO.name() + " - ";

	/**
	 * Prefix for warning output.
	 */
	private static final String WARNING_PREFIX = Level.WARNING.name() + " - ";

	/**
	 * Prefix for fatal error output.
	 */
	private static final String FATAL_PREFIX = Level.FATAL.name() + " - ";

	/**
	 * Ordinal that defines the logging level used.
	 */
	private static int myLevelOrdinal;

	/**
	 * Prints a message.
	 * 
	 * @param s
	 *            the message.
	 */
	private static void write(String s) {
		System.out.println(s);
	}

	/**
	 * Whether the output for the specified level is enabled.
	 * 
	 * @param level
	 *            the logging level.
	 * @return whether the output for the specified level is enabled.
	 */
	public static boolean isEnabledFor(Level level) {
		return level.ordinal() >= myLevelOrdinal;
	}

	/**
	 * Sets the logging level.
	 * 
	 * @param level
	 *            the logging level to be used.
	 */
	public static void setLevel(Level level) {
		myLevelOrdinal = level.ordinal();
	}

	/**
	 * Prints a debug message.
	 * 
	 * @param message
	 *            the message.
	 */
	public static void debug(String message) {
		write(DEBUG_PREFIX + message);
	}

	/**
	 * Prints a debug message and an exception.
	 * 
	 * @param message
	 *            the message.
	 * @param e
	 *            the exception.
	 */
	public static void debug(String message, Exception e) {
		write(DEBUG_PREFIX + message);
		e.printStackTrace();
	}

	/**
	 * Prints an info message.
	 * 
	 * @param message
	 *            the message.
	 */
	public static void info(String message) {
		write(INFO_PREFIX + message);
	}

	/**
	 * Prints an info message and an exception.
	 * 
	 * @param message
	 *            the message.
	 * @param e
	 *            the exception.
	 */
	public static void info(String message, Exception e) {
		write(INFO_PREFIX + message);
		e.printStackTrace();
	}

	/**
	 * Prints a warning message.
	 * 
	 * @param message
	 *            the message.
	 */
	public static void warning(String message) {
		write(WARNING_PREFIX + message);
	}

	/**
	 * Prints a warning message and an exception.
	 * 
	 * @param message
	 *            the message.
	 * @param e
	 *            the exception.
	 */
	public static void warning(String message, Exception e) {
		write(WARNING_PREFIX + message);
		e.printStackTrace();
	}

	/**
	 * Prints a fatal error message. Also exits the simulation.
	 * 
	 * @param message
	 *            the message.
	 */
	public static void fatal(String message) {
		write(FATAL_PREFIX + message);
		System.exit(-1);
	}

	/**
	 * Prints a fatal error message and an exception. Also exits the
	 * application.
	 * 
	 * @param message
	 *            the message.
	 * @param e
	 *            the exception.
	 */
	public static void fatal(String message, Exception e) {
		write(FATAL_PREFIX + message);
		e.printStackTrace();
		System.exit(-1);
	}

	/**
	 * Whether debugging output is enabled.
	 * 
	 * @return whether debugging output is enabled.
	 */
	public static boolean isDebugEnabled() {
		return isEnabledFor(Level.DEBUG);
	}

	/**
	 * Whether informational output is enabled.
	 * 
	 * @return whether informational output is enabled.
	 */
	public static boolean isInfoEnabled() {
		return isEnabledFor(Level.INFO);
	}

	/**
	 * Whether warning output is enabled.
	 * 
	 * @return whether warning output is enabled.
	 */
	public static boolean isWarningEnabled() {
		return isEnabledFor(Level.WARNING);
	}

	/**
	 * Whether fatal error output is enabled.
	 * 
	 * @return whether fatal error output is enabled.
	 */
	public static boolean isFatalEnabled() {
		return isEnabledFor(Level.FATAL);
	}
}