/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: util/EventLogger.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package core.util;

import gnu.trove.TIntObjectHashMap;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;

/**
 * Event trace logger for the simulator.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class EventLogger {
	/**
	 * Supported time units for the simulation and event logging.
	 * 
	 * CONVENTION: time units are listed in increasing order.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	public static enum TimeUnit {
		/**
		 * Nanoseconds.
		 */
		NANOSECONDS,
		/**
		 * Microseconds.
		 */
		MICROSECONDS,
		/**
		 * Milliseconds.
		 */
		MILLISECONDS,
		/**
		 * Seconds.
		 */
		SECONDS,
		/**
		 * Minutes.
		 */
		MINUTES,
		/**
		 * Hours.
		 */
		HOURS,
		/**
		 * Days.
		 */
		DAYS;
	};

	/**
	 * Scale factors between the different units supported (from days to
	 * nanoseconds).
	 */
	private static int unitScales[] = { 1000, 1000, 1000, 60, 60, 24, 1 };

	/**
	 * The simulation time unit.
	 */
	private static TimeUnit simTimeUnit = TimeUnit.MILLISECONDS;

	/**
	 * The logging time unit.
	 */
	private static TimeUnit logTimeUnit = TimeUnit.MILLISECONDS;

	/**
	 * Factor to convert simulation time units into the resolution chosen for
	 * trace file output.
	 */
	private static double conversionFactor = 1;

	/**
	 * Whether event logging is enables.
	 */
	private static boolean isEnabled = false;

	/**
	 * Base trace file name (without the LP ID extension).
	 */
	private static String baseFilename = null;

	/**
	 * List of loggers for the different LPs.
	 */
	private static TIntObjectHashMap<EventLogger> loggers;

	/**
	 * Whether all loggers write to the same file.
	 */
	private static final boolean ALL_ONE_FILE = false;

	/**
	 * Sets the time unit in which the simulation represents timestamps.
	 * 
	 * @param simulationTimeUnit
	 *            the input time unit.
	 */
	public static void setSimulationTimeUnit(TimeUnit simulationTimeUnit) {
		simTimeUnit = simulationTimeUnit;
		calcConversionFactor();
	}

	/**
	 * Sets the time unit in which to display timestamps in the trace file.
	 * 
	 * @param loggingTimeUnit
	 *            the output time unit.
	 */
	public static void setOutputTimeUnit(TimeUnit loggingTimeUnit) {
		logTimeUnit = loggingTimeUnit;
		calcConversionFactor();
	}

	/**
	 * Provides references to the event loggers of the different LPs. Only
	 * called at the beginning of the simulation. So the synchronized block is
	 * not expensive at all.
	 * 
	 * @param lpId
	 *            the LP ID of the calling LP.
	 * @return a reference to the logger for this LP.
	 */
	public static synchronized EventLogger getInstance(int lpId) {
		if (loggers == null) {
			loggers = new TIntObjectHashMap<EventLogger>();
		}
		if (ALL_ONE_FILE) {
			lpId = 0;
		}
		if (!loggers.containsKey(lpId)) {
			try {
				loggers.put(lpId, new EventLogger(lpId));
			} catch (IOException e) {
				if (CustomLogger.isFatalEnabled()) {
					CustomLogger.fatal("Could not create EventLogger!", e);
				}
			}
		}
		return loggers.get(lpId);
	}

	/**
	 * Whether event logging is enabled.
	 * 
	 * @return whether event logging is enabled.
	 */
	public static boolean isEnabled() {
		return isEnabled;
	}

	/**
	 * Enables or disables event logging.
	 * 
	 * @param enabled
	 *            whether to enable event logging.
	 */
	public static void setEnabled(boolean enabled) {
		EventLogger.isEnabled = enabled;
	}

	/**
	 * Sets the output trace file name.
	 * 
	 * @param filename
	 *            name of the trace file.
	 */
	public static void setOutput(String filename) {
		baseFilename = filename;
	}

	/**
	 * Closes all loggers.
	 */
	public static void closeAll() {
		if (loggers != null && isEnabled) {
			int[] logs = loggers.keys();
			for (int size = logs.length, i = 0; i < size; i++) {
				try {
					loggers.get(logs[i]).buffer_.close();
				} catch (IOException e) {
					if (CustomLogger.isWarningEnabled()) {
						CustomLogger.warning("Could not close EventLogger!", e);
					}
				}
			}
		}
	}

	/**
	 * Calculates a conversion factor to output timestamps in the chosen
	 * resolution.
	 */
	private static void calcConversionFactor() {
		if (simTimeUnit.ordinal() == logTimeUnit.ordinal()) {
			conversionFactor = 1;
		} else if (simTimeUnit.ordinal() > logTimeUnit.ordinal()) {
			int tmpUnit = simTimeUnit.ordinal();
			conversionFactor = 1;
			while (tmpUnit > logTimeUnit.ordinal()) {
				tmpUnit--;
				conversionFactor *= unitScales[tmpUnit];
			}
		} else {
			int tmpUnit = logTimeUnit.ordinal();
			conversionFactor = 1;
			while (tmpUnit > simTimeUnit.ordinal()) {
				tmpUnit--;
				conversionFactor *= unitScales[tmpUnit];
			}
			conversionFactor = 1 / conversionFactor;
		}
	}

	/**
	 * The trace file writer.
	 */
	private final BufferedWriter buffer_;

	/**
	 * Constructor. Creates an event logger for the specified LP.
	 * 
	 * @param lpId
	 *            the LP ID.
	 * @throws IOException
	 *             an exception, if one occurred.
	 */
	private EventLogger(int lpId) throws IOException {
		if (isEnabled) {
			buffer_ = new BufferedWriter(new FileWriter(baseFilename + "."
					+ lpId));
		} else {
			buffer_ = null;
		}
	}

	/**
	 * Writes information to the trace file.
	 * 
	 * @param time
	 *            the current simulation time
	 * @param moduleNumber
	 *            the module number (node ID) where the event occurred.
	 * @param eventId
	 *            the event ID.
	 * @param eventParam
	 *            the event parameter.
	 */
	@SuppressWarnings("rawtypes")
	public void trace(long time, int moduleNumber, Enum eventId,
			Object eventParam) {
		if (isEnabled) {
			StringBuilder string = new StringBuilder(80);
			string.append("(");
			addTimestamp(time, string).append(") ");
			string.append(
					eventId.getDeclaringClass().getDeclaringClass()
							.getSimpleName()).append("_");
			string.append(moduleNumber).append(" ");
			string.append(eventId).append(" ");
			string.append(eventParam);
			try {
				buffer_.write(string.toString());
				buffer_.newLine();
				buffer_.flush();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Adds a timestamp to the output string.
	 * 
	 * @param timestamp
	 *            the timestamp.
	 * @param string
	 *            the output string.
	 * @return the updated output string.
	 */
	private StringBuilder addTimestamp(long timestamp, StringBuilder string) {
		if (simTimeUnit == logTimeUnit) {
			string.append(timestamp);
		} else {
			string.append(timestamp * conversionFactor);
		}
		return string;
	}
}