/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/app/DataPacket.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.app;

import jist.swans.misc.Message;
import modules.wireless.scenario.NodeConfig.FlowPriority;

/**
 * Helper class to save information about a data transfer initialized by the
 * packet generator.
 * 
 * @author Patrick Peschlow
 * @author Daniel Plohmann
 * 
 * @version 1.0
 */
public class DataPacket extends Message {
	/**
	 * Default ID.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Size of the data.
	 */
	private final int size_;

	/**
	 * Destination node ID.
	 */
	private final int dstNode_;

	/**
	 * Source node ID.
	 */
	private final int srcNode_;

	/**
	 * Logical connection ID.
	 */
	private final int connId_;

	/**
	 * Flow priority (used in Mesh scenarios).
	 */
	private final FlowPriority priority_;

	/**
	 * Constructor.
	 * 
	 * @param size
	 *            data size.
	 * @param srcNode
	 *            source node ID.
	 * @param dstNode
	 *            destination node ID.
	 * @param connId
	 *            connection ID.
	 */
	public DataPacket(int size, int srcNode, int dstNode, int connId) {
		this(size, srcNode, dstNode, connId, FlowPriority.AC_BE);
	}

	/**
	 * Constructor.
	 * 
	 * @param size
	 *            data size.
	 * @param srcNode
	 *            source node ID.
	 * @param dstNode
	 *            destination node ID.
	 * @param connId
	 *            connection ID.
	 * @param priority
	 *            mesh flow priority.
	 */
	public DataPacket(int size, int srcNode, int dstNode, int connId,
			FlowPriority priority) {
		size_ = size;
		srcNode_ = srcNode;
		dstNode_ = dstNode;
		connId_ = connId;
		priority_ = priority;
	}

	/**
	 * @see jist.swans.misc.Message#equals(java.lang.Object)
	 */
	public boolean equals(Object o) {
		if (o == null) {
			return false;
		}
		if (!(this.getClass().equals(o.getClass()))) {
			return false;
		}

		DataPacket otherTransfer = (DataPacket) o;

		if (otherTransfer.size_ == this.size_
				&& otherTransfer.dstNode_ == this.dstNode_
				&& otherTransfer.srcNode_ == this.srcNode_
				&& otherTransfer.connId_ == this.connId_
				&& otherTransfer.priority_ == this.priority_) {
			return true;
		}
		return false;
	}

	/**
	 * @see jist.swans.misc.Message#hashCode()
	 */
	public int hashCode() {
		int result = 17;
		result = 37 * result + size_;
		result = 37 * result + dstNode_;
		result = 37 * result + srcNode_;
		result = 37 * result + connId_;
		result = 37 * result + priority_.hashCode();
		return result;
	}

	/**
	 * @see jist.swans.misc.Message#toString()
	 */
	public String toString() {
		return "DataPacket(" + srcNode_ + "->" + dstNode_ + ": " + connId_
				+ ")";
	}

	/**
	 * @see jist.swans.misc.Message#getSize()
	 */
	public int getSize() {
		return size_;
	}

	/**
	 * Provides the destination node ID.
	 * 
	 * @return the destination node ID.
	 */
	public int getDstNode() {
		return dstNode_;
	}

	/**
	 * Provides the connection ID.
	 * 
	 * @return the connection ID.
	 */
	public int getConnId() {
		return connId_;
	}

	/**
	 * Provides the flow priority.
	 * 
	 * @return the flow priority.
	 */
	public FlowPriority getPriority() {
		return priority_;
	}
}
