/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/app/PacketGenerator.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.app;

import jist.swans.Constants;
import jist.swans.trans.TransTcp;
import jist.swans.trans.TransUdp;
import modules.wireless.scenario.NodeConfig;
import modules.wireless.scenario.NodeConfig.FlowPriority;
import core.BaseModule;

/**
 * A simple packet generator for multiple traffic flows.
 * 
 * @author Patrick Peschlow
 * @author Christoph Sprenger
 * 
 * @version 1.0
 */
public class PacketGenerator extends BaseModule {
	/**
	 * The event types offered by this module.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	public static enum Events {
		/**
		 * Generates a packet for a TCP flow.
		 */
		SEND_TCP,
		/**
		 * Generates a packet for a UDP flow.
		 */
		SEND_UDP,
		/**
		 * Receives a packet of a TCP flow.
		 */
		RECEIVE_TCP,
		/**
		 * Receives a packet of a UDP flow.
		 */
		RECEIVE_UDP;
	}

	/**
	 * Type of a traffic flow (TCP or UDP).
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	public static enum ConnectionType {
		/**
		 * TCP flow.
		 */
		TCP,
		/**
		 * UDP flow.
		 */
		UDP;
	}

	/**
	 * The current simulation time.
	 */
	private long now_;

	/**
	 * Reference to my TCP transport layer module.
	 */
	private int myTransTcpLayerId_;

	/**
	 * Reference to my UDP transport layer module.
	 */
	private int myTransUdpLayerId_;

	/**
	 * The number of bytes received via TCP connections.
	 */
	private int numTcpBytesReceived_ = 0;

	/**
	 * The number of bytes sent via TCP connections.
	 */
	private int numTcpBytesSent_ = 0;

	/**
	 * The number of bytes received via UDP connections.
	 */
	private int numUdpBytesReceived_ = 0;

	/**
	 * The number of bytes sent via UDP connections.
	 */
	private int numUdpBytesSent_ = 0;

	/**
	 * Constructor. Creates a new packet generator for traffic flows
	 * ("connections").
	 * 
	 * @param number
	 *            node ID of the node this packet generator belongs to.
	 */
	public PacketGenerator(int number) {
		super(number);
	}

	/**
	 * @see core.BaseModule#activate(long, int, java.lang.Enum,
	 *      java.lang.Object)
	 */
	@Override
	@SuppressWarnings("rawtypes")
	public void activate(long execTime, int srcModuleId, Enum eventId,
			Object parameter) {
		now_ = execTime;
		if (eventId == Events.SEND_TCP) {
			Connection conn = (Connection) parameter;
			scheduleNewEvent(
					now_ + Constants.NET_DELAY,
					myTransTcpLayerId_,
					TransTcp.Events.RECEIVE_OUTGOING,
					new DataPacket(conn.getPacketSize(), number_, -1, conn
							.getId(), conn.getFlowPriority()));
			numTcpBytesSent_ += conn.getPacketSize();
			if (conn.getSendInterval() > 0) {
				scheduleNewEvent((now_ + conn.getSendInterval()), id_,
						Events.SEND_TCP, conn);
			}
		} else if (eventId == Events.SEND_UDP) {
			Connection conn = (Connection) parameter;
			scheduleNewEvent(
					now_ + Constants.NET_DELAY,
					myTransUdpLayerId_,
					TransUdp.Events.RECEIVE_OUTGOING,
					new DataPacket(conn.getPacketSize(), number_, conn
							.getDstNode(), 0, conn.getFlowPriority()));
			numUdpBytesSent_ += conn.getPacketSize();
			scheduleNewEvent((now_ + conn.getSendInterval()), id_,
					Events.SEND_UDP, conn);
		} else if (eventId == Events.RECEIVE_TCP) {
			DataPacket packet = (DataPacket) parameter;
			numTcpBytesReceived_ += packet.getSize();
		} else if (eventId == Events.RECEIVE_UDP) {
			DataPacket packet = (DataPacket) parameter;
			numUdpBytesReceived_ += packet.getSize();
		}
	}

	/**
	 * @see core.BaseModule#init()
	 */
	public void init() {
		if (NodeConfig.getTrafficInfoUDP() != null) {
			myTransUdpLayerId_ = findModuleId(TransUdp.class, number_);
		}
		if (NodeConfig.getTrafficInfoTCP() != null) {
			myTransTcpLayerId_ = findModuleId(TransTcp.class, number_);
		}
	}

	/**
	 * Convenience method to add TCP bytes of retransmissions.
	 * 
	 * @param numBytes
	 *            the number of bytes sent.
	 */
	public void addTcpBytesSent(int numBytes) {
		numTcpBytesSent_ += numBytes;
	}

	/**
	 * Provides the number of bytes received via TCP flows.
	 * 
	 * @return the number of bytes received via TCP flows.
	 */
	public int getTcpBytesReceived() {
		return numTcpBytesReceived_;
	}

	/**
	 * Provides the number of bytes sent via TCP flows.
	 * 
	 * @return the number of bytes sent via TCP flows.
	 */
	public int getTcpBytesSent() {
		return numTcpBytesSent_;
	}

	/**
	 * Provides the number of bytes received via UDP flows.
	 * 
	 * @return the number of bytes received via UDP flows.
	 */
	public int getUdpBytesReceived() {
		return numUdpBytesReceived_;
	}

	/**
	 * Provides the number of bytes sent via UDP flows.
	 * 
	 * @return the number of bytes sent via UDP flows.
	 */
	public int getUdpBytesSent() {
		return numUdpBytesSent_;
	}

	/**
	 * Adds a connection to this packet generator.
	 * 
	 * @param type
	 *            UDP or TCP.
	 * @param id
	 *            connection ID.
	 * @param startTime
	 *            time at which the first packet is generated.
	 * @param dstNode
	 *            destination node ID.
	 * @param packetSize
	 *            packet size.
	 * @param sendInterval
	 *            send interval.
	 */
	public void addConnection(ConnectionType type, int id, long startTime,
			int dstNode, int packetSize, long sendInterval) {
		addConnection(type, id, startTime, dstNode, packetSize, sendInterval,
				FlowPriority.AC_BE);
	}

	/**
	 * Adds a connection to this packet generator.
	 * 
	 * @param type
	 *            UDP or TCP.
	 * @param id
	 *            connection ID.
	 * @param startTime
	 *            time at which the first packet is generated.
	 * @param dstNode
	 *            destination node ID.
	 * @param packetSize
	 *            packet size.
	 * @param sendInterval
	 *            send interval.
	 * @param priority
	 *            flow priority.
	 */
	public void addConnection(ConnectionType type, int id, long startTime,
			int dstNode, int packetSize, long sendInterval,
			FlowPriority priority) {
		final Events eventId = (type == ConnectionType.TCP) ? Events.SEND_TCP
				: Events.SEND_UDP;
		final Connection conn = new Connection(id, dstNode, packetSize,
				sendInterval, priority);
		scheduleNewEvent(startTime, id_, eventId, conn);
	}

	/**
	 * Represents a logical connection (traffic flow).
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private static class Connection {
		/**
		 * Connection ID.
		 */
		private final int id_;

		/**
		 * Destination node ID.
		 */
		private final int dstNode_;

		/**
		 * Packet size.
		 */
		private final int packetSize_;

		/**
		 * Time interval between two consecutive packets generated.
		 */
		private final long sendInterval_;

		/**
		 * Flow priority.
		 */
		private final FlowPriority priority_;

		/**
		 * Constructor.
		 * 
		 * @param id
		 *            connection ID.
		 * @param dstNode
		 *            destination node ID.
		 * @param packetSize
		 *            packet size.
		 * @param sendInterval
		 *            send interval.
		 * @param priority
		 *            flow priority.
		 */
		private Connection(int id, int dstNode, int packetSize,
				long sendInterval, FlowPriority priority) {
			id_ = id;
			dstNode_ = dstNode;
			packetSize_ = packetSize;
			sendInterval_ = sendInterval;
			priority_ = priority;
		}

		/**
		 * @see java.lang.Object#toString()
		 */
		public String toString() {
			return "Connection_" + id_;
		}

		/**
		 * Provides the connection ID.
		 * 
		 * @return the connection ID.
		 */
		private int getId() {
			return id_;
		}

		/**
		 * Provides the destination node ID.
		 * 
		 * @return the destination node ID.
		 */
		private int getDstNode() {
			return dstNode_;
		}

		/**
		 * Provides the packet size.
		 * 
		 * @return the packet size.
		 */
		private int getPacketSize() {
			return packetSize_;
		}

		/**
		 * Provides the send interval.
		 * 
		 * @return the send interval.
		 */
		private long getSendInterval() {
			return sendInterval_;
		}

		/**
		 * Provides the flow priority.
		 * 
		 * @return the flow priority.
		 */
		private FlowPriority getFlowPriority() {
			return priority_;
		}
	}
}
