/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/lazy/LazyMac802_11.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.lazy;

import jist.swans.Constants;
import jist.swans.mac.Mac802_11;
import jist.swans.mac.MacAddress;
import jist.swans.mac.MacMessage;
import jist.swans.net.NetMessage;
import jist.swans.net.NetMessage.Ip;
import jist.swans.radio.RadioNoise;
import jist.swans.radio.RadioNoise.RadioMode;
import modules.wireless.lazy.LazyManager.HistoryEntry;
import modules.wireless.partition.TransmissionInfo;
import modules.wireless.partition.Partition;
import core.Config;

/**
 * This MAC layer implements, in conjunction with the LazyRadioNoise layer, a
 * lazy MAC state update for the IEEE 802.11 standard. It is based on the
 * original MAC layer.
 * 
 * @author Patrick Peschlow
 * @author Christoph Sprenger
 * 
 * @version 1.0
 */
public class LazyMac802_11 extends Mac802_11 {
	/**
	 * My lazy radio layer.
	 */
	private LazyRadio myLazyRadio_;

	/**
	 * My lazy partition.
	 */
	private LazyPartition myLazyPartition_;

	/**
	 * Whether I try to get activated because I want to transmit a frame.
	 */
	private boolean wantsTx_;

	/**
	 * Helper variable to update backoff during retrospective.
	 */
	private long storeBackoff_;

	/**
	 * Helper variable to store backoff state.
	 */
	private boolean storeHasBackoff_;

	/**
	 * Stores the eifs_ flag.
	 */
	private long storeEifs_;

	/**
	 * Stores the needEifs_ flag.
	 */
	private boolean storeNeedEifs_;

	/**
	 * Radio mode at the time that we got deactivated.
	 */
	private RadioMode storeRadioMode_;

	/**
	 * Stores whether we got busy while being deactivated.
	 */
	private boolean gotBusy_;

	/**
	 * Stores the last time when we got busy.
	 */
	private long storeTimeOfGotBusy_;

	/**
	 * Whether we had a backoff selected at the time we got busy.
	 */
	private boolean gotBusyHasBackoff_;

	/**
	 * Last time we were interested in DIFS.
	 */
	private long lastIdleSwitch_;

	/**
	 * Constructor.
	 * 
	 * @param number
	 *            the class-internal number of this MAC.
	 * @param networkLayerClass
	 *            class of the network layer.
	 */
	@SuppressWarnings("rawtypes")
	public LazyMac802_11(int number, Class networkLayerClass) {
		super(number, LazyRadio.class, networkLayerClass);
		assert localAddr_ == number_;
		wantsTx_ = false;
		storeBackoff_ = bo_;
		storeHasBackoff_ = false;
		gotBusy_ = false;
		storeTimeOfGotBusy_ = Long.MAX_VALUE;
		storeNeedEifs_ = false;
		storeEifs_ = 0;
		gotBusyHasBackoff_ = false;
		lastIdleSwitch_ = 0;
		storeRadioMode_ = RadioMode.IDLE;
	}

	/**
	 * @see jist.swans.mac.Mac802_11#init()
	 */
	@Override
	public void init() {
		super.init();
		assert myRadioLayer_ != null;
		myLazyRadio_ = (LazyRadio) myRadioLayer_;
		myLazyPartition_ = (LazyPartition) Partition
				.getPartitionOfNode(number_);
		assert myLazyPartition_ != null;
	}

	/**
	 * Receiving a new outgoing packet (event is scheduled by my network layer).
	 * This method may trigger an activation.
	 * 
	 * @param transfer
	 *            the packet to send
	 * @param nextHop
	 *            the next hop for the packet
	 */
	@Override
	protected void newTransfer(NetMessage.Ip transfer, int nextHop) {
		if (!isBusy() && !myLazyPartition_.isActive(myLazyRadio_)) {
			// I just became busy but am deactivated, so I have no idea about
			// the true state of the medium. In order not to "miss" my own
			// transmission, I need to assume that the medium is idle and I can
			// start a transmission immediately (this situation may arise, e.g.,
			// if EIFS just terminated and I have no backoff left).
			gotBusy_ = true;
			storeTimeOfGotBusy_ = now_;
			gotBusyHasBackoff_ = hasBackoff_;
			// Replace what's normally done in newTransfer().
			numDirectTransfers_++;
			currentPacket_ = transfer;
			currentPacketNextHop_ = nextHop;
			// Activate.
			wantsTx_ = true;
			myLazyPartition_.activateTxNode(now_, myLazyRadio_);
		} else {
			super.newTransfer(transfer, nextHop);
		}
	}

	/**
	 * Deactivate this node if it attempts to send but has nothing to send.
	 */
	@Override
	protected void startBackoffProcedure() {
		// First, initiate normal behavior (which continues while I am
		// deactivated).
		super.startBackoffProcedure();
		// Now see if I may get deactivated.
		if (myLazyPartition_.isActive(myLazyRadio_)) {
			// Avoid cases where I am busy and deactivation leads to the miss of
			// the END_RX event of an incoming signal. Then I might miss my own
			// transmission.
			if (isBusy() && (!isRadioIdle() || isWaitingEifs())) {
				return;
			}
			// Store relevant data before deactivation.
			wantsTx_ = false;
			storeBackoff_ = bo_;
			storeHasBackoff_ = hasBackoff_;
			storeEifs_ = eifs_;
			storeNeedEifs_ = needEifs_;
			storeRadioMode_ = myLazyRadio_.getMode();
			gotBusy_ = false;
			storeTimeOfGotBusy_ = Long.MAX_VALUE;
			gotBusyHasBackoff_ = false;
			if (isRadioIdle()) {
				lastIdleSwitch_ = now_;
			}
			// Deactivate.
			myLazyPartition_.deactivateNode(now_, myLazyRadio_);
		}
	}

	/**
	 * If we want to send a packet but are currently not active, we have to make
	 * a retrospection and get active, if still applicable then.
	 */
	@Override
	protected void sendPacket() {
		if (!myLazyPartition_.isActive(myLazyRadio_)) {
			// Not currently active, so get active and see if we can really
			// transmit the packet.
			wantsTx_ = true;
			myLazyPartition_.activateTxNode(now_, myLazyRadio_);
		} else {
			super.sendPacket();
		}
	}

	/**
	 * Starts a retrospective. Called when I try to get activated.
	 */
	void startRetro() {
		// If a timer is running, cancel it. At the end of the retrospective, we
		// may restart a timer with corrected information.
		cancelTimer();
		// Restore the state from the point of deactivation.
		bo_ = storeBackoff_;
		hasBackoff_ = storeHasBackoff_;
		eifs_ = storeEifs_;
		needEifs_ = storeNeedEifs_;
		myLazyRadio_.setMode(storeRadioMode_);
		if (!gotBusy_ && hasPacket()) {
			storeTimeOfGotBusy_ = 0;
		}
	}

	/**
	 * Ends a retrospective. Called when all history updates have been made.
	 * 
	 * @param now
	 *            the current simulation time
	 * @return whether I get activated
	 */
	boolean endRetro(long now) {
		// Finally, update the current time of the MAC.
		now_ = now;
		assert !isTimerRunning_;
		// Check if I got busy only during the final evaluation period.
		// PP: Maybe this can be simplified.
		if (gotBusy_ && storeTimeOfGotBusy_ < now) {
			if (!gotBusyHasBackoff_) {
				if (!isRadioIdle() || nav_ > storeTimeOfGotBusy_) {
					bo_ = ((long) (nextBoSample_ * cw_)) * SLOT_TIME;
					needBackoff_ = false;
					hasBackoff_ = true;
					presampleBackoff();
					gotBusyHasBackoff_ = true;
				}
			}
		}
		assert now_ == now;
		if (isRadioIdle() && !isWaitingNav() && !isWaitingEifs()) {
			// Final backoff and DIFS calculation.
			if (lastIdleSwitch_ < nav_) {
				lastIdleSwitch_ = nav_;
			}
			if (lastIdleSwitch_ < eifs_) {
				lastIdleSwitch_ = eifs_;
			}
			long diff;
			if (gotBusy_ && storeTimeOfGotBusy_ > lastIdleSwitch_) {
				// Got busy in this final interval.
				long backoffPassed = storeTimeOfGotBusy_ - lastIdleSwitch_;
				if (backoffPassed > 0) {
					// Test for special case.
					if (storeTimeOfGotBusy_ == now
							&& backoffPassed >= bo_ - RX_TX_TURNAROUND) {
						// Got busy right now and backoff reached zero.
						bo_ = 0;
					} else {
						bo_ -= (backoffPassed / SLOT_TIME) * SLOT_TIME;
					}
				}
				if (bo_ < 0) {
					bo_ = 0;
				}
				diff = now - storeTimeOfGotBusy_;
				lastIdleSwitch_ = storeTimeOfGotBusy_;
			} else {
				// Did not get busy in this final interval.
				diff = now - lastIdleSwitch_;
			}
			gotBusy_ = false;
			// Now see what all this results in.
			macMode_ = null;
			if (lastIdleSwitch_ == eifs_ || !isBusy()) {
				// We don't need to defer DIFS. Add it, it will be subtracted
				// again.
				diff += DIFS;
			}
			if (diff >= DIFS) {
				long backoffExpired = diff - DIFS;
				long remainingBackoff = bo_ - backoffExpired;
				if (remainingBackoff < 0) {
					remainingBackoff = 0;
				}
				// We finished DIFS, so only backoff remaining. Decrement it.
				if (remainingBackoff <= RX_TX_TURNAROUND) {
					bo_ = 0;
					if (isBusy()) {
						hasBackoff_ = false;
						super.sendPacket();
					} else {
						assert !wantsTx_;
						macMode_ = MacMode.SIDLE;
					}
					// Get active.
					assert macMode_ != null;
					return true;
				}
				assert remainingBackoff > RX_TX_TURNAROUND;
				hasBackoff_ = true;
				macMode_ = MacMode.SBO;
				// Find next higher multiple of the slot time.
				boStart_ = now - backoffExpired;
				startTimer(remainingBackoff - RX_TX_TURNAROUND);
				if (Config.usesProtocolLookahead()) {
					setNextPST(now_ + bo_);
				}
				if (!wantsTx_) {
					assert macMode_ != null;
					return true;
				}
			} else {
				// We are during DIFS.
				assert isBusy();
				final long remainingDifs = DIFS - diff;
				if (bo_ > 0) {
					// We have backoff left, so defer until end of DIFS.
					startTimer(remainingDifs);
					macMode_ = MacMode.SDIFS;
				} else {
					// No backoff left.
					if (remainingDifs > RX_TX_TURNAROUND) {
						// More than RX_TX_TURNAROUND left, so go on deferring.
						startTimer(remainingDifs - RX_TX_TURNAROUND);
						macMode_ = MacMode.SDIFS;
					} else {
						// We can transmit right now.
						hasBackoff_ = false;
						super.sendPacket();
						// Get active.
						assert macMode_ != null;
						return true;
					}
				}
				if (!wantsTx_) {
					assert macMode_ != null;
					return true;
				}
			}
		} else {
			// I cannot transmit right now.
			gotBusy_ = false;
			macMode_ = MacMode.SIDLE;
			startBackoffProcedure();
			if (!wantsTx_) {
				// I got activated because I start receiving something.
				assert macMode_ != null;
				return true;
			}
		}
		// We only reach this point if a transmission attempt triggered this
		// retrospection and we may not transmit now.
		assert wantsTx_;
		wantsTx_ = false;
		// Avoid cases where staying deactivated makes me miss the END_RX event
		// of an incoming signal.
		if (isBusy() && (!isRadioIdle() || isWaitingEifs())) {
			// Get active.
			assert macMode_ != null;
			return true;
		}
		// Stay deactivated.
		storeBackoff_ = bo_;
		storeHasBackoff_ = hasBackoff_;
		storeEifs_ = eifs_;
		storeNeedEifs_ = needEifs_;
		storeRadioMode_ = myLazyRadio_.getMode();
		assert macMode_ != null;
		return false;
	}

	/**
	 * Performs a step in the retrospective by evaluating the given history
	 * entry.
	 * 
	 * @param entry
	 *            the history entry to evaluate
	 */
	void updateRetro(HistoryEntry entry) {
		final LazySignal signal = entry.getSignal();
		if (signal.getTxNodeId() == number_) {
			// Own signal. Must be the end of my transmission before I got
			// deactivated.
			assert !entry.doesSignalBegin();
			myLazyRadio_
					.setMode(myLazyRadio_.isSensing() ? RadioNoise.RadioMode.SENSING
							: RadioNoise.RadioMode.IDLE);
		} else {
			// Foreign signal.
			final TransmissionInfo txInfo = signal.getTxInfo();
			if (!Partition.getMyField(myLazyRadio_).isReceived(number_, txInfo)) {
				return;
			}
			final long now = entry.getSimTime();
			// Check if we got busy during the current interval.
			// PP: Maybe we can (or have to) replace < by <= here
			if (gotBusy_ && storeTimeOfGotBusy_ < now) {
				if (!gotBusyHasBackoff_) {
					if (!isRadioIdle() || nav_ > storeTimeOfGotBusy_) {
						bo_ = ((long) (nextBoSample_ * cw_)) * SLOT_TIME;
						needBackoff_ = false;
						hasBackoff_ = true;
						presampleBackoff();
						gotBusyHasBackoff_ = true;
					}
				}
			}
			if (entry.doesSignalBegin()) {
				// Start of a foreign signal.
				final RadioMode recentMode = myLazyRadio_.getMode();
				myLazyRadio_.signalBegins(now, txInfo);
				// React if the mode changed from idle.
				if (recentMode == RadioMode.IDLE
						&& myLazyRadio_.getMode() != RadioMode.IDLE) {
					// We have switched from IDLE to SENSING or RECEIVING.
					updateBackoff(now);
				}
			} else {
				// End of a foreign signal.
				final RadioMode recentMode = myLazyRadio_.getMode();
				MacMessage msg = (MacMessage) myLazyRadio_.signalEnds(now,
						txInfo);
				if (msg != null) {
					// Our currently locked signal ended.
					assert signal.getDstNodeId() != number_;
					eifs_ = 0;
					needEifs_ = false;
					long newNav = -1;
					if (msg.getType() == MacMessage.TYPE_RTS) {
						// standard 2007, section 9.2.5.4
						final long ctsTime = MacMessage.Cts.SIZE
								* Constants.SECOND / bandwidth_
								+ SYNCHRONIZATION;
						newNav = now + SIFS + SIFS + ctsTime + SLOT_TIME
								+ SLOT_TIME + PHY_RX_START_DELAY;
					} else {
						if (msg.getDuration() > 0) {
							newNav = now + msg.getDuration();
						}
					}
					if (newNav > nav_) {
						nav_ = newNav;
					}
				} else if (myLazyRadio_.hasRcvError()) {
					if (myRadioLayer_.isSensing()) {
						needEifs_ = true;
					} else {
						eifs_ = now + EIFS;
					}
				}
				if (recentMode != RadioMode.IDLE
						&& myLazyRadio_.getMode() == RadioMode.IDLE) {
					// We have switched to IDLE.
					lastIdleSwitch_ = now;
					if (needEifs_) {
						eifs_ = now + EIFS;
						needEifs_ = false;
					}
				}
			}
		}
	}

	/**
	 * Updates the backoff timer after evaluating the current history entry.
	 * 
	 * @param now
	 *            the current simulation time
	 */
	private void updateBackoff(final long now) {
		if (nav_ > now) {
			// We had to wait NAV and couldn't count down backoff or DIFS.
			return;
		}
		if (eifs_ > now) {
			// We had to wait EIFS and couldn't count down backoff or DIFS.
			return;
		}
		// Did we have to wait NAV or EIFS earlier in this interval?
		if (lastIdleSwitch_ < nav_) {
			lastIdleSwitch_ = nav_;
		}
		if (lastIdleSwitch_ < eifs_) {
			lastIdleSwitch_ = eifs_;
		}
		// Do the actual calculation.
		long diff = now - lastIdleSwitch_;
		if (storeTimeOfGotBusy_ <= lastIdleSwitch_) {
			// We were busy from the beginning, so start DIFS first.
			if (lastIdleSwitch_ != eifs_) {
				diff -= DIFS;
			}
		} else if (storeTimeOfGotBusy_ < now) {
			// We start DIFS first, but only since we got busy.
			if (lastIdleSwitch_ != eifs_) {
				long difsPassed = now - storeTimeOfGotBusy_;
				if (difsPassed > DIFS) {
					difsPassed = DIFS;
				}
				assert difsPassed >= 0;
				diff -= difsPassed;
			}
			lastIdleSwitch_ = storeTimeOfGotBusy_;
		}
		// If instead, storeTimeOfGotBusy_ >= now, we don't wait DIFS at all.
		if (diff > 0) {
			// We have decremented backoff.
			bo_ -= (diff / SLOT_TIME) * SLOT_TIME;
		}
		if (bo_ < 0) {
			bo_ = 0;
		}
	}

	/**
	 * Handles a special case where no entries have been added to the signal
	 * history since my last deactivation.
	 */
	void handleSpecialCaseNoEntries() {
		if (wantsTx_) {
			wantsTx_ = false;
			if (gotBusy_) {
				// Called by newTransfer().
				NetMessage.Ip transfer = (Ip) currentPacket_;
				int nextHop = currentPacketNextHop_;
				currentPacket_ = null;
				currentPacketNextHop_ = MacAddress.NULL;
				numDirectTransfers_--;
				super.newTransfer(transfer, nextHop);
			} else {
				// Called by sendPacket().
				super.sendPacket();
			}
		}
	}
}
