/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/lazy/LazyManager.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.lazy;

import jist.swans.field.Field;
import jist.swans.mac.MacAddress;
import jist.swans.mac.MacMessage;
import jist.swans.radio.RadioNoise;
import modules.wireless.partition.TransmissionInfo;
import modules.wireless.partition.Partition;
import modules.wireless.scenario.WirelessNode;
import core.Config;
import core.util.CustomLogger;

/**
 * Central data structure used for Lazy MAC State Update. Controlled and
 * accessed by LazyPartitions.
 * 
 * @author Patrick Peschlow
 * @author Christoph Sprenger
 * 
 * @version 1.0
 */
class LazyManager {
	/**
	 * Initial size of the array holding the signal history.
	 */
	private static final int INIT_SIZE = 100;

	/**
	 * Maximum shift allowed before the history contents are shifted back in the
	 * internal array representation.
	 */
	private static final int MAX_SHIFT = 1000;

	/**
	 * Whether the number of active nodes should be monitored during the
	 * simulation.
	 */
	private static final boolean USE_MONITOR = true;

	/**
	 * ID of the partition this LazyLists instance belongs to.
	 */
	private final int partitionId_;

	/**
	 * Field object used to calculate signal powers.
	 */
	private final Field field_;

	/**
	 * Key: Node ID. Value: Whether this radio is active.
	 */
	private final boolean[] isActive_;

	/**
	 * Key: Node ID. Value: The last seen entry ID of this radio.
	 */
	private final int[] lastEntryIds_;

	/**
	 * A list of all radios of this partition.
	 */
	private LazyRadio[] radios_;

	/**
	 * The signal history.
	 */
	private HistoryEntry[] history_;

	/**
	 * The array index of the first signal in the history.
	 */
	private int firstIndex_;

	/**
	 * The first free array index of the history.
	 */
	private int firstFreeIndex_;

	/**
	 * The ID of the last history entry.
	 */
	private int currentEntryId_;

	/**
	 * Monitor for logging lazy activity.
	 */
	private LazyMonitor monitor_;

	/**
	 * The number of nodes of my partition.
	 */
	private int numMyNodes_;

	/**
	 * Constructor.
	 * 
	 * @param partitionId
	 *            the ID of the partition this LazyLists object belongs to
	 * @param numNodes
	 *            the total number of nodes of the simulation
	 * @param field
	 *            the Field object that shall be used for physical calculations
	 */
	LazyManager(int partitionId, int numNodes, Field field) {
		assert field != null;
		partitionId_ = partitionId;
		field_ = field;
		isActive_ = new boolean[numNodes];
		lastEntryIds_ = new int[numNodes];
		radios_ = null;
		history_ = new HistoryEntry[INIT_SIZE];
		firstIndex_ = 0;
		firstFreeIndex_ = 0;
		currentEntryId_ = 0;
		if (USE_MONITOR) {
			monitor_ = new LazyMonitor();
		}
	}

	/**
	 * Activate a lazy node as transmitter.
	 * 
	 * @param now
	 *            the current simulation time (at which the node is to be
	 *            activated)
	 * @param radio
	 *            radio to be activated
	 */
	void activateTxNode(long now, LazyRadio radio) {
		activateNode(now, radio);
	}

	/**
	 * Deactivate a lazy node.
	 * 
	 * @param now
	 *            the current simulation time (at which the node is to be
	 *            deactivated)
	 * @param radio
	 *            radio to be deactivated
	 */
	void deactivateNode(long now, LazyRadio radio) {
		assert isActive(radio);
		assert currentEntryId_ >= lastEntryIds_[radio.getNumber()];
		lastEntryIds_[radio.getNumber()] = currentEntryId_;
		setActive(radio, false);
		if (USE_MONITOR) {
			monitor_.monitorDeactivation(now);
		}
		assert !isActive(radio);
	}

	/**
	 * Whether the given is currently active.
	 * 
	 * @param radio
	 *            the radio of interest
	 * @return whether the given is currently active
	 */
	boolean isActive(LazyRadio radio) {
		return isActive_[radio.getNumber()];
	}

	/**
	 * Hands me a list of all nodes of my partition.
	 * 
	 * @param nodes
	 *            a list of all nodes of my partition.
	 */
	void setNodes(WirelessNode[] nodes) {
		assert nodes != null;
		numMyNodes_ = nodes.length;
		radios_ = new LazyRadio[numMyNodes_];
		for (int i = 0; i < numMyNodes_; i++) {
			radios_[i] = (LazyRadio) nodes[i].getRadio();
		}
	}

	/**
	 * New signal begins. The signal-start-event will be added to the signal
	 * history and all active lazy nodes will be notified by a RECEIVE event.
	 * 
	 * @param now
	 *            the current simulation time (at which the signal begins)
	 * @param txInfo
	 *            the TX information about the new signal
	 * @return new signal object returned to the partition
	 */
	LazySignal addSignal(long now, TransmissionInfo txInfo) {
		final int dstAddress = ((MacMessage) txInfo.getMsg()).getDst();
		if (MacAddress.ANY == dstAddress) {
			// Broadcast -> activate all radios.
			for (LazyRadio radio : radios_) {
				activateRxNode(now, radio, txInfo);
			}
		} else if (Partition.getPartitionIdOfNode(dstAddress) == partitionId_) {
			// Unicast -> activate destination radio if it belongs to this
			// partition.
			activateRxNode(now, (LazyRadio) WirelessNode.getNode(dstAddress)
					.getRadio(), txInfo);
		}
		// Clean up history.
		cleanHistoryUp();
		// Create new signal history entry.
		LazySignal signal = new LazySignal(now, txInfo);
		HistoryEntry entry = addHistoryEntry(signal, true);
		if (USE_MONITOR) {
			monitor_.monitorEvent();
		}
		// Notify active nodes immediately.
		int counter = 0;
		for (LazyRadio radio : radios_) {
			if (isActive(radio)) {
				counter++;
				final int rxNodeId = radio.getNumber();
				if (rxNodeId != txInfo.getTxNodeId()) {
					if (field_.isReceived(rxNodeId, txInfo)) {
						radio.activate(now, txInfo.getTxModuleId(),
								RadioNoise.Events.RECEIVE, txInfo);
					}
				}
			}
		}
		entry.incHasSeenCounter(counter);
		return signal;
	}

	/**
	 * Finish a signal. The signal-end-event will be added to the signal history
	 * and all active lazy nodes will be notified by an END_RECEIVE event.
	 * 
	 * @param now
	 *            the current simulation time (at which the signal ends)
	 * @param signal
	 *            signal to end
	 */
	void finishSignal(long now, LazySignal signal) {
		assert signal.getEndTime() == now;
		// Create new signal history entry and notify active nodes.
		HistoryEntry entry = addHistoryEntry(signal, false);
		if (USE_MONITOR) {
			monitor_.monitorEvent();
		}
		int counter = 0;
		for (LazyRadio radio : radios_) {
			if (isActive(radio)) {
				counter++;
				final int rxNodeId = radio.getNumber();
				if (rxNodeId != signal.getTxNodeId()) {
					final TransmissionInfo txInfo = signal.getTxInfo();
					if (field_.isReceived(rxNodeId, txInfo)) {
						radio.activate(now, signal.getTxModuleId(),
								RadioNoise.Events.END_RECEIVE, txInfo);
					}
				}
			}
		}
		entry.incHasSeenCounter(counter);
	}

	/**
	 * Sets the active/deactive state of the given radio.
	 * 
	 * @param radio
	 *            the radio that is about to be activated
	 * @param active
	 *            whether the radio state is set to active
	 */
	private void setActive(LazyRadio radio, boolean active) {
		isActive_[radio.getNumber()] = active;
	}

	/**
	 * Activate a lazy node as receiver of a signal. If the transmitter is too
	 * far away for the signal to be received by this node, it will not be
	 * activated. If the node is already active, nothing will happen. On
	 * activation, the MAC state of the node will be calculated using the signal
	 * history. The node will be notified of the activation.
	 * 
	 * @param now
	 *            the current simulation time (at which the node is activated)
	 * @param radio
	 *            the lazy radio to be activated
	 * @param txInfo
	 *            the transmission info of the currently received signal
	 */
	private void activateRxNode(long now, LazyRadio radio,
			TransmissionInfo txInfo) {
		// Ignore the transmitting node.
		final int rxNodeId = radio.getNumber();
		if (rxNodeId == txInfo.getTxNodeId()) {
			return;
		}
		// Check whether this node is already active.
		if (isActive(radio)) {
			return;
		}
		// Check with field if the signal is actually received by the node.
		if (!field_.isReceived(rxNodeId, txInfo)) {
			return;
		}
		// This node has to be activated.
		activateNode(now, radio);
	}

	/**
	 * Called when a node is to be activated.
	 * 
	 * @param now
	 *            the current simulation time
	 * @param radio
	 *            the lazy radio to be activated
	 */
	private void activateNode(long now, LazyRadio radio) {
		assert !isActive(radio);
		final int rxNodeId = radio.getNumber();
		boolean doActivate = true;
		if (firstIndex_ != firstFreeIndex_) {
			// Compute the first relevant entry for this node.
			int startIndex;
			final int diff = lastEntryIds_[rxNodeId]
					- history_[firstIndex_].getEntryId();
			if (diff < 0) {
				// Last event of this radio was before the signal history.
				assert diff == -1;
				startIndex = firstIndex_;
			} else {
				// Identify index of the first signal unprocessed by radio.
				startIndex = firstIndex_ + diff + 1;
			}
			if (startIndex < firstFreeIndex_) {
				radio.startRetro();
				// Iterate through signal history, calculating NAV and
				// EIFS-flag.
				int helper = -1;
				for (int i = startIndex; i < firstFreeIndex_; i++) {
					HistoryEntry entry = history_[i];
					if (entry.getSimTime() == now) {
						// Simultaneous events handling.
						helper = i;
						break;
					}
					radio.updateRetro(entry);
					entry.incHasSeenCounter(1);
				}
				doActivate = radio.endRetro(now);
				if (helper != -1) {
					for (int i = helper; i < firstFreeIndex_; i++) {
						HistoryEntry entry = history_[i];
						LazySignal signal = entry.getSignal();
						if (rxNodeId != signal.getTxNodeId()) {
							TransmissionInfo txInfo = signal.getTxInfo();
							if (field_.isReceived(rxNodeId, txInfo)) {
								radio.activate(
										now,
										txInfo.getTxModuleId(),
										entry.doesSignalBegin() ? RadioNoise.Events.RECEIVE
												: RadioNoise.Events.END_RECEIVE,
										txInfo);
							}
						}
						entry.incHasSeenCounter(1);
					}
				}
			} else {
				radio.handleSpecialCaseNoEntries();
			}
		} else {
			radio.handleSpecialCaseNoEntries();
		}
		if (doActivate) {
			setActive(radio, true);
			if (USE_MONITOR) {
				monitor_.monitorActivation(now);
			}
			assert isActive(radio);
		} else {
			lastEntryIds_[radio.getNumber()] = currentEntryId_;
			assert !isActive(radio);
		}
	}

	/**
	 * Removes entries that are not needed anymore from the signal history.
	 */
	private void cleanHistoryUp() {
		// PP NOTE: It may make sense to add clean up at zero-signal moments in
		// some way. This could improve memory usage.
		if (firstIndex_ == firstFreeIndex_) {
			return;
		}
		int index = firstIndex_ - 1;
		for (int i = firstIndex_; i < firstFreeIndex_; i++) {
			final HistoryEntry entry = history_[i];
			if (entry.haveSeen(numMyNodes_)) {
				index = i;
			} else {
				return;
			}
		}
		if (index >= firstIndex_) {
			firstIndex_ = index + 1;
			if (firstIndex_ > MAX_SHIFT) {
				shiftHistoryBack();
			}
		}
	}

	/**
	 * Adds a new entry to the signal history.
	 * 
	 * @param signal
	 *            the signal information to be added
	 * @param signalBegins
	 *            whether the signal currently begins (or ends)
	 * @return the new history entry
	 */
	private HistoryEntry addHistoryEntry(LazySignal signal, boolean signalBegins) {
		HistoryEntry entry = new HistoryEntry(++currentEntryId_, signal,
				signalBegins);
		history_[firstFreeIndex_++] = entry;
		// Ensure capacity and shift back.
		if (firstFreeIndex_ == history_.length) {
			HistoryEntry[] history = new HistoryEntry[history_.length * 2];
			final int size = firstFreeIndex_ - firstIndex_;
			System.arraycopy(history_, firstIndex_, history, 0, size);
			history_ = history;
			firstIndex_ = 0;
			firstFreeIndex_ = size;
		}
		return entry;
	}

	/**
	 * Shifts the contents of the signal history back in the array used for
	 * storing it. This method is called to avoid the array from growing.
	 */
	private void shiftHistoryBack() {
		final int size = firstFreeIndex_ - firstIndex_;
		System.arraycopy(history_, firstIndex_, history_, 0, size);
		for (int i = size; i < firstFreeIndex_; i++) {
			history_[i] = null;
		}
		firstIndex_ = 0;
		firstFreeIndex_ = size;
	}

	/**
	 * Prints statistics, if applicable.
	 */
	void printLazyStats() {
		if (USE_MONITOR) {
			monitor_.printLazyStats();
		}
	}

	/**
	 * An entry in the signal history.
	 * 
	 * @author Christoph Sprenger
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	static class HistoryEntry {
		/**
		 * Unique (increasing) ID of this entry.
		 */
		private final int entryId_;

		/**
		 * The signal this entry corresponds to (either beginning or end).
		 */
		private final LazySignal signal_;

		/**
		 * Whether this entry indicates the beginning of the signal.
		 */
		private final boolean signalBegins_;

		/**
		 * The simulation time of this entry.
		 */
		private final long simTime_;

		/**
		 * How many of my partition's nodes have already processed this entry.
		 */
		private int hasSeenCounter_;

		/**
		 * Constructor.
		 * 
		 * @param entryId
		 *            a unique ID for this entry.
		 * @param signal
		 *            signal this entry corresponds to
		 * @param signalBegins
		 *            whether this entry indicates the beginning of the signal
		 */
		private HistoryEntry(int entryId, LazySignal signal,
				boolean signalBegins) {
			entryId_ = entryId;
			signal_ = signal;
			signalBegins_ = signalBegins;
			simTime_ = (signalBegins_) ? signal_.getStartTime() : signal_
					.getEndTime();
			hasSeenCounter_ = 0;
		}

		/**
		 * Whether this entry indicates the beginning of the signal.
		 * 
		 * @return whether this entry indicates the beginning of the signal
		 */
		boolean doesSignalBegin() {
			return signalBegins_;
		}

		/**
		 * Provides the simulation time of this entry.
		 * 
		 * @return the simulation time of this entry
		 */
		long getSimTime() {
			return simTime_;
		}

		/**
		 * Provides the signal of this entry.
		 * 
		 * @return the signal of this entry
		 */
		LazySignal getSignal() {
			return signal_;
		}

		/**
		 * Provides the ID of this entry.
		 * 
		 * @return the ID of this entry
		 */
		private int getEntryId() {
			return entryId_;
		}

		/**
		 * Increments the counter by the specified amount.
		 * 
		 * @param numNodes
		 *            the increment to the counter
		 */
		private void incHasSeenCounter(int numNodes) {
			hasSeenCounter_ += numNodes;
		}

		/**
		 * Whether the specified number of nodes have already processed this
		 * signal.
		 * 
		 * @param numNodes
		 *            the number of nodes
		 * @return whether the specified number of nodes have already processed
		 *         this signal
		 */
		private boolean haveSeen(int numNodes) {
			return hasSeenCounter_ == numNodes;
		}

		/**
		 * @see java.lang.Object#toString()
		 */
		public String toString() {
			return "HistoryEntry_" + entryId_ + "_" + simTime_
					+ (signalBegins_ ? "_START_" : "_END_") + signal_;
		}
	}

	/**
	 * Helper class for monitoring activations and deactivations, if desired.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	private static class LazyMonitor {
		/**
		 * Current number of active nodes.
		 */
		private int numActiveNodes_;

		/**
		 * Last time a node was activated or deactivated.
		 */
		private long lastUpdateTime_;

		/**
		 * The weighted number of active nodes over time.
		 */
		private long weightedActiveTime_;

		/**
		 * The weighted number of active nodes at specific events
		 */
		private long weightedActiveEvents_;

		/**
		 * Counts the number of monitor events.
		 */
		private long numMonitorEvents_;

		/**
		 * Constructor.
		 */
		private LazyMonitor() {
			numActiveNodes_ = 0;
			lastUpdateTime_ = 0;
			weightedActiveTime_ = 0;
			weightedActiveEvents_ = 0;
			numMonitorEvents_ = 0;
		}

		/**
		 * Prints statistics.
		 */
		private void printLazyStats() {
			weightedActiveTime_ += (Config.getEndTime() - lastUpdateTime_)
					* numActiveNodes_;
			weightedActiveEvents_ /= numMonitorEvents_;
			assert Config.getStartTime() == 0;
			CustomLogger.info("av. num. active nodes:   "
					+ (weightedActiveTime_ /= Config.getEndTime()));
			CustomLogger.info("av. num. active nodes 2  "
					+ weightedActiveEvents_);
		}

		/**
		 * Called when a node is activated.
		 * 
		 * @param now
		 *            the time of the activation
		 */
		private void monitorActivation(long now) {
			if (now > lastUpdateTime_) {
				weightedActiveTime_ += (now - lastUpdateTime_)
						* numActiveNodes_;
				lastUpdateTime_ = now;
			}
			numActiveNodes_++;
		}

		/**
		 * Called when a node is deactivated.
		 * 
		 * @param now
		 *            the time of the deactivation
		 */
		private void monitorDeactivation(long now) {
			if (now > lastUpdateTime_) {
				weightedActiveTime_ += (now - lastUpdateTime_)
						* numActiveNodes_;
				lastUpdateTime_ = now;
			}
			numActiveNodes_--;
		}

		/**
		 * Monitors an event where the number of active nodes is of interest.
		 */
		private void monitorEvent() {
			weightedActiveEvents_ += numActiveNodes_;
			numMonitorEvents_++;
		}
	}
}
