/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/lazy/LazyParallelPartition.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.lazy;

import jist.swans.radio.RadioNoise;
import modules.wireless.partition.TransmissionInfo;
import modules.wireless.partition.ParallelPartition;
import modules.wireless.partition.Partition;
import core.Config;

/**
 * A partition to be used by parallel simulations using Lazy MAC State Update.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class LazyParallelPartition extends ParallelPartition implements
		LazyPartition {
	/**
	 * Setup node based partitions using lazy mac state update.
	 * 
	 * @param numPartitions
	 *            the requested number of partitions
	 * @param numNodes
	 *            the total number of nodes of the simulation
	 */
	public static void setup(int numPartitions, int numNodes) {
		assert partitions == null;
		nodeIdToPartitionId = new int[numNodes];
		partitions = new Partition[numPartitions];
		for (int i = 0; i < numPartitions; ++i) {
			new LazyParallelPartition(i, numNodes);
		}
	}

	/**
	 * Everything we need for Lazy MAC State Update.
	 */
	private final LazyManager myLazyManager_;

	/**
	 * Constructor.
	 * 
	 * @param partitionId
	 *            the id of this partition
	 * @param numNodes
	 *            the total number of nodes of the simulation
	 */
	private LazyParallelPartition(int partitionId, int numNodes) {
		super(partitionId);
		myLazyManager_ = new LazyManager(partitionId, numNodes, field_);
	}

	/**
	 * @see modules.wireless.partition.Partition#init()
	 */
	@Override
	public void init() {
		super.init();
		myLazyManager_.setNodes(nodes_);
	}

	/**
	 * @see modules.wireless.partition.ParallelPartition#scheduleEndReceiveEvent(long,
	 *      jist.swans.radio.RadioNoise,
	 *      modules.wireless.partition.TransmissionInfo)
	 */
	@Override
	public void scheduleEndReceiveEvent(long now, RadioNoise radio,
			TransmissionInfo txInfo) {
		// Do nothing in case of Lazy MAC (handled by LazyLists).
	}

	/**
	 * @see modules.wireless.partition.ParallelPartition#lpPropagation(long,
	 *      int, modules.wireless.partition.TransmissionInfo)
	 */
	@Override
	protected void lpPropagation(long now, int srcModuleId,
			TransmissionInfo txInfo) {
		// Create signal list entry (node activation is handled by LazyLists).
		final LazySignal signal = myLazyManager_.addSignal(now, txInfo);
		final long minPropDelay = Config.getPropagation().getDelay();
		final long shiftedBackTime = now - minPropDelay;
		scheduleNewEventWithSrcModuleIdAndSchedTime(now + txInfo.getDuration(),
				getId(), Events.END_LPPROPAGATION, signal, srcModuleId,
				shiftedBackTime);
	}

	/**
	 * @see modules.wireless.partition.ParallelPartition#endLpPropagation(long,
	 *      java.lang.Object)
	 */
	@Override
	protected void endLpPropagation(long now, Object parameter) {
		myLazyManager_.finishSignal(now, (LazySignal) parameter);
	}

	/**
	 * @see modules.wireless.lazy.LazyPartition#activateTxNode(long,
	 *      modules.wireless.lazy.LazyRadio)
	 */
	@Override
	public void activateTxNode(long now, LazyRadio radio) {
		myLazyManager_.activateTxNode(now, radio);
	}

	/**
	 * @see modules.wireless.lazy.LazyPartition#deactivateNode(long,
	 *      modules.wireless.lazy.LazyRadio)
	 */
	@Override
	public void deactivateNode(long now, LazyRadio radio) {
		myLazyManager_.deactivateNode(now, radio);
	}

	/**
	 * @see modules.wireless.lazy.LazyPartition#isActive(modules.wireless.lazy.LazyRadio)
	 */
	public boolean isActive(LazyRadio radio) {
		return myLazyManager_.isActive(radio);
	}
}
