/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/lazy/LazyRadio.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.lazy;

import jist.swans.radio.RadioNoise;
import modules.wireless.lazy.LazyManager.HistoryEntry;

/**
 * Extends the detailed radio noise module for simulations using Lazy MAC State
 * Update.
 * 
 * @author Patrick Peschlow
 * @author Christoph Sprenger
 * 
 * @version 1.0
 */
public class LazyRadio extends RadioNoise {
	/**
	 * The lazy MAC object belonging to this radio.
	 */
	private LazyMac802_11 myLazyMac_;

	/**
	 * Constructor.
	 * 
	 * @param number
	 *            the node ID of this node
	 */
	public LazyRadio(int number) {
		super(number, LazyMac802_11.class);
	}

	/**
	 * @see jist.swans.radio.RadioNoise#init()
	 */
	@Override
	public void init() {
		super.init();
		myLazyMac_ = (LazyMac802_11) myMac_;
	}

	/**
	 * Sets the radio mode.
	 * 
	 * @param mode
	 *            the new mode.
	 */
	void setMode(RadioMode mode) {
		mode_ = mode;
	}

	/**
	 * Whether the last reception had bit errors.
	 * 
	 * @return whether the last reception had bit errors
	 */
	boolean hasRcvError() {
		return hasRcvError_;
	}

	/**
	 * Starts a retrospective. Called when I try to get activated.
	 */
	void startRetro() {
		myLazyMac_.startRetro();
	}

	/**
	 * Ends a retrospective. Called when all history updates have been made.
	 * 
	 * @param now
	 *            the current simulation time
	 * @return whether I get activated
	 */
	boolean endRetro(long now) {
		return myLazyMac_.endRetro(now);
	}

	/**
	 * Performs a step in the retrospective by evaluating the given history
	 * entry.
	 * 
	 * @param entry
	 *            the history entry to evaluate
	 */
	void updateRetro(HistoryEntry entry) {
		myLazyMac_.updateRetro(entry);
	}

	/**
	 * Handles a special case where no entries have been added to the signal
	 * history since my last deactivation.
	 */
	void handleSpecialCaseNoEntries() {
		myLazyMac_.handleSpecialCaseNoEntries();
	}
}
