/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/lazy/LazySignal.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.lazy;

import jist.swans.Constants;
import jist.swans.mac.MacMessage;
import modules.wireless.partition.TransmissionInfo;

/**
 * Data structure for a signal stored in the signal history.
 * 
 * @author Patrick Peschlow
 * @author Christoph Sprenger
 * 
 * @version 1.0
 */
public class LazySignal {
	/**
	 * The TransmissionInfo corresponding to this signal.
	 */
	private final TransmissionInfo txInfo_;

	/**
	 * The start time of (the reception of) this signal.
	 */
	private final long startTime_;

	/**
	 * The end time of (the reception of) this signal.
	 */
	private final long endTime_;

	/**
	 * The NAV associated with this signal.
	 */
	private final long nav_;

	/**
	 * Constructor.
	 * 
	 * @param startTime
	 *            the time of reception of the first bit (= now)
	 * @param txInfo
	 *            the transmission information
	 */
	LazySignal(long startTime, TransmissionInfo txInfo) {
		txInfo_ = txInfo;
		startTime_ = startTime;
		endTime_ = startTime + txInfo.getDuration();
		nav_ = endTime_ + ((MacMessage) txInfo.getMsg()).getDuration()
				+ Constants.EPSILON_DELAY; // see Mac802_11
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return txInfo_.toString();
	}

	/**
	 * Provides the start time of this signal.
	 * 
	 * @return the start time of this signal.
	 */
	long getStartTime() {
		return startTime_;
	}

	/**
	 * Provides the end time of this signal.
	 * 
	 * @return the end time of this signal.
	 */
	long getEndTime() {
		return endTime_;
	}

	/**
	 * Provides the NAV information of this signal.
	 * 
	 * @return the NAV information of this signal.
	 */
	long getNav() {
		return nav_;
	}

	/**
	 * Provides the node ID of the transmitter of this signal.
	 * 
	 * @return the node ID of the transmitter of this signal.
	 */
	int getTxNodeId() {
		return txInfo_.getTxNodeId();
	}

	/**
	 * Provides the node ID of the destination of the message.
	 * 
	 * @return the node ID of the destination of the message.
	 */
	int getDstNodeId() {
		return ((MacMessage) txInfo_.getMsg()).getDst();
	}

	/**
	 * Provides the module ID of the transmitting radio of this signal.
	 * 
	 * @return the module ID of the transmitting radio of this signal.
	 */
	int getTxModuleId() {
		return txInfo_.getTxModuleId();
	}

	/**
	 * Provides the transmission information of this signal.
	 * 
	 * @return the transmission information of this signal.
	 */
	TransmissionInfo getTxInfo() {
		return txInfo_;
	}
}
