/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/mesh/DummyNode.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.mesh;

import gnu.trove.TIntIntHashMap;

import java.util.ArrayList;
import java.util.List;

import jist.swans.net.DummyNetIp;
import jist.swans.net.PacketLoss;
import jist.swans.trans.TcpSocket;
import jist.swans.trans.TransTcp;
import jist.swans.trans.TransUdp;
import modules.wireless.app.PacketGenerator;
import modules.wireless.app.PacketGenerator.ConnectionType;
import modules.wireless.scenario.DummyModule;
import modules.wireless.scenario.NodeConfig;
import modules.wireless.scenario.NodeConfig.FlowPriority;
import modules.wireless.scenario.WirelessNode;
import rng.RandomManager;
import core.BaseModule;
import core.ModuleRoot;

/**
 * Dummy node simulating endpoints for client connections to the internet.
 * 
 * @author Patrick Peschlow
 * @author Daniel Plohmann
 * 
 * @version 1.0
 */
public class DummyNode {
	/**
	 * A list of all dummy nodes of the simulation.
	 */
	private static List<DummyNode> dummyNodes = new ArrayList<DummyNode>();

	/**
	 * Maps dummy ("internet") addresses to node IDs.
	 */
	private static TIntIntHashMap dummyToIndex = new TIntIntHashMap();

	/**
	 * Get a reference to the dummy node with a specific number.
	 * 
	 * @param dummyNodeId
	 *            the ID of the dummy node.
	 * @return dummy node with the specified number.
	 */
	public static DummyNode getNode(int dummyNodeId) {
		int index = dummyToIndex.get(dummyNodeId);
		assert dummyNodes.get(index) != null;
		return dummyNodes.get(index);
	}

	/**
	 * Creates a new dummy node.
	 * 
	 * @param relatedClientNode
	 *            the ID of the other connection endpoint node.
	 * @param dummyId
	 *            the dummy node ID for this dummy.
	 * @param partitionId
	 *            ID of the partition this dummy is managed by.
	 * @param portalNodeId
	 *            ID of the portal of this dummy.
	 * @return the dummy ID.
	 */
	public static int addDummy(int relatedClientNode, int dummyId,
			int partitionId, int portalNodeId) {
		assert dummyId == dummyNodes.size();
		DummyNode newDummy = new DummyNode(relatedClientNode, dummyId,
				partitionId, portalNodeId);
		dummyNodes.add(newDummy);
		final int number = newDummy.getNumber();
		dummyToIndex.put(number, dummyId);
		WirelessNode.registerDummy(portalNodeId, number, newDummy.net_.getId());
		return number;
	}

	/**
	 * Creates a new remote dummy node.
	 * 
	 * @param relatedClientNode
	 *            the ID of the other connection endpoint node.
	 * @param dummyId
	 *            the dummy node ID for this dummy.
	 * @return the dummy ID.
	 */
	public static int addRemoteDummy(int relatedClientNode, int dummyId) {
		assert dummyId == dummyNodes.size();

		// Network.
		DummyModule.registerDummyModule();
		// Transport.
		DummyModule.registerDummyModule();
		DummyModule.registerDummyModule();
		RandomManager.getInstance().getUniform01Substream();
		// Packet generator.
		DummyModule.registerDummyModule();

		dummyNodes.add(null);
		final int number = 2000000 + dummyId;
		dummyToIndex.put(number, dummyId);
		return number;
	}

	/**
	 * Provides the number of bytes sent via TCP flows of all dummies.
	 * 
	 * @return the number of bytes sent via TCP flows.
	 */
	public static long getTcpBytesSent() {
		long bytes = 0;
		for (DummyNode node : dummyNodes) {
			if (node != null) {
				bytes += node.pgen_.getTcpBytesSent();
			}
		}
		return bytes;
	}

	/**
	 * Provides the number of bytes received via TCP flows of all dummies.
	 * 
	 * @return the number of bytes received via TCP flows.
	 */
	public static long getTcpBytesReceived() {
		long bytes = 0;
		for (DummyNode node : dummyNodes) {
			if (node != null) {
				bytes += node.pgen_.getTcpBytesReceived();
			}
		}
		return bytes;
	}

	/**
	 * Provides the number of bytes sent via UDP flows of all dummies.
	 * 
	 * @return the number of bytes sent via UDP flows.
	 */
	public static long getUdpBytesSent() {
		long bytes = 0;
		for (DummyNode node : dummyNodes) {
			if (node != null) {
				bytes += node.pgen_.getUdpBytesSent();
			}
		}
		return bytes;
	}

	/**
	 * Provides the number of bytes received via UDP flows of all dummies.
	 * 
	 * @return the number of bytes received via UDP flows.
	 */
	public static int getUdpBytesReceived() {
		int bytes = 0;
		for (DummyNode node : dummyNodes) {
			if (node != null) {
				bytes += node.pgen_.getUdpBytesReceived();
			}
		}
		return bytes;
	}

	/**
	 * Provides the number of transmissions sent by the IP layers of all
	 * dummies.
	 * 
	 * @return the number of transmissions sent by the IP layers of all dummies.
	 */
	public static int getDummyTransmissions() {
		int transmissions = 0;
		for (DummyNode node : dummyNodes) {
			if (node != null) {
				transmissions += node.net_.getNumTransmissions();
			}
		}
		return transmissions;
	}

	/**
	 * The ID of this dummy.
	 */
	private final int number_;

	/**
	 * The id of the partition this dummy is associated with.
	 */
	private final int partitionId_;

	/**
	 * List of the modules of the dummy.
	 */
	private final List<BaseModule> modules_;

	/**
	 * The packet generator of this dummy.
	 */
	private final PacketGenerator pgen_;

	/**
	 * The UDP module of this dummy.
	 */
	private final TransUdp udp_;

	/**
	 * The TCP socket module of this dummy.
	 */
	private TcpSocket srcSock_;

	/**
	 * The TCP module of this dummy.
	 */
	private final TransTcp tcp_;

	/**
	 * The IP layer of this dummy.
	 */
	private final DummyNetIp net_;

	/**
	 * Creates a new dummy node.
	 * 
	 * @param relatedClientNode
	 *            the ID of the other connection endpoint node.
	 * @param dummyId
	 *            the dummy node ID for this dummy.
	 * @param partitionId
	 *            my partition ID.
	 * @param portalNodeId
	 *            ID of my portal.
	 */
	private DummyNode(int relatedClientNode, int dummyId, int partitionId,
			int portalNodeId) {
		number_ = 2000000 + dummyId;
		partitionId_ = partitionId;

		modules_ = new ArrayList<BaseModule>();

		// Create network layer.
		final PacketLoss.Zero in = new PacketLoss.Zero();
		final PacketLoss.Zero out = new PacketLoss.Zero();
		net_ = new DummyNetIp(number_, portalNodeId, in, out);
		modules_.add(net_);

		// Create transport layer.
		udp_ = new TransUdp(number_, DummyNetIp.class, PacketGenerator.class);
		modules_.add(udp_);
		tcp_ = new TransTcp(number_, DummyNetIp.class, PacketGenerator.class);
		modules_.add(tcp_);

		// Create packet generator.
		pgen_ = new PacketGenerator(number_);
		modules_.add(pgen_);

		// Set LP allocation IDs.
		for (BaseModule module : modules_) {
			ModuleRoot.getRoot().allocateModuleToLp(module, partitionId);
		}
	}

	/**
	 * Provides the node ID of this dummy.
	 * 
	 * @return the node ID of this dummy.
	 */
	public int getNumber() {
		return number_;
	}

	/**
	 * Provides the packet generator of this dummy.
	 * 
	 * @return the packet generator of this dummy.
	 */
	public PacketGenerator getPgen() {
		return pgen_;
	}

	/**
	 * Initializes a UDP traffic flow between two nodes. The packet generator of
	 * the sending node will be notified of the new traffic flow.
	 * 
	 * @param connectionId
	 *            ID identifying the connection.
	 * @param dstNodeId
	 *            ID of the destination node.
	 * @param startTime
	 *            simulation time the first packet will be sent.
	 * @param packetSize
	 *            packet size of each packet generated.
	 * @param sendInterval
	 *            send interval.
	 * @param priority
	 *            flow priority.
	 */
	public void startUdpFlow(int connectionId, int dstNodeId, long startTime,
			int packetSize, long sendInterval, FlowPriority priority) {
		assert NodeConfig.getTrafficInfoUDP() != null;
		pgen_.addConnection(ConnectionType.UDP, connectionId, startTime,
				dstNodeId, packetSize, sendInterval, priority);
	}

	/**
	 * Creates a TCP traffic source.
	 * 
	 * @param connectionId
	 *            ID identifying the connection.
	 * @param srcNodeId
	 *            ID of the source node.
	 * @param dstNodeId
	 *            ID of the destination node.
	 * @param srcPort
	 *            source port.
	 * @param dstPort
	 *            destination port.
	 * @param startTime
	 *            simulation time the first packet will be sent.
	 * @param packetSize
	 *            packet size of each packet generated.
	 * @param sendInterval
	 *            send interval.
	 * @param priority
	 *            flow priority.
	 */
	public void createTcpSource(int connectionId, int srcNodeId, int dstNodeId,
			int srcPort, int dstPort, long startTime, int packetSize,
			long sendInterval, FlowPriority priority) {
		assert NodeConfig.getTrafficInfoTCP() != null;
		srcSock_ = new TcpSocket(connectionId, srcNodeId, dstNodeId, srcPort,
				dstPort, tcp_, packetSize, sendInterval == 0);
		ModuleRoot.getRoot().allocateModuleToLp(srcSock_, partitionId_);
		modules_.add(srcSock_);
		pgen_.addConnection(ConnectionType.TCP, connectionId, startTime,
				dstNodeId, packetSize, sendInterval, priority);
	}

	/**
	 * Creates a TCP traffic sink.
	 * 
	 * @param connectionId
	 *            ID identifying the connection.
	 * @param srcNodeId
	 *            ID of the source node.
	 * @param dstNodeId
	 *            ID of the destination node.
	 * @param packetSize
	 *            packet size of each packet generated.
	 * @param srcPort
	 *            source port.
	 * @param dstPort
	 *            destination port.
	 */
	public void createTcpSink(int connectionId, int srcNodeId, int dstNodeId,
			int srcPort, int dstPort, int packetSize) {
		assert NodeConfig.getTrafficInfoTCP() != null;
		TcpSocket dstSock = new TcpSocket(connectionId, dstNodeId, srcNodeId,
				dstPort, srcPort, tcp_, packetSize, false);
		ModuleRoot.getRoot().allocateModuleToLp(dstSock, partitionId_);
		modules_.add(dstSock);
	}

}
