/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/mesh/MeshPortal.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.mesh;

import gnu.trove.TIntIntHashMap;
import jist.swans.Constants;
import jist.swans.mac.MacMessage;
import jist.swans.net.NetIp;
import jist.swans.net.NetIpAP;
import jist.swans.net.NetMessage;
import rng.RandomManager;
import rng.RandomStream;
import core.BaseModule;
import core.EventParameterTuple;

/**
 * Simple mesh portal implementation.
 * 
 * @author Patrick Peschlow
 * @author Daniel Plohmann
 * 
 * @version 1.0
 */
public class MeshPortal extends BaseModule {
	/**
	 * The event types offered by this module.
	 * 
	 * @author Patrick Peschlow
	 * @author Daniel Plohmann
	 * 
	 * @version 1.0
	 */
	public static enum Events {
		/**
		 * An incoming packet has to be forwarded.
		 */
		INCOMING_PACKET,
		/**
		 * An outgoing packet has to be forwarded.
		 */
		OUTGOING_PACKET;
	}

	/**
	 * Mean delay for communication between portals and internet dummies.
	 */
	private static final long INTERNET_DELAY_MEAN = 5 * Constants.MICRO_SECOND;

	/**
	 * Standard deviation of the delay for communication between portals and
	 * internet dummies.
	 */
	private static final long INTERNET_DELAY_STD = 1 * Constants.MICRO_SECOND;

	/**
	 * Minimum delay for communication between portals and internet dummies.
	 */
	private static final long INTERNET_DELAY_MIN = 5 * Constants.MICRO_SECOND;

	/**
	 * The current simulation time.
	 */
	private long now_;

	/**
	 * Node ID of my AP network layer.
	 */
	private int myNetIpAPId_;

	/**
	 * Key: Dummy ID. Value: Module ID of the network layer of that dummy.
	 */
	private TIntIntHashMap dummyIdToNetModuleId_;

	/**
	 * RNG used for calculating delays.
	 */
	private RandomStream rn_;

	/**
	 * Constructor.
	 * 
	 * @param number
	 *            ID of the portal.
	 */
	public MeshPortal(int number) {
		super(number);
		dummyIdToNetModuleId_ = new TIntIntHashMap();
		rn_ = RandomManager.getInstance().getNormalSubstream(
				MeshPortal.INTERNET_DELAY_MEAN, MeshPortal.INTERNET_DELAY_STD);

	}

	/**
	 * @see core.BaseModule#activate(long, int, java.lang.Enum,
	 *      java.lang.Object)
	 */
	@SuppressWarnings("rawtypes")
	@Override
	public void activate(long execTime, int srcModuleId, Enum eventId,
			Object parameter) {
		now_ = execTime;
		if (eventId == Events.INCOMING_PACKET) {
			EventParameterTuple param = (EventParameterTuple) parameter;
			NetMessage msg = (NetMessage) param.getParam1();
			forwardIncoming(msg);
		} else if (eventId == Events.OUTGOING_PACKET) {
			EventParameterTuple param = (EventParameterTuple) parameter;
			NetMessage msg = (NetMessage) param.getParam1();
			forwardOutgoing(msg);
		}
	}

	/**
	 * @see core.BaseModule#init()
	 */
	@Override
	public void init() {
		myNetIpAPId_ = findModuleId(NetIpAP.class, number_);
	}

	/**
	 * Registers a new dummy with this portal.
	 * 
	 * @param dummyId
	 *            the internet ID of the dummy.
	 * @param netModuleId
	 *            the module ID of the network module of the dummy.
	 */
	public void registerDummy(int dummyId, int netModuleId) {
		dummyIdToNetModuleId_.put(dummyId, netModuleId);
	}

	/**
	 * Whether the specified dummy is managed by this portal.
	 * 
	 * @param dummyId
	 *            the dummy internet ID.
	 * @return whether the specified dummy is managed by this portal.
	 */
	public boolean hasDummy(int dummyId) {
		return dummyIdToNetModuleId_.containsKey(dummyId);
	}

	/**
	 * Forwards a packet.
	 * 
	 * @param msg
	 *            the IP packet to forward.
	 */
	private void forwardIncoming(NetMessage msg) {
		final long delay = calculateDelay();
		MacMessage wrap = new MacMessage.Data(0, 0, 0, msg);
		scheduleNewEvent(now_ + delay, myNetIpAPId_,
				NetIpAP.Events.RECEIVE_INCOMING, wrap);
	}

	/**
	 * Forwards a packet.
	 * 
	 * @param msg
	 *            the IP packet to forward.
	 */
	private void forwardOutgoing(NetMessage msg) {
		final long delay = calculateDelay();
		assert hasDummy(msg.getDst().getIntIP());
		int dummyNetId = dummyIdToNetModuleId_.get(msg.getDst().getIntIP());
		MacMessage wrap = new MacMessage.Data(0, 0, 0, msg);
		scheduleNewEvent(now_ + delay, dummyNetId,
				NetIp.Events.RECEIVE_INCOMING, wrap);
	}

	/**
	 * Calculate normally distributed delay, with a minimum delay.
	 * 
	 * @return long delay
	 */
	private long calculateDelay() {
		return (long) Math.max(rn_.nextDouble(), INTERNET_DELAY_MIN);
	}

}
