/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/partition/DistributedPartition.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.partition;

import jist.swans.radio.RadioNoise;
import modules.wireless.scenario.DummyModule;
import core.Config;

/**
 * A partition to be used in distributed simulations.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class DistributedPartition extends ParallelPartition {
	/**
	 * Constant indicating a broadcast to all partitions in a distributed
	 * simulation.
	 */
	public static final int DIST_PARTITION_BROADCAST = -1;

	/**
	 * Sets up the specified number of partitions for the specified total number
	 * of nodes.
	 * 
	 * @param numPartitions
	 *            the number of partitions.
	 * @param numNodes
	 *            the number of nodes of the simulation.
	 * @param firstPartitionId
	 *            the first ID of the sequence of partitions on this host.
	 * @param lastPartitionId
	 *            the last ID of the sequence of partitions on this host.
	 */
	public static void setup(int numPartitions, int numNodes,
			int firstPartitionId, int lastPartitionId) {
		assert (partitions == null);
		nodeIdToPartitionId = new int[numNodes];
		partitions = new Partition[numPartitions];
		for (int i = 0; i < numPartitions; i++) {
			if (i < firstPartitionId || i > lastPartitionId) {
				DummyModule.registerDummyModule();
			} else {
				new DistributedPartition(i);
			}
		}
	}

	/**
	 * Constructor.
	 * 
	 * @param partitionId
	 *            the ID of this partition.
	 */
	protected DistributedPartition(int partitionId) {
		super(partitionId);
	}

	/**
	 * @see modules.wireless.partition.Partition#propagateReception(jist.swans.radio.RadioNoise,
	 *      modules.wireless.partition.TransmissionInfo)
	 */
	@Override
	public void propagateReception(RadioNoise radio, TransmissionInfo txInfo) {
		// The signal reception is propagated to all LPs by means of an
		// LPPROPAGATION event. It is scheduled with the minimum propagation
		// delay.
		final long shiftedTime = txInfo.getStartTime()
				+ Config.getPropagation().getDelay();
		// In distributed simulation, only a single event is created and sent to
		// all partitions via the server. Each partition (LP) then creates its
		// own copy of the event.
		scheduleNewEventWithSrcModuleIdAndSchedTime(shiftedTime,
				DIST_PARTITION_BROADCAST, Partition.Events.LPPROPAGATION,
				txInfo, radio.getId(), 0);
	}
}
