/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/partition/ParallelPartition.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.partition;

import jist.swans.Constants;
import jist.swans.misc.Location;
import jist.swans.radio.RadioNoise;
import modules.wireless.scenario.NodeMovementInfo;
import core.Config;
import core.Config.Propagation;

/**
 * A partition to be used in parallel simulations.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class ParallelPartition extends Partition {
	/**
	 * Sets up the specified number of partitions for the specified total number
	 * of nodes.
	 * 
	 * @param numPartitions
	 *            the number of partitions.
	 * 
	 * @param numNodes
	 *            the number of nodes of the simulation.
	 */
	public static void setup(int numPartitions, int numNodes) {
		assert (partitions == null);
		nodeIdToPartitionId = new int[numNodes];
		partitions = new Partition[numPartitions];
		for (int i = 0; i < numPartitions; i++) {
			new ParallelPartition(i);
		}
	}

	/**
	 * Constructor.
	 * 
	 * @param partitionId
	 *            the ID of this partition.
	 */
	protected ParallelPartition(int partitionId) {
		super(partitionId);
	}

	/**
	 * @see modules.wireless.partition.Partition#propagateReception(jist.swans.radio.RadioNoise,
	 *      modules.wireless.partition.TransmissionInfo)
	 */
	@Override
	public void propagateReception(RadioNoise radio, TransmissionInfo txInfo) {
		// The signal reception is propagated to all LPs by means of an
		// LPPROPAGATION event. It is scheduled with the minimum propagation
		// delay.
		final long shiftedTime = txInfo.getStartTime()
				+ Config.getPropagation().getDelay();
		for (Partition partition : partitions) {
			scheduleNewEventWithSrcModuleIdAndSchedTime(shiftedTime,
					partition.getId(), Partition.Events.LPPROPAGATION, txInfo,
					radio.getId(), 0);
		}
	}

	/**
	 * @see modules.wireless.partition.Partition#scheduleEndReceiveEvent(long,
	 *      jist.swans.radio.RadioNoise,
	 *      modules.wireless.partition.TransmissionInfo)
	 */
	@Override
	public void scheduleEndReceiveEvent(long now, RadioNoise radio,
			TransmissionInfo txInfo) {
		if (Config.getPropagation() == Propagation.INDIVIDUAL) {
			// The END_RECEIVE events are scheduled by the receive events and
			// thus have a delay of the transmission duration.
			radio.scheduleNewEvent(now + txInfo.getDuration(), radio.getId(),
					RadioNoise.Events.END_RECEIVE, txInfo);
		}
		// In case of common propagation delay, we do nothing. The END_RECEIVE
		// events are scheduled by the END_LPPROPAGATION event.
	}

	/**
	 * @see modules.wireless.partition.Partition#lpPropagation(long, int,
	 *      modules.wireless.partition.TransmissionInfo)
	 */
	@Override
	protected void lpPropagation(long now, int txModuleId,
			TransmissionInfo txInfo) {
		if (Config.getPropagation() == Propagation.INDIVIDUAL) {
			lpPropagationIndividualDelay(now, txModuleId, txInfo);
		} else {
			lpPropagationCommonDelay(now, txModuleId, txInfo);
		}
	}

	/**
	 * @see modules.wireless.partition.Partition#endLpPropagation(long,
	 *      java.lang.Object)
	 */
	@Override
	protected void endLpPropagation(long now, Object parameter) {
		// Only called in simulations with common propagation delay. The
		// endReceive methods are called directly (END_RECEIVE event bundling)
		// since the transmission duration has already passed.
		assert Config.getPropagation() == Propagation.COMMON;
		final TransmissionInfo txInfo = (TransmissionInfo) parameter;
		for (int i = 0; i < nodes_.length; i++) {
			final RadioNoise rxRadio = nodes_[i].getRadio();
			final RadioNoise rxMeshRadio = nodes_[i].getMeshRadio();
			// Skip the signal's source.
			final int txModuleId = txInfo.getTxModuleId();
			if ((rxRadio.getId() == txModuleId) && (rxMeshRadio == null)) {
				continue;
			}
			// Beyond distance limit?
			if (!field_.isReceived(rxRadio.getNumber(), txInfo)) {
				continue;
			}
			if (rxRadio.getId() != txModuleId) {
				rxRadio.activate(now, rxRadio.getId(),
						RadioNoise.Events.END_RECEIVE, txInfo);
			}
			if ((rxMeshRadio != null) && (rxMeshRadio.getId() != txModuleId)) {
				rxMeshRadio.activate(now, rxMeshRadio.getId(),
						RadioNoise.Events.END_RECEIVE, txInfo);
			}
		}
	}

	/**
	 * Special lpPropagation method for simulations with individual propagation
	 * delay.
	 * 
	 * @param now
	 *            the current simulation time.
	 * @param txModuleId
	 *            the module ID of the transmitter.
	 * @param txInfo
	 *            the transmission information.
	 */
	private void lpPropagationIndividualDelay(long now, int txModuleId,
			TransmissionInfo txInfo) {
		// The RECEIVE events are scheduled with the individual propagation
		// delay minus the minimum propagation delay that has already passed.
		final Location txLoc = txInfo.getTxNodeLoc();
		final NodeMovementInfo nmi = NodeMovementInfo.getInfo();
		final long minPropDelay = Config.getPropagation().getDelay();
		final long shiftedBackTime = now - minPropDelay;
		// Calculate the signal power for each radio.
		for (int i = 0; i < nodes_.length; i++) {
			final RadioNoise rxRadio = nodes_[i].getRadio();
			final RadioNoise rxMeshRadio = nodes_[i].getMeshRadio();
			// Skip the signal's source.
			if ((rxRadio.getId() == txModuleId) && (rxMeshRadio == null)) {
				continue;
			}
			// Beyond distance limit?
			final int rxNodeId = rxRadio.getNumber();
			if (!field_.isReceived(rxNodeId, txInfo)) {
				continue;
			}
			// Notify node.
			final Location rxLoc = nmi.getLocation(rxNodeId, now);
			long dstPropagationDelay = (long) (Location.distance(txLoc, rxLoc)
					/ (Constants.SPEED_OF_LIGHT) * (Constants.SECOND));
			if (dstPropagationDelay < minPropDelay) {
				dstPropagationDelay = minPropDelay;
			}
			if (rxRadio.getId() != txModuleId) {
				scheduleNewEventWithSrcModuleIdAndSchedTime(shiftedBackTime
						+ dstPropagationDelay, rxRadio.getId(),
						RadioNoise.Events.RECEIVE, txInfo, txModuleId,
						shiftedBackTime);
			}
			if ((rxMeshRadio != null) && (rxMeshRadio.getId() != txModuleId)) {
				scheduleNewEventWithSrcModuleIdAndSchedTime(shiftedBackTime
						+ dstPropagationDelay, rxMeshRadio.getId(),
						RadioNoise.Events.RECEIVE, txInfo, txModuleId,
						shiftedBackTime);
			}
		}
	}

	/**
	 * Special lpPropagation method for simulations with common propagation
	 * delay.
	 * 
	 * @param now
	 *            the current simulation time.
	 * @param txModuleId
	 *            the module ID of the transmitter.
	 * @param txInfo
	 *            the transmission information.
	 */
	private void lpPropagationCommonDelay(long now, int txModuleId,
			TransmissionInfo txInfo) {
		// The receive events are activated directly (RECEIVE event bundling)
		// because the (common) propagation delay has already passed. Also, the
		// END_LPPROPAGATION event is scheduled, with a delay that equals the
		// transmission duration.
		// Calculate the signal power for each radio.
		for (int i = 0; i < nodes_.length; i++) {
			final RadioNoise rxRadio = nodes_[i].getRadio();
			final RadioNoise rxMeshRadio = nodes_[i].getMeshRadio();
			// Skip the signal's source.
			if ((rxRadio.getId() == txModuleId) && (rxMeshRadio == null)) {
				continue;
			}
			// Beyond distance limit?
			if (!field_.isReceived(rxRadio.getNumber(), txInfo)) {
				continue;
			}
			// Notify node.
			if (rxRadio.getId() != txModuleId) {
				rxRadio.activate(now, txModuleId, RadioNoise.Events.RECEIVE,
						txInfo);
			}
			if ((rxMeshRadio != null) && (rxMeshRadio.getId() != txModuleId)) {
				rxMeshRadio.activate(now, txModuleId,
						RadioNoise.Events.RECEIVE, txInfo);
			}
		}
		final long minPropDelay = Config.getPropagation().getDelay();
		final long shiftedBackTime = now - minPropDelay;
		scheduleNewEventWithSrcModuleIdAndSchedTime(now + txInfo.getDuration(),
				getId(), Events.END_LPPROPAGATION, txInfo, txModuleId,
				shiftedBackTime);
	}
}
