/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/partition/Partition.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.partition;

import java.util.List;

import jist.swans.Constants;
import jist.swans.field.Fading;
import jist.swans.field.Field;
import jist.swans.field.PathLoss;
import jist.swans.radio.BERTable;
import jist.swans.radio.RadioNoise;
import modules.wireless.scenario.NodeConfig;
import modules.wireless.scenario.NodeMovementInfo;
import modules.wireless.scenario.WirelessNode;
import core.BaseModule;
import core.Config;
import core.Config.Movement;
import core.ModuleRoot;

/**
 * Base class for partitions.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public abstract class Partition extends BaseModule {
	/**
	 * The events offered by a partition.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	public static enum Events {
		/**
		 * Indicates the start of a transmission to other partitions.
		 */
		LPPROPAGATION,
		/**
		 * Indicates the end of a transmission to other partitions.
		 */
		END_LPPROPAGATION,
		/**
		 * Position update event, used only in simulations with discrete
		 * movement update.
		 */
		MOVEMENTUPDATE
	}

	/**
	 * Key: Node ID. Value: The ID of the partition this node is assigned to.
	 */
	protected static int nodeIdToPartitionId[];

	/**
	 * Key: Partition ID. Value: The partition with that ID.
	 */
	protected static Partition partitions[];

	/**
	 * Provides the partition the given radio is assigned to.
	 * 
	 * @param radio
	 *            the radio of interest.
	 * @return the partition the given radio is assigned to.
	 */
	public static Partition getMyPartition(RadioNoise radio) {
		return partitions[nodeIdToPartitionId[radio.getNumber()]];
	}

	/**
	 * Provides the field the given radio is assigned to.
	 * 
	 * @param radio
	 *            the radio of interest.
	 * @return the partition the given radio is assigned to.
	 */
	public static Field getMyField(RadioNoise radio) {
		return getMyPartition(radio).field_;
	}

	/**
	 * Provides the partition to the specified ID.
	 * 
	 * @param partitionId
	 *            the partition ID
	 * @return the partition to the specified ID.
	 */
	public static Partition getPartition(int partitionId) {
		assert partitions.length > partitionId;
		return partitions[partitionId];
	}

	/**
	 * Provides the ID of the partition the given node is assigned to.
	 * 
	 * @param nodeId
	 *            the ID of the node of interest.
	 * @return the IF of the partition the given node is assigned to.
	 */
	public static int getPartitionIdOfNode(int nodeId) {
		assert nodeIdToPartitionId != null;
		assert nodeId < nodeIdToPartitionId.length;
		return nodeIdToPartitionId[nodeId];
	}

	/**
	 * Provides the partition the given node is assigned to.
	 * 
	 * @param nodeId
	 *            the ID of the node of interest.
	 * @return the partition the given node is assigned to.
	 */
	public static Partition getPartitionOfNode(int nodeId) {
		return partitions[getPartitionIdOfNode(nodeId)];
	}

	/**
	 * The field that contains the nodes of this partition.
	 */
	protected final Field field_;

	/**
	 * List of the nodes of this partition (created by WirelessNode).
	 */
	protected WirelessNode[] nodes_;

	/**
	 * Constructor.
	 * 
	 * @param partitionId
	 *            the ID of this partition.
	 */
	protected Partition(int partitionId) {
		super(partitionId);
		assert partitions != null;
		assert partitionId < partitions.length;
		ModuleRoot.getRoot().allocateModuleToLp(this, partitionId);
		partitions[partitionId] = this;
		// Create field.
		Fading fading;
		try {
			fading = (Fading) Config.getFadingModel().getFadingClass()
					.newInstance();
		} catch (InstantiationException e) {
			throw new RuntimeException("Could not create fading model!");
		} catch (IllegalAccessException e) {
			throw new RuntimeException("Could not create fading model!");
		}
		PathLoss pathloss;
		try {
			pathloss = (PathLoss) Config.getPathlossModel().getPathlossClass()
					.newInstance();
		} catch (InstantiationException e) {
			throw new RuntimeException("Could not create pathloss model!");
		} catch (IllegalAccessException e) {
			throw new RuntimeException("Could not create pathloss model!");
		}
		final BERTable berTable = new BERTable();
		final double distanceLimit = NodeConfig.getDistanceLimit();
		field_ = new Field(fading, pathloss, berTable, distanceLimit);
	}

	/**
	 * @see core.BaseModule#init()
	 */
	@Override
	public void init() {
		// Get nodes for this partition.
		List<WirelessNode> myNodes = WirelessNode.getNodesOfPartition(number_);
		nodes_ = new WirelessNode[myNodes.size()];
		nodes_ = myNodes.toArray(nodes_);

		// Populate our part of the static radio-partition list.
		for (int i = 0; i < nodes_.length; i++) {
			nodeIdToPartitionId[nodes_[i].getNumber()] = number_;
		}
		// Schedule movement update event, if necessary.
		if (Config.getMovement() == Movement.DISCRETE) {
			scheduleNewEvent(Config.getStartTime(), id_, Events.MOVEMENTUPDATE,
					null);
		}
	}

	/**
	 * @see core.BaseModule#activate(long, int, java.lang.Enum,
	 *      java.lang.Object)
	 */
	@Override
	@SuppressWarnings("rawtypes")
	public void activate(long execTime, int srcModuleId, Enum eventId,
			Object parameter) {
		if (eventId == Events.MOVEMENTUPDATE) {
			movementUpdate(execTime);
		} else if (eventId == Events.LPPROPAGATION) {
			lpPropagation(execTime, srcModuleId, (TransmissionInfo) parameter);
		} else if (eventId == Events.END_LPPROPAGATION) {
			endLpPropagation(execTime, parameter);
		}
	}

	/**
	 * Used in parallel simulations with protocol lookahead in order to
	 * calculate the next safe time.
	 * 
	 * @param lastSafeTime
	 *            the last safe time.
	 * @return the next safe time estimate.
	 */
	public long calculateNextSyncPoint(long lastSafeTime) {
		long min = Long.MAX_VALUE;
		for (WirelessNode node : nodes_) {
			final long nodePST = node.getNextPST(lastSafeTime);
			if (nodePST < min) {
				min = nodePST;
			}
		}
		return min;
	}

	/**
	 * Called when a movement update event happens. Updates the positions of all
	 * nodes.
	 * 
	 * @param now
	 *            the current simulation time.
	 */
	protected void movementUpdate(long now) {
		assert Config.getMovement() == Movement.DISCRETE;

		// Update locations.
		NodeMovementInfo nmi = NodeMovementInfo.getInfo();
		for (int i = 0; i < nodes_.length; i++) {
			nmi.updateLocation(nodes_[i].getNumber(), now);
		}

		// Schedule next location update.
		long nextUpdateTime = now + Config.getMovementInterval();
		if (nextUpdateTime + Constants.NANO_SECOND < Config.getEndTime()) {
			// Make sure the event is executed before the SimController
			// does the update but last in this interval (via schedTime).
			scheduleNewEventWithSchedTime(nextUpdateTime, getId(),
					Events.MOVEMENTUPDATE, null, Long.MAX_VALUE);
		}
	}

	/**
	 * Called when a partition receives information about the start of a
	 * transmitted signal.
	 * 
	 * @param now
	 *            the current simulation time.
	 * @param txModuleId
	 *            the module ID of the transmitter.
	 * @param txInfo
	 *            the transmission information.
	 */
	protected abstract void lpPropagation(long now, int txModuleId,
			TransmissionInfo txInfo);

	/**
	 * Called when a partition receives information about a the end of a
	 * transmitted signal.
	 * 
	 * @param now
	 *            the current simulation time.
	 * @param parameter
	 *            the information required to finish the transmission correctly.
	 */
	protected abstract void endLpPropagation(long now, Object parameter);

	/**
	 * Called when a radio starts transmitting a signal, in order to propagate
	 * the information to all partitions and their receivers.
	 * 
	 * @param radio
	 *            the transmitting radio.
	 * @param txInfo
	 *            the transmission information.
	 */
	public abstract void propagateReception(RadioNoise radio,
			TransmissionInfo txInfo);

	/**
	 * Called when a radio starts receiving a signal, in order to schedule its
	 * end of reception event.
	 * 
	 * @param now
	 *            the current simulation time.
	 * @param radio
	 *            the receiving radio.
	 * @param txInfo
	 *            the transmission information.
	 */
	public abstract void scheduleEndReceiveEvent(long now, RadioNoise radio,
			TransmissionInfo txInfo);
};
