/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/partition/SinglePartition.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.partition;

import jist.swans.Constants;
import jist.swans.misc.Location;
import jist.swans.radio.RadioNoise;
import modules.wireless.scenario.NodeMovementInfo;
import core.Config;
import core.Config.Propagation;

/**
 * A partition to be used in sequential simulations.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class SinglePartition extends Partition {
	/**
	 * Sets up a partition with the specified number of nodes.
	 * 
	 * @param numNodes
	 *            the number of nodes of the simulation.
	 */
	public static void setup(int numNodes) {
		assert (partitions == null);
		nodeIdToPartitionId = new int[numNodes];
		partitions = new Partition[1];
		new SinglePartition(0);
	}

	/**
	 * Constructor.
	 * 
	 * @param partitionId
	 *            the id of this partition.
	 */
	protected SinglePartition(int partitionId) {
		super(partitionId);
	}

	/**
	 * @see modules.wireless.partition.Partition#propagateReception(jist.swans.radio.RadioNoise,
	 *      modules.wireless.partition.TransmissionInfo)
	 */
	@Override
	public void propagateReception(RadioNoise radio, TransmissionInfo txInfo) {
		final long now = txInfo.getStartTime();
		final Location txLoc = txInfo.getTxNodeLoc();
		if (Config.getPropagation() == Propagation.INDIVIDUAL) {
			// The RECEIVE events are scheduled with a delay that equals the
			// individual propagation delay.
			final int txModuleId = txInfo.getTxModuleId();
			final long minDelay = Config.getPropagation().getDelay();
			final NodeMovementInfo nmi = NodeMovementInfo.getInfo();
			// Calculate the signal power for each radio.
			for (int i = 0; i < nodes_.length; i++) {
				final RadioNoise rxRadio = nodes_[i].getRadio();
				final RadioNoise rxMeshRadio = nodes_[i].getMeshRadio();
				// Skip the signal's source.
				if ((rxRadio.getId() == txModuleId) && (rxMeshRadio == null)) {
					continue;
				}
				// Beyond distance limit?
				final int rxNodeId = rxRadio.getNumber();
				if (!field_.isReceived(rxNodeId, txInfo)) {
					continue;
				}
				// Notify node.
				final Location rxLoc = nmi.getLocation(rxNodeId, now);
				long dstPropagationDelay = (long) (Location.distance(txLoc,
						rxLoc) / (Constants.SPEED_OF_LIGHT) * (Constants.SECOND));
				if (dstPropagationDelay < minDelay) {
					dstPropagationDelay = minDelay;
				}
				if (rxRadio.getId() != txModuleId) {
					radio.scheduleNewEvent(now + dstPropagationDelay,
							rxRadio.getId(), RadioNoise.Events.RECEIVE, txInfo);
				}
				if ((rxMeshRadio != null)
						&& (rxMeshRadio.getId() != txModuleId)) {
					radio.scheduleNewEvent(now + dstPropagationDelay,
							rxMeshRadio.getId(), RadioNoise.Events.RECEIVE,
							txInfo);
				}
			}
		} else {
			final long commonRcvTime = now + Config.getPropagation().getDelay();
			scheduleNewEventWithSrcModuleIdAndSchedTime(commonRcvTime, getId(),
					Partition.Events.LPPROPAGATION, txInfo, radio.getId(), 0);
		}
	}

	/**
	 * @see modules.wireless.partition.Partition#scheduleEndReceiveEvent(long,
	 *      jist.swans.radio.RadioNoise,
	 *      modules.wireless.partition.TransmissionInfo)
	 */
	@Override
	public void scheduleEndReceiveEvent(long now, RadioNoise radio,
			TransmissionInfo txInfo) {
		if (Config.getPropagation() == Propagation.INDIVIDUAL) {
			// The END_RECEIVE events are scheduled by the receive events and
			// thus have a delay of the transmission duration.
			radio.scheduleNewEvent(now + txInfo.getDuration(), radio.getId(),
					RadioNoise.Events.END_RECEIVE, txInfo);
		}
		// In case of common propagation delay, we do nothing. The END_RECEIVE
		// events are scheduled by the END_LPPROPAGATION event.
	}

	/**
	 * @see modules.wireless.partition.Partition#lpPropagation(long, int,
	 *      modules.wireless.partition.TransmissionInfo)
	 */
	@Override
	protected void lpPropagation(long now, int txModuleId,
			TransmissionInfo txInfo) {
		assert Config.getPropagation() == Config.Propagation.COMMON;
		// Calculate the signal power for each radio.
		for (int i = 0; i < nodes_.length; i++) {
			final RadioNoise rxRadio = nodes_[i].getRadio();
			final RadioNoise rxMeshRadio = nodes_[i].getMeshRadio();
			// Skip the signal's source.
			if ((rxRadio.getId() == txModuleId) && (rxMeshRadio == null)) {
				continue;
			}
			// Beyond distance limit?
			if (!field_.isReceived(rxRadio.getNumber(), txInfo)) {
				continue;
			}
			if (rxRadio.getId() != txModuleId) {
				rxRadio.activate(now, txModuleId, RadioNoise.Events.RECEIVE,
						txInfo);
			}
			if ((rxMeshRadio != null) && (rxMeshRadio.getId() != txModuleId)) {
				rxMeshRadio.activate(now, txModuleId,
						RadioNoise.Events.RECEIVE, txInfo);
			}
		}
		// Correct the scheduling time for parallel simulation.
		final long minPropDelay = Config.getPropagation().getDelay();
		final long shiftedBackTime = now - minPropDelay;
		scheduleNewEventWithSrcModuleIdAndSchedTime(now + txInfo.getDuration(),
				getId(), Events.END_LPPROPAGATION, txInfo, txModuleId,
				shiftedBackTime);
	}

	/**
	 * @see modules.wireless.partition.Partition#endLpPropagation(long,
	 *      java.lang.Object)
	 */
	@Override
	protected void endLpPropagation(long now, Object parameter) {
		// Only called in simulations with common propagation delay. The
		// endReceive methods are called directly (END_RECEIVE event bundling)
		// since the transmission duration has already passed.
		assert Config.getPropagation() == Propagation.COMMON;
		final TransmissionInfo txInfo = (TransmissionInfo) parameter;
		final int txModuleId = txInfo.getTxModuleId();
		for (int i = 0; i < nodes_.length; i++) {
			final RadioNoise rxRadio = nodes_[i].getRadio();
			final RadioNoise rxMeshRadio = nodes_[i].getMeshRadio();
			// Skip the signal's source.
			if ((rxRadio.getId() == txModuleId) && (rxMeshRadio == null)) {
				continue;
			}
			// Beyond distance limit?
			if (!field_.isReceived(rxRadio.getNumber(), txInfo)) {
				continue;
			}
			if (rxRadio.getId() != txModuleId) {
				rxRadio.activate(now, rxRadio.getId(),
						RadioNoise.Events.END_RECEIVE, txInfo);
			}
			if ((rxMeshRadio != null) && (rxMeshRadio.getId() != txModuleId)) {
				rxMeshRadio.activate(now, rxMeshRadio.getId(),
						RadioNoise.Events.END_RECEIVE, txInfo);
			}
		}
	}
}
