/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/partition/TransmissionInfo.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.partition;

import java.io.Serializable;

import jist.swans.field.Field;
import jist.swans.misc.Location;
import jist.swans.misc.Message;
import modules.wireless.scenario.NodeMovementInfo;

/**
 * Helper class that stores all relevant information of a transmission, which is
 * to be used by the partitions and nodes.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class TransmissionInfo extends Object implements Serializable {
	/**
	 * Default ID.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * The transmitted message.
	 */
	private final Message msg_;

	/**
	 * The start time of the transmission.
	 */
	private final long startTime_;

	/**
	 * The duration of the transmission.
	 */
	private final long duration_;

	/**
	 * The node ID of the transmitter.
	 */
	private final int txNodeId_;

	/**
	 * The module ID of the transmitting module.
	 */
	private final int txModuleId_;

	/**
	 * The location of the transmitting node.
	 */
	private final Location txNodeLoc_;

	/**
	 * The receive signal powers of all nodes that may possibly receive the
	 * signal.
	 */
	private final double[] signalPowers_mW_;

	/**
	 * Constructor. Note that the signal powers are filled in later.
	 * 
	 * @param msg
	 *            the transmitted message.
	 * @param startTime
	 *            the start time of the transmission.
	 * @param duration
	 *            the duration of the transmission.
	 * @param txNodeId
	 *            the node ID of the transmitter.
	 * @param txModuleId
	 *            the module ID of the transmitting module.
	 */
	public TransmissionInfo(Message msg, long startTime, long duration,
			int txNodeId, int txModuleId) {
		msg_ = msg;
		startTime_ = startTime;
		duration_ = duration;
		txNodeId_ = txNodeId;
		txModuleId_ = txModuleId;
		txNodeLoc_ = NodeMovementInfo.getInfo()
				.getLocation(txNodeId, startTime);
		signalPowers_mW_ = new double[NodeMovementInfo.getInfo().getNodeCount()];
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object o) {
		if (o == this) {
			return true;
		}
		if (o == null || !(o instanceof TransmissionInfo)) {
			return false;
		}
		final TransmissionInfo other = (TransmissionInfo) o;
		if (startTime_ != other.startTime_) {
			return false;
		}
		if (txModuleId_ != other.txModuleId_) {
			return false;
		}
		return true;
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + (int) (startTime_ ^ (startTime_ >>> 32));
		result = prime * result + txModuleId_;
		return result;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "TransmissionInfo(" + msg_.toString() + "," + startTime_ + ","
				+ duration_ + ")";
	}

	/**
	 * Provides the message transmitted.
	 * 
	 * @return the message transmitted.
	 */
	public Message getMsg() {
		return msg_;
	}

	/**
	 * Provides the start time of the transmission.
	 * 
	 * @return the start time of the transmission.
	 */
	public long getStartTime() {
		return startTime_;
	}

	/**
	 * Provides the duration of the transmission.
	 * 
	 * @return the duration of the transmission.
	 */
	public long getDuration() {
		return duration_;
	}

	/**
	 * Provides the node ID of the transmitter.
	 * 
	 * @return the node ID of the transmitter.
	 */
	public int getTxNodeId() {
		return txNodeId_;
	}

	/**
	 * Provides the module ID of the transmitter.
	 * 
	 * @return the module ID of the transmitter.
	 */
	public int getTxModuleId() {
		return txModuleId_;
	}

	/**
	 * Provides the location of the transmitter.
	 * 
	 * @return the location of the transmitter.
	 */
	public Location getTxNodeLoc() {
		return txNodeLoc_;
	}

	/**
	 * Returns (and, if necessary, calculates, the receive power of the signal
	 * at the specified node. This method does not have to be thread-safe, since
	 * each node is managed by exactly one partition. That's also the reason why
	 * several partitions may operate on the same TransmissionInfo object.
	 * 
	 * @param field
	 *            the field instance used to calculate signal powers.
	 * @param rxNodeId
	 *            the receiving node.
	 * @return the receive power of the signal at the specified node.
	 */
	public double getPower(Field field, int rxNodeId) {
		if (signalPowers_mW_[rxNodeId] != 0.0) {
			return signalPowers_mW_[rxNodeId];
		}
		final Location rxLoc = NodeMovementInfo.getInfo().getLocation(rxNodeId,
				startTime_);
		final double power_mW = field.calculateSignalPower_mW(rxNodeId,
				txNodeLoc_, rxLoc);
		signalPowers_mW_[rxNodeId] = power_mW;
		return power_mW;
	}
}
