/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/scenario/CreateWorld.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.scenario;

import java.util.Arrays;
import java.util.List;

import jist.swans.Constants;
import jist.swans.field.Field;
import jist.swans.misc.Location;
import jist.swans.radio.RadioInfo;
import modules.wireless.lazy.LazyDistributedPartition;
import modules.wireless.lazy.LazyParallelPartition;
import modules.wireless.lazy.LazySinglePartition;
import modules.wireless.mesh.DummyNode;
import modules.wireless.partition.DistributedPartition;
import modules.wireless.partition.ParallelPartition;
import modules.wireless.partition.Partition;
import modules.wireless.partition.SinglePartition;
import modules.wireless.scenario.NodeConfig.FlowDirection;
import modules.wireless.scenario.NodeConfig.FlowPriority;
import modules.wireless.scenario.NodeConfig.FlowType;
import modules.wireless.scenario.NodeConfig.MeshTrafficInfoTCP;
import modules.wireless.scenario.NodeConfig.MeshTrafficInfoUDP;
import modules.wireless.scenario.NodeConfig.RandomTrafficInfoTCP;
import modules.wireless.scenario.NodeConfig.RandomTrafficInfoUDP;
import modules.wireless.scenario.NodeConfig.ScenarioType;
import modules.wireless.scenario.WirelessNode.NodeType;
import rng.RandomManager;
import rng.RandomStreamInt;
import core.Config;
import core.Config.SimType;
import core.distribution.StatsPacket;
import core.sync.AbstractSimController;
import core.sync.DistributedSimController;
import core.sync.ParallelSimController;
import core.util.CustomLogger;

/**
 * This class creates the world for the simulation. It creates partitions and
 * nodes as defined by the user. Primary reason for this class is to reduce the
 * overhead in the init file.
 * 
 * Note: The order of creation (partitions, lazy lists, nodes) is important!
 * 
 * @author Patrick Peschlow
 * @author Christoph Sprenger
 * 
 * @version 1.0
 */
public class CreateWorld implements SimulationCallback {
	/**
	 * Constructor.
	 */
	public CreateWorld() {
		// Empty.
	}

	/**
	 * Constructor. This class is used to set up the scenario after the init
	 * script has been evaluated.
	 */
	public void setupSimulation() {
		assert Config.getSimType() == SimType.SEQUENTIAL
				|| Config.getSimType() == SimType.PARALLEL;

		commonSetupBeforeNodeSetup();

		// Create partitions and nodes.
		switch (Config.getSimType()) {
		case SEQUENTIAL:
			setupSequential();
			break;
		case PARALLEL:
			setupParallel();
			break;
		default:
			throw new RuntimeException("Simulation type error!");
		}

		commonSetupAfterNodeSetup();
	}

	/**
	 * Sets up a distributed simulation.
	 * 
	 * @param firstPartitionId
	 *            the ID of the first partition on this host.
	 * @param numLocalPartitions
	 *            the number of partitions on this host.
	 */
	public void setupSimulationDistributed(int firstPartitionId,
			int numLocalPartitions) {
		assert Config.getSimType() == SimType.DISTRIBUTED;
		commonSetupBeforeNodeSetup();

		// Create partitions and nodes.
		setupDistributed(firstPartitionId, numLocalPartitions);

		commonSetupAfterNodeSetup();
	}

	/**
	 * Called before the nodes are set up.
	 */
	private void commonSetupBeforeNodeSetup() {
		Config.setSimulationCallback(this);

		NodeMovementInfo.init();

		RadioInfo.setRadioInfo(Constants.FREQUENCY_DEFAULT, 6000000,
				NodeConfig.getTransmitPower(), Constants.GAIN_DEFAULT,
				Constants.GAIN_DEFAULT,
				Field.fromDB(Constants.SENSITIVITY_DEFAULT),
				Field.fromDB(Constants.THRESHOLD_DEFAULT),
				Constants.TEMPERATURE_DEFAULT,
				Constants.TEMPERATURE_FACTOR_DEFAULT,
				Constants.AMBIENT_NOISE_DEFAULT);
	}

	/**
	 * Called after the nodes have been set up.
	 */
	private void commonSetupAfterNodeSetup() {
		final int nodeCount = NodeMovementInfo.getInfo().getNodeCount();
		// Initialize traffic flows (must be done after nodes are created).
		createTcpTraffic(nodeCount);
		createUdpTraffic(nodeCount);

		// Substreams for fading on each partition.
		Config.getFadingModel().setupRngs();

		// Substreams for shadowing on each partition.
		Config.getPathlossModel().setupRngs();

		initialSimOutput();
	}

	/**
	 * Prints initial information about the simulation scenario.
	 */
	private void initialSimOutput() {
		if (CustomLogger.isInfoEnabled()) {
			NodeMovementInfo nmi = NodeMovementInfo.getInfo();
			CustomLogger.info("Partitions:          "
					+ Config.getNumPartitions());
			CustomLogger.info("Nodes created:       " + nmi.getNodeCount());
			if (NodeConfig.getScenarioType() == ScenarioType.MESH) {
				CustomLogger.info("AP Nodes (Portals):  "
						+ NodeConfig.getNumberAPs() + " ("
						+ NodeConfig.getNumberPortals() + ")");
			}
			double simTime = Config.getEndTime() / Constants.SECOND;
			CustomLogger.info("Duration:            " + simTime + " seconds");
			Location areaSize = nmi.getAreaSize();
			CustomLogger.info("Area size:           " + areaSize.getX() + " x "
					+ areaSize.getY());
			CustomLogger.info("tcp flows:           "
					+ ((NodeConfig.getTrafficInfoTCP() == null) ? 0
							: NodeConfig.getNumFlowsTCP()));
			CustomLogger.info("udp flows:           "
					+ ((NodeConfig.getTrafficInfoUDP() == null) ? 0
							: NodeConfig.getNumFlowsUDP()));
			if (NodeConfig.getScenarioType() == ScenarioType.MESH) {
				FlowPriority flow = FlowPriority.AC_BK;
				if (NodeConfig.getTrafficInfoTCP() != null) {
					List<RandomTrafficInfoTCP> trafficList = NodeConfig
							.getTrafficInfoTCP();
					MeshTrafficInfoTCP traffic = (MeshTrafficInfoTCP) trafficList
							.get(0);
					flow = traffic.getPriority();
				}
				if (NodeConfig.getTrafficInfoUDP() != null) {
					List<RandomTrafficInfoUDP> trafficList = NodeConfig
							.getTrafficInfoUDP();
					MeshTrafficInfoUDP traffic = (MeshTrafficInfoUDP) trafficList
							.get(0);
					flow = traffic.getPriority();
				}
				CustomLogger.info("flow priority:       " + flow);
			}
			CustomLogger.info("transmit power:      "
					+ NodeConfig.getTransmitPower() + " dBm");
			CustomLogger.info("seed:                " + Config.getRandomSeed());
			CustomLogger.info("RxTx-Turnaround      "
					+ NodeConfig.getRxTxTurnaround());
			CustomLogger
					.info("propagation delay:   " + Config.getPropagation());
			CustomLogger.info("movement:            " + Config.getMovement());
			CustomLogger.info("lazy mac 802.11:     " + Config.usesLazyMac());
			CustomLogger.info("simulation type:     " + Config.getSimType());
			CustomLogger.info("controller:          "
					+ Config.getSimType().getSimControllerClass()
							.getSimpleName());
			CustomLogger.info("lookahead:           " + Config.getLookahead());
		}
	}

	/**
	 * Sets up a sequential simulation.
	 */
	private void setupSequential() {
		final int numNodes = NodeMovementInfo.getInfo().getNodeCount();
		// Setup partition.
		if (Config.usesLazyMac()) {
			LazySinglePartition.setup(numNodes);
		} else {
			SinglePartition.setup(numNodes);
		}
		// Setup nodes.
		final int numVirtualPartitions = Config.getNumVirtualPartitions();
		final int numAPs = NodeConfig.getNumberAPs();
		final int numPortals = NodeConfig.getNumberPortals();
		WirelessNode.initNodes(numNodes, numAPs, numPortals);
		if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MANET) {
			for (int nodeId = 0; nodeId < numNodes; nodeId++) {
				final int rngId = nodeId % numVirtualPartitions;
				WirelessNode.registerNode(nodeId, 0, rngId);
			}
		} else if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH) {
			assert numAPs > 0 && numAPs <= numNodes;
			final int[] portalIds = new int[numAPs];
			Arrays.fill(portalIds, WirelessNode.NO_PORTAL);
			final RandomStreamInt rngPortals = RandomManager.getInstance()
					.getUniformIntSubstream(0, numAPs - 1);
			for (int portalId = 0; portalId < numPortals; portalId++) {
				int ap;
				do {
					ap = rngPortals.nextInt();
				} while (portalIds[ap] != WirelessNode.NO_PORTAL);
				portalIds[ap] = portalId;
			}
			for (int nodeId = 0; nodeId < numNodes; nodeId++) {
				final int rngId = nodeId % numVirtualPartitions;
				if (nodeId < numAPs) {
					WirelessNode.registerNode(nodeId, 0, rngId, NodeType.AP,
							portalIds[nodeId]);
				} else {
					WirelessNode.registerNode(nodeId, 0, rngId,
							NodeType.CLIENT, WirelessNode.NO_PORTAL);
				}
			}
			setupAssociation(numNodes, numAPs);
		}
	}

	/**
	 * Sets up a parallel simulation.
	 */
	private void setupParallel() {
		final int numNodes = NodeMovementInfo.getInfo().getNodeCount();
		// Setup partitions.
		final int numPartitions = Config.getNumPartitions();
		if (Config.usesLazyMac()) {
			LazyParallelPartition.setup(numPartitions, numNodes);
		} else {
			ParallelPartition.setup(numPartitions, numNodes);
		}
		// Choose portal APs.
		final int numAPs = NodeConfig.getNumberAPs();
		final int numPortals = NodeConfig.getNumberPortals();
		final int[] portalIds = new int[numAPs];
		Arrays.fill(portalIds, WirelessNode.NO_PORTAL);
		if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH) {
			assert numAPs > 0 && numAPs <= numNodes;
			final RandomStreamInt rngPortals = RandomManager.getInstance()
					.getUniformIntSubstream(0, numAPs - 1);
			for (int portalId = 0; portalId < numPortals; portalId++) {
				int ap;
				do {
					ap = rngPortals.nextInt();
				} while (portalIds[ap] != WirelessNode.NO_PORTAL);
				portalIds[ap] = portalId;
			}
		}
		// Setup nodes.
		final int numVirtualPartitions = Config.getNumVirtualPartitions();
		WirelessNode.initNodes(numNodes, numAPs, numPortals);
		final int[] rngIdToPartitionId = new int[numVirtualPartitions];
		for (int nodeId = 0; nodeId < numNodes; nodeId++) {
			final int partitionId = nodeId % numPartitions;
			final int rngId = nodeId % numVirtualPartitions;
			if (nodeId < numVirtualPartitions) {
				rngIdToPartitionId[rngId] = partitionId;
			} else {
				if (rngIdToPartitionId[rngId] != partitionId) {
					throw new RuntimeException(
							"Error in virtual partition to partition assignment!");
				}
			}
			if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MANET) {
				WirelessNode.registerNode(nodeId, partitionId, rngId);
			} else if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH) {
				if (nodeId < numAPs) {
					WirelessNode.registerNode(nodeId, partitionId, rngId,
							NodeType.AP, portalIds[nodeId]);
				} else {
					WirelessNode.registerNode(nodeId, partitionId, rngId,
							NodeType.CLIENT, WirelessNode.NO_PORTAL);
				}
			}
		}
		if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH) {
			setupAssociation(numNodes, numAPs);
		}
	}

	/**
	 * Sets up a distributed simulation.
	 * 
	 * @param firstPartitionId
	 *            the ID of the first partition on this host.
	 * @param numLocalPartitions
	 *            the number of partitions on this host.
	 */
	private void setupDistributed(int firstPartitionId, int numLocalPartitions) {
		final int numNodes = NodeMovementInfo.getInfo().getNodeCount();
		final int lastPartitionId = firstPartitionId + numLocalPartitions - 1;
		// Setup partitions.
		final int numPartitions = Config.getNumPartitions();
		if (Config.usesLazyMac()) {
			LazyDistributedPartition.setup(numPartitions, numNodes);
		} else {
			DistributedPartition.setup(numPartitions, numNodes,
					firstPartitionId, lastPartitionId);
		}
		// Choose portal APs.
		final int numAPs = NodeConfig.getNumberAPs();
		final int numPortals = NodeConfig.getNumberPortals();
		final int[] portalIds = new int[numAPs];
		Arrays.fill(portalIds, WirelessNode.NO_PORTAL);
		if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH) {
			assert numAPs > 0 && numAPs <= numNodes;
			final RandomStreamInt rngPortals = RandomManager.getInstance()
					.getUniformIntSubstream(0, numAPs - 1);
			for (int portalId = 0; portalId < numPortals; portalId++) {
				int ap;
				do {
					ap = rngPortals.nextInt();
				} while (portalIds[ap] != WirelessNode.NO_PORTAL);
				portalIds[ap] = portalId;
			}
		}
		// Setup nodes.
		final int numVirtualPartitions = Config.getNumVirtualPartitions();
		WirelessNode.initNodes(numNodes, numAPs, numPortals);
		final int[] rngIdToPartitionId = new int[numVirtualPartitions];
		for (int nodeId = 0; nodeId < numNodes; nodeId++) {
			final int partitionId = nodeId % numPartitions;
			final int rngId = nodeId % numVirtualPartitions;
			if (nodeId < numVirtualPartitions) {
				rngIdToPartitionId[rngId] = partitionId;
			} else {
				if (rngIdToPartitionId[rngId] != partitionId) {
					throw new RuntimeException(
							"Error in virtual partition to partition assignment!");
				}
			}
			if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MANET) {
				if (firstPartitionId <= partitionId
						&& partitionId <= lastPartitionId) {
					WirelessNode.registerNode(nodeId, partitionId, rngId);
				} else {
					DistInitHelperNode.registerRemoteNode(nodeId, partitionId,
							rngId);
				}
			} else if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH) {
				assert NodeConfig.getNumberAPs() <= numNodes;
				if (nodeId < numAPs) {
					if (firstPartitionId <= partitionId
							&& partitionId <= lastPartitionId) {
						WirelessNode.registerNode(nodeId, partitionId, rngId,
								NodeType.AP, portalIds[nodeId]);
					} else {
						DistInitHelperNode.registerRemoteNode(nodeId,
								partitionId, rngId, NodeType.AP,
								portalIds[nodeId]);
					}
				} else {
					if (firstPartitionId <= partitionId
							&& partitionId <= lastPartitionId) {
						WirelessNode.registerNode(nodeId, partitionId, rngId,
								NodeType.CLIENT, WirelessNode.NO_PORTAL);
					} else {
						DistInitHelperNode.registerRemoteNode(nodeId,
								partitionId, rngId, NodeType.CLIENT,
								WirelessNode.NO_PORTAL);
					}
				}
			}
		}
		if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH) {
			setupAssociation(numNodes, numAPs);
		}
	}

	/**
	 * Sets up the association between mesh clients and APs.
	 * 
	 * @param nodeCount
	 *            the total number of nodes of the simulation.
	 * @param numAPs
	 *            the total number of APs of the simulation.
	 */
	private void setupAssociation(int nodeCount, int numAPs) {
		Location[] apLocations = new Location[numAPs];
		NodeMovementInfo nmi = NodeMovementInfo.getInfo();
		for (int nodeId = 0; nodeId < nodeCount; nodeId++) {
			WirelessNode node = WirelessNode.getNode(nodeId);
			if (nodeId < numAPs) {
				apLocations[nodeId] = nmi.getLocation(nodeId, 0);
			} else {
				Location nodeLocation = nmi.getLocation(nodeId, 0);
				double minDist = Double.MAX_VALUE;
				int associatedApId = 0;
				for (int apId = 0; apId < numAPs; apId++) {
					double distance = Location.distanceSqr(nodeLocation,
							apLocations[apId]);
					if (distance < minDist) {
						minDist = distance;
						associatedApId = apId;
					}
				}
				WirelessNode.setAssociation(nodeId, associatedApId);
				node.associateAp(associatedApId);
				WirelessNode.getNode(associatedApId).associateClient(nodeId);
			}
		}
	}

	/**
	 * Creates random TCP traffic between nodes, as specified by the input file.
	 * 
	 * @param nodeCount
	 *            the total number of nodes of the simulation.
	 */
	private void createTcpTraffic(int nodeCount) {
		final List<RandomTrafficInfoTCP> trafficTCP = NodeConfig
				.getTrafficInfoTCP();
		if (trafficTCP == null) {
			return;
		}
		// Substreams for setting up traffic connections.
		final RandomStreamInt rngSrcNodes = RandomManager.getInstance()
				.getUniformIntSubstream(0, nodeCount - 1);
		final RandomStreamInt rngDstNodes = RandomManager.getInstance()
				.getUniformIntSubstream(0, nodeCount - 1);
		if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MANET) {
			createTcpTrafficManet(trafficTCP, rngSrcNodes, rngDstNodes);
		} else if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH) {
			createTcpTrafficMesh(trafficTCP, rngSrcNodes, rngDstNodes);
		}
	}

	/**
	 * Creates random TCP traffic for mesh scenarios.
	 * 
	 * @param trafficTCP
	 *            the TCP traffic info.
	 * @param rngSrcNodes
	 *            the RNG for selecting source nodes.
	 * @param rngDstNodes
	 *            the RNG for selecting destination nodes.
	 */
	private void createTcpTrafficMesh(
			final List<RandomTrafficInfoTCP> trafficTCP,
			final RandomStreamInt rngSrcNodes, final RandomStreamInt rngDstNodes) {
		int connectionId = 0;
		int numDummies = 0;
		// Loop over traffic packages.
		for (RandomTrafficInfoTCP traffic : trafficTCP) {
			final RandomStreamInt rngStartTimes = RandomManager.getInstance()
					.getUniformIntSubstream(0, (int) traffic.getMaxStartTime());
			final RandomStreamInt rngPorts = RandomManager.getInstance()
					.getUniformIntSubstream(0, 1024);
			final MeshTrafficInfoTCP meshTraffic = (MeshTrafficInfoTCP) traffic;
			// Create flows for this traffic package.
			final int numFlows = (meshTraffic.getDirection() == FlowDirection.BOTH) ? meshTraffic
					.getNumFlows() / 2 : meshTraffic.getNumFlows();
			for (int i = 0; i < numFlows; i++) {
				int srcNodeId = 0;
				int dstNodeId = 0;
				if (meshTraffic.getType() == FlowType.CLIENT_TO_CLIENT) {
					// Select endpoints.
					do {
						srcNodeId = rngSrcNodes.nextInt();
					} while (srcNodeId < NodeConfig.getNumberAPs());
					do {
						dstNodeId = rngDstNodes.nextInt();
					} while (dstNodeId == srcNodeId
							|| dstNodeId < NodeConfig.getNumberAPs());
					// Create flow(s).
					final int startTime = rngStartTimes.nextInt();
					final int srcPort = rngPorts.nextInt();
					final int dstPort = rngPorts.nextInt();
					WirelessNode.getNode(srcNodeId).createTcpSource(
							connectionId, srcNodeId, dstNodeId, srcPort,
							dstPort, startTime, meshTraffic.getPacketSize(),
							meshTraffic.getSendInterval(),
							meshTraffic.getPriority());
					WirelessNode.getNode(dstNodeId).createTcpSink(connectionId,
							srcNodeId, dstNodeId, srcPort, dstPort,
							meshTraffic.getPacketSize());
					connectionId++;
					if (meshTraffic.getDirection() == FlowDirection.BOTH) {
						final int otherStartTime = rngStartTimes.nextInt();
						final int otherSrcPort = rngPorts.nextInt();
						final int otherDstPort = rngPorts.nextInt();
						WirelessNode.getNode(dstNodeId).createTcpSource(
								connectionId, dstNodeId, srcNodeId,
								otherSrcPort, otherDstPort, otherStartTime,
								meshTraffic.getPacketSize(),
								meshTraffic.getSendInterval(),
								meshTraffic.getPriority());
						WirelessNode.getNode(srcNodeId).createTcpSink(
								connectionId, dstNodeId, srcNodeId,
								otherSrcPort, otherDstPort,
								meshTraffic.getPacketSize());
						connectionId++;
					}
				} else if (meshTraffic.getType() == FlowType.CLIENT_TO_PORTAL) {
					// Select endpoints.
					do {
						srcNodeId = rngSrcNodes.nextInt();
					} while (srcNodeId < NodeConfig.getNumberAPs());
					final int associatedApId = WirelessNode
							.getAssociatedAp(srcNodeId);
					int portalNodeId;
					if (WirelessNode.isMeshPortal(associatedApId)) {
						// Is the associated AP already a portal?
						portalNodeId = associatedApId;
					} else {
						int portalId = numDummies
								% NodeConfig.getNumberPortals();
						portalNodeId = WirelessNode.getPortalNodeId(portalId);
					}
					final int partitionId = WirelessNode.getNode(portalNodeId)
							.getPartitionId();
					final boolean isLocalPortal = (partitionId != DistInitHelperNode.REMOTE_PARTITION);
					if (isLocalPortal) {
						dstNodeId = DummyNode.addDummy(srcNodeId, numDummies,
								partitionId, portalNodeId);
					} else {
						// This dummy has to be instantiated on a remote host.
						dstNodeId = DummyNode.addRemoteDummy(srcNodeId,
								numDummies);
					}
					numDummies++;
					// Create flow(s).
					final int startTime = rngStartTimes.nextInt();
					final int srcPort = rngPorts.nextInt();
					final int dstPort = rngPorts.nextInt();
					WirelessNode.getNode(srcNodeId).createTcpSource(
							connectionId, srcNodeId, dstNodeId, srcPort,
							dstPort, startTime, meshTraffic.getPacketSize(),
							meshTraffic.getSendInterval(),
							meshTraffic.getPriority());
					if (isLocalPortal) {
						DummyNode.getNode(dstNodeId).createTcpSink(
								connectionId, srcNodeId, dstNodeId, srcPort,
								dstPort, meshTraffic.getPacketSize());
					}
					connectionId++;
					if (meshTraffic.getDirection() == FlowDirection.BOTH) {
						final int otherStartTime = rngStartTimes.nextInt();
						final int otherSrcPort = rngPorts.nextInt();
						final int otherDstPort = rngPorts.nextInt();
						if (isLocalPortal) {
							DummyNode.getNode(dstNodeId).createTcpSource(
									connectionId, dstNodeId, srcNodeId,
									otherSrcPort, otherDstPort, otherStartTime,
									meshTraffic.getPacketSize(),
									meshTraffic.getSendInterval(),
									meshTraffic.getPriority());
						}
						WirelessNode.getNode(srcNodeId).createTcpSink(
								connectionId, dstNodeId, srcNodeId,
								otherSrcPort, otherDstPort,
								meshTraffic.getPacketSize());
						connectionId++;
					}
				} else if (meshTraffic.getType() == FlowType.PORTAL_TO_CLIENT) {
					// Select endpoints.
					do {
						dstNodeId = rngDstNodes.nextInt();
					} while (dstNodeId < NodeConfig.getNumberAPs());
					final int associatedApId = WirelessNode
							.getAssociatedAp(dstNodeId);
					int portalNodeId;
					if (WirelessNode.isMeshPortal(associatedApId)) {
						// Is the associated AP already a portal?
						portalNodeId = associatedApId;
					} else {
						int portalId = numDummies
								% NodeConfig.getNumberPortals();
						portalNodeId = WirelessNode.getPortalNodeId(portalId);
					}
					final int partitionId = WirelessNode.getNode(portalNodeId)
							.getPartitionId();
					final boolean isLocalDummy = (partitionId != DistInitHelperNode.REMOTE_PARTITION);
					if (isLocalDummy) {
						srcNodeId = DummyNode.addDummy(dstNodeId, numDummies,
								partitionId, portalNodeId);
					} else {
						// This dummy has to be instantiated on a remote host.
						srcNodeId = DummyNode.addRemoteDummy(dstNodeId,
								numDummies);
					}
					numDummies++;
					// Create flow(s).
					final int startTime = rngStartTimes.nextInt();
					final int srcPort = rngPorts.nextInt();
					final int dstPort = rngPorts.nextInt();
					if (isLocalDummy) {
						DummyNode.getNode(srcNodeId).createTcpSource(
								connectionId, srcNodeId, dstNodeId, srcPort,
								dstPort, startTime,
								meshTraffic.getPacketSize(),
								meshTraffic.getSendInterval(),
								meshTraffic.getPriority());
					}
					WirelessNode.getNode(dstNodeId).createTcpSink(connectionId,
							srcNodeId, dstNodeId, srcPort, dstPort,
							meshTraffic.getPacketSize());
					connectionId++;
					if (meshTraffic.getDirection() == FlowDirection.BOTH) {
						final int otherSrcPort = rngPorts.nextInt();
						final int otherDstPort = rngPorts.nextInt();
						final int otherStartTime = rngStartTimes.nextInt();
						WirelessNode.getNode(dstNodeId).createTcpSource(
								connectionId, dstNodeId, srcNodeId,
								otherSrcPort, otherDstPort, otherStartTime,
								meshTraffic.getPacketSize(),
								meshTraffic.getSendInterval(),
								meshTraffic.getPriority());
						if (isLocalDummy) {
							DummyNode.getNode(srcNodeId).createTcpSink(
									connectionId, dstNodeId, srcNodeId,
									otherSrcPort, otherDstPort,
									meshTraffic.getPacketSize());
						}
						connectionId++;
					}
				}
			}
		}
	}

	/**
	 * Creates random TCP traffic for MANET scenarios.
	 * 
	 * @param trafficTCP
	 *            the TCP traffic info.
	 * @param rngSrcNodes
	 *            the RNG for selecting source nodes.
	 * @param rngDstNodes
	 *            the RNG for selecting destination nodes.
	 */
	private void createTcpTrafficManet(
			final List<RandomTrafficInfoTCP> trafficTCP,
			final RandomStreamInt rngSrcNodes, final RandomStreamInt rngDstNodes) {
		for (RandomTrafficInfoTCP traffic : trafficTCP) {
			final RandomStreamInt rngStartTimes = RandomManager.getInstance()
					.getUniformIntSubstream(0, (int) traffic.getMaxStartTime());
			final RandomStreamInt rngPorts = RandomManager.getInstance()
					.getUniformIntSubstream(0, 1024);
			int numFlows = traffic.getNumFlows();
			for (int connectionId = 0; connectionId < numFlows; connectionId++) {
				final int srcNodeId = rngSrcNodes.nextInt();
				int dstNodeId;
				do {
					dstNodeId = rngDstNodes.nextInt();
				} while (dstNodeId == srcNodeId);
				final long startTime = rngStartTimes.nextInt();
				final int srcPort = rngPorts.nextInt();
				final int dstPort = rngPorts.nextInt();
				WirelessNode.getNode(srcNodeId).createTcpSource(connectionId,
						srcNodeId, dstNodeId, srcPort, dstPort, startTime,
						traffic.getPacketSize(), traffic.getSendInterval(),
						FlowPriority.AC_BE);
				WirelessNode.getNode(dstNodeId).createTcpSink(connectionId,
						srcNodeId, dstNodeId, srcPort, dstPort,
						traffic.getPacketSize());
			}
		}
	}

	/**
	 * Creates random UDP traffic between nodes, as specified by the input file.
	 * 
	 * @param nodeCount
	 *            the total number of nodes of the simulation.
	 */
	private void createUdpTraffic(final int nodeCount) {
		final List<RandomTrafficInfoUDP> trafficUDP = NodeConfig
				.getTrafficInfoUDP();
		if (trafficUDP == null) {
			return;
		}
		// Substreams for setting up traffic connections.
		final RandomStreamInt rngSrcNodes = RandomManager.getInstance()
				.getUniformIntSubstream(0, nodeCount - 1);
		final RandomStreamInt rngDstNodes = RandomManager.getInstance()
				.getUniformIntSubstream(0, nodeCount - 1);
		if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MANET) {
			createUdpTrafficManet(trafficUDP, rngSrcNodes, rngDstNodes);
		} else if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH) {
			createUdpTrafficMesh(trafficUDP, rngSrcNodes, rngDstNodes);
		}
	}

	/**
	 * Creates random UDP traffic for mesh scenarios.
	 * 
	 * @param trafficUDP
	 *            the UDP traffic info.
	 * @param rngSrcNodes
	 *            the RNG for selecting source nodes.
	 * @param rngDstNodes
	 *            the RNG for selecting destination nodes.
	 */
	private void createUdpTrafficMesh(
			final List<RandomTrafficInfoUDP> trafficUDP,
			final RandomStreamInt rngSrcNodes, final RandomStreamInt rngDstNodes) {
		int connectionId = 0;
		int numDummies = 0;
		// Loop over traffic packages.
		for (RandomTrafficInfoUDP traffic : trafficUDP) {
			final RandomStreamInt rngStartTimes = RandomManager.getInstance()
					.getUniformIntSubstream(0, (int) traffic.getMaxStartTime());
			final MeshTrafficInfoUDP meshTraffic = (MeshTrafficInfoUDP) traffic;
			// Create flows for this traffic package.
			final int numFlows = (meshTraffic.getDirection() == FlowDirection.BOTH) ? meshTraffic
					.getNumFlows() / 2 : meshTraffic.getNumFlows();
			for (int i = 0; i < numFlows; i++) {
				int srcNodeId = 0;
				int dstNodeId = 0;
				if (meshTraffic.getType() == FlowType.CLIENT_TO_CLIENT) {
					// Select endpoints.
					do {
						srcNodeId = rngSrcNodes.nextInt();
					} while (srcNodeId < NodeConfig.getNumberAPs());
					do {
						dstNodeId = rngDstNodes.nextInt();
					} while (dstNodeId == srcNodeId
							|| dstNodeId < NodeConfig.getNumberAPs());
					// Set up flow(s).
					final int startTime = rngStartTimes.nextInt();
					WirelessNode.getNode(srcNodeId).startUdpFlow(connectionId,
							dstNodeId, startTime, meshTraffic.getPacketSize(),
							meshTraffic.getSendInterval(),
							meshTraffic.getPriority());
					connectionId++;
					if (meshTraffic.getDirection() == FlowDirection.BOTH) {
						final int otherStartTime = rngStartTimes.nextInt();
						WirelessNode.getNode(dstNodeId).startUdpFlow(
								connectionId, srcNodeId, otherStartTime,
								meshTraffic.getPacketSize(),
								meshTraffic.getSendInterval(),
								meshTraffic.getPriority());
						connectionId++;
					}
				} else if (meshTraffic.getType() == FlowType.CLIENT_TO_PORTAL) {
					// Select endpoints.
					do {
						srcNodeId = rngSrcNodes.nextInt();
					} while (srcNodeId < NodeConfig.getNumberAPs());
					final int associatedApId = WirelessNode
							.getAssociatedAp(srcNodeId);
					int portalNodeId;
					if (WirelessNode.isMeshPortal(associatedApId)) {
						// Is the associated AP already a portal?
						portalNodeId = associatedApId;
					} else {
						int portalId = numDummies
								% NodeConfig.getNumberPortals();
						portalNodeId = WirelessNode.getPortalNodeId(portalId);
					}
					final int partitionId = WirelessNode.getNode(portalNodeId)
							.getPartitionId();
					final boolean isLocalPortal = (partitionId != DistInitHelperNode.REMOTE_PARTITION);
					if (isLocalPortal) {
						dstNodeId = DummyNode.addDummy(srcNodeId, numDummies,
								partitionId, portalNodeId);
					} else {
						// This dummy has to be instantiated on a remote host.
						dstNodeId = DummyNode.addRemoteDummy(srcNodeId,
								numDummies);
					}
					numDummies++;
					// Set up flow(s).
					final int startTime = rngStartTimes.nextInt();
					WirelessNode.getNode(srcNodeId).startUdpFlow(connectionId,
							dstNodeId, startTime, meshTraffic.getPacketSize(),
							meshTraffic.getSendInterval(),
							meshTraffic.getPriority());
					connectionId++;
					if (meshTraffic.getDirection() == FlowDirection.BOTH) {
						final int otherStartTime = rngStartTimes.nextInt();
						if (isLocalPortal) {
							DummyNode.getNode(dstNodeId).startUdpFlow(
									connectionId, srcNodeId, otherStartTime,
									meshTraffic.getPacketSize(),
									meshTraffic.getSendInterval(),
									meshTraffic.getPriority());
						}
						connectionId++;
					}
				} else if (meshTraffic.getType() == FlowType.PORTAL_TO_CLIENT) {
					// Select endpoints.
					do {
						dstNodeId = rngDstNodes.nextInt();
					} while (dstNodeId < NodeConfig.getNumberAPs());
					final int associatedApId = WirelessNode
							.getAssociatedAp(dstNodeId);
					int portalNodeId;
					if (WirelessNode.isMeshPortal(associatedApId)) {
						// Is the associated AP already a portal?
						portalNodeId = associatedApId;
					} else {
						int portalId = numDummies
								% NodeConfig.getNumberPortals();
						portalNodeId = WirelessNode.getPortalNodeId(portalId);
					}
					final int partitionId = WirelessNode.getNode(portalNodeId)
							.getPartitionId();
					final boolean isLocalDummy = (partitionId != DistInitHelperNode.REMOTE_PARTITION);
					if (isLocalDummy) {
						srcNodeId = DummyNode.addDummy(dstNodeId, numDummies,
								partitionId, portalNodeId);
					} else {
						// This dummy has to be instantiated on a remote host.
						srcNodeId = DummyNode.addRemoteDummy(dstNodeId,
								numDummies);
					}
					numDummies++;
					// Set up flow(s).
					final int startTime = rngStartTimes.nextInt();
					if (isLocalDummy) {
						DummyNode.getNode(srcNodeId).startUdpFlow(connectionId,
								dstNodeId, startTime,
								meshTraffic.getPacketSize(),
								meshTraffic.getSendInterval(),
								meshTraffic.getPriority());
					}
					connectionId++;
					if (meshTraffic.getDirection() == FlowDirection.BOTH) {
						final int otherStartTime = rngStartTimes.nextInt();
						WirelessNode.getNode(dstNodeId).startUdpFlow(
								connectionId, srcNodeId, otherStartTime,
								meshTraffic.getPacketSize(),
								meshTraffic.getSendInterval(),
								meshTraffic.getPriority());
						connectionId++;
					}
				}
			}
		}
	}

	/**
	 * Creates random UDP traffic for MANET scenarios.
	 * 
	 * @param trafficUDP
	 *            the UDP traffic info.
	 * @param rngSrcNodes
	 *            the RNG for selecting source nodes.
	 * @param rngDstNodes
	 *            the RNG for selecting destination nodes.
	 */
	private void createUdpTrafficManet(
			final List<RandomTrafficInfoUDP> trafficUDP,
			final RandomStreamInt rngSrcNodes, final RandomStreamInt rngDstNodes) {
		for (RandomTrafficInfoUDP traffic : trafficUDP) {
			final RandomStreamInt rngStartTimes = RandomManager.getInstance()
					.getUniformIntSubstream(0, (int) traffic.getMaxStartTime());
			int numFlows = traffic.getNumFlows();
			for (int connectionId = 0; connectionId < numFlows; connectionId++) {
				final long startTime = rngStartTimes.nextInt();
				final int srcNode = rngSrcNodes.nextInt();
				int dstNode;
				do {
					dstNode = rngDstNodes.nextInt();
				} while (dstNode == srcNode);
				WirelessNode.getNode(srcNode).startUdpFlow(connectionId,
						dstNode, startTime, traffic.getPacketSize(),
						traffic.getSendInterval(), FlowPriority.AC_BE);
			}
		}
	}

	/**
	 * @see modules.wireless.scenario.SimulationCallback#endSimulation(core.sync.AbstractSimController)
	 */
	@Override
	public void endSimulation(AbstractSimController simCtrl) {
		long tcpRecv = WirelessNode.getAllTcpBytesReceived();
		long udpRecv = WirelessNode.getAllUdpBytesReceived();
		long simDur = Config.getEndTime() - Config.getStartTime();
		int throughNet = (int) ((tcpRecv + udpRecv) / ((double) simDur / jist.swans.Constants.SECOND));
		double dropPer = WirelessNode.getAllDroppedPackets() * 100.0
				/ WirelessNode.getAllIncomingPackets();
		int trans = WirelessNode.getAllNumBroadcasts()
				+ WirelessNode.getAllNumUnicasts();
		double broadPer = WirelessNode.getAllNumBroadcasts() * 100.0 / trans;
		long numSyncs = (simCtrl instanceof ParallelSimController) ? ((ParallelSimController) simCtrl)
				.getNumSyncs()
				: ((simCtrl instanceof DistributedSimController) ? ((DistributedSimController) simCtrl)
						.getNumSyncs() : 0);
		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("radio inc/dropped:       "
					+ WirelessNode.getAllIncomingPackets() + " / "
					+ WirelessNode.getAllDroppedPackets());
			CustomLogger.info("radio drop%:             " + dropPer);
			CustomLogger.info("captured signals:        "
					+ WirelessNode.getAllCapturedSignals());
			if (NodeConfig.getScenarioType() == ScenarioType.MESH) {
				tcpRecv += DummyNode.getTcpBytesReceived();
				udpRecv += DummyNode.getUdpBytesReceived();
				long tcpSend = DummyNode.getTcpBytesSent()
						+ WirelessNode.getAllTcpBytesSent();
				long udpSend = DummyNode.getUdpBytesSent()
						+ WirelessNode.getAllUdpBytesSent();
				throughNet = (int) ((tcpRecv + udpRecv) / ((double) simDur / jist.swans.Constants.SECOND));

				CustomLogger.info("tcp send/receive:        " + tcpSend + " / "
						+ tcpRecv);
				CustomLogger.info("udp send/receive:        " + udpSend + " / "
						+ udpRecv);
				CustomLogger.info("throughput (net):        " + throughNet
						+ " bytes/sec");
			} else {
				CustomLogger.info("tcp send/receive:        "
						+ WirelessNode.getAllTcpBytesSent() + " / " + tcpRecv);
				CustomLogger.info("udp send/receive:        "
						+ WirelessNode.getAllUdpBytesSent() + " / " + udpRecv);
				CustomLogger.info("throughput (net):        " + throughNet
						+ " bytes/sec");
			}
			CustomLogger.info("bytes sent (mac):        "
					+ WirelessNode.getAllMacBytesSent());
			CustomLogger.info("MacQueueOffered:         "
					+ WirelessNode.getAllMacQueueOffered());
			CustomLogger.info("MacQueueDropped:         "
					+ WirelessNode.getAllMacQueueDropped());
			CustomLogger.info("MacDirectOffered:        "
					+ WirelessNode.getAllDirectTransfers());
			CustomLogger.info("transmissions:           " + trans);
			CustomLogger.info("broadcasts:              "
					+ WirelessNode.getAllNumBroadcasts());
			CustomLogger.info("unicasts:                "
					+ WirelessNode.getAllNumUnicasts());
			CustomLogger.info("broadcast%:              " + broadPer);
			if (NodeConfig.getScenarioType() == ScenarioType.MANET) {
				CustomLogger.info("routing fwBroadcasts:    "
						+ WirelessNode.getAllFwBroadcasts());
				CustomLogger.info("routing rcvBroadcasts:   "
						+ WirelessNode.getAllRcvBroadcasts());
			} else if (NodeConfig.getScenarioType() == ScenarioType.MESH) {
				CustomLogger.info("802.11 transmissions:    "
						+ WirelessNode.getAll802_11Transmissions());
				CustomLogger.info("802.11s transmissions:   "
						+ WirelessNode.getAll802_11sTransmissions());
				CustomLogger
						.info("802.11s transmissions%:  "
								+ 100.0
								* WirelessNode.getAll802_11sTransmissions()
								/ (WirelessNode.getAll802_11sTransmissions() + WirelessNode
										.getAll802_11Transmissions()));
				CustomLogger.info("Dummy transmissions:     "
						+ DummyNode.getDummyTransmissions());
				CustomLogger.info("HWMP Queue Drops:        "
						+ WirelessNode.getAllHwmpDrops());
				CustomLogger.info("number of beacons:       "
						+ WirelessNode.getAllNumBeacons());
			}
			CustomLogger.info("number of syncs:         " + numSyncs);
			if (Config.usesLazyMac()
					&& Config.getSimType() == Config.SimType.SEQUENTIAL) {
				LazySinglePartition partition = (LazySinglePartition) Partition
						.getPartition(0);
				partition.printLazyStats();
			}

			if (Config.getSimType() == SimType.DISTRIBUTED) {
				DistributedSimController distSimCtrl = (DistributedSimController) simCtrl;
				StatsPacket statsPacket = distSimCtrl.getStatsPacket();
				statsPacket.radioInc = WirelessNode.getAllIncomingPackets();
				statsPacket.radioDrop = WirelessNode.getAllDroppedPackets();
				statsPacket.captSignals = WirelessNode.getAllCapturedSignals();
				statsPacket.tcpRecv = tcpRecv;
				statsPacket.udpRecv = udpRecv;
				statsPacket.tcpSend = WirelessNode.getAllTcpBytesSent();
				statsPacket.udpSend = WirelessNode.getAllUdpBytesSent();
				statsPacket.throughNet = throughNet;
				statsPacket.macBytesSend = WirelessNode.getAllMacBytesSent();
				statsPacket.macQueueOffered = WirelessNode
						.getAllMacQueueOffered();
				statsPacket.macQueueDropped = WirelessNode
						.getAllMacQueueDropped();
				statsPacket.directOffered = WirelessNode
						.getAllDirectTransfers();
				statsPacket.transmissions = trans;
				statsPacket.broadcasts = WirelessNode.getAllNumBroadcasts();
				statsPacket.unicasts = WirelessNode.getAllNumUnicasts();
				if (NodeConfig.getScenarioType() == ScenarioType.MANET) {
					statsPacket.routingRcvBroadcasts = WirelessNode
							.getAllRcvBroadcasts();
					statsPacket.routingFwBroadcasts = WirelessNode
							.getAllFwBroadcasts();
				} else if (NodeConfig.getScenarioType() == ScenarioType.MESH) {
					statsPacket.transmissions802_11 = WirelessNode
							.getAll802_11Transmissions();
					statsPacket.transmissions802_11s = WirelessNode
							.getAll802_11sTransmissions();
					statsPacket.dummyTransmissions = DummyNode
							.getDummyTransmissions();
					statsPacket.hwmpDrops = WirelessNode.getAllHwmpDrops();
					statsPacket.numBeacons = WirelessNode.getAllNumBeacons();
				}
			}
		}
	}
}
