/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/scenario/DistInitHelperNode.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.scenario;

import modules.wireless.scenario.NodeConfig.FlowPriority;
import rng.RandomManager;
import core.Config;

/**
 * Helper node for initializing distributed simulations.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class DistInitHelperNode extends WirelessNode {
	/**
	 * Constant indicating a remote partition.
	 */
	public static final int REMOTE_PARTITION = -1;

	/**
	 * Helper node to manage the correct initialization of scenarios in
	 * distributed simulations with remote nodes.
	 */
	private static DistInitHelperNode helperNode;

	/**
	 * Registers a remote node for MANET scenarios.
	 * 
	 * @param nodeId
	 *            ID of this node.
	 * @param partitionId
	 *            ID of the partition this node belongs to.
	 * @param rngId
	 *            ID of the virtual partition this node belongs to.
	 */
	public static void registerRemoteNode(int nodeId, int partitionId, int rngId) {
		if (helperNode == null) {
			// Includes registration.
			helperNode = new DistInitHelperNode();
		}
		assert nodes != null;
		assert nodeId < nodes.length;
		assert NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MANET;
		nodes[nodeId] = helperNode;
		helperNode.createDummyNode(nodeId, partitionId, rngId);
	}

	/**
	 * Registers a remote node for mesh scenarios.
	 * 
	 * @param nodeId
	 *            ID of this node.
	 * @param partitionId
	 *            ID of the partition this node belongs to.
	 * @param rngId
	 *            ID of the virtual partition this node belongs to.
	 * @param type
	 *            type of this node.
	 * @param portalId
	 *            a portal ID, if this node is a mesh portal, otherwise
	 *            NO_PORTAL.
	 */
	public static void registerRemoteNode(int nodeId, int partitionId,
			int rngId, NodeType type, int portalId) {
		if (helperNode == null) {
			// Includes registration.
			helperNode = new DistInitHelperNode();
		}
		assert nodes != null;
		assert nodeId < nodes.length;
		assert NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH;
		if (portalId != NO_PORTAL) {
			isMeshPortal[nodeId] = true;
			portalIdToNodeId.put(portalId, nodeId);
		}
		nodes[nodeId] = helperNode;
		helperNode.createDummyNode(nodeId, partitionId, rngId, type);
	}

	/**
	 * Constructor.
	 */
	DistInitHelperNode() {
		// Empty.
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "DistInitHelperNode";
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getRngId()
	 */
	@Override
	public int getRngId() {
		assert false;
		return -1;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getPartitionId()
	 */
	@Override
	public int getPartitionId() {
		return REMOTE_PARTITION;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getNodeType()
	 */
	@Override
	public NodeType getNodeType() {
		assert false;
		return null;
	}

	/**
	 * Registers a remote node for MANET scenarios.
	 * 
	 * @param nodeId
	 *            ID of this node.
	 * @param partitionId
	 *            ID of the partition this node belongs to.
	 * @param rngId
	 *            ID of the virtual partition this node belongs to.
	 */
	void createDummyNode(int nodeId, int partitionId, int rngId) {
		// MAC / lazy MAC.
		DummyModule.registerDummyModule();
		DummyModule.registerDummyModule();
		RandomManager.getInstance().getUniform01Substream();

		// Network.
		DummyModule.registerDummyModule();

		// Routing.
		switch (NodeConfig.getRoutingProtocol()) {
		case AODV:
			// Fall through.
		case DSR:
			DummyModule.registerDummyModule();
			RandomManager.getInstance().getUniformIntSubstream(0, 1);
			break;
		case NONE:
			break;
		default:
			throw new RuntimeException("unknown routing protocol!");
		}

		// Transport.
		if (NodeConfig.getTrafficInfoUDP() != null) {
			DummyModule.registerDummyModule();
		}
		if (NodeConfig.getTrafficInfoTCP() != null) {
			DummyModule.registerDummyModule();
			RandomManager.getInstance().getUniform01Substream();
		}

		// Packet generator.
		DummyModule.registerDummyModule();
	}

	/**
	 * Registers a remote node for mesh scenarios.
	 * 
	 * @param nodeId
	 *            ID of this node.
	 * @param partitionId
	 *            ID of the partition this node belongs to.
	 * @param rngId
	 *            ID of the virtual partition this node belongs to.
	 * @param type
	 *            type of this node.
	 */
	void createDummyNode(int nodeId, int partitionId, int rngId, NodeType type) {
		// Radio and MAC layer.
		DummyModule.registerDummyModule();
		DummyModule.registerDummyModule();
		RandomManager.getInstance().getUniform01Substream();

		if (!Config.usesLazyMac() && type == NodeType.AP) {
			// Mesh MAC and radio.
			DummyModule.registerDummyModule();
			DummyModule.registerDummyModule();
			RandomManager.getInstance().getUniform01Substream();
			// HWMP.
			DummyModule.registerDummyModule();
			RandomManager.getInstance().getUniformIntSubstream(0, 1);
			// Portal.
			if (isMeshPortal[nodeId]) {
				DummyModule.registerDummyModule();
				RandomManager.getInstance().getNormalSubstream(0, 1);
			}
		}

		// Network layer.
		DummyModule.registerDummyModule();

		// Create routing module.
		switch (NodeConfig.getRoutingProtocol()) {
		case AODV:
			// Fall through.
		case DSR:
			DummyModule.registerDummyModule();
			RandomManager.getInstance().getUniformIntSubstream(0, 1);
			break;
		case NONE:
			break;
		default:
			throw new RuntimeException("unknown routing protocol!");
		}

		// Transport.
		if (NodeConfig.getTrafficInfoUDP() != null) {
			DummyModule.registerDummyModule();
		}
		if (NodeConfig.getTrafficInfoTCP() != null) {
			DummyModule.registerDummyModule();
			RandomManager.getInstance().getUniform01Substream();
		}

		// Packet generator.
		DummyModule.registerDummyModule();
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#startUdpFlow(int, int, long,
	 *      int, long, modules.wireless.scenario.NodeConfig.FlowPriority)
	 */
	@Override
	void startUdpFlow(int connectionId, int dstNodeId, long startTime,
			int packetSize, long sendInterval, FlowPriority priority) {
		// Do nothing.
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#createTcpSource(int, int,
	 *      int, int, int, long, int, long,
	 *      modules.wireless.scenario.NodeConfig.FlowPriority)
	 */
	@Override
	void createTcpSource(int connectionId, int srcNodeId, int dstNodeId,
			int srcPort, int dstPort, long startTime, int packetSize,
			long sendInterval, FlowPriority priority) {
		// Do nothing.
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#createTcpSink(int, int, int,
	 *      int, int, int)
	 */
	@Override
	void createTcpSink(int connectionId, int srcNodeId, int dstNodeId,
			int srcPort, int dstPort, int packetSize) {
		// Do nothing.
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#associateAp(int)
	 */
	@Override
	void associateAp(int apId) {
		// Do nothing.
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#associateClient(int)
	 */
	@Override
	void associateClient(int clientId) {
		// Do nothing.
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getTcpBytesSent()
	 */
	@Override
	protected long getTcpBytesSent() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getTcpBytesReceived()
	 */
	@Override
	protected long getTcpBytesReceived() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getUdpBytesSent()
	 */
	@Override
	protected long getUdpBytesSent() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getUdpBytesReceived()
	 */
	@Override
	protected long getUdpBytesReceived() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getMacBytesSent()
	 */
	@Override
	protected long getMacBytesSent() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getMacQueueOffered()
	 */
	@Override
	protected int getMacQueueOffered() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getMacQueueDropped()
	 */
	@Override
	protected int getMacQueueDropped() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getDirectTransfers()
	 */
	@Override
	protected int getDirectTransfers() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getIncomingPackets()
	 */
	@Override
	protected int getIncomingPackets() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getDroppedPackets()
	 */
	@Override
	protected int getDroppedPackets() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getCapturedSignals()
	 */
	@Override
	protected int getCapturedSignals() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getNumBroadcasts()
	 */
	@Override
	protected int getNumBroadcasts() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getNumUnicasts()
	 */
	@Override
	protected int getNumUnicasts() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getFwBroadcasts()
	 */
	@Override
	protected int getFwBroadcasts() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getRcvBroadcasts()
	 */
	@Override
	protected int getRcvBroadcasts() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#get802_11Transmissions()
	 */
	@Override
	protected int get802_11Transmissions() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#get802_11sTransmissions()
	 */
	@Override
	protected int get802_11sTransmissions() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getHwmpDrops()
	 */
	@Override
	protected int getHwmpDrops() {
		return 0;
	}

	/**
	 * @see modules.wireless.scenario.WirelessNode#getNumBeacons()
	 */
	@Override
	protected int getNumBeacons() {
		return 0;
	}
}