/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/scenario/NodeConfig.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.scenario;

import java.util.ArrayList;
import java.util.List;

import jist.swans.Constants;

/**
 * Configuration class for scenario-related (wireless node-related) stuff.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class NodeConfig {
	/**
	 * Offset for internet addresses (used by the mesh model).
	 */
	public static final int INTERNET_ADDRESS_OFFSET = 2000000;

	/**
	 * Offset for mesh network addresses (used by the mesh model).
	 */
	public static final int MESH_ADDRESS_OFFSET = 1000000;

	/**
	 * Whether the simulation has already started.
	 */
	private static boolean isRunning = false;

	/**
	 * Routing protocol used for ad hoc routing.
	 */
	private static RoutingProtocol routingProtocol = RoutingProtocol.AODV;

	/**
	 * Transmit power of each node.
	 * 
	 * Some orientation for TwoRay path loss:<br>
	 * RX range | TX power<br>
	 * 100m | -0.94792 (CS until ~270m)<br>
	 * 200m | 5.07268 (CS until ~400m)<br>
	 * 300m | 11.04120<br>
	 * ~376m | 15.0 (Constants.TRANSMIT_DEFAULT)<br>
	 * 400m | 16.03875 (CS until ~700m)<br>
	 * 500m | 19.91515<br>
	 * 600m | 23.08240<br>
	 * 700m | 25.76028<br>
	 * 800m | 28.07995<br>
	 * 1420m | 38.04789<br>
	 */
	private static double transmitPower_dBm = Constants.TRANSMIT_DEFAULT;

	/**
	 * A distance limit for transmissions (optional).
	 */
	private static double distanceLimit = 700;

	/**
	 * Whether to use the distance limit.
	 */
	private static boolean useDistanceLimit = false;

	/**
	 * The RxTxTurnaround to be used by IEEE 802.11.
	 */
	private static long rxTxTurnaround = Constants.MICRO_SECOND;

	/**
	 * The TCP traffic information.
	 */
	private static List<RandomTrafficInfoTCP> trafficInfoTCP = null;

	/**
	 * The UDP traffic information.
	 */
	private static List<RandomTrafficInfoUDP> trafficInfoUDP = null;

	/**
	 * The scenario type.
	 */
	private static ScenarioType scenarioType = null;

	/**
	 * Number of access points (used in mesh model).
	 */
	private static int numberAPs = 0;

	/**
	 * Number of portals (used in mesh model).
	 */
	private static int numberPortals = 0;

	/**
	 * Used flow priorities (used in mesh model).
	 */
	private static boolean[] usedFlowPriorities = null;

	/**
	 * Maximum used flow priority (used in mesh model).
	 */
	private static int maxUsedFlowPriorities = -1;

	/**
	 * Called when the simulation starts (i.e., everything is configured).
	 */
	public static void simulationStarts() {
		isRunning = true;
	}

	/**
	 * Sets the routing protocol.
	 * 
	 * @param arg
	 *            the routing protocol.
	 */
	public static void setRoutingProtocol(RoutingProtocol arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		routingProtocol = arg;
	}

	/**
	 * Sets the transmit power (in dBm).
	 * 
	 * @param arg
	 *            the transmit power.
	 */
	public static void setTransmitPower_dBm(double arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		transmitPower_dBm = arg;
	}

	/**
	 * Sets the (optional) distance limit.
	 * 
	 * @param arg
	 *            the distance limit.
	 */
	public static void setDistanceLimit(double arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		distanceLimit = arg;
		useDistanceLimit = true;
	}

	/**
	 * Sets the RxTxTurnaround (in simulation time units).
	 * 
	 * @param arg
	 *            the RxTxTurnaround.
	 */
	public static void setRxTxTurnaround(long arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		rxTxTurnaround = arg;
	}

	/**
	 * Sets the TCP traffic information. Called by the groovy script.
	 * 
	 * @param arg
	 *            the TCP traffic information.
	 */
	public static void addTrafficTCP(RandomTrafficInfoTCP arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		if (arg.getMaxStartTime() < 0 || arg.getNumFlows() < 1
				|| arg.getPacketSize() < 0 || arg.getSendInterval() < 0) {
			throw new RuntimeException("Weird TCP traffic info: " + arg);
		}
		if (trafficInfoTCP == null) {
			trafficInfoTCP = new ArrayList<RandomTrafficInfoTCP>();
		}
		trafficInfoTCP.add(arg);
	}

	/**
	 * Sets the UDP traffic information. Called by the groovy script.
	 * 
	 * @param arg
	 *            the UDP traffic information.
	 */
	public static void addTrafficUDP(RandomTrafficInfoUDP arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		if (arg.getMaxStartTime() < 0 || arg.getNumFlows() < 1
				|| arg.getPacketSize() < 0 || arg.getSendInterval() <= 0) {
			throw new RuntimeException("Weird UDP traffic info: " + arg);
		}
		if (trafficInfoUDP == null) {
			trafficInfoUDP = new ArrayList<RandomTrafficInfoUDP>();
		}
		trafficInfoUDP.add(arg);
	}

	/**
	 * Sets the scenario type.
	 * 
	 * @param arg
	 *            the scenario type.
	 */
	public static void setScenarioType(ScenarioType arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		scenarioType = arg;
	}

	/**
	 * Sets the number of access points (used by mesh model).
	 * 
	 * @param arg
	 *            the number of access points.
	 */
	public static void setNumAPs(int arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		numberAPs = arg;
	}

	/**
	 * Sets the number of portals (used by mesh model).
	 * 
	 * @param arg
	 *            the number of portals.
	 */
	public static void setNumPortals(int arg) {
		if (isRunning) {
			invalidConfigChange(arg);
		}
		assert numberPortals <= numberAPs;
		numberPortals = arg;
	}

	/**
	 * Provides the routing protocol.
	 * 
	 * @return the routing protocol.
	 */
	public static RoutingProtocol getRoutingProtocol() {
		if (routingProtocol == null) {
			throw new RuntimeException(
					"Forgot to set routing protocol by calling class "
							+ NodeConfig.class.getSimpleName() + " ?");
		}
		return routingProtocol;
	}

	/**
	 * Provides the transmit power of each node in dBm.
	 * 
	 * @return the transmit power.
	 */
	public static double getTransmitPower() {
		return transmitPower_dBm;
	}

	/**
	 * Provides the (optional) distance limit.
	 * 
	 * @return the distance limit.
	 */
	public static double getDistanceLimit() {
		return distanceLimit;
	}

	/**
	 * Whether the distance limit is used.
	 * 
	 * @return whether the distance limit is used.
	 */
	public static boolean useDistanceLimit() {
		return useDistanceLimit;
	}

	/**
	 * Provides the RxTxTurnaround.
	 * 
	 * @return the RxTxTurnaround.
	 */
	public static long getRxTxTurnaround() {
		return rxTxTurnaround;
	}

	/**
	 * Provides the TCP traffic information.
	 * 
	 * @return the TCP traffic information.
	 */
	public static List<RandomTrafficInfoTCP> getTrafficInfoTCP() {
		return trafficInfoTCP;
	}

	/**
	 * Provides the UDP traffic information.
	 * 
	 * @return the UDP traffic information.
	 */
	public static List<RandomTrafficInfoUDP> getTrafficInfoUDP() {
		return trafficInfoUDP;
	}

	/**
	 * Provides the scenario type.
	 * 
	 * @return the scenario type.
	 */
	public static ScenarioType getScenarioType() {
		if (scenarioType == null) {
			throw new RuntimeException(
					"Forgot to set scenario type by calling class "
							+ NodeConfig.class.getSimpleName() + " ?");
		}
		return scenarioType;
	}

	/**
	 * Provides the number of access points.
	 * 
	 * @return the number of access points.
	 */
	public static int getNumberAPs() {
		return numberAPs;
	}

	/**
	 * Provides the number of portals.
	 * 
	 * @return the number of portals.
	 */
	public static int getNumberPortals() {
		return numberPortals;
	}

	/**
	 * Provides the number of TCP flows.
	 * 
	 * @return the number of TCP flows.
	 */
	public static int getNumFlowsTCP() {
		if (trafficInfoTCP == null) {
			return 0;
		}
		int number = 0;
		for (RandomTrafficInfoTCP flowInfo : trafficInfoTCP) {
			number += flowInfo.getNumFlows();
		}
		return number;
	}

	/**
	 * Provides the number of UDP flows.
	 * 
	 * @return the number of UDP flows.
	 */
	public static int getNumFlowsUDP() {
		if (trafficInfoUDP == null) {
			return 0;
		}
		int number = 0;
		for (RandomTrafficInfoUDP flowInfo : trafficInfoUDP) {
			number += flowInfo.getNumFlows();
		}
		return number;
	}

	/**
	 * Provides the flow priority for TCP traffic flows.
	 * 
	 * @return the flow priority.
	 */
	public static FlowPriority getFlowPriority() {
		if (trafficInfoTCP == null) {
			return FlowPriority.AC_BK;
		}
		FlowPriority result = FlowPriority.AC_BK;
		for (RandomTrafficInfoTCP flowInfo : trafficInfoTCP) {
			MeshTrafficInfoTCP meshTraffic = (MeshTrafficInfoTCP) flowInfo;
			result = meshTraffic.getPriority();
			break;
		}
		return result;
	}

	/**
	 * Provides an array of used flow priorities.
	 * 
	 * @return an array of used flow priorities.
	 */
	public static boolean[] getUsedFlowPriorities() {
		if (usedFlowPriorities == null) {
			usedFlowPriorities = new boolean[4];
			usedFlowPriorities[1] = true;
			if (trafficInfoTCP != null) {
				for (RandomTrafficInfoTCP trafficInfo : trafficInfoTCP) {
					if (trafficInfo instanceof MeshTrafficInfoTCP) {
						MeshTrafficInfoTCP meshTrafficTCP = (MeshTrafficInfoTCP) trafficInfo;
						switch (meshTrafficTCP.priority_) {
						case AC_BK:
							usedFlowPriorities[0] = true;
							break;
						case AC_BE:
							usedFlowPriorities[1] = true;
							break;
						case AC_VI:
							usedFlowPriorities[2] = true;
							break;
						case AC_VO:
							usedFlowPriorities[3] = true;
							break;
						}
					}
				}
			}
			if (trafficInfoUDP != null) {
				for (RandomTrafficInfoUDP trafficInfo : trafficInfoUDP) {
					if (trafficInfo instanceof MeshTrafficInfoUDP) {
						MeshTrafficInfoUDP meshTrafficUDP = (MeshTrafficInfoUDP) trafficInfo;
						switch (meshTrafficUDP.priority_) {
						case AC_BK:
							usedFlowPriorities[0] = true;
							break;
						case AC_BE:
							usedFlowPriorities[1] = true;
							break;
						case AC_VI:
							usedFlowPriorities[2] = true;
							break;
						case AC_VO:
							usedFlowPriorities[3] = true;
							break;
						}
					}
				}
			}
		}
		return usedFlowPriorities;
	}

	/**
	 * Provides the highest used flow priority.
	 * 
	 * @return the highest used flow priority.
	 */
	public static int getMaxUsedFlowPriority() {
		if (usedFlowPriorities == null) {
			getUsedFlowPriorities();
		}
		if (maxUsedFlowPriorities < 0) {
			for (int i = usedFlowPriorities.length - 1; i >= 0; i--) {
				if (usedFlowPriorities[i]) {
					maxUsedFlowPriorities = i;
					break;
				}
			}
		}
		return maxUsedFlowPriorities;
	}

	/**
	 * Prints an error message in case of an attempt to change the configuration
	 * during the simulation.
	 * 
	 * @param arg
	 *            the modified configuration parameter.
	 */
	private static void invalidConfigChange(Object arg) {
		throw new RuntimeException(
				"Attempt to change node configuration parameters during the simulation: "
						+ arg);
	}

	/**
	 * Routing protocol type used in MANET scenarios.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	public static enum RoutingProtocol {
		/**
		 * Ad-hoc On-demand Distance Vector Routing.
		 */
		AODV,
		/**
		 * Dynamic Source Routing.
		 */
		DSR,
		/**
		 * No multi-hop routing (used by mesh clients).
		 */
		NONE;
	};

	/**
	 * Scenario type: MANET or Mesh network simulation.
	 * 
	 * @author Patrick Peschlow
	 * 
	 * @version 1.0
	 */
	public static enum ScenarioType {
		/**
		 * MANET.
		 */
		MANET,
		/**
		 * Mesh.
		 */
		MESH;
	};

	/**
	 * Flow type (where are the two endpoints located) in mesh scenarios.
	 * 
	 * @author Patrick Peschlow
	 * @author Daniel Plohmann
	 * 
	 * @version 1.0
	 */
	public static enum FlowType {
		/**
		 * Client to portal.
		 */
		CLIENT_TO_PORTAL,
		/**
		 * Portal to client.
		 */
		PORTAL_TO_CLIENT,
		/**
		 * Client to client.
		 */
		CLIENT_TO_CLIENT;
	};

	/**
	 * Traffic flow direction in mesh scenarios.
	 * 
	 * @author Patrick Peschlow
	 * @author Daniel Plohmann
	 * 
	 * @version 1.0
	 */
	public static enum FlowDirection {
		/**
		 * Unidirectional.
		 */
		SINGLE,
		/**
		 * Bidirectional.
		 */
		BOTH;
	};

	/**
	 * Flow priority type in mesh scenarios.
	 * 
	 * @author Patrick Peschlow
	 * @author Daniel Plohmann
	 * 
	 * @version 1.0
	 */
	public static enum FlowPriority {
		/**
		 * Access Category: Background.
		 */
		AC_BK,
		/**
		 * Access Category: Best Effort.
		 */
		AC_BE,
		/**
		 * Access Category: Video.
		 */
		AC_VI,
		/**
		 * Access Category: Voice.
		 */
		AC_VO;
	};

	/**
	 * Information about TCP traffic flows to set up randomly between nodes.
	 * 
	 * @author Patrick Peschlow
	 * @author Andreas Voss
	 * 
	 * @version 1.0
	 */
	public static class RandomTrafficInfoTCP {
		/**
		 * The number of flows to create.
		 */
		protected final int numFlows_;

		/**
		 * The maximum start time of all flows.
		 */
		private final long maxStartTime_;

		/**
		 * The interval to wait before the next packet is generated (for TCP,
		 * "0" means saturated senders).
		 */
		private final long sendInterval_;

		/**
		 * The packet (payload) size.
		 */
		private final int packetSize_;

		/**
		 * Constructor.
		 * 
		 * @param numFlows
		 *            the number of flows to create.
		 * @param maxStartTime
		 *            the maximum start time.
		 * @param sendInterval
		 *            the interval to wait between consecutive packets (for TCP,
		 *            "0" means saturated senders).
		 * @param packetSize
		 *            the packet size.
		 */
		public RandomTrafficInfoTCP(int numFlows, long maxStartTime,
				long sendInterval, int packetSize) {
			numFlows_ = numFlows;
			maxStartTime_ = maxStartTime;
			sendInterval_ = sendInterval;
			packetSize_ = packetSize;
		}

		/**
		 * Provides the number of flows.
		 * 
		 * @return the number of flows.
		 */
		public int getNumFlows() {
			return numFlows_;
		}

		/**
		 * Provides the maximum start time.
		 * 
		 * @return the maximum start time.
		 */
		public long getMaxStartTime() {
			return maxStartTime_;
		}

		/**
		 * Provides the send interval.
		 * 
		 * @return the send interval.
		 */
		public long getSendInterval() {
			return sendInterval_;
		}

		/**
		 * Provides the packet size.
		 * 
		 * @return the packet size.
		 */
		public int getPacketSize() {
			return packetSize_;
		}
	}

	/**
	 * Information about UDP traffic flows to set up randomly between nodes.
	 * 
	 * @author Patrick Peschlow
	 * @author Andreas Voss
	 * 
	 * @version 1.0
	 */
	public static class RandomTrafficInfoUDP {
		/**
		 * The number of flows to create.
		 */
		protected final int numFlows_;

		/**
		 * The maximum start time of all flows.
		 */
		private final long maxStartTime_;

		/**
		 * The interval to wait before the next packet is generated.
		 */
		private final long sendInterval_;

		/**
		 * The packet (payload) size.
		 */
		private final int packetSize_;

		/**
		 * Constructor.
		 * 
		 * @param numFlows
		 *            the number of flows to create.
		 * @param maxStartTime
		 *            the maximum start time.
		 * @param sendInterval
		 *            the interval to wait between consecutive packets.
		 * @param packetSize
		 *            the packet size.
		 */
		public RandomTrafficInfoUDP(int numFlows, long maxStartTime,
				long sendInterval, int packetSize) {
			numFlows_ = numFlows;
			maxStartTime_ = maxStartTime;
			sendInterval_ = sendInterval;
			packetSize_ = packetSize;
		}

		/**
		 * Provides the number of flows.
		 * 
		 * @return the number of flows.
		 */
		public int getNumFlows() {
			return numFlows_;
		}

		/**
		 * Provides the maximum start time.
		 * 
		 * @return the maximum start time.
		 */
		public long getMaxStartTime() {
			return maxStartTime_;
		}

		/**
		 * Provides the send interval.
		 * 
		 * @return the send interval.
		 */
		public long getSendInterval() {
			return sendInterval_;
		}

		/**
		 * Provides the packet size.
		 * 
		 * @return the packet size.
		 */
		public int getPacketSize() {
			return packetSize_;
		}
	}

	/**
	 * Information about UDP traffic flows in mesh scenarios.
	 * 
	 * @author Patrick Peschlow
	 * @author Daniel Plohmann
	 * 
	 * @version 1.0
	 */
	public static class MeshTrafficInfoUDP extends RandomTrafficInfoUDP {
		/**
		 * Flow type.
		 */
		private final FlowType type_;

		/**
		 * Flow direction.
		 */
		private final FlowDirection direction_;

		/**
		 * Flow priority.
		 */
		private final FlowPriority priority_;

		/**
		 * Constructor.
		 * 
		 * @param flowType
		 *            the flow type.
		 * @param flowDirection
		 *            the flow direction.
		 * @param numFlows
		 *            the number of flows to create.
		 * @param maxStartTime
		 *            the maximum start time.
		 * @param sendInterval
		 *            the interval to wait between consecutive packets.
		 * @param packetSize
		 *            the packet size.
		 */
		public MeshTrafficInfoUDP(FlowType flowType,
				FlowDirection flowDirection, int numFlows, long maxStartTime,
				long sendInterval, int packetSize) {
			super(numFlows, maxStartTime, sendInterval, packetSize);
			type_ = flowType;
			direction_ = flowDirection;
			priority_ = FlowPriority.AC_BE;
		}

		/**
		 * Constructor.
		 * 
		 * @param flowType
		 *            the flow type.
		 * @param flowDirection
		 *            the flow direction.
		 * @param priority
		 *            the flow priority.
		 * @param numFlows
		 *            the number of flows to create.
		 * @param maxStartTime
		 *            the maximum start time.
		 * @param sendInterval
		 *            the interval to wait between consecutive packets.
		 * @param packetSize
		 *            the packet size.
		 */
		public MeshTrafficInfoUDP(FlowType flowType,
				FlowDirection flowDirection, FlowPriority priority,
				int numFlows, long maxStartTime, long sendInterval,
				int packetSize) {
			super(numFlows, maxStartTime, sendInterval, packetSize);
			type_ = flowType;
			direction_ = flowDirection;
			priority_ = priority;
		}

		/**
		 * @see modules.wireless.scenario.NodeConfig.RandomTrafficInfoUDP#getNumFlows()
		 */
		@Override
		public int getNumFlows() {
			if (direction_ == FlowDirection.BOTH) {
				return 2 * numFlows_;
			} else {
				return numFlows_;
			}
		}

		/**
		 * Provides the flow type.
		 * 
		 * @return the flow type.
		 */
		public FlowType getType() {
			return type_;
		}

		/**
		 * Provides the flow direction.
		 * 
		 * @return the flow direction.
		 */
		public FlowDirection getDirection() {
			return direction_;
		}

		/**
		 * Provides the flow priority.
		 * 
		 * @return the flow priority.
		 */
		public FlowPriority getPriority() {
			return priority_;
		}
	}

	/**
	 * Information about TCP traffic flows in mesh scenarios.
	 * 
	 * @author Patrick Peschlow
	 * @author Daniel Plohmann
	 * 
	 * @version 1.0
	 */
	public static class MeshTrafficInfoTCP extends RandomTrafficInfoTCP {
		/**
		 * Flow type.
		 */
		private final FlowType type_;

		/**
		 * Flow direction.
		 */
		private final FlowDirection direction_;

		/**
		 * Flow priority.
		 */
		private final FlowPriority priority_;

		/**
		 * Constructor.
		 * 
		 * @param flowType
		 *            the flow type.
		 * @param flowDirection
		 *            the flow direction.
		 * @param numFlows
		 *            the number of flows to create.
		 * @param maxStartTime
		 *            the maximum start time.
		 * @param sendInterval
		 *            the interval to wait between consecutive packets.
		 * @param packetSize
		 *            the packet size.
		 */
		public MeshTrafficInfoTCP(FlowType flowType,
				FlowDirection flowDirection, int numFlows, long maxStartTime,
				long sendInterval, int packetSize) {
			super(numFlows, maxStartTime, sendInterval, packetSize);
			type_ = flowType;
			direction_ = flowDirection;
			priority_ = FlowPriority.AC_BE;
		}

		/**
		 * Constructor.
		 * 
		 * @param flowType
		 *            the flow type.
		 * @param flowDirection
		 *            the flow direction.
		 * @param priority
		 *            the flow priority.
		 * @param numFlows
		 *            the number of flows to create.
		 * @param maxStartTime
		 *            the maximum start time.
		 * @param sendInterval
		 *            the interval to wait between consecutive packets.
		 * @param packetSize
		 *            the packet size.
		 */
		public MeshTrafficInfoTCP(FlowType flowType,
				FlowDirection flowDirection, FlowPriority priority,
				int numFlows, long maxStartTime, long sendInterval,
				int packetSize) {
			super(numFlows, maxStartTime, sendInterval, packetSize);
			type_ = flowType;
			direction_ = flowDirection;
			this.priority_ = priority;
		}

		/**
		 * @see modules.wireless.scenario.NodeConfig.RandomTrafficInfoUDP#getNumFlows()
		 */
		@Override
		public int getNumFlows() {
			if (direction_ == FlowDirection.BOTH) {
				return 2 * numFlows_;
			} else {
				return numFlows_;
			}
		}

		/**
		 * Provides the flow type.
		 * 
		 * @return the flow type.
		 */
		public FlowType getType() {
			return type_;
		}

		/**
		 * Provides the flow direction.
		 * 
		 * @return the flow direction.
		 */
		public FlowDirection getDirection() {
			return direction_;
		}

		/**
		 * Provides the flow priority.
		 * 
		 * @return the flow priority.
		 */
		public FlowPriority getPriority() {
			return priority_;
		}
	}
}
