/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/scenario/NodeMovementInfo.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.scenario;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import jist.swans.Constants;
import jist.swans.misc.Location;
import core.Config;
import core.Config.Movement;
import core.util.CustomLogger;

/**
 * Base class for node movement information (NMI) implementations.
 * 
 * @author Patrick Peschlow
 * @author Alexander Marchal
 * 
 * @version 1.0
 */
public abstract class NodeMovementInfo {
	/**
	 * The unique NMI used by the simulation.
	 */
	private static NodeMovementInfo nmi = null;

	/**
	 * Initializes the NMI.
	 */
	public static void init() {
		assert nmi == null;
		nmi = (Config.getMovement() == Movement.CONTINUOUS) ? new ContNodeMovementInfo()
				: new DisNodeMovementInfo();
	}

	/**
	 * Provides a reference to the unique NMI.
	 * 
	 * @return a reference to the NMI.
	 */
	public static NodeMovementInfo getInfo() {
		assert nmi != null;
		return nmi;
	}

	/**
	 * The simulated area dimensions.
	 */
	protected Location areaSize_;

	/**
	 * Simulated area dimension information: [0, x] [0, y]
	 * 
	 * @return the simulated area dimensions.
	 */
	public Location getAreaSize() {
		return areaSize_;
	}

	/**
	 * Provides the location of a node at the specified time.
	 * 
	 * @param node
	 *            the node ID.
	 * @param time
	 *            the time.
	 * @return the location of the node at the specified time.
	 */
	abstract public Location getLocation(int node, long time);

	/**
	 * Update location of a node.
	 * 
	 * @param node
	 *            the node ID.
	 * @param time
	 *            the current simulation time.
	 */
	abstract public void updateLocation(int node, long time);

	/**
	 * Provides the total number of nodes of the simulation.
	 * 
	 * @return the number of nodes.
	 */
	abstract public int getNodeCount();

	/**
	 * Continuous NMI. Uses pre-calculated mobility based on BonnMotion input.
	 * 
	 * @author Patrick Peschlow
	 * @author Alexander Marchal
	 * 
	 * @version 1.0
	 */
	private static class ContNodeMovementInfo extends NodeMovementInfo {
		/**
		 * The number of nodes in the simulation.
		 */
		private int nodeCount_;

		/**
		 * Key: Node ID. Value: Ordered list of waypoints for this node.
		 */
		private List<LocationTimeInfo>[] waypoints_;

		/**
		 * Key: Node ID. Value: true if the node does not move, false otherwise.
		 */
		private boolean[] isStatic_;

		/**
		 * Constructor.
		 */
		@SuppressWarnings("unchecked")
		private ContNodeMovementInfo() {
			String wayPointFile = Config.getMovementFile();
			String paramsFile = wayPointFile.substring(0,
					wayPointFile.lastIndexOf('.'))
					+ ".params";
			parseParams(new File(paramsFile));
			waypoints_ = new List[nodeCount_];
			isStatic_ = new boolean[nodeCount_];
			parseMovementFile(new File(wayPointFile), Config.getEndTime());
		}

		/**
		 * @see modules.wireless.scenario.NodeMovementInfo#getLocation(int,
		 *      long)
		 */
		@Override
		public Location getLocation(int node, long time) {
			return calcLocation(node, time);
		}

		/**
		 * @see modules.wireless.scenario.NodeMovementInfo#updateLocation(int,
		 *      long)
		 */
		@Override
		public void updateLocation(int node, long time) {
			assert false : "Location update not performed with ContNodeMovementInfo.";
		}

		/**
		 * @see modules.wireless.scenario.NodeMovementInfo#getNodeCount()
		 */
		@Override
		public int getNodeCount() {
			return nodeCount_;
		}

		/**
		 * Calculates the location of a node at the specified point in time.
		 * 
		 * @param node
		 *            the node ID of the node.
		 * @param time
		 *            the simulation time of interest.
		 * @return the location of the node.
		 */
		protected Location calcLocation(int node, long time) {
			if (isStatic_[node]) {
				return waypoints_[node].get(0).getLocation();
			}
			assert node < nodeCount_ && waypoints_[node] != null;
			final List<LocationTimeInfo> ltiList = waypoints_[node];
			final int size = ltiList.size();
			assert size > 1;
			LocationTimeInfo first = ltiList.get(0);
			LocationTimeInfo second = null;
			for (int i = 1; i < size; i++) {
				LocationTimeInfo lti = ltiList.get(i);
				if (lti.time_ <= time) {
					first = lti;
				} else {
					second = lti;
					break;
				}
			}
			assert second != null;

			// Calculate current location.
			final long scaledTime = time - first.getTime();
			final long scaledEndTime = second.getTime() - first.getTime();
			final double perc = (double) scaledTime / (double) scaledEndTime;

			final double dx = second.getLocation().getX()
					- first.getLocation().getX();
			final double dy = second.getLocation().getY()
					- first.getLocation().getY();

			final double currentX = first.getLocation().getX() + perc * dx;
			final double currentY = first.getLocation().getY() + perc * dy;

			return new Location(currentX, currentY);
		}

		/**
		 * Parses the Bonnmotion waypoint file.
		 * 
		 * @param wayPointFile
		 *            the Bonnmotion waypoint file.
		 * @param endTime
		 *            the end time of the simulation.
		 */
		private void parseMovementFile(File wayPointFile, long endTime) {
			BufferedReader reader = null;
			try {
				reader = new BufferedReader(new FileReader(wayPointFile));
			} catch (FileNotFoundException e) {
				if (CustomLogger.isFatalEnabled()) {
					CustomLogger.fatal("Could not find node file "
							+ wayPointFile, e);
				}
			}
			assert reader != null;

			String buffer = null;
			int nodeId = 0;
			try {
				while ((buffer = reader.readLine()) != null) {
					List<LocationTimeInfo> wayPoints = new ArrayList<LocationTimeInfo>(
							10);
					isStatic_[nodeId] = true;
					StringTokenizer st = new StringTokenizer(buffer, " ");
					// Check for missing "0.0" entry.
					if (st.hasMoreTokens()) {
						final double time = Double.valueOf(st.nextToken());
						assert st.hasMoreTokens();
						final double x = Float.valueOf(st.nextToken());
						assert st.hasMoreTokens();
						final double y = Float.valueOf(st.nextToken());
						Location location = new Location(x, y);
						if (time != 0.0) {
							// Entry is missing.
							wayPoints.add(new LocationTimeInfo(location, 0));
						}
						wayPoints.add(new LocationTimeInfo(location,
								convertBonnmotionTime(time)));
					}
					// Get all other entries.
					while (st.hasMoreTokens()) {
						final double time = Double.valueOf(st.nextToken());
						assert st.hasMoreTokens();
						final double x = Float.valueOf(st.nextToken());
						assert st.hasMoreTokens();
						final double y = Float.valueOf(st.nextToken());
						Location location = new Location(x, y);
						if (isStatic_[nodeId]) {
							Location lastLocation = wayPoints.get(
									wayPoints.size() - 1).getLocation();
							if (!location.equals(lastLocation)) {
								isStatic_[nodeId] = false;
							}
						}
						wayPoints.add(new LocationTimeInfo(location,
								convertBonnmotionTime(time)));
					}
					assert !wayPoints.isEmpty();
					if (isStatic_[nodeId]) {
						if (wayPoints.size() > 1) {
							for (int i = wayPoints.size() - 1; i > 0; i--) {
								wayPoints.remove(i);
							}
						}
						assert wayPoints.size() == 1;
					} else {
						// Check whether the last entry is missing.
						LocationTimeInfo lti = wayPoints
								.get(wayPoints.size() - 1);
						if (lti.getTime() < endTime) {
							wayPoints.add(new LocationTimeInfo(lti
									.getLocation(), endTime));
						}
						assert wayPoints.size() >= 2;
						assert wayPoints.get(wayPoints.size() - 2).getTime() < wayPoints
								.get(wayPoints.size() - 1).getTime();
					}
					waypoints_[nodeId++] = wayPoints;
				}
			} catch (NumberFormatException e) {
				if (CustomLogger.isFatalEnabled()) {
					CustomLogger.fatal("Strange format of node file content!",
							e);
				}
			} catch (IOException e) {
				if (CustomLogger.isFatalEnabled()) {
					CustomLogger.fatal("Unknown IO error!", e);
				}
			}
		}

		/**
		 * Parses the Bonnmotion params file.
		 * 
		 * @param paramsFile
		 *            the Bonnmotion params file.
		 */
		private void parseParams(File paramsFile) {
			BufferedReader reader = null;
			try {
				reader = new BufferedReader(new FileReader(paramsFile));
			} catch (FileNotFoundException e) {
				if (CustomLogger.isFatalEnabled()) {
					CustomLogger.fatal("Could not find parameter file "
							+ paramsFile, e);
				}
			}
			assert reader != null;

			String line, param;
			try {
				double area_x = 0.0;
				double area_y = 0.0;
				while ((line = reader.readLine()) != null) {
					StringTokenizer st = new StringTokenizer(line, "=");
					param = st.nextToken();
					if ("x".compareTo(param) == 0) {
						area_x = Double.valueOf(st.nextToken());
					} else if ("y".compareTo(param) == 0) {
						area_y = Double.valueOf(st.nextToken());
					} else if ("nn".compareTo(param) == 0) {
						nodeCount_ = Integer.valueOf(st.nextToken());
					} else {
						st.nextToken(); // skip parameter value
					}
				}
				assert area_x != 0.0 && area_y != 0.0;
				areaSize_ = new Location(area_x, area_y);
			} catch (IOException e) {
				if (CustomLogger.isFatalEnabled()) {
					CustomLogger.fatal("IO error while parsing parameter file "
							+ paramsFile, e);
				}
			}

		}

		/**
		 * Converts a time value contained in a Bonnmotion movement file (which
		 * is a double precision floating-point value containing time in
		 * seconds) to a long int value in the time resolution used for the
		 * simulation.
		 * 
		 * @param bmTime
		 *            the Bonnmotion time value in seconds.
		 * @return the long value in the time resolution used for the
		 *         simulation.
		 */
		private long convertBonnmotionTime(double bmTime) {
			final double nanosDouble = bmTime * Constants.SECOND;
			final long nanosLong = (long) nanosDouble;
			return nanosLong;
		}

		/**
		 * Helper class for internal storage of waypoint information. Stores a
		 * location at a specified time.
		 * 
		 * @author Patrick Peschlow
		 */
		private static class LocationTimeInfo {
			/**
			 * The location.
			 */
			private final Location location_;

			/**
			 * The time.
			 */
			private final long time_;

			/**
			 * Constructor.
			 * 
			 * @param location
			 *            the location.
			 * @param time
			 *            the time.
			 */
			private LocationTimeInfo(Location location, long time) {
				location_ = location;
				time_ = time;
			}

			/**
			 * Provides the location.
			 * 
			 * @return the location.
			 */
			private Location getLocation() {
				return location_;
			}

			/**
			 * Provides the time.
			 * 
			 * @return the time.
			 */
			private long getTime() {
				return time_;
			}
		}
	}

	/**
	 * Discrete NMI. Assumes that nodes do not change LPs and that each LP is
	 * only interested in (i.e., only has to query) the locations of its
	 * associated nodes.
	 * 
	 * @author Patrick Peschlow
	 * @author Alexander Marchal
	 * 
	 * @version 1.0
	 */
	private static class DisNodeMovementInfo extends NodeMovementInfo {
		/**
		 * The ContNodeMovementInfo used to store the waypoints.
		 */
		private final ContNodeMovementInfo contNmi_;

		/**
		 * The total number of nodes in the simulation.
		 */
		private final int nodeCount_;

		/**
		 * The previous location of each node.
		 */
		private final Location[] previousLocation_;

		/**
		 * The current location of each node.
		 */
		private final Location[] currentLocation_;

		/**
		 * Last time the locations have been updated.
		 */
		private long lastUpdateTime_;

		/**
		 * Constructor.
		 */
		private DisNodeMovementInfo() {
			contNmi_ = new ContNodeMovementInfo();
			areaSize_ = contNmi_.getAreaSize();
			nodeCount_ = contNmi_.getNodeCount();
			lastUpdateTime_ = 0;
			previousLocation_ = new Location[nodeCount_];
			currentLocation_ = new Location[nodeCount_];
			for (int nodeId = 0; nodeId < nodeCount_; nodeId++) {
				previousLocation_[nodeId] = contNmi_.calcLocation(nodeId, 0);
				currentLocation_[nodeId] = contNmi_.calcLocation(nodeId, 0);
			}
		}

		/**
		 * @see modules.wireless.scenario.NodeMovementInfo#getLocation(int,
		 *      long)
		 */
		@Override
		public Location getLocation(int node, long time) {
			if (time < lastUpdateTime_) {
				return previousLocation_[node];
			} else {
				return currentLocation_[node];
			}
		}

		/**
		 * @see modules.wireless.scenario.NodeMovementInfo#updateLocation(int,
		 *      long)
		 */
		@Override
		public void updateLocation(int node, long time) {
			lastUpdateTime_ = time;
			previousLocation_[node] = currentLocation_[node];
			currentLocation_[node] = contNmi_.calcLocation(node, time);
		}

		/**
		 * @see modules.wireless.scenario.NodeMovementInfo#getNodeCount()
		 */
		@Override
		public int getNodeCount() {
			return nodeCount_;
		}
	}
}
