/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: modules/wireless/scenario/WirelessNode.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package modules.wireless.scenario;

import gnu.trove.TIntIntHashMap;

import java.util.ArrayList;
import java.util.List;

import jist.swans.mac.HWMP;
import jist.swans.mac.Mac802_11;
import jist.swans.mac.Mac802_11AP;
import jist.swans.mac.Mac802_11Client;
import jist.swans.mac.Mac802_11s;
import jist.swans.net.NetIp;
import jist.swans.net.NetIpAP;
import jist.swans.net.PacketLoss;
import jist.swans.radio.RadioMesh;
import jist.swans.radio.RadioNoise;
import jist.swans.route.RouteAodv;
import jist.swans.route.RouteDsr;
import jist.swans.route.RouteInterface;
import jist.swans.trans.TcpSocket;
import jist.swans.trans.TransTcp;
import jist.swans.trans.TransUdp;
import modules.wireless.app.PacketGenerator;
import modules.wireless.app.PacketGenerator.ConnectionType;
import modules.wireless.lazy.LazyMac802_11;
import modules.wireless.lazy.LazyRadio;
import modules.wireless.mesh.MeshPortal;
import modules.wireless.scenario.NodeConfig.FlowPriority;
import core.BaseModule;
import core.Config;
import core.ModuleRoot;

/**
 * A wireless node that bundles and is responsible for all its modules.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class WirelessNode {
	/**
	 * The type of this node (relevant only for Mesh simulations).
	 * 
	 * @author Patrick Peschlow
	 * @author Daniel Plohmann
	 * 
	 * @version 1.0
	 */
	public static enum NodeType {
		/**
		 * MANET node.
		 */
		MANET,
		/**
		 * Mesh AP.
		 */
		AP,
		/**
		 * Mesh client.
		 */
		CLIENT;
	}

	/**
	 * A list of all nodes of the simulation.
	 */
	protected static WirelessNode[] nodes;

	/**
	 * Maps portal IDs to node IDs in mesh scenarios.
	 */
	protected static TIntIntHashMap portalIdToNodeId;

	/**
	 * Whether a node is a mesh portal.
	 */
	protected static boolean[] isMeshPortal;

	/**
	 * Keeps track of associations of mesh clients to mesh APs.
	 */
	private static int[] associatedAps;

	/**
	 * Constant indicating that a mesh AP is not a portal.
	 */
	static final int NO_PORTAL = -1;

	/**
	 * Creates a list of all nodes belonging to the specified partition.
	 * 
	 * @param partitionId
	 *            the ID of the partition requesting its list of nodes.
	 * @return list of all nodes belonging to the specified partition.
	 */
	public static List<WirelessNode> getNodesOfPartition(int partitionId) {
		List<WirelessNode> nodesOfPartition = new ArrayList<WirelessNode>();
		for (WirelessNode node : nodes) {
			if (node.getPartitionId() == partitionId) {
				nodesOfPartition.add(node);
			}
		}
		assert !nodesOfPartition.isEmpty();
		return nodesOfPartition;
	}

	/**
	 * Provides a reference to the node with the specific ID.
	 * 
	 * @param nodeId
	 *            the ID of the requested node.
	 * @return the node with the specified ID.
	 */
	public static WirelessNode getNode(int nodeId) {
		assert nodes != null;
		assert nodeId < nodes.length;
		return nodes[nodeId];
	}

	/**
	 * Provides the node ID of the mesh portal with the specified portal ID.
	 * 
	 * @param portalId
	 *            the portal ID of the requested mesh portal.
	 * @return the ID of the mesh portal with the specified ID.
	 */
	static int getPortalNodeId(int portalId) {
		assert portalIdToNodeId.containsKey(portalId);
		return portalIdToNodeId.get(portalId);
	}

	/**
	 * Initializes the node array before all nodes are created.
	 * 
	 * @param numNodes
	 *            the total number of nodes.
	 * @param numAps
	 *            the number of mesh APs.
	 * @param numPortals
	 *            the number of mesh portals.
	 */
	static void initNodes(int numNodes, int numAps, int numPortals) {
		assert nodes == null;
		nodes = new WirelessNode[numNodes];
		if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH) {
			portalIdToNodeId = new TIntIntHashMap(numPortals);
			isMeshPortal = new boolean[numAps];
			associatedAps = new int[numNodes];
		}
	}

	/**
	 * Creates a new node.
	 * 
	 * @param nodeId
	 *            ID of this node.
	 * @param partitionId
	 *            ID of the partition this node belongs to.
	 * @param rngId
	 *            ID of the virtual partition this node belongs to.
	 */
	static void registerNode(int nodeId, int partitionId, int rngId) {
		assert nodes != null;
		assert nodeId < nodes.length;
		assert NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MANET;
		nodes[nodeId] = new WirelessNode(nodeId, partitionId, rngId);
	}

	/**
	 * Creates a new node.
	 * 
	 * @param nodeId
	 *            ID of this node.
	 * @param partitionId
	 *            ID of the partition this node belongs to.
	 * @param rngId
	 *            ID of the virtual partition this node belongs to.
	 * @param type
	 *            type of this node.
	 * @param portalId
	 *            a portal ID, if this node is a mesh portal, otherwise
	 *            NO_PORTAL.
	 */
	static void registerNode(int nodeId, int partitionId, int rngId,
			NodeType type, int portalId) {
		assert nodes != null;
		assert nodeId < nodes.length;
		assert NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH;
		if (portalId != NO_PORTAL) {
			isMeshPortal[nodeId] = true;
			portalIdToNodeId.put(portalId, nodeId);
		}
		nodes[nodeId] = new WirelessNode(nodeId, partitionId, rngId, type);
	}

	/**
	 * Registers a new dummy and its network layer module ID with a portal.
	 * 
	 * @param portalNodeId
	 *            the portal ID.
	 * @param dummyId
	 *            the internet ID of the dummy.
	 * @param netModuleId
	 *            the module ID of the network module of the dummy.
	 */
	public static void registerDummy(int portalNodeId, int dummyId,
			int netModuleId) {
		getNode(portalNodeId).meshPortal_.registerDummy(dummyId, netModuleId);
	}

	/**
	 * Whether this node is a mesh portal.
	 * 
	 * @param nodeId
	 *            ID of this node.
	 * 
	 * @return whether this node is a mesh portal.
	 */
	static boolean isMeshPortal(int nodeId) {
		assert nodeId < isMeshPortal.length;
		return isMeshPortal[nodeId];
	}

	/**
	 * Sets the association of a remote mesh client node to an AP.
	 * 
	 * @param nodeId
	 *            the node ID.
	 * @param apId
	 *            the AP ID.
	 */
	static void setAssociation(int nodeId, int apId) {
		associatedAps[nodeId] = apId;
	}

	/**
	 * Provides the association of a remote mesh client node.
	 * 
	 * @param nodeId
	 *            the node ID.
	 * @return the associated AP ID.
	 */
	static int getAssociatedAp(int nodeId) {
		return associatedAps[nodeId];
	}

	/**
	 * Provides the number of bytes sent via TCP flows of all nodes.
	 * 
	 * @return the number of bytes sent via TCP flows.
	 */
	static long getAllTcpBytesSent() {
		long bytes = 0;
		for (WirelessNode node : nodes) {
			bytes += node.getTcpBytesSent();
		}
		return bytes;
	}

	/**
	 * Provides the number of bytes received via TCP flows of all nodes.
	 * 
	 * @return the number of bytes received via TCP flows.
	 */
	static long getAllTcpBytesReceived() {
		long bytes = 0;
		for (WirelessNode node : nodes) {
			bytes += node.getTcpBytesReceived();
		}
		return bytes;
	}

	/**
	 * Provides the number of bytes sent via UDP flows of all nodes.
	 * 
	 * @return the number of bytes sent via UDP flows.
	 */
	static long getAllUdpBytesSent() {
		long bytes = 0;
		for (WirelessNode node : nodes) {
			bytes += node.getUdpBytesSent();
		}
		return bytes;
	}

	/**
	 * Provides the number of bytes received via UDP flows of all nodes.
	 * 
	 * @return the number of bytes received via UDP flows.
	 */
	static long getAllUdpBytesReceived() {
		long bytes = 0;
		for (WirelessNode node : nodes) {
			bytes += node.getUdpBytesReceived();
		}
		return bytes;
	}

	/**
	 * Provides the number of bytes sent by the MAC layers of all nodes.
	 * 
	 * @return the number of bytes sent by the MAC layers of all nodes.
	 */
	static long getAllMacBytesSent() {
		long bytes = 0;
		for (WirelessNode node : nodes) {
			bytes += node.getMacBytesSent();
		}
		return bytes;
	}

	/**
	 * Provides the number of packets offered to the frame queues of the MAC
	 * layers of all nodes.
	 * 
	 * @return the number of packets offered to the frame queues of the MAC
	 *         layers of all nodes.
	 */
	static int getAllMacQueueOffered() {
		int transfers = 0;
		for (WirelessNode node : nodes) {
			transfers += node.getMacQueueOffered();
		}
		return transfers;
	}

	/**
	 * Provides the number of packets dropped because of overflowing frame
	 * queues of the MAC layers of all nodes.
	 * 
	 * @return the number of packets dropped because of overflowing frame queues
	 *         of the MAC layers of all nodes.
	 */
	static int getAllMacQueueDropped() {
		int transfers = 0;
		for (WirelessNode node : nodes) {
			transfers += node.getMacQueueDropped();
		}
		return transfers;
	}

	/**
	 * Provides the number of packets offered to idle MAC layers of all nodes.
	 * 
	 * @return the number of packets offered to idle MAC layers of all nodes.
	 */
	static int getAllDirectTransfers() {
		int transfers = 0;
		for (WirelessNode node : nodes) {
			transfers += node.getDirectTransfers();
		}
		return transfers;
	}

	/**
	 * Provides the number of incoming packets of the radio layers of all nodes.
	 * 
	 * @return the number of incoming packets of the radio layers of all nodes.
	 */
	static int getAllIncomingPackets() {
		int packets = 0;
		for (WirelessNode node : nodes) {
			packets += node.getIncomingPackets();
		}
		return packets;
	}

	/**
	 * Provides the number of dropped packets of the radio layers of all nodes.
	 * 
	 * @return the number of dropped packets of the radio layers of all nodes.
	 */
	static int getAllDroppedPackets() {
		int packets = 0;
		for (WirelessNode node : nodes) {
			packets += node.getDroppedPackets();
		}
		return packets;
	}

	/**
	 * Provides the number of captured signals of the radio layers of all nodes.
	 * 
	 * @return the number of captured signals of the radio layers of all nodes.
	 */
	static int getAllCapturedSignals() {
		int captured = 0;
		for (WirelessNode node : nodes) {
			captured += node.getCapturedSignals();
		}
		return captured;
	}

	/**
	 * Provides the number of broadcasts sent by the MAC layers of all nodes.
	 * 
	 * @return the number of broadcasts sent by the MAC layers of all nodes.
	 */
	static int getAllNumBroadcasts() {
		int broadcasts = 0;
		for (WirelessNode node : nodes) {
			broadcasts += node.getNumBroadcasts();
		}
		return broadcasts;
	}

	/**
	 * Provides the number of unicasts sent by the MAC layers of all nodes.
	 * 
	 * @return the number of unicasts sent by the MAC layers of all nodes.
	 */
	static int getAllNumUnicasts() {
		int unicasts = 0;
		for (WirelessNode node : nodes) {
			unicasts += node.getNumUnicasts();
		}
		return unicasts;
	}

	/**
	 * Provides the number of forwarded broadcasts sent by the routing modules
	 * of all nodes.
	 * 
	 * @return the number of forwarded broadcasts sent by the routing modules of
	 *         all nodes.
	 */
	static int getAllFwBroadcasts() {
		int fwBroadcasts = 0;
		if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MANET) {
			for (WirelessNode node : nodes) {
				fwBroadcasts += node.getFwBroadcasts();
			}
		}
		return fwBroadcasts;
	}

	/**
	 * Provides the number of received broadcasts sent by the routing modules of
	 * all nodes.
	 * 
	 * @return the number of received broadcasts sent by the routing modules of
	 *         all nodes.
	 */
	static int getAllRcvBroadcasts() {
		int rcvBroadcasts = 0;
		if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MANET) {
			for (WirelessNode node : nodes) {
				rcvBroadcasts += node.getRcvBroadcasts();
			}
		}
		return rcvBroadcasts;
	}

	/**
	 * Provides the number of frames sent by the MAC layers of all nodes.
	 * 
	 * @return the number of frames sent by the MAC layers of all nodes.
	 */
	static int getAll802_11Transmissions() {
		int transmissions = 0;
		for (WirelessNode node : nodes) {
			transmissions += node.get802_11Transmissions();
		}
		return transmissions;
	}

	/**
	 * Provides the number of frames sent by the mesh MAC layers of all nodes.
	 * 
	 * @return the number of frames sent by the mesh MAC layers of all nodes.
	 */
	static int getAll802_11sTransmissions() {
		int transmissions = 0;
		for (WirelessNode node : nodes) {
			transmissions += node.get802_11sTransmissions();
		}
		return transmissions;
	}

	/**
	 * Provides the number of packets dropped by the HWMP modules of all nodes.
	 * 
	 * @return the number of packets dropped by the HWMP modules of all nodes.
	 */
	static int getAllHwmpDrops() {
		int drops = 0;
		for (WirelessNode node : nodes) {
			drops += node.getHwmpDrops();
		}
		return drops;
	}

	/**
	 * Provides the number of beacons sent by all APs.
	 * 
	 * @return the number of beacons sent by all APs.
	 */
	static int getAllNumBeacons() {
		int numBeacons = 0;
		if (NodeConfig.getScenarioType() == NodeConfig.ScenarioType.MESH) {
			for (WirelessNode node : nodes) {
				numBeacons += node.getNumBeacons();
			}
		}
		return numBeacons;
	}

	/**
	 * The ID of this node.
	 */
	private final int number_;

	/**
	 * The global RNG ID (virtual partition) of this node.
	 */
	private final int rngId_;

	/**
	 * The ID of the partition this node is associated with.
	 */
	private final int partitionId_;

	/**
	 * List of the modules of the node. Includes PHY, MAC, etc.
	 */
	private final List<BaseModule> modules_;

	/**
	 * The packet generator of this node.
	 */
	private final PacketGenerator pgen_;

	/**
	 * The UDP module of this node.
	 */
	private TransUdp udp_;

	/**
	 * The TCP socket module of this node.
	 */
	private TcpSocket srcSock_;

	/**
	 * The TCP module of this node.
	 */
	private TransTcp tcp_;

	/**
	 * The IP layer of this node.
	 */
	private final NetIp net_;

	/**
	 * The IP routing module of this node.
	 */
	private final RouteInterface route_;

	/**
	 * The MAC layer of this node.
	 */
	private final Mac802_11 mac_;

	/**
	 * The radio layer of this node.
	 */
	private final RadioNoise radio_;

	/**
	 * The (optional) mesh MAC layer of this node.
	 */
	private final Mac802_11s meshMac_;

	/**
	 * The (optional) HWMP module of this node.
	 */
	private final HWMP hwmp_;

	/**
	 * The (optional) mesh radio layer of this node.
	 */
	private final RadioNoise meshRadio_;

	/**
	 * The (optional) mesh portal of this node.
	 */
	private final MeshPortal meshPortal_;

	/**
	 * The type of this node.
	 */
	private final NodeType type_;

	/**
	 * Creates a new node.
	 * 
	 * @param nodeId
	 *            ID of this node.
	 * @param partitionId
	 *            ID of the partition this node belongs to.
	 * @param rngId
	 *            ID of the virtual partition this node belongs to.
	 */
	@SuppressWarnings("rawtypes")
	private WirelessNode(int nodeId, int partitionId, int rngId) {
		number_ = nodeId;
		partitionId_ = partitionId;
		rngId_ = rngId;
		type_ = NodeType.MANET;

		modules_ = new ArrayList<BaseModule>();

		// Create radio and MAC layer.
		Class macClass = null;
		if (Config.usesLazyMac()) {
			macClass = LazyMac802_11.class;
			radio_ = new LazyRadio(number_);
			mac_ = new LazyMac802_11(number_, NetIp.class);
		} else {
			macClass = Mac802_11.class;
			radio_ = new RadioNoise(number_, macClass);
			mac_ = new Mac802_11(number_, RadioNoise.class, NetIp.class);
		}
		addModule(radio_);
		addModule(mac_);

		meshMac_ = null;
		hwmp_ = null;
		meshRadio_ = null;
		meshPortal_ = null;

		// Create network layer.
		final PacketLoss.Zero in = new PacketLoss.Zero();
		final PacketLoss.Zero out = new PacketLoss.Zero();
		net_ = new NetIp(number_, macClass, in, out);
		addModule(net_);

		// Create routing module.
		switch (NodeConfig.getRoutingProtocol()) {
		case AODV:
			route_ = new RouteAodv(number_, net_);
			addModule(route_);
			break;
		case DSR:
			route_ = new RouteDsr(number_, net_);
			addModule(route_);
			break;
		case NONE:
			// Mesh simulation, no routing in clients needed.
			route_ = null;
			break;
		default:
			throw new RuntimeException("unknown routing protocol!");
		}

		// Create transport layer.
		if (NodeConfig.getTrafficInfoUDP() != null) {
			udp_ = new TransUdp(number_, NetIp.class, PacketGenerator.class);
			addModule(udp_);
		}
		if (NodeConfig.getTrafficInfoTCP() != null) {
			tcp_ = new TransTcp(number_, NetIp.class, PacketGenerator.class);
			addModule(tcp_);
		}

		// Create packet generator.
		pgen_ = new PacketGenerator(number_);
		addModule(pgen_);

		// Set LP (partition) allocation IDs.
		for (BaseModule module : modules_) {
			ModuleRoot.getRoot().allocateModuleToLp(module, partitionId);
		}
	}

	/**
	 * Creates a new node.
	 * 
	 * @param nodeId
	 *            ID of this node.
	 * @param partitionId
	 *            ID of the partition this node belongs to.
	 * @param rngId
	 *            ID of the virtual partition this node belongs to.
	 * @param type
	 *            type of this node.
	 */
	@SuppressWarnings("rawtypes")
	private WirelessNode(int nodeId, int partitionId, int rngId, NodeType type) {
		number_ = nodeId;
		partitionId_ = partitionId;
		rngId_ = rngId;
		type_ = type;

		modules_ = new ArrayList<BaseModule>();

		// Create radio and MAC layer.
		Class macClass = null;
		if (Config.usesLazyMac()) {
			// Lazy MAC currently only implemented with MANET scenarios.
			assert type == NodeType.MANET;
			macClass = LazyMac802_11.class;
			radio_ = new LazyRadio(number_);
			mac_ = new LazyMac802_11(number_, NetIp.class);
			meshMac_ = null;
			hwmp_ = null;
			meshRadio_ = null;
			meshPortal_ = null;
		} else {
			switch (type) {
			case MANET:
				macClass = Mac802_11.class;
				radio_ = new RadioNoise(number_, macClass);
				mac_ = new Mac802_11(number_, RadioNoise.class, NetIp.class);
				meshMac_ = null;
				hwmp_ = null;
				meshRadio_ = null;
				meshPortal_ = null;
				break;
			case AP:
				macClass = Mac802_11AP.class;
				radio_ = new RadioNoise(number_, macClass);
				mac_ = new Mac802_11AP(number_, RadioNoise.class, NetIpAP.class);
				Class meshMacClass = Mac802_11s.class;
				meshRadio_ = new RadioMesh(number_, meshMacClass);
				addModule(meshRadio_);
				meshMac_ = new Mac802_11s(number_, RadioMesh.class,
						NetIpAP.class);
				addModule(meshMac_);
				hwmp_ = new HWMP(number_, isMeshPortal[number_]);
				addModule(hwmp_);
				if (isMeshPortal[number_]) {
					meshPortal_ = new MeshPortal(number_);
					addModule(meshPortal_);
				} else {
					meshPortal_ = null;
				}
				break;
			case CLIENT:
				macClass = Mac802_11Client.class;
				radio_ = new RadioNoise(number_, macClass);
				mac_ = new Mac802_11Client(number_, RadioNoise.class,
						NetIp.class);
				meshMac_ = null;
				hwmp_ = null;
				meshRadio_ = null;
				meshPortal_ = null;
				break;
			default:
				throw new RuntimeException("unknown node type!");
			}
		}
		assert radio_ != null;
		assert mac_ != null;
		addModule(radio_);
		addModule(mac_);

		// Create network layer.
		final PacketLoss.Zero in = new PacketLoss.Zero();
		final PacketLoss.Zero out = new PacketLoss.Zero();
		switch (type) {
		case MANET:
			// Fall through.
		case CLIENT:
			net_ = new NetIp(number_, macClass, in, out);
			break;
		case AP:
			net_ = new NetIpAP(number_, macClass, in, out,
					isMeshPortal[number_]);
			break;
		default:
			throw new RuntimeException("unknown network layer!");
		}
		addModule(net_);

		// Create routing module.
		switch (NodeConfig.getRoutingProtocol()) {
		case AODV:
			route_ = new RouteAodv(number_, net_);
			addModule(route_);
			break;
		case DSR:
			route_ = new RouteDsr(number_, net_);
			addModule(route_);
			break;
		case NONE:
			// Mesh simulation, no routing in clients needed.
			route_ = null;
			break;
		default:
			throw new RuntimeException("unknown routing protocol!");
		}

		// Create transport layer.
		if (NodeConfig.getTrafficInfoUDP() != null) {
			udp_ = new TransUdp(number_, NetIp.class, PacketGenerator.class);
			addModule(udp_);
		}
		if (NodeConfig.getTrafficInfoTCP() != null) {
			tcp_ = new TransTcp(number_, NetIp.class, PacketGenerator.class);
			addModule(tcp_);
		}

		// Create packet generator.
		pgen_ = new PacketGenerator(number_);
		addModule(pgen_);

		// Set LP (partition) allocation IDs.
		for (BaseModule module : modules_) {
			ModuleRoot.getRoot().allocateModuleToLp(module, partitionId);
		}
	}

	/**
	 * Default constructor. Only to be called by subclasses.
	 */
	protected WirelessNode() {
		number_ = -1;
		rngId_ = -1;
		partitionId_ = -1;
		modules_ = null;
		pgen_ = null;
		udp_ = null;
		srcSock_ = null;
		tcp_ = null;
		net_ = null;
		route_ = null;
		mac_ = null;
		radio_ = null;
		meshMac_ = null;
		hwmp_ = null;
		meshRadio_ = null;
		meshPortal_ = null;
		type_ = null;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "Node" + number_;
	}

	/**
	 * Provides the node ID of this node.
	 * 
	 * @return the node ID of this node.
	 */
	public int getNumber() {
		return number_;
	}

	/**
	 * Provides the RNG ID (virtual partition) of this node.
	 * 
	 * @return the RNG ID (virtual partition) of this node.
	 */
	public int getRngId() {
		return rngId_;
	}

	/**
	 * Provides the partition ID of this node.
	 * 
	 * @return the partition ID of this node.
	 */
	public int getPartitionId() {
		return partitionId_;
	}

	/**
	 * Provides the radio layer of this node.
	 * 
	 * @return the radio layer of this node.
	 */
	public RadioNoise getRadio() {
		return radio_;
	}

	/**
	 * Provides the mesh radio layer of this node.
	 * 
	 * @return the mesh radio layer of this node.
	 */
	public RadioNoise getMeshRadio() {
		return meshRadio_;
	}

	/**
	 * Provides the MAC layer of this node.
	 * 
	 * @return the MAC layer of this node.
	 */
	public Mac802_11 getMac() {
		return mac_;
	}

	/**
	 * Provides the mesh MAC layer of this node.
	 * 
	 * @return the mesh MAC layer of this node.
	 */
	public Mac802_11s getMeshMac() {
		return meshMac_;
	}

	/**
	 * Provides the IP layer of this node.
	 * 
	 * @return the IP layer of this node.
	 */
	public NetIp getNet() {
		return net_;
	}

	/**
	 * Provides the UDP module of this node.
	 * 
	 * @return the UDP module of this node.
	 */
	public TransUdp getUdp() {
		return udp_;
	}

	/**
	 * Provides the TCP module of this node.
	 * 
	 * @return the TCP module of this node.
	 */
	public TransTcp getTcp() {
		return tcp_;
	}

	/**
	 * Provides the packet generator of this node.
	 * 
	 * @return the packet generator of this node.
	 */
	public PacketGenerator getPgen() {
		return pgen_;
	}

	/**
	 * Provides the type of this node.
	 * 
	 * @return the type of this node.
	 */
	public NodeType getNodeType() {
		return type_;
	}

	/**
	 * Provides the next time this node might possibly send something. Used for
	 * lookahead optimization.
	 * 
	 * @param lastSafeTime
	 *            the largest possible timestamp of an already executed event.
	 * @return the earliest possible time of this node's next transmission.
	 */
	public long getNextPST(long lastSafeTime) {
		long macPST = mac_.getNextPST(lastSafeTime);
		if (type_ == NodeType.AP) {
			macPST = Math.min(macPST, meshMac_.getNextPST(lastSafeTime));
		}
		return macPST;
	}

	/**
	 * Adds a module to my list of modules. Should only be called internally.
	 * 
	 * @param module
	 *            the module to add.
	 */
	protected void addModule(BaseModule module) {
		modules_.add(module);
	}

	/**
	 * Initializes a UDP traffic flow between two nodes. The packet generator of
	 * the sending node will be notified of the new traffic flow.
	 * 
	 * @param connectionId
	 *            ID identifying the connection.
	 * @param dstNodeId
	 *            ID of the destination node.
	 * @param startTime
	 *            simulation time the first packet will be sent.
	 * @param packetSize
	 *            packet size of each packet generated.
	 * @param sendInterval
	 *            send interval.
	 * @param priority
	 *            flow priority.
	 */
	void startUdpFlow(int connectionId, int dstNodeId, long startTime,
			int packetSize, long sendInterval, FlowPriority priority) {
		assert NodeConfig.getTrafficInfoUDP() != null;
		pgen_.addConnection(ConnectionType.UDP, connectionId, startTime,
				dstNodeId, packetSize, sendInterval, priority);
	}

	/**
	 * Creates a TCP traffic source.
	 * 
	 * @param connectionId
	 *            ID identifying the connection.
	 * @param srcNodeId
	 *            ID of the source node.
	 * @param dstNodeId
	 *            ID of the destination node.
	 * @param srcPort
	 *            source port.
	 * @param dstPort
	 *            destination port.
	 * @param startTime
	 *            simulation time the first packet will be sent.
	 * @param packetSize
	 *            packet size of each packet generated.
	 * @param sendInterval
	 *            send interval.
	 * @param priority
	 *            flow priority.
	 */
	void createTcpSource(int connectionId, int srcNodeId, int dstNodeId,
			int srcPort, int dstPort, long startTime, int packetSize,
			long sendInterval, FlowPriority priority) {
		assert NodeConfig.getTrafficInfoTCP() != null;
		srcSock_ = new TcpSocket(connectionId, srcNodeId, dstNodeId, srcPort,
				dstPort, getTcp(), packetSize, sendInterval == 0);
		ModuleRoot.getRoot().allocateModuleToLp(srcSock_, getPartitionId());
		addModule(srcSock_);
		pgen_.addConnection(ConnectionType.TCP, connectionId, startTime,
				dstNodeId, packetSize, sendInterval, priority);
	}

	/**
	 * Creates a TCP traffic sink.
	 * 
	 * @param connectionId
	 *            ID identifying the connection.
	 * @param srcNodeId
	 *            ID of the source node.
	 * @param dstNodeId
	 *            ID of the destination node.
	 * @param packetSize
	 *            packet size of each packet generated.
	 * @param srcPort
	 *            source port.
	 * @param dstPort
	 *            destination port.
	 */
	void createTcpSink(int connectionId, int srcNodeId, int dstNodeId,
			int srcPort, int dstPort, int packetSize) {
		assert NodeConfig.getTrafficInfoTCP() != null;
		TcpSocket dstSock = new TcpSocket(connectionId, dstNodeId, srcNodeId,
				dstPort, srcPort, getTcp(), packetSize, false);
		ModuleRoot.getRoot().allocateModuleToLp(dstSock, getPartitionId());
		addModule(dstSock);
	}

	/**
	 * Associates this client with an AP.
	 * 
	 * @param apId
	 *            the AP ID.
	 */
	void associateAp(int apId) {
		assert type_ == NodeType.CLIENT;
		((Mac802_11Client) getMac()).setAssociatedAP(apId);
	}

	/**
	 * Associates a new client to this AP.
	 * 
	 * @param clientId
	 *            the client node ID.
	 */
	void associateClient(int clientId) {
		assert type_ == NodeType.AP;
		((Mac802_11AP) getMac()).addAssociatedClient(clientId);
		((NetIpAP) getNet()).addAssociatedClientIPAddr(clientId);
	}

	/**
	 * Provides the number of bytes sent via TCP flows of this node.
	 * 
	 * @return the number of bytes sent via TCP flows of this node.
	 */
	protected long getTcpBytesSent() {
		return pgen_.getTcpBytesSent();
	}

	/**
	 * Provides the number of bytes received via TCP flows of this node.
	 * 
	 * @return the number of bytes received via TCP flows of this node.
	 */
	protected long getTcpBytesReceived() {
		return pgen_.getTcpBytesReceived();
	}

	/**
	 * Provides the number of bytes sent via UDP flows of this node.
	 * 
	 * @return the number of bytes sent via UDP flows of this node.
	 */
	protected long getUdpBytesSent() {
		return pgen_.getUdpBytesSent();
	}

	/**
	 * Provides the number of bytes received via UDP flows of this node.
	 * 
	 * @return the number of bytes received via UDP flows of this node.
	 */
	protected long getUdpBytesReceived() {
		return pgen_.getUdpBytesReceived();
	}

	/**
	 * Provides the number of bytes sent by the MAC layers of this node.
	 * 
	 * @return the number of bytes sent by the MAC layers of this node.
	 */
	protected long getMacBytesSent() {
		return mac_.getNumBytesSent();
	}

	/**
	 * Provides the number of packets offered to the frame queues of the MAC
	 * layers of this node.
	 * 
	 * @return the number of packets offered to the frame queues of the MAC
	 *         layers of this node.
	 */
	protected int getMacQueueOffered() {
		return mac_.getNumMacQueueOffered();
	}

	/**
	 * Provides the number of packets dropped because of overflowing frame
	 * queues of the MAC layers of this node.
	 * 
	 * @return the number of packets dropped because of overflowing frame queues
	 *         of the MAC layers of this node.
	 */
	protected int getMacQueueDropped() {
		return mac_.getNumMacQueueDropped();
	}

	/**
	 * Provides the number of packets offered to idle MAC layers of this node.
	 * 
	 * @return the number of packets offered to idle MAC layers of this node.
	 */
	protected int getDirectTransfers() {
		return mac_.getNumDirectTransfers();
	}

	/**
	 * Provides the number of incoming packets of the radio layers of this node.
	 * 
	 * @return the number of incoming packets of the radio layers of this node.
	 */
	protected int getIncomingPackets() {
		return radio_.getIncomingPackets();
	}

	/**
	 * Provides the number of dropped packets of the radio layers of this node.
	 * 
	 * @return the number of dropped packets of the radio layers of this node.
	 */
	protected int getDroppedPackets() {
		return radio_.getDroppedPackets();
	}

	/**
	 * Provides the number of captured signals of the radio layers of this node.
	 * 
	 * @return the number of captured signals of the radio layers of this node.
	 */
	protected int getCapturedSignals() {
		return radio_.getCapturedSignals();
	}

	/**
	 * Provides the number of broadcasts sent by the MAC layers of this node.
	 * 
	 * @return the number of broadcasts sent by the MAC layers of this node.
	 */
	protected int getNumBroadcasts() {
		assert mac_ != null;
		int broadcasts = mac_.getNumBroadcasts();
		if (meshMac_ != null) {
			broadcasts += meshMac_.getNumBroadcasts();
		}
		return broadcasts;
	}

	/**
	 * Provides the number of unicasts sent by the MAC layers of this node.
	 * 
	 * @return the number of unicasts sent by the MAC layers of this node.
	 */
	protected int getNumUnicasts() {
		assert mac_ != null;
		int unicasts = mac_.getNumUnicasts();
		if (meshMac_ != null) {
			unicasts += meshMac_.getNumUnicasts();
		}
		return unicasts;
	}

	/**
	 * Provides the number of forwarded broadcasts sent by the routing modules
	 * of this node.
	 * 
	 * @return the number of forwarded broadcasts sent by the routing modules of
	 *         all nodes.
	 */
	protected int getFwBroadcasts() {
		assert route_ != null;
		return route_.getFwBroadcasts();
	}

	/**
	 * Provides the number of received broadcasts sent by the routing modules of
	 * all nodes.
	 * 
	 * @return the number of received broadcasts sent by the routing modules of
	 *         all nodes.
	 */
	protected int getRcvBroadcasts() {
		assert route_ != null;
		return route_.getRcvBroadcasts();
	}

	/**
	 * Provides the number of frames sent by the MAC layers of this node.
	 * 
	 * @return the number of frames sent by the MAC layers of this node.
	 */
	protected int get802_11Transmissions() {
		assert mac_ != null;
		int transmissions = mac_.getNumUnicasts();
		transmissions += mac_.getNumBroadcasts();
		return transmissions;
	}

	/**
	 * Provides the number of frames sent by the mesh MAC layers of this node.
	 * 
	 * @return the number of frames sent by the mesh MAC layers of this node.
	 */
	protected int get802_11sTransmissions() {
		if (meshMac_ != null) {
			int transmissions = meshMac_.getNumUnicasts();
			transmissions += meshMac_.getNumBroadcasts();
			return transmissions;
		}
		return 0;
	}

	/**
	 * Provides the number of packets dropped by the HWMP modules of this node.
	 * 
	 * @return the number of packets dropped by the HWMP modules of this node.
	 */
	protected int getHwmpDrops() {
		if (hwmp_ != null) {
			return hwmp_.getHwmpDrops();
		}
		return 0;
	}

	/**
	 * Provides the number of beacons sent by all APs.
	 * 
	 * @return the number of beacons sent by all APs.
	 */
	protected int getNumBeacons() {
		assert NodeConfig.getNumberAPs() > 0;
		if (getNodeType() == NodeType.AP) {
			Mac802_11AP ap = (Mac802_11AP) getMac();
			return ap.getNumBeacons();
		}
		return 0;
	}
}