/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: rng/DefaultRandomStreamNormal.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package rng;

import extern.RngStream;

/**
 * A normal random number stream based on the RngStream class.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public class DefaultRandomStreamNormal implements RandomStream {
	/**
	 * The RngStream used to implement this random stream.
	 */
	final RngStream gen_;

	/**
	 * The mean of the distribution.
	 */
	final double mean_;

	/**
	 * The standard deviation of the distribution.
	 */
	final double stddev_;

	/**
	 * Whether the next random number is already available (helper variable to
	 * implement the polar method).
	 */
	private boolean isAvailable_;

	/**
	 * The stored next random number, if available (helper variable to implement
	 * the polar method).
	 */
	private double stored_;

	/**
	 * Constructor.
	 * 
	 * @param gen
	 *            the RNG used to implement this stream.
	 * @param mean
	 *            the mean of the distribution.
	 * @param stddev
	 *            the standard deviation of the distribution.
	 */
	DefaultRandomStreamNormal(RngStream gen, double mean, double stddev) {
		gen_ = gen;
		mean_ = mean;
		stddev_ = stddev;
		isAvailable_ = false;
		stored_ = 0;
	}

	/**
	 * @see rng.RandomStream#nextDouble()
	 */
	public double nextDouble() {
		if (isAvailable_) {
			isAvailable_ = false;
			return mean_ + stddev_ * stored_;
		} else {
			double v1, v2, s;
			do {
				v1 = 2 * gen_.randU01() - 1;
				v2 = 2 * gen_.randU01() - 1;
				s = v1 * v1 + v2 * v2;
			} while (s > 1.0 || s == 0.0);
			double multiplier = Math.sqrt(-2.0 * Math.log(s) / s);
			stored_ = v1 * multiplier;
			isAvailable_ = true;
			return mean_ + stddev_ * v2 * multiplier;
		}
	}
}
