/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: rng/RandomManager.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package rng;

import java.lang.reflect.Constructor;

import core.Config;

/**
 * Abstract base class for random number manager implementations.
 * 
 * @author Patrick Peschlow
 * 
 * @version 1.0
 */
public abstract class RandomManager {
	/**
	 * The RNG manager class used by the simulation.
	 */
	private static RandomManager manager;

	/**
	 * Provides a reference to (and, if necessary, creates) the RNG manager.
	 * 
	 * @return a reference to the RNG manager.
	 */
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public static RandomManager getInstance() {
		if (manager == null) {
			try {
				Class managerClass = Config.getRandom().getRandomManagerClass();
				assert RandomManager.class.isAssignableFrom(managerClass) : managerClass;
				Constructor constructor = managerClass
						.getConstructor(int.class);
				manager = (RandomManager) constructor.newInstance(Config
						.getRandomSeed());
			} catch (Exception e) {
				throw new RuntimeException(
						"Exception during RandomManager init", e);
			}
		}
		return manager;
	}

	/**
	 * Provides a random number stream with a standard uniform distribution.
	 * 
	 * @return a random number stream with a standard uniform distribution.
	 */
	public abstract RandomStream getUniform01Substream();

	/**
	 * Provides a random number stream with a uniform distribution between a
	 * lower and upper bound.
	 * 
	 * @param lower
	 *            the lower bound for the distribution.
	 * @param upper
	 *            the upper bound for the distribution.
	 * @return a random number stream with a uniform standard distribution
	 *         between lower and upper.
	 */
	public abstract RandomStream getUniformSubstream(double lower, double upper);

	/**
	 * Provides a random number stream with a uniform distribution of integers
	 * between a lower and upper bound.
	 * 
	 * @param lower
	 *            the lower bound for the distribution.
	 * @param upper
	 *            the upper bound for the distribution.
	 * @return a random number stream with a uniform standard distribution of
	 *         integers between lower and upper.
	 */
	public abstract RandomStreamInt getUniformIntSubstream(int lower, int upper);

	/**
	 * Provides a random number stream with a standard normal distribution.
	 * 
	 * @return a random number stream with a standard normal distribution.
	 */
	public abstract RandomStream getNormal01Substream();

	/**
	 * Provides a random number stream with a normal distribution with the
	 * specified mean and standard deviation.
	 * 
	 * @param mean
	 *            the mean of the normal distribution.
	 * @param stddev
	 *            the standard deviation of the normal distribution.
	 * @return a random number stream with a standard normal distribution with
	 *         the specified mean and standard deviation.
	 */
	public abstract RandomStream getNormalSubstream(double mean, double stddev);
}
