/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: run/StartRemoteHost.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package run;

import core.distribution.SimClient;
import core.util.CustomLogger;
import jargs.gnu.CmdLineParser;

/**
 * Starts a remote host for distributed simulation.
 * 
 * @author Patrick Peschlow
 * @author Tobias Honecker
 * @author Thomas Saftig
 * @author Daniel Heupel
 * 
 * @version 1.0
 */
public class StartRemoteHost {
	/**
	 * The main() method for starting a remote host.
	 * 
	 * @param args
	 *            the program arguments.
	 */
	public static void main(String[] args) {
		CmdLineParser parser = new CmdLineParser();
		CmdLineParser.Option debug = parser.addStringOption('d', "debug");
		CmdLineParser.Option initiator = parser.addStringOption('i',
				"initiator");
		CmdLineParser.Option port = parser.addIntegerOption('p', "port");
		CmdLineParser.Option numLps = parser.addIntegerOption('n', "numLps");

		try {
			parser.parse(args);
		} catch (CmdLineParser.OptionException e) {
			System.err.println(e);
			printUsage();
		}

		String initiatorHostname = (String) parser.getOptionValue(initiator);
		int initiatorPort = (Integer) parser.getOptionValue(port,
				StartSimCtrl.DEFAULT_PORT);
		int numLocalLps = (Integer) parser.getOptionValue(numLps, 0);
		String debugLevel = (String) parser.getOptionValue(debug,
				CustomLogger.Level.INFO.name());

		initCustomLogger(debugLevel);

		SimClient client = new SimClient(initiatorPort, initiatorHostname,
				numLocalLps);
		client.runSimulation();
	}

	/**
	 * Initializes the custom logger.
	 * 
	 * @param debugLevel
	 *            the debug level used for the simulation run.
	 */
	private static void initCustomLogger(String debugLevel) {
		CustomLogger.setLevel(CustomLogger.Level.valueOf(debugLevel));
		System.out.println("Initialized CustomLogger to level " + debugLevel);
	}

	/**
	 * Prints usage information.
	 */
	private static void printUsage() {
		String usage = "Usage:\n" + "run.StartRemoteHost\n"
				+ "<{-i,--initiator} a_string> name of initializing host\n"
				+ "[{-p,--port} a_number] port of the initializing host\n"
				+ "[{-d,--debug} a_number] debug level\n"
				+ "[{-n,--numLps} a_number] number of local LPs\n";
		die(usage);
	}

	/**
	 * Terminate the program.
	 * 
	 * @param msg
	 *            custom error message.
	 */
	private static void die(String msg) {
		System.out.println(msg);
		System.err.println(msg);
		System.exit(-1);
	}
}