/*
 * HiPWiNS - High Performance Wireless Network Simulator
 * File: run/StartSimCtrl.java
 * 
 * Copyright 2010 Institute of Computer Science 4, University of Bonn, Germany
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package run;

import groovy.lang.Binding;
import groovy.lang.GroovyShell;
import jargs.gnu.CmdLineParser;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import modules.wireless.scenario.CreateWorld;
import core.Config;
import core.Config.SimType;
import core.ModuleRoot;
import core.distribution.SimMaster;
import core.sync.AbstractSimController;
import core.util.CustomLogger;
import core.util.EventLogger;

/**
 * Contains the <code>main()</code> method for starting the simulator.
 * 
 * @author Patrick Peschlow
 * @author Tobias Honecker
 * 
 * @version 1.0
 */
public class StartSimCtrl {
	/**
	 * Default server listener port for distributed simulation.
	 */
	public static final int DEFAULT_PORT = 11000;

	/**
	 * The main() method for starting the simulator.
	 * 
	 * @param args
	 *            the program arguments.
	 */
	public static void main(String[] args) {
		System.out.println("Starting HiPWiNS.");

		if (args.length == 0) {
			printUsage();
		}

		CmdLineParser parser = new CmdLineParser();
		CmdLineParser.Option initFileOption = parser.addStringOption('f',
				"file");
		CmdLineParser.Option debugLevelOption = parser.addStringOption('d',
				"debug");
		CmdLineParser.Option initPortOption = parser.addIntegerOption('p',
				"port");
		try {
			parser.parse(args);
		} catch (CmdLineParser.OptionException e) {
			System.err.println(e);
			printUsage();
		}
		String initFile = (String) parser.getOptionValue(initFileOption);
		String debugLevel = (String) parser.getOptionValue(debugLevelOption,
				CustomLogger.Level.INFO.name());
		int initPort = ((Integer) parser.getOptionValue(initPortOption,
				DEFAULT_PORT)).intValue();

		initCustomLogger(debugLevel);
		initSimulation(initFile, initPort);
	}

	/**
	 * Initializes the simulation.
	 * 
	 * @param initFile
	 *            the name of the initialization file.
	 * @param initPort
	 *            the port where the server listens for connections.
	 */
	private static void initSimulation(String initFile, int initPort) {
		// Initialize the central instances.
		ModuleRoot root = ModuleRoot.getRoot();
		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("Created ModuleRoot.");
		}

		// Read configuration script.
		String initFileContents = null;
		if (initFile.endsWith(".groovy")) {
			if (CustomLogger.isInfoEnabled()) {
				CustomLogger.info("Reading groovy init file " + initFile);
			}
			FileInputStream fis = null;
			byte[] buffer = new byte[(int) new File(initFile).length()];
			try {
				fis = new FileInputStream(initFile);
				fis.read(buffer);
			} catch (IOException e) {
				if (CustomLogger.isFatalEnabled()) {
					CustomLogger.fatal("Couldn't read init file!", e);
				}
			} finally {
				try {
					if (fis != null) {
						fis.close();
					}
				} catch (IOException e) {
					if (CustomLogger.isWarningEnabled()) {
						CustomLogger
								.warning(
										"Couldn't close FileInputStream for init file!",
										e);
					}
				}
			}
			initFileContents = new String(buffer);
		} else {
			if (CustomLogger.isFatalEnabled()) {
				CustomLogger
						.fatal("Unsupported init file type! (Only .groovy files supported.)");
			}
		}
		assert initFileContents != null;

		// Evaluate init file string.
		new GroovyShell(new Binding()).evaluate(initFileContents);
		if (CustomLogger.isInfoEnabled()) {
			CustomLogger.info("Evaluated init file.");
		}

		// Configuration sanity check.
		Config.sanityCheck();

		if (Config.getSimType() == SimType.DISTRIBUTED) {
			// Distribute init file to clients.
			new SimMaster(initPort).runSimulation(initFileContents);
		} else {
			// Create all modules of the simulation.
			new CreateWorld().setupSimulation();

			// Call the init methods of all modules.
			root.init();

			// Instantiate the simulation controller.
			AbstractSimController simCtrl;
			try {
				simCtrl = (AbstractSimController) Config.getSimType()
						.getSimControllerClass().newInstance();
			} catch (Exception e) {
				throw new RuntimeException("Exception during SimCtrl init", e);
			}

			// Initiate the simulation.
			simCtrl.startSim();

			// User defined actions at the end of the simulation.
			if (Config.getCallback() != null) {
				Config.getCallback().endSimulation(simCtrl);
			}

			EventLogger.closeAll();
		}
	}

	/**
	 * Initializes the custom logger.
	 * 
	 * @param debugLevel
	 *            the debug level used for the simulation run.
	 */
	private static void initCustomLogger(String debugLevel) {
		CustomLogger.setLevel(CustomLogger.Level.valueOf(debugLevel));
		System.out.println("Initialized CustomLogger to level " + debugLevel);
	}

	/**
	 * Prints usage information.
	 */
	private static void printUsage() {
		String usage = "Usage:\n"
				+ "run.StartSimCtrl\n"
				+ "<{-f,--file} a_string> init file name (.groovy or .java file)\n"
				+ "[{-d,--debug} a_string] debug level (ALL, DEBUG, INFO, WARNING, FATAL)\n"
				+ "[{-p,--port} a_number> port on which to listen (for distributed simulation)\n";
		die(usage);
	}

	/**
	 * Terminate the program.
	 * 
	 * @param msg
	 *            custom error message.
	 */
	private static void die(String msg) {
		System.out.println(msg);
		System.err.println(msg);
		System.exit(-1);
	}
}